@@ -1,8 +1,8 @@
-/* $Id: minissdp.c,v 1.95 2019/05/02 10:08:14 nanard Exp $ */
+/* $Id: minissdp.c,v 1.103 2021/05/21 22:05:17 nanard Exp $ */
 /* vim: tabstop=4 shiftwidth=4 noexpandtab
  * MiniUPnP project
  * http://miniupnp.free.fr/ or https://miniupnp.tuxfamily.org/
- * (c) 2006-2019 Thomas Bernard
+ * (c) 2006-2021 Thomas Bernard
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
@@ -197,6 +197,10 @@ OpenAndConfSSDPReceiveSocket(int ipv6)
 	{
 		syslog(LOG_WARNING, "setsockopt(udp, SO_REUSEADDR): %m");
 	}
+	if (setsockopt(s, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on)) < 0)
+	{
+		syslog(LOG_WARNING, "setsockopt(udp, SO_REUSEPORT): %m");
+	}
 #ifdef IP_RECVIF
 	/* BSD */
 	if(!ipv6) {
@@ -205,8 +209,7 @@ OpenAndConfSSDPReceiveSocket(int ipv6)
 			syslog(LOG_WARNING, "setsockopt(udp, IP_RECVIF): %m");
 		}
 	}
-#endif /* IP_RECVIF */
-#ifdef IP_PKTINFO
+#elif defined(IP_PKTINFO) /* IP_RECVIF */
 	/* Linux */
 	if(!ipv6) {
 		if(setsockopt(s, IPPROTO_IP, IP_PKTINFO, &on, sizeof(on)) < 0)
@@ -550,22 +553,27 @@ SendSSDPResponse(int s, const struct sockaddr * addr,
 		"CONFIGID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
 		"\r\n",
 #ifdef ENABLE_HTTP_DATE
-		http_date,
+		http_date,								/* DATE: */
 #endif
-		st_len, st, suffix,
-		uuidvalue, st_is_uuid ? "" : "::",
-		st_is_uuid ? 0 : st_len, st, suffix,
-		host, (unsigned int)http_port,
+		st_len, st, suffix,						/* ST: */
+		uuidvalue, st_is_uuid ? "" : "::",		/* USN: 2/5 */
+		st_is_uuid ? 0 : st_len, st, suffix,	/* USN: 3/5 */
+#ifdef DYNAMIC_OS_VERSION
+		os_version,								/* SERVER: */
+#endif
+		host, (unsigned int)http_port,			/* LOCATION: */
 #ifdef RANDOMIZE_URLS
-		random_url,
+		random_url,								/* LOCATION: 3/3 */
 #endif	/* RANDOMIZE_URLS */
 #ifdef ENABLE_HTTPS
-		host, (unsigned int)https_port,
+		host, (unsigned int)https_port,			/* SECURELOCATION.UPNP.ORG */
 #ifdef RANDOMIZE_URLS
-		random_url,
+		random_url,								/* SECURELOCATION.UPNP.ORG 3/3 */
 #endif	/* RANDOMIZE_URLS */
 #endif	/* ENABLE_HTTPS */
-		upnp_bootid, upnp_bootid, upnp_configid);
+		upnp_bootid,							/* 01-NLS: */
+		upnp_bootid,							/* BOOTID.UPNP.ORG: */
+		upnp_configid);							/* CONFIGID.UPNP.ORG: */
 	if(l<0)
 	{
 		syslog(LOG_ERR, "%s: snprintf failed %m",
@@ -684,6 +692,9 @@ SendSSDPNotify(int s, const struct sockaddr * dest, socklen_t dest_len,
 		random_url,
 #endif	/* RANDOMIZE_URLS */
 #endif	/* ENABLE_HTTPS */
+#ifdef DYNAMIC_OS_VERSION
+		os_version,						/* SERVER: */
+#endif
 		nt, suffix,						/* NT: */
 		usn1, usn2, usn3, suffix,		/* USN: */
 		upnp_bootid,					/* 01-NLS: */
@@ -799,7 +810,8 @@ SendSSDPNotifies(int s, const char * host, unsigned short http_port,
 			               known_service_types[i].s, /* ver_str,	USN: */
 			               lifetime);
 			/* for devices, also send NOTIFY on the uuid */
-			if(0==memcmp(known_service_types[i].s,
+			if(i > 0 &&	/* only known_service_types[0].s is shorter than "urn:schemas-upnp-org:device" */
+			   0==memcmp(known_service_types[i].s,
 			             "urn:schemas-upnp-org:device", sizeof("urn:schemas-upnp-org:device")-1)) {
 				SendSSDPNotify(s, (struct sockaddr *)&sockname, sockname_len, dest_str,
 				               host, http_port,
@@ -867,24 +879,12 @@ ProcessSSDPRequest(int s, unsigned short http_port)
 	struct sockaddr_in sendername;
 #endif
 	int source_ifindex = -1;
-#ifdef IP_PKTINFO
-	char cmbuf[CMSG_SPACE(sizeof(struct in_pktinfo))];
-	struct iovec iovec = {
-		.iov_base = bufr,
-		.iov_len = sizeof(bufr)
-	};
-	struct msghdr mh = {
-		.msg_name = &sendername,
-		.msg_namelen = sizeof(sendername),
-		.msg_iov = &iovec,
-		.msg_iovlen = 1,
-		.msg_control = cmbuf,
-		.msg_controllen = sizeof(cmbuf)
-	};
-	struct cmsghdr *cmptr;
-#endif /* IP_PKTINFO */
+#if defined(IP_RECVIF) || defined(IP_PKTINFO)
 #ifdef IP_RECVIF
 	char cmbuf[CMSG_SPACE(sizeof(struct sockaddr_dl))];
+#else /* IP_PKTINFO */
+	char cmbuf[CMSG_SPACE(sizeof(struct in_pktinfo))];
+#endif
 	struct iovec iovec = {
 		.iov_base = bufr,
 		.iov_len = sizeof(bufr)
@@ -898,9 +898,7 @@ ProcessSSDPRequest(int s, unsigned short http_port)
 		.msg_controllen = sizeof(cmbuf)
 	};
 	struct cmsghdr *cmptr;
-#endif /* IP_RECVIF */
 
-#if defined(IP_RECVIF) || defined(IP_PKTINFO)
 	n = recvmsg(s, &mh, 0);
 #else
 	socklen_t len_r;
@@ -925,7 +923,15 @@ ProcessSSDPRequest(int s, unsigned short http_port)
 	for(cmptr = CMSG_FIRSTHDR(&mh); cmptr != NULL; cmptr = CMSG_NXTHDR(&mh, cmptr))
 	{
 		syslog(LOG_DEBUG, "level=%d type=%d", cmptr->cmsg_level, cmptr->cmsg_type);
-#ifdef IP_PKTINFO
+#ifdef IP_RECVIF
+		if(cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_RECVIF)
+		{
+			struct sockaddr_dl *sdl;	/* fields : len, family, index, type, nlen, alen, slen, data */
+			sdl = (struct sockaddr_dl *)CMSG_DATA(cmptr);
+			syslog(LOG_DEBUG, "sdl_index = %d  %s", sdl->sdl_index, link_ntoa(sdl));
+			source_ifindex = sdl->sdl_index;
+		}
+#elif defined(IP_PKTINFO) /* IP_RECVIF */
 		if(cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_PKTINFO)
 		{
 			struct in_pktinfo * pi;	/* fields : ifindex, spec_dst, addr */
@@ -943,15 +949,6 @@ ProcessSSDPRequest(int s, unsigned short http_port)
 			source_ifindex = pi6->ipi6_ifindex;
 		}
 #endif /* defined(ENABLE_IPV6) && defined(IPV6_RECVPKTINFO) */
-#ifdef IP_RECVIF
-		if(cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_RECVIF)
-		{
-			struct sockaddr_dl *sdl;	/* fields : len, family, index, type, nlen, alen, slen, data */
-			sdl = (struct sockaddr_dl *)CMSG_DATA(cmptr);
-			syslog(LOG_DEBUG, "sdl_index = %d  %s", sdl->sdl_index, link_ntoa(sdl));
-			source_ifindex = sdl->sdl_index;
-		}
-#endif /* IP_RECVIF */
 	}
 #endif /* defined(IP_RECVIF) || defined(IP_PKTINFO) */
 #ifdef ENABLE_HTTPS
@@ -1007,11 +1004,16 @@ ProcessSSDPData(int s, const char *bufr, int n,
 	/* get the string representation of the sender address */
 	sockaddr_to_string(sender, sender_str, sizeof(sender_str));
 	lan_addr = get_lan_for_peer(sender);
-	if(source_if >= 0)
+	if(source_if > 0)
 	{
 		if(lan_addr != NULL)
 		{
+#ifndef MULTIPLE_EXTERNAL_IP
+			if(lan_addr->index != (unsigned)source_if && lan_addr->index != 0
+			   && !(lan_addr->add_indexes & (1UL << (source_if - 1))))
+#else
 			if(lan_addr->index != (unsigned)source_if && lan_addr->index != 0)
+#endif
 			{
 				syslog(LOG_WARNING, "interface index not matching %u != %d", lan_addr->index, source_if);
 			}
@@ -1088,7 +1090,7 @@ ProcessSSDPData(int s, const char *bufr, int n,
 #endif /* defined(UPNP_STRICT) || defined(DELAY_MSEARCH_RESPONSE) */
 #if defined(UPNP_STRICT)
 			/* Fix UDA-1.2.10 Man header empty or invalid */
-			else if((i < n - 4) && (strncasecmp(bufr+i, "man:", 3) == 0))
+			else if((i < n - 4) && (strncasecmp(bufr+i, "man:", 4) == 0))
 			{
 				const char * man;
 				int man_len;
@@ -1136,7 +1138,13 @@ ProcessSSDPData(int s, const char *bufr, int n,
 			syslog(LOG_INFO, "SSDP M-SEARCH from %s ST: %.*s",
 			       sender_str, st_len, st);
 			/* find in which sub network the client is */
+#ifdef ENABLE_IPV6
+			if((sender->sa_family == AF_INET) ||
+			   (sender->sa_family == AF_INET6 &&
+			    IN6_IS_ADDR_V4MAPPED(&((struct sockaddr_in6 *)sender)->sin6_addr)))
+#else
 			if(sender->sa_family == AF_INET)
+#endif
 			{
 				if (lan_addr == NULL)
 				{
@@ -1148,7 +1156,7 @@ ProcessSSDPData(int s, const char *bufr, int n,
 				announced_host = lan_addr->str;
 			}
 #ifdef ENABLE_IPV6
-			else
+			else if(sender->sa_family == AF_INET6)
 			{
 				/* IPv6 address with brackets */
 #ifdef UPNP_STRICT
@@ -1188,6 +1196,13 @@ ProcessSSDPData(int s, const char *bufr, int n,
 #endif
 			}
 #endif
+			else
+			{
+				syslog(LOG_ERR,
+				       "Unknown address family %d for client %s",
+				       sender->sa_family, sender_str);
+				return;
+			}
 			/* Responds to request with a device as ST header */
 			for(i = 0; known_service_types[i].s; i++)
 			{
@@ -1371,7 +1386,9 @@ SendSSDPbyebye(int s, const struct sockaddr * dest, socklen_t destlen,
 	             dest_str, SSDP_PORT,		/* HOST : */
 	             nt, suffix,				/* NT: */
 	             usn1, usn2, usn3, suffix,	/* USN: */
-	             upnp_bootid, upnp_bootid, upnp_configid);
+	             upnp_bootid,				/* 01-NLS: */
+	             upnp_bootid,				/* BOOTID.UPNP.ORG: */
+	             upnp_configid);			/* CONFIGID.UPNP.ORG: */
 	if(l<0)
 	{
 		syslog(LOG_ERR, "%s: snprintf error", "SendSSDPbyebye()");
@@ -1459,7 +1476,8 @@ SendSSDPGoodbye(int * sockets, int n_sockets)
 			                      known_service_types[i].s, ver_str,	/* NT: */
 			                      known_service_types[i].uuid, "::",
 			                      known_service_types[i].s); /* ver_str, USN: */
-			if(0==memcmp(known_service_types[i].s,
+			if(i > 0 &&	/* only known_service_types[0].s is shorter than "urn:schemas-upnp-org:device" */
+			   0==memcmp(known_service_types[i].s,
 			             "urn:schemas-upnp-org:device", sizeof("urn:schemas-upnp-org:device")-1))
 			{
 				ret += SendSSDPbyebye(sockets[j],
@@ -1533,9 +1551,22 @@ SubmitServicesToMiniSSDPD(const char * host, unsigned short port) {
 		CODELENGTH(l, p);
 		memcpy(p, strbuf, l);
 		p += l;
+#ifdef DYNAMIC_OS_VERSION
+		l = snprintf(strbuf, sizeof(strbuf), MINIUPNPD_SERVER_STRING,
+		             os_version);
+		if(l<0) {
+			syslog(LOG_WARNING, "SubmitServicesToMiniSSDPD: snprintf %m");
+			continue;
+		} else if((unsigned)l>=sizeof(strbuf)) {
+			l = sizeof(strbuf) - 1;
+		}
+		CODELENGTH(l, p);
+		memcpy(p, strbuf, l);
+#else
 		l = (int)strlen(MINIUPNPD_SERVER_STRING);
 		CODELENGTH(l, p);
 		memcpy(p, MINIUPNPD_SERVER_STRING, l);
+#endif
 		p += l;
 		l = snprintf(strbuf, sizeof(strbuf), "http://%s:%u" ROOTDESC_PATH,
 		             host, (unsigned int)port);
