@@ -1,8 +1,8 @@
-/* $Id: miniupnpd.c,v 1.235 2019/05/21 08:39:43 nanard Exp $ */
+/* $Id: miniupnpd.c,v 1.250 2021/05/21 22:04:34 nanard Exp $ */
 /* vim: tabstop=4 shiftwidth=4 noexpandtab
  * MiniUPnP project
- * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
- * (c) 2006-2019 Thomas Bernard
+ * http://miniupnp.free.fr/ or https://miniupnp.tuxfamily.org/
+ * (c) 2006-2021 Thomas Bernard
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
@@ -44,11 +44,24 @@
 /* for BSD's sysctl */
 #include <sys/sysctl.h>
 #endif
+#ifdef HAS_LIBCAP
+#include <sys/capability.h>
+#endif
+#ifdef HAS_LIBCAP_NG
+#include <cap-ng.h>
+#endif
 
 /* unix sockets */
 #ifdef USE_MINIUPNPDCTL
 #include <sys/un.h>
 #endif
+#ifdef ENABLE_HTTPS
+#include <openssl/crypto.h>
+#endif
+
+#ifdef DYNAMIC_OS_VERSION
+#include <sys/utsname.h>
+#endif
 
 #ifdef TOMATO
 #include <sys/stat.h>
@@ -790,7 +803,7 @@ sigusr1(int sig)
 #if 0
 	/* calling syslog() is forbidden in signal handler according to
 	 * signal(3) */
-	syslog(LOG_INFO, "received signal %d, public ip address change", sig);
+	syslog(LOG_INFO, "received signal %d, public IP address change", sig);
 #endif
 
 	should_send_public_address_change_notif = 1;
@@ -896,7 +909,7 @@ struct runtime_vars {
 /* parselanaddr()
  * parse address with mask
  * ex: 192.168.1.1/24 or 192.168.1.1/255.255.255.0
- * When MULTIPLE_EXTERNAL_IP is enabled, the ip address of the
+ * When MULTIPLE_EXTERNAL_IP is enabled, the IP address of the
  * external interface associated with the lan subnet follows.
  * ex : 192.168.1.1/24 81.21.41.11
  *
@@ -906,10 +919,10 @@ struct runtime_vars {
  *    0 : ok
  *   -1 : error */
 static int
-parselanaddr(struct lan_addr_s * lan_addr, const char * str)
+parselanaddr(struct lan_addr_s * lan_addr, const char * str, int debug_flag)
 {
 	const char * p;
-	int n;
+	unsigned int n;
 	char tmp[16];
 
 	memset(lan_addr, 0, sizeof(struct lan_addr_s));
@@ -926,6 +939,7 @@ parselanaddr(struct lan_addr_s * lan_addr, const char * str)
 		             &lan_addr->addr, &lan_addr->mask) < 0) {
 #ifdef ENABLE_IPV6
 			fprintf(stderr, "interface \"%s\" has no IPv4 address\n", str);
+			syslog(LOG_NOTICE, "interface \"%s\" has no IPv4 address\n", str);
 			lan_addr->str[0] = '\0';
 			lan_addr->addr.s_addr = htonl(0x00000000u);
 			lan_addr->mask.s_addr = htonl(0xffffffffu);
@@ -944,6 +958,13 @@ parselanaddr(struct lan_addr_s * lan_addr, const char * str)
 		if(!inet_aton(lan_addr->str, &lan_addr->addr))
 			goto parselan_error;
 	}
+	if(!addr_is_reserved(&lan_addr->addr)) {
+		INIT_PRINT_ERR("Error: LAN address contains public IP address : %s\n", lan_addr->str);
+		INIT_PRINT_ERR("Public IP address can be configured via ext_ip= option\n");
+		INIT_PRINT_ERR("LAN address should contain private address, e.g. from 192.168. block\n");
+		INIT_PRINT_ERR("Listening on public IP address is a security issue\n");
+		return -1;
+	}
 	if(*p == '/')
 	{
 		const char * q = ++p;
@@ -955,7 +976,7 @@ parselanaddr(struct lan_addr_s * lan_addr, const char * str)
 			while(*p && (*p=='.' || isdigit(*p)))
 				p++;
 			n = p - q;
-			if(n>15)
+			if(n >= sizeof(tmp))
 				goto parselan_error;
 			memcpy(tmp, q, n);
 			tmp[n] = '\0';
@@ -981,7 +1002,7 @@ parselanaddr(struct lan_addr_s * lan_addr, const char * str)
 	while(*p && isspace(*p))
 		p++;
 	if(*p) {
-		/* parse the exteral ip address to associate with this subnet */
+		/* parse the exteral IP address to associate with this subnet */
 		n = 0;
 		while(p[n] && !isspace(*p))
 			n++;
@@ -990,42 +1011,65 @@ parselanaddr(struct lan_addr_s * lan_addr, const char * str)
 			lan_addr->ext_ip_str[n] = '\0';
 			if(!inet_aton(lan_addr->ext_ip_str, &lan_addr->ext_ip_addr)) {
 				/* error */
-				fprintf(stderr, "Error parsing address : %s\n", lan_addr->ext_ip_str);
+				INIT_PRINT_ERR("Error parsing address : %s\n", lan_addr->ext_ip_str);
 				return -1;
 			}
 			if(addr_is_reserved(&lan_addr->ext_ip_addr)) {
 				/* error */
-				fprintf(stderr, "Error: option ext_ip address contains reserved / private address : %s\n", lan_addr->ext_ip_str);
+				INIT_PRINT_ERR("Error: option ext_ip address contains reserved / private address : %s\n", lan_addr->ext_ip_str);
 				return -1;
 			}
 		}
 	}
-#endif
-	if(lan_addr->ifname[0] != '\0')
-	{
-		lan_addr->index = if_nametoindex(lan_addr->ifname);
-		if(lan_addr->index == 0)
-			fprintf(stderr, "Cannot get index for network interface %s",
-			        lan_addr->ifname);
-	}
-#ifdef ENABLE_IPV6
-	else
-	{
-		fprintf(stderr,
-		        "Error: please specify LAN network interface by name instead of IPv4 address : %s\n",
-		        str);
-		return -1;
-	}
 #else
-	else
-	{
-		syslog(LOG_NOTICE, "it is advised to use network interface name instead of %s", str);
+	while(*p) {
+		/* skip spaces */
+		while(*p && isspace(*p))
+			p++;
+		if(*p) {
+			unsigned int index;
+			n = 0;
+			while(p[n] && !isspace(p[n]) && n < sizeof(tmp)) {
+				tmp[n] = p[n];
+				n++;
+			}
+			if(n >= sizeof(tmp)) {
+				INIT_PRINT_ERR("Cannot parse '%s'\n", p);
+				break;
+			}
+			tmp[n] = '\0';
+			index = if_nametoindex(tmp);
+			if(index == 0) {
+				fprintf(stderr, "Cannot get index for network interface %s\n",
+				        tmp);
+				syslog(LOG_WARNING, "Cannot get index for network interface %s\n",
+				        tmp);
+			} else {
+				lan_addr->add_indexes |= (1UL << (index - 1));
+			}
+			p += n;
+		}
 	}
 #endif
+	if(lan_addr->ifname[0] != '\0') {
+		lan_addr->index = if_nametoindex(lan_addr->ifname);
+		if(lan_addr->index == 0) {
+			fprintf(stderr, "Cannot get index for network interface %s\n",
+			        lan_addr->ifname);
+			syslog(LOG_WARNING, "Cannot get index for network interface %s\n",
+			        lan_addr->ifname);
+		}
+	} else {
+#ifdef ENABLE_IPV6
+		INIT_PRINT_ERR("Error: please specify LAN network interface by name instead of IPv4 address : %s\n", str);
+		return -1;
+#else
+		syslog(LOG_NOTICE, "it is advised to use network interface name instead of %s", str);
+#endif
+	}
 	return 0;
 parselan_error:
-	fprintf(stderr, "Error parsing address/mask (or interface name) : %s\n",
-	        str);
+	INIT_PRINT_ERR("Error parsing address/mask (or interface name) : %s\n", str);
 	return -1;
 }
 
@@ -1054,11 +1098,20 @@ int update_ext_ip_addr_from_stun(int init)
 
 	if ((init || disable_port_forwarding) && !restrictive_nat) {
 		if (addr_is_reserved(&if_addr))
-			syslog(LOG_INFO, "STUN: ext interface %s with IP address %s is now behind unrestricted NAT 1:1 with public IP address %s: Port forwarding is now enabled", ext_if_name, if_addr_str, ext_addr_str);
+			syslog(LOG_INFO, "STUN: ext interface %s with IP address %s is now behind unrestricted full-cone NAT 1:1 with public IP address %s and firewall does not block incoming connections set by miniunnpd", ext_if_name, if_addr_str, ext_addr_str);
 		else
-			syslog(LOG_INFO, "STUN: ext interface %s has now public IP address %s: Port forwarding is now enabled", ext_if_name, if_addr_str);
+			syslog(LOG_INFO, "STUN: ext interface %s has now public IP address %s and firewall does not block incoming connections set by miniunnpd", ext_if_name, if_addr_str);
+		syslog(LOG_INFO, "Port forwarding is now enabled");
 	} else if ((init || !disable_port_forwarding) && restrictive_nat) {
-		syslog(LOG_WARNING, "STUN: ext interface %s with IP address %s is now behind restrictive NAT with public IP address %s: Port forwarding is now impossible", ext_if_name, if_addr_str, ext_addr_str);
+		if (addr_is_reserved(&if_addr)) {
+			syslog(LOG_WARNING, "STUN: ext interface %s with private IP address %s is now behind restrictive or symmetric NAT with public IP address %s which does not support port forwarding", ext_if_name, if_addr_str, ext_addr_str);
+			syslog(LOG_WARNING, "NAT on upstream router blocks incoming connections set by miniupnpd");
+			syslog(LOG_WARNING, "Turn off NAT on upstream router or change it to full-cone NAT 1:1 type");
+		} else {
+			syslog(LOG_WARNING, "STUN: ext interface %s has now public IP address %s but firewall filters incoming connections set by miniunnpd", ext_if_name, if_addr_str);
+			syslog(LOG_WARNING, "Check configuration of firewall on local machine and also on upstream router");
+		}
+		syslog(LOG_WARNING, "Port forwarding is now disabled");
 	} else {
 		syslog(LOG_INFO, "STUN: ... done");
 	}
@@ -1113,8 +1166,11 @@ static int
 init(int argc, char * * argv, struct runtime_vars * v)
 {
 	int i;
+#ifndef NO_BACKGROUND_NO_PIDFILE
 	int pid;
+#endif
 	int debug_flag = 0;
+	int verbosity_level = 0;	/* for determining setlogmask() */
 	int openlog_option;
 	struct in_addr addr;
 	struct sigaction sa;
@@ -1130,9 +1186,20 @@ init(int argc, char * * argv, struct runtime_vars * v)
 	/* only print usage if -h is used */
 	for(i=1; i<argc; i++)
 	{
-		if(0 == strcmp(argv[i], "-h"))
+		if(0 == strcmp(argv[i], "-h") || 0 == strcmp(argv[i], "--help"))
 			goto print_usage;
+		if(0 == strcmp(argv[i], "-d"))
+			debug_flag = 1;
 	}
+
+	openlog_option = LOG_PID|LOG_CONS;
+	if(debug_flag)
+	{
+		openlog_option |= LOG_PERROR;	/* also log on stderr */
+	}
+
+	openlog("miniupnpd", openlog_option, LOG_MINIUPNPD);
+
 #ifndef DISABLE_CONFIG_FILE
 	/* first check if "-f" option is used */
 	for(i=2; i<argc; i++)
@@ -1163,11 +1230,14 @@ init(int argc, char * * argv, struct runtime_vars * v)
 #ifndef DISABLE_CONFIG_FILE
 	/* read options file first since
 	 * command line arguments have final say */
-	if(readoptionsfile(optionsfile) < 0)
+	if(readoptionsfile(optionsfile, debug_flag) < 0)
 	{
 		/* only error if file exists or using -f */
 		if(access(optionsfile, F_OK) == 0 || options_flag)
-			fprintf(stderr, "Error reading configuration file %s\n", optionsfile);
+		{
+			INIT_PRINT_ERR("Error reading configuration file %s\n", optionsfile);
+			return 1;
+		}
 	}
 	else
 	{
@@ -1200,18 +1270,18 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				lan_addr = (struct lan_addr_s *) malloc(sizeof(struct lan_addr_s));
 				if (lan_addr == NULL)
 				{
-					fprintf(stderr, "malloc(sizeof(struct lan_addr_s)): %m");
-					break;
+					INIT_PRINT_ERR("malloc(sizeof(struct lan_addr_s)): %m");
+					return 1;
 				}
-				if(parselanaddr(lan_addr, ary_options[i].value) != 0)
+				if(parselanaddr(lan_addr, ary_options[i].value, debug_flag) != 0)
 				{
-					fprintf(stderr, "can't parse \"%s\" as a valid "
+					INIT_PRINT_ERR("can't parse \"%s\" as a valid "
 #ifndef ENABLE_IPV6
 					        "LAN address or "
 #endif
 					        "interface name\n", ary_options[i].value);
 					free(lan_addr);
-					break;
+					return 1;
 				}
 				LIST_INSERT_HEAD(&lan_addrs, lan_addr, list);
 				break;
@@ -1219,9 +1289,14 @@ init(int argc, char * * argv, struct runtime_vars * v)
 			case UPNPIPV6_LISTENING_IP:
 				if (inet_pton(AF_INET6, ary_options[i].value, &ipv6_bind_addr) < 1)
 				{
-					fprintf(stderr, "can't parse \"%s\" as valid IPv6 listening address", ary_options[i].value);
+					INIT_PRINT_ERR("can't parse \"%s\" as valid IPv6 listening address", ary_options[i].value);
+					return 1;
 				}
 				break;
+			case UPNPIPV6_DISABLE:
+				if(strcmp(ary_options[i].value, "yes") == 0)
+					SETFLAG(IPV6DISABLEDMASK);
+				break;
 #endif /* ENABLE_IPV6 */
 			case UPNPPORT:
 				v->port = atoi(ary_options[i].value);
@@ -1268,15 +1343,15 @@ init(int argc, char * * argv, struct runtime_vars * v)
 #endif	/* ENABLE_MANUFACTURER_INFO_CONFIGURATION */
 #ifdef USE_NETFILTER
 			case UPNPFORWARDCHAIN:
-				miniupnpd_forward_chain = ary_options[i].value;
+				set_rdr_name(RDR_FORWARD_CHAIN_NAME, ary_options[i].value);
 				break;
 			case UPNPNATCHAIN:
-				miniupnpd_nat_chain = ary_options[i].value;
+				set_rdr_name(RDR_NAT_PREROUTING_CHAIN_NAME, ary_options[i].value);
 				break;
 			case UPNPNATPOSTCHAIN:
-				miniupnpd_nat_postrouting_chain = ary_options[i].value;
+				set_rdr_name(RDR_NAT_POSTROUTING_CHAIN_NAME, ary_options[i].value);
 				break;
-#endif	/* USE_NETFILTER */
+#endif    /* USE_NETFILTER */
 			case UPNPNOTIFY_INTERVAL:
 				v->notify_interval = atoi(ary_options[i].value);
 				break;
@@ -1375,25 +1450,26 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				if (strcmp(ary_options[i].value, "yes") == 0)
 					SETFLAG(FORCEIGDDESCV1MASK);
 				else if (strcmp(ary_options[i].value, "no") != 0 ) {
-					fprintf(stderr, "force_igd_desc_v1 can only be yes or no\n");
+					INIT_PRINT_ERR("force_igd_desc_v1 can only be yes or no\n");
+					return 1;
 				}
 				break;
 #endif
 			default:
-				fprintf(stderr, "Unknown option in file %s\n",
+				INIT_PRINT_ERR("Unknown option in file %s\n",
 				        optionsfile);
 			}
 		}
 #ifdef ENABLE_PCP
 		/* if lifetimes are inverse */
 		if (min_lifetime >= max_lifetime) {
-			fprintf(stderr, "Minimum lifetime (%lu) is greater than or equal to maximum lifetime (%lu).\n", min_lifetime, max_lifetime);
-			fprintf(stderr, "Check your configuration file.\n");
+			INIT_PRINT_ERR("Minimum lifetime (%lu) is greater than or equal to maximum lifetime (%lu).\n", min_lifetime, max_lifetime);
+			INIT_PRINT_ERR("Check your configuration file.\n");
 			return 1;
 		}
 #endif	/* ENABLE_PCP */
 		if (GETFLAG(PERFORMSTUNMASK) && !ext_stun_host) {
-			fprintf(stderr, "You must specify ext_stun_host= when ext_perform_stun=yes\n");
+			INIT_PRINT_ERR("You must specify ext_stun_host= when ext_perform_stun=yes\n");
 			return 1;
 		}
 	}
@@ -1404,10 +1480,23 @@ init(int argc, char * * argv, struct runtime_vars * v)
 	{
 		if(argv[i][0]!='-')
 		{
-			fprintf(stderr, "Unknown option: %s\n", argv[i]);
+			INIT_PRINT_ERR("Unknown option: %s\n", argv[i]);
+			goto print_usage;
 		}
 		else switch(argv[i][1])
 		{
+		case 'v':
+			{
+				int j;
+				for (j = 1; argv[i][j] != '\0'; j++)
+					verbosity_level++;
+			}
+			break;
+#ifdef ENABLE_IPV6
+		case '4':
+			SETFLAG(IPV6DISABLEDMASK);
+			break;
+#endif
 #ifdef IGD_V2
 		case '1':
 			SETFLAG(FORCEIGDDESCV1MASK);
@@ -1416,8 +1505,10 @@ init(int argc, char * * argv, struct runtime_vars * v)
 		case 'b':
 			if(i+1 < argc) {
 				upnp_bootid = (unsigned int)strtoul(argv[++i], NULL, 10);
-			} else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'o':
 			if(i+1 < argc) {
@@ -1433,49 +1524,63 @@ init(int argc, char * * argv, struct runtime_vars * v)
 					}
 				} else
 					use_ext_ip_addr = argv[i];
-			} else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 't':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				v->notify_interval = atoi(argv[++i]);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'r':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				v->clean_ruleset_interval = atoi(argv[++i]);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'u':
 			if(i+1 < argc) {
 				strncpy(uuidvalue_igd+5, argv[++i], strlen(uuidvalue_igd+5) + 1);
 				complete_uuidvalues();
-			} else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #ifdef ENABLE_MANUFACTURER_INFO_CONFIGURATION
 		case 'z':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				strncpy(friendly_name, argv[++i], FRIENDLY_NAME_MAX_LEN);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			friendly_name[FRIENDLY_NAME_MAX_LEN-1] = '\0';
 			break;
 #endif	/* ENABLE_MANUFACTURER_INFO_CONFIGURATION */
 		case 's':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				strncpy(serialnumber, argv[++i], SERIALNUMBER_MAX_LEN);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			serialnumber[SERIALNUMBER_MAX_LEN-1] = '\0';
 			break;
 		case 'm':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				strncpy(modelnumber, argv[++i], MODELNUMBER_MAX_LEN);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			modelnumber[MODELNUMBER_MAX_LEN-1] = '\0';
 			break;
 #ifdef ENABLE_NATPMP
@@ -1501,55 +1606,67 @@ init(int argc, char * * argv, struct runtime_vars * v)
 			SETFLAG(SECUREMODEMASK);
 			break;
 		case 'i':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				ext_if_name = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #ifdef ENABLE_IPV6
 		case 'I':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				ext_if_name6 = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #endif
 #ifdef USE_PF
 		case 'q':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				queue = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'T':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				tag = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #endif	/* USE_PF */
 		case 'p':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				v->port = atoi(argv[++i]);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #ifdef ENABLE_HTTPS
 		case 'H':
-			if(i+1 < argc)
+			if(i+1 < argc) {
 				v->https_port = atoi(argv[++i]);
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 #endif	/* ENABLE_HTTPS */
 #ifdef ENABLE_NFQUEUE
 		case 'Q':
-			if(i+1<argc)
-			{
+			if(i+1<argc) {
 				nfqueue = atoi(argv[++i]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
 			}
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
 			break;
 		case 'n':
 			if (i+1 < argc) {
@@ -1557,36 +1674,42 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				if(n_nfqix < MAX_LAN_ADDR) {
 					nfqix[n_nfqix++] = if_nametoindex(argv[i]);
 				} else {
-					fprintf(stderr,"Too many nfq interfaces. Ignoring %s\n", argv[i]);
+					INIT_PRINT_ERR( "Too many nfq interfaces. Ignoring %s\n", argv[i]);
 				}
 			} else {
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
 			}
 			break;
 #endif	/* ENABLE_NFQUEUE */
+#ifndef NO_BACKGROUND_NO_PIDFILE
 		case 'P':
 			if(i+1 < argc)
 				pidfilename = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
-		case 'd':
-			debug_flag = 1;
+#endif
+		case 'd':	/* discarding */
 			break;
 		case 'w':
 			if(i+1 < argc)
 				presurl = argv[++i];
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'B':
-			if(i+2<argc)
-			{
+			if(i+2<argc) {
 				downstream_bitrate = strtoul(argv[++i], 0, 0);
 				upstream_bitrate = strtoul(argv[++i], 0, 0);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes two argument.\n", argv[i][1]);
+				goto print_usage;
 			}
-			else
-				fprintf(stderr, "Option -%c takes two arguments.\n", argv[i][1]);
 			break;
 		case 'a':
 #ifndef MULTIPLE_EXTERNAL_IP
@@ -1596,18 +1719,18 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				lan_addr = (struct lan_addr_s *) malloc(sizeof(struct lan_addr_s));
 				if (lan_addr == NULL)
 				{
-					fprintf(stderr, "malloc(sizeof(struct lan_addr_s)): %m");
-					break;
+					INIT_PRINT_ERR("malloc(sizeof(struct lan_addr_s)): %m");
+					return 1;
 				}
-				if(parselanaddr(lan_addr, argv[i]) != 0)
+				if(parselanaddr(lan_addr, argv[i], debug_flag) != 0)
 				{
-					fprintf(stderr, "can't parse \"%s\" as a valid "
+					INIT_PRINT_ERR("can't parse \"%s\" as a valid "
 #ifndef ENABLE_IPV6
 					        "LAN address or "
 #endif	/* #ifndef ENABLE_IPV6 */
 					        "interface name\n", argv[i]);
 					free(lan_addr);
-					break;
+					return 1;
 				}
 				/* check if we already have this address */
 				for(lan_addr2 = lan_addrs.lh_first; lan_addr2 != NULL; lan_addr2 = lan_addr2->list.le_next)
@@ -1617,33 +1740,34 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				}
 				if (lan_addr2 == NULL)
 					LIST_INSERT_HEAD(&lan_addrs, lan_addr, list);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
 			}
-			else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
 #else	/* #ifndef MULTIPLE_EXTERNAL_IP */
 			if(i+2 < argc)
 			{
 				char *val = calloc((strlen(argv[i+1]) + strlen(argv[i+2]) + 2), sizeof(char));
 				if (val == NULL)
 				{
-					fprintf(stderr, "memory allocation error for listen address storage\n");
-					break;
+					INIT_PRINT_ERR("memory allocation error for listen address storage\n");
+					return 1;
 				}
 				sprintf(val, "%s %s", argv[i+1], argv[i+2]);
 
 				lan_addr = (struct lan_addr_s *) malloc(sizeof(struct lan_addr_s));
 				if (lan_addr == NULL)
 				{
-					fprintf(stderr, "malloc(sizeof(struct lan_addr_s)): %m");
+					INIT_PRINT_ERR("malloc(sizeof(struct lan_addr_s)): %m");
 					free(val);
-					break;
+					return 1;
 				}
-				if(parselanaddr(lan_addr, val) != 0)
+				if(parselanaddr(lan_addr, val, debug_flag) != 0)
 				{
-					fprintf(stderr, "can't parse \"%s\" as a valid LAN address or interface name\n", val);
+					INIT_PRINT_ERR("can't parse \"%s\" as a valid LAN address or interface name\n", val);
 					free(lan_addr);
 					free(val);
-					break;
+					return 1;
 				}
 				/* check if we already have this address */
 				for(lan_addr2 = lan_addrs.lh_first; lan_addr2 != NULL; lan_addr2 = lan_addr2->list.le_next)
@@ -1656,9 +1780,10 @@ init(int argc, char * * argv, struct runtime_vars * v)
 
 				free(val);
 				i+=2;
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
 			}
-			else
-				fprintf(stderr, "Option -%c takes two arguments.\n", argv[i][1]);
 #endif 	/* #ifndef MULTIPLE_EXTERNAL_IP */
 			break;
 		case 'A':
@@ -1666,23 +1791,28 @@ init(int argc, char * * argv, struct runtime_vars * v)
 				void * tmp;
 				tmp = realloc(upnppermlist, sizeof(struct upnpperm) * (num_upnpperm+1));
 				if(tmp == NULL) {
-					fprintf(stderr, "memory allocation error for permission\n");
+					INIT_PRINT_ERR("memory allocation error for permission\n");
+					return 1;
 				} else {
 					upnppermlist = tmp;
 					if(read_permission_line(upnppermlist + num_upnpperm, argv[++i]) >= 0) {
 						num_upnpperm++;
 					} else {
-						fprintf(stderr, "Permission rule parsing error :\n%s\n", argv[i]);
+						INIT_PRINT_ERR("Permission rule parsing error :\n%s\n", argv[i]);
+						return 1;
 					}
 				}
-			} else
-				fprintf(stderr, "Option -%c takes one argument.\n", argv[i][1]);
+			} else {
+				INIT_PRINT_ERR("Option -%c takes one argument.\n", argv[i][1]);
+				goto print_usage;
+			}
 			break;
 		case 'f':
 			i++;	/* discarding, the config file is already read */
 			break;
 		default:
-			fprintf(stderr, "Unknown option: %s\n", argv[i]);
+			INIT_PRINT_ERR("Unknown option: %s\n", argv[i]);
+			goto print_usage;
 		}
 	}
 	if(!ext_if_name || !lan_addrs.lh_first) {
@@ -1697,21 +1827,22 @@ init(int argc, char * * argv, struct runtime_vars * v)
 #endif
 
 	if (use_ext_ip_addr && GETFLAG(PERFORMSTUNMASK)) {
-		fprintf(stderr, "Error: options ext_ip= and ext_perform_stun=yes cannot be specified together\n");
+		INIT_PRINT_ERR("Error: options ext_ip= and ext_perform_stun=yes cannot be specified together\n");
 		return 1;
 	}
 
 	if (use_ext_ip_addr) {
 		if (inet_pton(AF_INET, use_ext_ip_addr, &addr) != 1) {
-			fprintf(stderr, "Error: option ext_ip contains invalid address %s\n", use_ext_ip_addr);
+			INIT_PRINT_ERR("Error: option ext_ip contains invalid address %s\n", use_ext_ip_addr);
 			return 1;
 		}
 		if (addr_is_reserved(&addr)) {
-			fprintf(stderr, "Error: option ext_ip contains reserved / private address %s, not public routable\n", use_ext_ip_addr);
+			INIT_PRINT_ERR("Error: option ext_ip contains reserved / private address %s, not public routable\n", use_ext_ip_addr);
 			return 1;
 		}
 	}
 
+#ifndef NO_BACKGROUND_NO_PIDFILE
 	if(debug_flag)
 	{
 		pid = getpid();
@@ -1727,26 +1858,32 @@ init(int argc, char * * argv, struct runtime_vars * v)
 		pid = daemonize();
 #endif
 	}
-
-	openlog_option = LOG_PID|LOG_CONS;
-	if(debug_flag)
-	{
-		openlog_option |= LOG_PERROR;	/* also log on stderr */
-	}
-
-	openlog("miniupnpd", openlog_option, LOG_MINIUPNPD);
+#endif
 
 	if(!debug_flag)
 	{
-		/* speed things up and ignore LOG_INFO and LOG_DEBUG */
-		setlogmask(LOG_UPTO(LOG_NOTICE));
+		switch (verbosity_level)
+		{
+		case 0:
+			/* speed things up and ignore LOG_INFO and LOG_DEBUG */
+			setlogmask(LOG_UPTO(LOG_NOTICE));
+			break;
+		case 1:
+			/* ignore LOG_DEBUG */
+			setlogmask(LOG_UPTO(LOG_INFO));
+			break;
+		case 2:
+			setlogmask(LOG_UPTO(LOG_DEBUG));
+		}
 	}
 
+#ifndef NO_BACKGROUND_NO_PIDFILE
 	if(checkforrunning(pidfilename) < 0)
 	{
 		syslog(LOG_ERR, "MiniUPnPd is already running. EXITING");
 		return 1;
 	}
+#endif
 
 #ifdef TOMATO
 	syslog(LOG_NOTICE, "version " MINIUPNPD_VERSION " started");
@@ -1823,8 +1960,10 @@ init(int argc, char * * argv, struct runtime_vars * v)
 #endif
 #endif
 
+#ifndef NO_BACKGROUND_NO_PIDFILE
 	if(writepidfile(pidfilename, pid) < 0)
 		pidfilename = NULL;
+#endif
 
 #ifdef ENABLE_LEASEFILE
 	/*remove(lease_file);*/
@@ -1839,13 +1978,15 @@ init(int argc, char * * argv, struct runtime_vars * v)
 	return 0;
 print_usage:
 	fprintf(stderr, "Usage:\n\t"
+	        "%s --version\n\t"
+	        "%s --help\n\t"
 	        "%s "
 #ifndef DISABLE_CONFIG_FILE
 			"[-f config_file] "
 #endif
 			"[-i ext_ifname] "
 #ifdef ENABLE_IPV6
-			"[-I ext_ifname6] "
+			"[-I ext_ifname6] [-4] "
 #endif
 			"[-o ext_ip]\n"
 #ifndef MULTIPLE_EXTERNAL_IP
@@ -1856,7 +1997,7 @@ print_usage:
 #ifdef ENABLE_HTTPS
 			" [-H https_port]"
 #endif
-			" [-p port] [-d]"
+			" [-p port] [-d] [-v]"
 #if defined(USE_PF) || defined(USE_IPF)
 			" [-L]"
 #endif
@@ -1867,7 +2008,10 @@ print_usage:
 			"\n"
 			/*"[-l logfile] " not functionnal */
 			"\t\t[-u uuid] [-s serial] [-m model_number] \n"
-			"\t\t[-t notify_interval] [-P pid_filename] "
+			"\t\t[-t notify_interval] "
+#ifndef NO_BACKGROUND_NO_PIDFILE
+			"[-P pid_filename] "
+#endif
 #ifdef ENABLE_MANUFACTURER_INFO_CONFIGURATION
 			"[-z fiendly_name]"
 #endif
@@ -1885,10 +2029,15 @@ print_usage:
 			"\n"
 	        "\nNotes:\n\tThere can be one or several listening_ips.\n"
 	        "\tNotify interval is in seconds. Default is 30 seconds.\n"
+#ifndef NO_BACKGROUND_NO_PIDFILE
 			"\tDefault pid file is '%s'.\n"
+#endif
 			"\tDefault config file is '%s'.\n"
 			"\tWith -d miniupnpd will run as a standard program.\n"
 			"\t-o argument is either an IPv4 address or \"STUN:host[:port]\".\n"
+#ifdef ENABLE_IPV6
+			"\t-4 disable IPv6\n"
+#endif
 #if defined(USE_PF) || defined(USE_IPF)
 			"\t-L sets packet log in pf and ipf on.\n"
 #endif
@@ -1917,8 +2066,13 @@ print_usage:
 #ifdef IGD_V2
 			"\t-1 force reporting IGDv1 in rootDesc *use with care*\n"
 #endif
-			"\t-h prints this help and quits.\n"
-	        "", argv[0], pidfilename, DEFAULT_CONFIG);
+			"\t-v enables LOG_INFO messages, -vv LOG_DEBUG as well (default with -d)\n"
+			"\t-h / --help prints this help and quits.\n"
+	        "", argv[0], argv[0], argv[0],
+#ifndef NO_BACKGROUND_NO_PIDFILE
+			pidfilename,
+#endif
+			DEFAULT_CONFIG);
 	return 1;
 }
 
@@ -1979,6 +2133,39 @@ main(int argc, char * * argv)
 	unsigned int next_pinhole_ts;
 #endif
 
+	for(i = 0; i < argc; i++) {
+		if(strcmp(argv[i], "version") == 0 || strcmp(argv[i], "--version") == 0) {
+			puts("miniupnpd " MINIUPNPD_VERSION
+#ifdef MINIUPNPD_GIT_REF
+			     " " MINIUPNPD_GIT_REF
+#endif
+			     " " __DATE__ );
+#ifdef USE_PF
+			puts("using pf backend");
+#endif
+#ifdef USE_IPF
+			puts("using ipf backend");
+#endif
+#ifdef USE_IPFW
+			puts("using ipfw backend");
+#endif
+#ifdef USE_IPTABLES
+			puts("using netfilter(iptables) backend");
+#endif
+#ifdef USE_NFTABLES
+			puts("using netfilter(nftables) backend");
+#endif
+#ifdef ENABLE_HTTPS
+#ifdef OPENSSL_VERSION
+			puts(OpenSSL_version(OPENSSL_VERSION));
+#else
+			puts(SSLeay_version(SSLEAY_VERSION));
+#endif
+#endif
+			return 0;
+		}
+	}
+	memset(&v, 0, sizeof(v));
 	if(init(argc, argv, &v) != 0)
 		return 1;
 #ifdef ENABLE_HTTPS
@@ -2055,17 +2242,31 @@ main(int argc, char * * argv)
 		} else if (addr_is_reserved(&addr)) {
 			syslog(LOG_INFO, "Reserved / private IP address %s on ext interface %s: Port forwarding is impossible", if_addr, ext_if_name);
 			syslog(LOG_INFO, "You are probably behind NAT, enable option ext_perform_stun=yes to detect public IP address");
+			syslog(LOG_INFO, "Or use ext_ip= / -o option to declare public IP address");
+			syslog(LOG_INFO, "Public IP address is required by UPnP/PCP/PMP protocols and clients do not work without it");
 			disable_port_forwarding = 1;
 		}
 	}
 
+#ifdef DYNAMIC_OS_VERSION
+	{
+		struct utsname utsname;
+		if (uname(&utsname) < 0) {
+			syslog(LOG_ERR, "uname(): %m");
+			os_version = strdup("unknown");
+		} else {
+			os_version = strdup(utsname.release);
+		}
+	}
+#endif /* DYNAMIC_OS_VERSION */
+
 	if(GETFLAG(ENABLEUPNPMASK))
 	{
 		unsigned short listen_port;
 		listen_port = (v.port > 0) ? v.port : 0;
 		/* open socket for HTTP connections. Listen on the 1st LAN address */
 #ifdef ENABLE_IPV6
-		shttpl = OpenAndConfHTTPSocket(&listen_port, 1);
+		shttpl = OpenAndConfHTTPSocket(&listen_port, !GETFLAG(IPV6DISABLEDMASK));
 #else /* ENABLE_IPV6 */
 		shttpl = OpenAndConfHTTPSocket(&listen_port);
 #endif /* ENABLE_IPV6 */
@@ -2091,7 +2292,7 @@ main(int argc, char * * argv)
 		/* https */
 		listen_port = (v.https_port > 0) ? v.https_port : 0;
 #ifdef ENABLE_IPV6
-		shttpsl = OpenAndConfHTTPSocket(&listen_port, 1);
+		shttpsl = OpenAndConfHTTPSocket(&listen_port, !GETFLAG(IPV6DISABLEDMASK));
 #else /* ENABLE_IPV6 */
 		shttpsl = OpenAndConfHTTPSocket(&listen_port);
 #endif /* ENABLE_IPV6 */
@@ -2229,6 +2430,69 @@ main(int argc, char * * argv)
 	}
 #endif
 
+	/* drop privileges */
+#ifdef HAS_PLEDGE
+	/* mcast ? unix ? */
+	if (pledge("stdio inet pf", NULL) < 0) {
+		syslog(LOG_ERR, "pledge(): %m");
+		return 1;
+	}
+#endif /* HAS_PLEDGE */
+#ifdef HAS_LIBCAP
+	{
+		cap_t caps = cap_get_proc();
+		if (caps == NULL) {
+			syslog(LOG_ERR, "cap_get_proc(): %m");
+		} else {
+			static const cap_value_t cap_list[3] = { CAP_NET_BROADCAST, CAP_NET_ADMIN, CAP_NET_RAW };
+			char * txt_caps = cap_to_text(caps, NULL);
+			if (txt_caps == NULL) {
+				syslog(LOG_ERR, "cap_to_text(): %m");
+			} else {
+				syslog(LOG_DEBUG, "capabilities %s", txt_caps);
+				if (cap_free(txt_caps) < 0) {
+					syslog(LOG_ERR, "cap_free(): %m");
+				}
+			}
+			if (cap_clear(caps) < 0) {
+				syslog(LOG_ERR, "cap_clear(): %m");
+			}
+			if (cap_set_flag(caps, CAP_PERMITTED, sizeof(cap_list)/sizeof(cap_list[0]), cap_list, CAP_SET) < 0) {
+				syslog(LOG_ERR, "cap_set_flag(): %m");
+			}
+			if (cap_set_flag(caps, CAP_EFFECTIVE, sizeof(cap_list)/sizeof(cap_list[0]), cap_list, CAP_SET) < 0) {
+				syslog(LOG_ERR, "cap_set_flag(): %m");
+			}
+			txt_caps = cap_to_text(caps, NULL);
+			if (txt_caps == NULL) {
+				syslog(LOG_ERR, "cap_to_text(): %m");
+			} else {
+				syslog(LOG_DEBUG, "capabilities %s", txt_caps);
+				if (cap_free(txt_caps) < 0) {
+					syslog(LOG_ERR, "cap_free(): %m");
+				}
+			}
+			if (cap_set_proc(caps) < 0) {
+				syslog(LOG_ERR, "cap_set_proc(): %m");
+			}
+			if (cap_free(caps) < 0) {
+				syslog(LOG_ERR, "cap_free(): %m");
+			}
+		}
+	}
+#endif /* HAS_LIBCAP */
+#ifdef HAS_LIBCAP_NG
+	capng_setpid(getpid());
+	capng_clear(CAPNG_SELECT_BOTH);
+	if (capng_updatev(CAPNG_ADD, CAPNG_EFFECTIVE|CAPNG_PERMITTED, CAP_NET_BROADCAST, CAP_NET_ADMIN, CAP_NET_RAW, -1) < 0) {
+		syslog(LOG_ERR, "capng_updatev() failed");
+	} else {
+		if (capng_apply(CAPNG_SELECT_BOTH) < 0) {
+			syslog(LOG_ERR, "capng_apply() failed");
+		}
+	}
+#endif /* HAS_LIBCAP_NG */
+
 	/* main loop */
 	while(!quitting)
 	{
@@ -2263,6 +2527,8 @@ main(int argc, char * * argv)
 					} else if (!disable_port_forwarding && reserved) {
 						syslog(LOG_INFO, "Reserved / private IP address %s on ext interface %s: Port forwarding is impossible", if_addr, ext_if_name);
 						syslog(LOG_INFO, "You are probably behind NAT, enable option ext_perform_stun=yes to detect public IP address");
+						syslog(LOG_INFO, "Or use ext_ip= / -o option to declare public IP address");
+						syslog(LOG_INFO, "Public IP address is required by UPnP/PCP/PMP protocols and clients do not work without it");
 					}
 					disable_port_forwarding = reserved;
 				}
@@ -2876,10 +3142,12 @@ shutdown:
 	}
 
 	/* remove pidfile */
+#ifndef NO_BACKGROUND_NO_PIDFILE
 	if(pidfilename && (unlink(pidfilename) < 0))
 	{
 		syslog(LOG_ERR, "Failed to remove pidfile %s: %m", pidfilename);
 	}
+#endif
 
 	/* delete lists */
 	while(lan_addrs.lh_first != NULL)
@@ -2896,11 +3164,17 @@ shutdown:
 	free(snatpmp);
 #endif
 	free(snotify);
-	closelog();
+
+	shutdown_redirect();
+
 #ifndef DISABLE_CONFIG_FILE
+	/* in some case shutdown_redirect() may need the option values */
 	freeoptions();
 #endif
-
+#ifdef DYNAMIC_OS_VERSION
+	free(os_version);
+#endif
+	closelog();
 	return 0;
 }
 
