@@ -166,14 +166,11 @@ function reticuleDesignCheck()
 	{
 		setReticuleButton(4, _("Design (F4)"), "image_design_up.png", "image_design_down.png");
 		setMiniMap(true);
-		setDesign(true);
 	}
 	else
 	{
 		setReticuleButton(4, _("Design - construct HQ first"), "", "");
 		setMiniMap(false);
-		// Will enable templates that are researched whenever the reticule buttons update.
-		setDesign(false);
 	}
 }
 
@@ -235,6 +232,9 @@ function reticuleUpdate(obj, eventType)
 
 function setupGame()
 {
+	//Use dark fog for campaign
+	setRevealStatus(false);
+
 	if (tilesetType === "ARIZONA")
 	{
 		setCampaignNumber(1);
@@ -261,11 +261,11 @@ function setupGame()
 
 	// Disable by default
 	setMiniMap(false);
-	setDesign(false);
+	// Enable all templates
+	setDesign(true);
 
+	showInterface(); // init buttons. This MUST come before setting the reticule button data
 	setMainReticule();
-	showInterface();
-	hackPlayIngameAudio();
 	mainReticule = true;
 	queue("resetPower", 1000);
 }
@@ -278,11 +278,12 @@ function eventGameLoaded()
 function eventGameInit()
 {
 	setupGame();
+	setTimer("autoSave", 10*60*1000);
 }
 
 function setLimits()
 {
-	setDroidLimit(selectedPlayer, 100, DROID_ANY);
+	setDroidLimit(selectedPlayer, 101, DROID_ANY); //note: the transporter is a unit you own
 	setDroidLimit(selectedPlayer, 10, DROID_COMMAND);
 	setDroidLimit(selectedPlayer, 15, DROID_CONSTRUCT);
 
@@ -384,6 +385,14 @@ function eventStructureDemolish(struct, droid)
 	}
 }
 
+function eventStructureUpgradeStarted(struct)
+{
+	if (struct.player === selectedPlayer)
+	{
+		reticuleUpdate(struct, TRANSFER_LIKE_EVENT);
+	}
+}
+
 function eventDestroyed(victim)
 {
 	if (victim.player === selectedPlayer)
@@ -411,28 +420,7 @@ function eventTransporterLanded(transport)
 
 function eventResearched(research, structure, player)
 {
-	//debug("RESEARCH : " + research.fullname + "(" + research.name + ") for " + player);
-	// iterate over all results
-	for (var i = 0; i < research.results.length; i++)
-	{
-		var v = research.results[i];
-		//debug("    RESULT : class=" + v['class'] + " parameter=" + v['parameter'] + " value=" + v['value'] + " filter=" + v['filterParameter'] + " filterparam=" + v['filterParameter']);
-		for (var cname in Upgrades[player][v['class']]) // iterate over all components of this type
-		{
-			var parameter = v['parameter'];
-			var ctype = v['class'];
-			var filterparam = v['filterParameter'];
-			if ('filterParameter' in v && Stats[ctype][cname][filterparam] != v['filterValue']) // more specific filter
-			{
-				continue;
-			}
-			if (Stats[ctype][cname][parameter] > 0) // only applies if stat has above zero value already
-			{
-				Upgrades[player][ctype][cname][parameter] += Math.ceil(Stats[ctype][cname][parameter] * v['value'] / 100);
-				//debug("      upgraded " + cname + " to " + Upgrades[player][ctype][cname][parameter] + " by " + Math.ceil(Stats[ctype][cname][parameter] * v['value'] / 100));
-			}
-		}
-	}
+	// NOTE: Research upgrades are handled by the C++ core game engine since 4.1.0
 }
 
 var lastHitTime = 0;
