@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
  */
 
 #include "frame.h"
+#include "wztime.h"
 
 #include <string.h>
 #include <stdio.h>
@@ -33,10 +34,25 @@
 #include "wzapp.h"
 #include <map>
 #include <string>
+#include <regex>
 
-#ifdef WZ_OS_LINUX
+#if defined(WZ_OS_LINUX) && defined(__GLIBC__)
 #include <execinfo.h>  // Nonfatal runtime backtraces.
-#endif //WZ_OS_LINUX
+#endif // defined(WZ_OS_LINUX) && defined(__GLIBC__)
+
+#if defined(WZ_OS_UNIX)
+# include <fcntl.h>
+# ifndef _POSIX_C_SOURCE
+#  define _POSIX_C_SOURCE 1
+# endif
+# ifndef _XOPEN_SOURCE
+#  define _XOPEN_SOURCE
+# endif
+# ifndef _POSIX_SOURCE
+#  define _POSIX_SOURCE
+# endif
+# include <stdio.h>
+#endif
 
 #define MAX_LEN_LOG_LINE 512
 
@@ -91,6 +107,9 @@ static const char *code_part_names[] =
 	"input",
 	"popup",
 	"console",
+	"activity",
+	"research",
+	"savegame",
 	"last"
 };
 
@@ -187,6 +206,24 @@ void debug_callback_file(void **data, const char *outputBuffer)
 	}
 }
 
+static WzString replaceUsernameInPath(WzString path)
+{
+	// Some common forms:
+	// C:\Users\<USERNAME>\... (Windows Vista+)
+	// C:\Documents and Settings\<USERNAME>\... (Windows XP)
+	// /Users/<USERNAME>/... (macOS)
+	// /home/<USERNAME>/... (Linux)
+
+	// First pass, do a simple string regex replacement
+	const auto win_re = std::regex("^[A-Za-z]:\\\\(Users|Documents and Settings)\\\\[^\\\\]+", std::regex_constants::ECMAScript);
+	auto result = std::regex_replace(path.toUtf8(), win_re, "[WIN_USER_FOLDER]");
+	const auto unix_re = std::regex("^\\/(Users|home)\\/[^\\/]+", std::regex_constants::ECMAScript);
+	result = std::regex_replace(result, unix_re, "[USER_HOME]");
+
+	// POSSIBLE FUTURE TODO: Do OS-specific handling where we query the "home"/user folder?
+	return WzString::fromUtf8(result);
+}
+
 char WZ_DBGFile[PATH_MAX] = {0};	//Used to save path of the created log file
 /**
  * Setup the file callback
@@ -210,13 +247,13 @@ bool debug_callback_file_init(void **data)
 	int wstr_len = MultiByteToWideChar(CP_UTF8, 0, WZDebugfilename.toUtf8().c_str(), -1, NULL, 0);
 	if (wstr_len <= 0)
 	{
-		fprintf(stderr, "Could not not convert string from UTF-8; MultiByteToWideChar failed with error %d: %s\n", GetLastError(), WZDebugfilename.toUtf8().c_str());
+		fprintf(stderr, "Could not not convert string from UTF-8; MultiByteToWideChar failed with error %lu: %s\n", GetLastError(), WZDebugfilename.toUtf8().c_str());
 		return false;
 	}
 	std::vector<wchar_t> wstr_filename(wstr_len, 0);
 	if (MultiByteToWideChar(CP_UTF8, 0, WZDebugfilename.toUtf8().c_str(), -1, &wstr_filename[0], wstr_len) == 0)
 	{
-		fprintf(stderr, "Could not not convert string from UTF-8; MultiByteToWideChar[2] failed with error %d: %s\n", GetLastError(), WZDebugfilename.toUtf8().c_str());
+		fprintf(stderr, "Could not not convert string from UTF-8; MultiByteToWideChar[2] failed with error %lu: %s\n", GetLastError(), WZDebugfilename.toUtf8().c_str());
 		return false;
 	}
 	FILE *const logfile = _wfopen(&wstr_filename[0], L"w");
@@ -228,9 +265,17 @@ bool debug_callback_file_init(void **data)
 		fprintf(stderr, "Could not open %s for appending!\n", WZDebugfilename.toUtf8().c_str());
 		return false;
 	}
+#if defined(WZ_OS_UNIX)
+	int fd = fileno(logfile);
+	if (fd != -1)
+	{
+		fcntl(fd, F_SETFD, FD_CLOEXEC);
+	}
+#endif
 	snprintf(WZ_DBGFile, sizeof(WZ_DBGFile), "%s", WZDebugfilename.toUtf8().c_str());
 	setbuf(logfile, nullptr);
-	fprintf(logfile, "--- Starting log [%s]---\n", WZDebugfilename.toUtf8().c_str());
+	WzString fileNameStripped = replaceUsernameInPath(WZDebugfilename);
+	fprintf(logfile, "--- Starting log [%s]---\n", fileNameStripped.toUtf8().c_str());
 	*data = logfile;
 
 	return true;
@@ -499,12 +544,12 @@ void _debug(int line, code_part part, const char *function, const char *str, ...
 	if (!repeated)
 	{
 		time_t rawtime;
-		struct tm *timeinfo;
+		struct tm timeinfo = {};
 		char ourtime[15];		//HH:MM:SS
 
 		time(&rawtime);
-		timeinfo = localtime(&rawtime);
-		strftime(ourtime, 15, "%H:%M:%S", timeinfo);
+		timeinfo = getLocalTime(rawtime);
+		strftime(ourtime, 15, "%H:%M:%S", &timeinfo);
 
 		// Assemble the outputBuffer:
 		ssprintf(outputBuffer, "%-8s|%s: %s", code_part_names[part], ourtime, useInputBuffer1 ? inputBuffer[1] : inputBuffer[0]);
@@ -523,13 +568,13 @@ void _debug(int line, code_part part, const char *function, const char *str, ...
 		{
 			if (wzIsFullscreen())
 			{
-				wzToggleFullscreen();
+				wzChangeWindowMode(WINDOW_MODE::windowed);
 			}
 #if defined(WZ_OS_WIN)
-			char wbuf[512];
+			char wbuf[1024];
 			ssprintf(wbuf, "%s\n\nPlease check the file (%s) in your configuration directory for more details. \
 				\nDo not forget to upload the %s file, WZdebuginfo.txt and the warzone2100.rpt files in your bug reports at https://github.com/Warzone2100/warzone2100/issues/new!", useInputBuffer1 ? inputBuffer[1] : inputBuffer[0], WZ_DBGFile, WZ_DBGFile);
-			MessageBoxA(NULL, wbuf, "Warzone has terminated unexpectedly", MB_OK | MB_ICONERROR);
+			wzDisplayDialog(Dialog_Error, "Warzone has terminated unexpectedly", wbuf);
 #elif defined(WZ_OS_MAC)
 			char wbuf[1024];
 			ssprintf(wbuf, "%s\n\nPlease check your logs and attach them along with a bug report. Thanks!", useInputBuffer1 ? inputBuffer[1] : inputBuffer[0]);
@@ -564,7 +609,7 @@ void _debug(int line, code_part part, const char *function, const char *str, ...
 			}
 #else
 			const char *popupBuf = useInputBuffer1 ? inputBuffer[1] : inputBuffer[0];
-			wzFatalDialog(popupBuf);
+			wzDisplayDialog(Dialog_Error, "Warzone has terminated unexpectedly", popupBuf);
 #endif
 		}
 
@@ -572,15 +617,7 @@ void _debug(int line, code_part part, const char *function, const char *str, ...
 		// This is a popup dialog used for times when the error isn't fatal, but we still need to notify user what is going on.
 		if (part == LOG_POPUP)
 		{
-#if defined(WZ_OS_WIN)
-			char wbuf[512];
-			ssprintf(wbuf, "A non fatal error has occurred.\n\n%s\n\n", useInputBuffer1 ? inputBuffer[1] : inputBuffer[0]);
-			MessageBoxA(NULL,
-			            wbuf,
-			            "Warzone has detected a problem.", MB_OK | MB_ICONINFORMATION);
-#elif defined(WZ_OS_MAC)
-			cocoaShowAlert("Warzone has detected a problem.", inputBuffer[useInputBuffer1 ? 1 : 0], 0, "OK", NULL);
-#endif
+			wzDisplayDialog(Dialog_Information, "Warzone has detected a problem.", inputBuffer[useInputBuffer1 ? 1 : 0]);
 		}
 
 	}
@@ -589,7 +626,7 @@ void _debug(int line, code_part part, const char *function, const char *str, ...
 
 void _debugBacktrace(code_part part)
 {
-#ifdef WZ_OS_LINUX
+#if defined(WZ_OS_LINUX) && defined(__GLIBC__)
 	void *btv[20];
 	unsigned num = backtrace(btv, sizeof(btv) / sizeof(*btv));
 	char **btc = backtrace_symbols(btv, num);
@@ -613,3 +650,16 @@ void debugDisableAssert()
 {
 	assertEnabled = false;
 }
+
+#include <sstream>
+
+void _debug_multiline(int line, code_part part, const char *function, const std::string &multilineString)
+{
+	std::stringstream ss(multilineString);
+	std::string to;
+
+	while(std::getline(ss,to,'\n')){
+		_debug(line, part, function, "%s", to.c_str());
+	}
+}
+
