@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1992-2007  Trolltech ASA.
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -544,7 +544,9 @@
 #    include <unistd.h>
 #    include <sys/param.h>
 #    include <w32api.h>
-#    define _WIN32_IE IE5
+#    if !defined(_WIN32_IE)
+#      define _WIN32_IE IE5
+#    endif
 // Required for alloca
 #    include <malloc.h>
 
@@ -582,6 +584,8 @@
 
 #    define PATH_MAX MAX_PATH
 
+#  endif /* WZ_CC_MSVC */
+
 // These are useless for MSVC builds, since we don't populate them / use them at this time.
 #ifndef PACKAGE_DISTRIBUTOR
 # define PACKAGE_DISTRIBUTOR "UNKNOWN"
@@ -593,9 +597,6 @@
 # define PACKAGE "Warzone"
 #endif
 
-
-#  endif /* WZ_CC_MSVC */
-
 /* Make sure that PATH_MAX is large enough to use as the size for return
  * buffers for Windows API calls
  */
@@ -695,4 +696,9 @@ static inline char *_WZ_ASSERT_ARRAY_EXPR_FUNCTION(T *&)
 	#define MSVC_PRAGMA(x)
 #endif
 
+// Gfx backend option support
+#if defined(WZ_OS_WIN)
+# define WZ_BACKEND_DIRECTX
+#endif
+
 #endif /* WZGLOBAL_H */
