@@ -1,6 +1,6 @@
 /*
  *	This file is part of Warzone 2100.
- *	Copyright (C) 2018-2019  Warzone 2100 Project
+ *	Copyright (C) 2018-2020  Warzone 2100 Project
  *
  *	Warzone 2100 is free software; you can redistribute it and/or modify
  *	it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #include <cctype>
 #include <sstream>
 #include <iomanip>
+#include <limits>
 #include <utfcpp/source/utf8.h>
 #include <utf8proc/utf8proc.h>
 
@@ -89,11 +90,13 @@ WzString::WzString(const char * str, int size /*= -1*/)
 	if (str == nullptr) { return; }
 	if (size < 0)
 	{
-		size = strlen(str);
+		size_t strLen = strlen(str);
+		ASSERT_OR_RETURN(, strLen <= static_cast<size_t>(std::numeric_limits<int>::max()), "String length (%zu) exceeds maximum support length: %d", strLen, std::numeric_limits<int>::max());
+		size = static_cast<int>(strLen);
 	}
-	ASSERT(utf8::is_valid(str, str + size), "Input text is not valid UTF-8");
+	ASSERT(utf8::is_valid(str, str + size), "Input text is not valid UTF-8 (size: %d): %s", size, str);
 	try {
-		utf8::replace_invalid(str, str + size, back_inserter(_utf8String), '?');
+		utf8::replace_invalid(str, str + size, std::back_inserter(_utf8String), '?');
 	}
 	catch (const std::exception &e) {
 		// Likely passed an incomplete UTF-8 sequence
@@ -109,7 +112,8 @@ WzString WzString::fromUtf8(const char *str, int size /*= -1*/)
 
 WzString WzString::fromUtf8(const std::string &str)
 {
-	return WzString(str.c_str(), str.length());
+	ASSERT_OR_RETURN(WzString(), str.length() <= static_cast<size_t>(std::numeric_limits<int>::max()), "String length (%zu) exceeds maximum support length: %d", str.length(), std::numeric_limits<int>::max());
+	return WzString(str.c_str(), static_cast<int>(str.length()));
 }
 
 WzString WzString::fromUtf16(const std::vector<uint16_t>& utf16)
@@ -140,14 +144,14 @@ const std::string& WzString::toStdString() const
 	return toUtf8();
 }
 
-const std::vector<uint16_t> WzString::toUtf16() const
+std::vector<uint16_t> WzString::toUtf16() const
 {
 	std::vector<uint16_t> utf16result;
 	utf8::utf8to16(_utf8String.begin(), _utf8String.end(), back_inserter(utf16result));
 	return utf16result;
 }
 
-const std::vector<uint32_t> WzString::toUtf32() const
+std::vector<uint32_t> WzString::toUtf32() const
 {
 	std::vector<uint32_t> utf32result;
 	utf8::utf8to32(_utf8String.begin(), _utf8String.end(), back_inserter(utf32result));
@@ -190,7 +194,17 @@ int WzString::length() const
 	return (int)len_distance;
 }
 
-const WzUniCodepoint WzString::at(int position) const
+WzString WzString::substr(size_t start, size_t length) const
+{
+	auto begin = _utf8String.begin();
+	_utf8_advance(begin, start, _utf8String.end());
+
+	auto end = begin;
+	_utf8_advance(end, length, _utf8String.end());
+	return WzString(std::string(begin, end));
+}
+
+WzUniCodepoint WzString::at(int position) const
 {
 	auto it = _utf8String.begin();
 	_utf8_advance(it, position, _utf8String.end());
@@ -219,16 +233,6 @@ WzString& WzString::insert(size_t position, const WzString &str)
 {
 	auto it = _utf8String.begin();
 	_utf8_advance(it, position, _utf8String.end());
-	if (it == _utf8String.end())
-	{
-		size_t distance = it - _utf8String.begin();
-		if (distance > position)
-		{
-			// TODO: To match QString behavior, we need to extend the string?
-			ASSERT(it != _utf8String.end(), "Cannot find position in string prior to end of string.");
-		}
-		// deliberately fall-through
-	}
 	_utf8String.insert(it, str._utf8String.begin(), str._utf8String.end());
 	return *this;
 }
@@ -237,15 +241,6 @@ WzString& WzString::insert(size_t i, WzUniCodepoint c)
 {
 	auto it = _utf8String.begin();
 	_utf8_advance(it, i, _utf8String.end());
-	if (it == _utf8String.end())
-	{
-		size_t distance = it - _utf8String.begin();
-		if (distance > i)
-		{
-			// TODO: To match QString behavior, we need to extend the string?
-			ASSERT(it != _utf8String.end(), "Cannot find position in string prior to end of string.");
-		}
-	}
 	auto cUtf8Codepoints = WzString::fromCodepoint(c);
 	_utf8String.insert(it, cUtf8Codepoints._utf8String.begin(), cUtf8Codepoints._utf8String.end());
 	return *this;
@@ -398,7 +393,7 @@ std::vector<WzString> WzString::split(const WzString & delimiter) const
 	auto it = _utf8String.begin();
 	auto it_currenttokenstart = _utf8String.begin();
 	bool lastCodepointWasDelimiterEnd = false;
-	uint32_t codepoint = 0;
+	uint32_t codepoint;
 	while (it != _utf8String.end())
 	{
 		codepoint = utf8::next(it, _utf8String.end());
@@ -432,7 +427,7 @@ std::vector<WzString> WzString::split(const WzString & delimiter) const
 	{
 		// string ends with delimiter (and delimiter is not an empty string)
 		// append an empty string to the end of the list
-		tokens.push_back(WzString());
+		tokens.emplace_back();
 	}
 	return tokens;
 }
@@ -552,7 +547,7 @@ WzString& WzString::operator+=(const char* str)
 	return *this;
 }
 
-const WzString WzString::operator+(const WzString &other) const
+WzString WzString::operator+(const WzString &other) const
 {
 	WzString newString = *this;
 	newString.append(other);
@@ -560,7 +555,7 @@ const WzString WzString::operator+(const WzString &other) const
 }
 
 // NOTE: The char * should be valid UTF-8.
-const WzString WzString::operator+(const char* other) const
+WzString WzString::operator+(const char* other) const
 {
 	WzString newString = *this;
 	newString.append(other);
@@ -580,7 +575,7 @@ WzString& WzString::operator=(const WzUniCodepoint& ch)
 	return *this;
 }
 
-WzString& WzString::operator=(WzString&& other)
+WzString& WzString::operator=(WzString&& other) noexcept
 {
 	if (this != &other)
 	{
