@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -41,6 +41,9 @@
 #include "imd.h" // for imd structures
 #include "tex.h" // texture page loading
 
+#include <glm/vec4.hpp>
+using Vector4f = glm::vec4;
+
 // Scale animation numbers from int to float
 #define INT_SCALE       1000
 
@@ -63,6 +66,14 @@ void modelShutdown()
 	models.clear();
 }
 
+void enumerateLoadedModels(const std::function<void (const std::string& modelName, iIMDShape& model)>& func)
+{
+	for (auto& keyvaluepair : models)
+	{
+		func(keyvaluepair.first, keyvaluepair.second);
+	}
+}
+
 static bool tryLoad(const WzString &path, const WzString &filename)
 {
 	if (PHYSFS_exists(path + filename))
@@ -100,9 +111,10 @@ const std::string &modelName(iIMDShape *model)
 iIMDShape *modelGet(const WzString &filename)
 {
 	WzString name(filename.toLower());
-	if (models.count(name.toStdString()) > 0)
+	auto it = models.find(name.toStdString());
+	if (it != models.end())
 	{
-		return &models.at(name.toStdString()); // cached
+		return &it->second; // cached
 	}
 	else if (tryLoad("structs/", name) || tryLoad("misc/", name) || tryLoad("effects/", name)
 	         || tryLoad("components/prop/", name) || tryLoad("components/weapons/", name)
@@ -140,7 +152,7 @@ static bool AtEndOfFile(const char *CurPos, const char *EndOfFile)
  * \param s Pointer to shape level
  * \return false on error (memory allocation failure/bad file format), true otherwise
  */
-static bool _imd_load_polys(const WzString &filename, const char **ppFileData, iIMDShape *s, int pieVersion)
+static bool _imd_load_polys(const WzString &filename, const char **ppFileData, iIMDShape *s, int pieVersion, const uint32_t npoints)
 {
 	const char *pFileData = *ppFileData;
 
@@ -161,13 +173,19 @@ static bool _imd_load_polys(const WzString &filename, const char **ppFileData, i
 
 		poly->flags = flags;
 		ASSERT_OR_RETURN(false, npnts == 3, "Invalid polygon size (%d)", npnts);
-		if (sscanf(pFileData, "%d %d %d%n", &poly->pindex[0], &poly->pindex[1], &poly->pindex[2], &cnt) != 3)
+		if (sscanf(pFileData, "%" PRIu32 "%" PRIu32 "%" PRIu32 "%n", &poly->pindex[0], &poly->pindex[1], &poly->pindex[2], &cnt) != 3)
 		{
 			debug(LOG_ERROR, "failed reading triangle, point %d", i);
 			return false;
 		}
 		pFileData += cnt;
 
+		// sanity check
+		for (size_t pIdx = 0; pIdx < 3; pIdx++)
+		{
+			ASSERT_OR_RETURN(false, (poly->pindex[pIdx] < npoints), "Point index (%" PRIu32 ") exceeds max index (%" PRIu32 ")", poly->pindex[pIdx], (npoints - 1));
+		}
+
 		// calc poly normal
 		{
 			Vector3f p0, p1, p2;
@@ -232,7 +250,7 @@ static bool _imd_load_polys(const WzString &filename, const char **ppFileData, i
 					poly->texAnim.x /= OLD_TEXTURE_SIZE_FIX;
 					poly->texAnim.y /= OLD_TEXTURE_SIZE_FIX;
 				}
-				framesPerLine = 1 / poly->texAnim.x;
+				framesPerLine = static_cast<int>(1 / poly->texAnim.x);
 			}
 			else
 			{
@@ -327,29 +345,29 @@ static void _imd_calc_bounds(iIMDShape &s)
 	{
 		if (p.x > s.max.x)
 		{
-			s.max.x = p.x;
+			s.max.x = static_cast<int>(p.x);
 		}
 		if (p.x < s.min.x)
 		{
-			s.min.x = p.x;
+			s.min.x = static_cast<int>(p.x);
 		}
 
 		if (p.y > s.max.y)
 		{
-			s.max.y = p.y;
+			s.max.y = static_cast<int>(p.y);
 		}
 		if (p.y < s.min.y)
 		{
-			s.min.y = p.y;
+			s.min.y = static_cast<int>(p.y);
 		}
 
 		if (p.z > s.max.z)
 		{
-			s.max.z = p.z;
+			s.max.z = static_cast<int>(p.z);
 		}
 		if (p.z < s.min.z)
 		{
-			s.min.z = p.z;
+			s.min.z = static_cast<int>(p.z);
 		}
 
 		// for tight sphere calculations
@@ -402,7 +420,7 @@ static void _imd_calc_bounds(iIMDShape &s)
 	zmax = MAX(s.max.z, -s.min.z);
 
 	s.radius = MAX(xmax, MAX(ymax, zmax));
-	s.sradius = sqrtf(xmax * xmax + ymax * ymax + zmax * zmax);
+	s.sradius = static_cast<int>(sqrtf(xmax * xmax + ymax * ymax + zmax * zmax));
 
 // START: tight bounding sphere
 
@@ -444,9 +462,9 @@ static void _imd_calc_bounds(iIMDShape &s)
 	}
 
 	// dia1, dia2 diameter of initial sphere
-	cen.x = (dia1.x + dia2.x) / 2.;
-	cen.y = (dia1.y + dia2.y) / 2.;
-	cen.z = (dia1.z + dia2.z) / 2.;
+	cen.x = (dia1.x + dia2.x) / 2.f;
+	cen.y = (dia1.y + dia2.y) / 2.f;
+	cen.z = (dia1.z + dia2.z) / 2.f;
 
 	// calc initial radius
 	dx = dia2.x - cen.x;
@@ -475,9 +493,9 @@ static void _imd_calc_bounds(iIMDShape &s)
 			rad_sq = rad * rad;
 			old_to_new = old_to_p - rad;
 			// centre of new sphere
-			cen.x = (rad * cen.x + old_to_new * p.x) / old_to_p;
-			cen.y = (rad * cen.y + old_to_new * p.y) / old_to_p;
-			cen.z = (rad * cen.z + old_to_new * p.z) / old_to_p;
+			cen.x = static_cast<float>((rad * cen.x + old_to_new * p.x) / old_to_p);
+			cen.y = static_cast<float>((rad * cen.y + old_to_new * p.y) / old_to_p);
+			cen.z = static_cast<float>((rad * cen.z + old_to_new * p.z) / old_to_p);
 		}
 	}
 
@@ -486,7 +504,7 @@ static void _imd_calc_bounds(iIMDShape &s)
 // END: tight bounding sphere
 }
 
-static bool _imd_load_points(const char **ppFileData, iIMDShape &s, int npoints)
+static bool _imd_load_points(const char **ppFileData, iIMDShape &s, uint32_t npoints)
 {
 	//load the points then pass through a second time to setup bounding datavalues
 	s.points.resize(npoints);
@@ -543,43 +561,165 @@ static std::vector<gfx_api::gfxFloat> vertices;
 static std::vector<gfx_api::gfxFloat> normals;
 static std::vector<gfx_api::gfxFloat> texcoords;
 static std::vector<gfx_api::gfxFloat> tangents;
+static std::vector<gfx_api::gfxFloat> bitangents;
 static std::vector<uint16_t> indices; // size is npolys * 3 * numFrames
 static uint16_t vertexCount = 0;
 
-static inline uint16_t addVertex(iIMDShape &s, int i, const iIMDPoly *p, int frameidx)
+static bool ReadNormals(const char **ppFileData, std::vector<Vector3f> &pie_level_normals)
+{
+   const char *pFileData = *ppFileData;
+   int dataCnt;
+
+   for (Vector3f &normal : pie_level_normals)
+   {
+	   if (sscanf(pFileData, "%f %f %f%n", &normal.x, &normal.y, &normal.z, &dataCnt) != 3)
+	   {
+		   debug(LOG_ERROR, "File corrupt - could not read normals");
+		   return false;
+	   }
+	   pFileData += dataCnt;
+   }
+
+   *ppFileData = pFileData;
+
+   return true;
+}
+
+static bool _imd_load_normals(const char **ppFileData, std::vector<Vector3f> &pie_level_normals, uint32_t num_normal_lines)
+{
+   // We only support triangles!
+   pie_level_normals.resize(static_cast<size_t>(num_normal_lines * 3));
+
+   if (ReadNormals(ppFileData, pie_level_normals) == false)
+   {
+	   pie_level_normals.clear();
+	   return false;
+   }
+
+   return true;
+}
+
+static inline uint16_t addVertex(iIMDShape &s, size_t i, const iIMDPoly *p, int frameidx, size_t polyIdx, std::vector<Vector3f> &pie_level_normals)
 {
 	// if texture animation flag is present, fetch animation coordinates for this polygon
 	// otherwise just show the first set of texel coordinates
 	int frame = (p->flags & iV_IMD_TEXANIM) ? frameidx : 0;
 
-	// See if we already have this defined, if so, return reference to it.
-	for (uint16_t j = 0; j < vertexCount; j++)
-	{
-		if (texcoords[j * 2 + 0] == p->texCoord[frame * 3 + i].x
-		    && texcoords[j * 2 + 1] == p->texCoord[frame * 3 + i].y
-		    && vertices[j * 3 + 0] == s.points[p->pindex[i]].x
-		    && vertices[j * 3 + 1] == s.points[p->pindex[i]].y
-		    && vertices[j * 3 + 2] == s.points[p->pindex[i]].z
-		    && normals[j * 3 + 0] == p->normal.x
-		    && normals[j * 3 + 1] == p->normal.y
-		    && normals[j * 3 + 2] == p->normal.z)
+	const Vector3f* normal;
+
+ 	// Do not weld for for models with normals, those are presumed to be correct. Otherwise, it will break tangents
+ 	if (pie_level_normals.empty())
+ 	{
+		normal = &p->normal;
+		// See if we already have this defined, if so, return reference to it.
+		for (uint16_t j = 0; j < vertexCount; j++)
 		{
-			return j;
+			if (texcoords[j * 2 + 0] == p->texCoord[frame * 3 + i].x
+ 			    && texcoords[j * 2 + 1] == p->texCoord[frame * 3 + i].y
+ 			    && vertices[j * 3 + 0] == s.points[p->pindex[i]].x
+ 			    && vertices[j * 3 + 1] == s.points[p->pindex[i]].y
+ 			    && vertices[j * 3 + 2] == s.points[p->pindex[i]].z
+ 			    && normals[j * 3 + 0] == normal->x
+ 			    && normals[j * 3 + 1] == normal->y
+ 			    && normals[j * 3 + 2] == normal->z)
+ 			{
+ 				return j;
+ 			}
 		}
 	}
+	else
+	{
+		normal = &pie_level_normals[polyIdx * 3 + i];
+	}
+
 	// We don't have it, add it.
-	normals.emplace_back(p->normal.x);
-	normals.emplace_back(p->normal.y);
-	normals.emplace_back(p->normal.z);
+	normals.emplace_back(normal->x);
+ 	normals.emplace_back(normal->y);
+ 	normals.emplace_back(normal->z);
 	texcoords.emplace_back(p->texCoord[frame * 3 + i].x);
 	texcoords.emplace_back(p->texCoord[frame * 3 + i].y);
 	vertices.emplace_back(s.points[p->pindex[i]].x);
 	vertices.emplace_back(s.points[p->pindex[i]].y);
 	vertices.emplace_back(s.points[p->pindex[i]].z);
 	vertexCount++;
+
 	return vertexCount - 1;
 }
 
+void calculateTangentsForTriangle(const uint16_t ia, const uint16_t ib, const uint16_t ic)
+{
+   // This will work as long as vecs are packed (which is default in glm)
+   const Vector3f* verticesAsVec3 = reinterpret_cast<const Vector3f*>(vertices.data());
+   const Vector2f* texcoordsAsVec2 = reinterpret_cast<const Vector2f*>(texcoords.data());
+   Vector4f* tangentsAsVec4 = reinterpret_cast<Vector4f*>(tangents.data());
+   Vector3f* bitangentsAsVec3 = reinterpret_cast<Vector3f*>(bitangents.data());
+
+   // Shortcuts for vertices
+   const Vector3f& va(verticesAsVec3[ia]);
+   const Vector3f& vb(verticesAsVec3[ib]);
+   const Vector3f& vc(verticesAsVec3[ic]);
+
+   // Shortcuts for UVs
+   const Vector2f& uva(texcoordsAsVec2[ia]);
+   const Vector2f& uvb(texcoordsAsVec2[ib]);
+   const Vector2f& uvc(texcoordsAsVec2[ic]);
+
+   // Edges of the triangle : postion delta
+   const Vector3f deltaPos1 = vb - va;
+   const Vector3f deltaPos2 = vc - va;
+
+   // UV delta
+   const Vector2f deltaUV1 = uvb - uva;
+   const Vector2f deltaUV2 = uvc - uva;
+
+   // check for nan
+   float r = deltaUV1.x * deltaUV2.y - deltaUV1.y * deltaUV2.x;
+   if (r != 0.f)
+	   r = 1.f / r;
+
+   const Vector4f tangent(Vector3f(deltaPos1 * deltaUV2.y - deltaPos2 * deltaUV1.y) * r, 0.f);
+   const Vector3f bitangent(Vector3f(deltaPos2 * deltaUV1.x - deltaPos1 * deltaUV2.x) * r);
+
+   tangentsAsVec4[ia] += tangent;
+   tangentsAsVec4[ib] += tangent;
+   tangentsAsVec4[ic] += tangent;
+
+   bitangentsAsVec3[ia] += bitangent;
+   bitangentsAsVec3[ib] += bitangent;
+   bitangentsAsVec3[ic] += bitangent;
+}
+
+void finishTangentsGeneration()
+{
+   // This will work as long as vecs are packed (which is default in glm)
+   const Vector3f* normalsAsVec3 = reinterpret_cast<const Vector3f*>(normals.data());
+   Vector4f* tangentsAsVec4 = reinterpret_cast<Vector4f*>(tangents.data());
+   const Vector3f* bitangentsAsVec3 = reinterpret_cast<const Vector3f*>(bitangents.data());
+
+   Vector3f t;
+
+   for (auto i = 0; i < vertexCount; ++i)
+   {
+	   const Vector3f& n = normalsAsVec3[i];
+	   const Vector3f& b = bitangentsAsVec3[i];
+	   t = tangentsAsVec4[i].xyz();
+
+	   // Gram-Schmidt orthogonalize
+	   t = glm::normalize(t - n * glm::dot(n, t));
+
+	   // Calculate handedness
+	   if (glm::dot(glm::cross(n, t), b) < 0.f)
+	   {
+		   tangentsAsVec4[i] = Vector4f(-t, 1.f);
+	   }
+	   else
+	   {
+		   tangentsAsVec4[i] = Vector4f(-t, -1.f);
+	   }
+   }
+}
+
+
 /*!
  * Load shape levels recursively
  * \param ppFileData Pointer to the data (usually read from a file)
@@ -593,7 +733,7 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 {
 	const char *pFileData = *ppFileData;
 	char buffer[PATH_MAX] = {'\0'};
-	int cnt = 0, n = 0, i;
+	int cnt = 0, n = 0, scanResult = 0;
 	float dummy;
 
 	if (nlevels == 0)
@@ -609,16 +749,18 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 	}
 	ASSERT(models.count(key) == 0, "Duplicate model load for %s!", key.c_str());
 	iIMDShape &s = models[key]; // create entry and return reference
+	s.pShadowPoints = &s.points;
+	s.pShadowPolys = &s.polys;
 
-	i = sscanf(pFileData, "%255s %n", buffer, &cnt);
-	ASSERT_OR_RETURN(nullptr, i == 1, "Bad directive following LEVEL");
+	scanResult = sscanf(pFileData, "%255s %n", buffer, &cnt);
+	ASSERT_OR_RETURN(nullptr, scanResult == 1, "Bad directive following LEVEL");
 
 	// Optionally load and ignore deprecated MATERIALS directive
 	if (strcmp(buffer, "MATERIALS") == 0)
 	{
-		i = sscanf(pFileData, "%255s %f %f %f %f %f %f %f %f %f %f%n", buffer, &dummy, &dummy, &dummy, &dummy,
+		scanResult = sscanf(pFileData, "%255s %f %f %f %f %f %f %f %f %f %f%n", buffer, &dummy, &dummy, &dummy, &dummy,
 		           &dummy, &dummy, &dummy, &dummy, &dummy, &dummy, &cnt);
-		ASSERT_OR_RETURN(nullptr, i == 11, "Bad MATERIALS directive");
+		ASSERT_OR_RETURN(nullptr, scanResult == 11, "Bad MATERIALS directive");
 		debug(LOG_WARNING, "MATERIALS directive no longer supported!");
 		pFileData += cnt;
 	}
@@ -635,8 +777,8 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 		pFileData += cnt;
 	}
 
-	int npoints = 0;
-	if (sscanf(pFileData, "%255s %d%n", buffer, &npoints, &cnt) != 2)
+	uint32_t npoints = 0;
+	if (sscanf(pFileData, "%255s %" PRIu32 "%n", buffer, &npoints, &cnt) != 2)
 	{
 		debug(LOG_ERROR, "_imd_load_level(2): file corrupt");
 		return nullptr;
@@ -649,20 +791,38 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 
 	_imd_load_points(&pFileData, s, npoints);
 
-	int npolys = 0;
-	if (sscanf(pFileData, "%255s %d%n", buffer, &npolys, &cnt) != 2)
+	uint32_t npolys = 0;
+	if (sscanf(pFileData, "%255s %" PRIu32 "%n", buffer, &npolys, &cnt) != 2)
 	{
 		debug(LOG_ERROR, "_imd_load_level(3): file corrupt");
 		return nullptr;
 	}
 	pFileData += cnt;
+
+	std::vector<Vector3f> pie_level_normals;
+
+	// It could be optional normals directive
+ 	if (strcmp(buffer, "NORMALS") == 0)
+ 	{
+ 		_imd_load_normals(&pFileData, pie_level_normals, npolys);
+
+ 		// Attemps to read polys again
+ 		if (sscanf(pFileData, "%255s %" PRIu32 "%n", buffer, &npolys, &cnt) != 2)
+ 		{
+ 			debug(LOG_ERROR, "_imd_load_level(3a): file corrupt");
+ 			return nullptr;
+ 		}
+		debug(LOG_3D, "imd[_load_level] = normals %d", static_cast<int>(pie_level_normals.size()));
+ 		pFileData += cnt;
+ 	}
+
 	s.polys.resize(npolys);
 
 	ASSERT_OR_RETURN(nullptr, strcmp(buffer, "POLYGONS") == 0, "Expecting 'POLYGONS' directive, got: %s", buffer);
 
-	_imd_load_polys(filename, &pFileData, &s, pieVersion);
+	_imd_load_polys(filename, &pFileData, &s, pieVersion, npoints);
 
-	// optional stuff : levels, object animations, connectors
+	// optional stuff : levels, object animations, connectors, shadow points + polys
 	s.objanimframes = 0;
 	while (!AtEndOfFile(pFileData, FileDataEnd)) // check for end of file (give or take white space)
 	{
@@ -675,7 +835,7 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 
 		if (strcmp(buffer, "LEVEL") == 0)	// check for next level
 		{
-			debug(LOG_3D, "imd[_load_level] = npoints %d, npolys %d", npoints, npolys);
+			debug(LOG_3D, "imd[_load_level] = npoints %" PRIu32 ", npolys %" PRIu32 "", npoints, npolys);
 			s.next = _imd_load_level(filename, &pFileData, FileDataEnd, nlevels - 1, pieVersion, level + 1);
 		}
 		else if (strcmp(buffer, "CONNECTORS") == 0)
@@ -709,12 +869,36 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 				s.objanimdata[i].pos.x = pos.x / INT_SCALE;
 				s.objanimdata[i].pos.y = pos.z / INT_SCALE;
 				s.objanimdata[i].pos.z = pos.y / INT_SCALE;
-				s.objanimdata[i].rot.pitch = -(rot.x * DEG_1 / INT_SCALE);
-				s.objanimdata[i].rot.direction = -(rot.z * DEG_1 / INT_SCALE);
-				s.objanimdata[i].rot.roll = -(rot.y * DEG_1 / INT_SCALE);
+				s.objanimdata[i].rot.pitch = static_cast<uint16_t>(-(rot.x * DEG_1 / INT_SCALE));
+				s.objanimdata[i].rot.direction = static_cast<uint16_t>(-(rot.z * DEG_1 / INT_SCALE));
+				s.objanimdata[i].rot.roll = static_cast<uint16_t>(-(rot.y * DEG_1 / INT_SCALE));
 				pFileData += cnt;
 			}
 		}
+		else if (strcmp(buffer, "SHADOWPOINTS") == 0)
+		{
+			ASSERT_OR_RETURN(nullptr, n > 0, "Invalid 'SHADOW_POINTS' count, got: %d", n);
+			uint32_t nShadowPoints = static_cast<uint32_t>(n);
+
+			iIMDShape tmpShadowShape;
+			_imd_load_points(&pFileData, tmpShadowShape, nShadowPoints);
+
+			s.altShadowPoints = std::move(tmpShadowShape.points);
+			s.pShadowPoints = &s.altShadowPoints;
+		}
+		else if (strcmp(buffer, "SHADOWPOLYGONS") == 0)
+		{
+			ASSERT_OR_RETURN(nullptr, s.altShadowPoints.size() > 0, "'SHADOW_POLYGONS' must follow a non-empty SHADOW_POINTS section");
+			ASSERT_OR_RETURN(nullptr, n > 0, "Invalid 'SHADOW_POLYGONS' count, got: %d", n);
+			uint32_t nShadowPolys = static_cast<uint32_t>(n);
+
+			iIMDShape tmpShadowShape;
+			tmpShadowShape.polys.resize(nShadowPolys);
+			_imd_load_polys(filename, &pFileData, &tmpShadowShape, PIE_FLOAT_VER, static_cast<uint32_t>(s.altShadowPoints.size()));
+
+			s.altShadowPolys = std::move(tmpShadowShape.polys);
+			s.pShadowPolys = &s.altShadowPolys;
+		}
 		else
 		{
 			debug(LOG_ERROR, "(_load_level) unexpected directive %s %d", buffer, n);
@@ -722,20 +906,65 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 		}
 	}
 
+	// Sanity checks
+ 	if (!pie_level_normals.empty())
+ 	{
+ 		if (s.polys.size() * 3 != pie_level_normals.size())
+ 		{
+ 			debug(LOG_ERROR, "imd[_load_level] = got %" PRIu32 " npolys, but there are only %zu normals! Discarding normals...",
+ 			      npolys, pie_level_normals.size());
+ 			// This will force usage of calculated normals in addVertex
+ 			pie_level_normals.clear();
+ 		}
+ 	}
+	if (!s.altShadowPoints.empty())
+	{
+		if (s.altShadowPolys.empty())
+		{
+			debug(LOG_ERROR, "imd[_load_level] = %zu SHADOWPOINTS specified, but there are no SHADOWPOLYGONS! Discarding shadow points...",
+				  s.altShadowPoints.size());
+			s.altShadowPoints.clear();
+			s.altShadowPolys.clear();
+			s.pShadowPoints = &s.points;
+			s.pShadowPolys = &s.polys;
+		}
+	}
+
 	// FINALLY, massage the data into what can stream directly to OpenGL
 	vertexCount = 0;
 	for (int k = 0; k < MAX(1, s.numFrames); k++)
 	{
 		// Go through all polygons for each frame
-		for (const iIMDPoly &p : s.polys)
+		for (size_t npol = 0; npol < s.polys.size(); ++npol)
 		{
+			const iIMDPoly& p = s.polys[npol];
 			// Do we already have the vertex data for this polygon?
-			indices.emplace_back(addVertex(s, 0, &p, k));
-			indices.emplace_back(addVertex(s, 1, &p, k));
-			indices.emplace_back(addVertex(s, 2, &p, k));
+			indices.emplace_back(addVertex(s, 0, &p, k, npol, pie_level_normals));
+ 			indices.emplace_back(addVertex(s, 1, &p, k, npol, pie_level_normals));
+ 			indices.emplace_back(addVertex(s, 2, &p, k, npol, pie_level_normals));
 		}
 	}
 
+	s.vertexCount = vertexCount;
+
+	// Tangents are optional, only if normals were loaded and passed sanity check above
+ 	if (!pie_level_normals.empty())
+ 	{
+ 		tangents.resize(vertexCount * 4);
+ 		bitangents.resize(vertexCount * 3);
+
+ 		for (size_t i = 0; i < indices.size(); i += 3)
+ 			calculateTangentsForTriangle(indices[i], indices[i+1], indices[i+2]);
+ 		finishTangentsGeneration();
+
+ 		if (!tangents.empty())
+ 		{
+ 			if (!s.buffers[VBO_TANGENT])
+ 				s.buffers[VBO_TANGENT] = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::vertex_buffer);
+ 			s.buffers[VBO_TANGENT]->upload(tangents.size() * sizeof(gfx_api::gfxFloat), tangents.data());
+ 		}
+ 	}
+
 	if (!s.buffers[VBO_VERTEX])
 		s.buffers[VBO_VERTEX] = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::vertex_buffer);
 	s.buffers[VBO_VERTEX]->upload(vertices.size() * sizeof(gfx_api::gfxFloat), vertices.data());
@@ -752,12 +981,12 @@ static iIMDShape *_imd_load_level(const WzString &filename, const char **ppFileD
 		s.buffers[VBO_TEXCOORD] = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::vertex_buffer);
 	s.buffers[VBO_TEXCOORD]->upload(texcoords.size() * sizeof(gfx_api::gfxFloat), texcoords.data());
 
-	glBindBuffer(GL_ARRAY_BUFFER, 0); // unbind
-
 	indices.resize(0);
 	vertices.resize(0);
 	texcoords.resize(0);
 	normals.resize(0);
+	tangents.resize(0);
+	bitangents.resize(0);
 
 	*ppFileData = pFileData;
 
@@ -775,11 +1004,12 @@ static void iV_ProcessIMD(const WzString &filename, const char **ppFileData, con
 {
 	const char *pFileData = *ppFileData;
 	char buffer[PATH_MAX], texfile[PATH_MAX], normalfile[PATH_MAX], specfile[PATH_MAX];
-	int cnt, nlevels;
+	int cnt, nlevels, interpolate;
 	UDWORD level;
 	int32_t imd_version;
 	uint32_t imd_flags;
 	bool bTextured = false;
+	bool readInterpolate = false;
 	iIMDShape *objanimpie[ANIM_EVENT_COUNT];
 
 	memset(normalfile, 0, sizeof(normalfile));
@@ -814,6 +1044,18 @@ static void iV_ProcessIMD(const WzString &filename, const char **ppFileData, con
 	}
 	pFileData += cnt;
 
+	// Read interpolation -- optional
+	if (sscanf(pFileData, "%255s %d%n", buffer, &interpolate, &cnt) != 2)
+	{
+		debug(LOG_ERROR, "%s: Expecting INTERPOLATE: %s", filename.toUtf8().c_str(), buffer);
+		return;
+	}
+	if (strncmp(buffer, "INTERPOLATE", 11) == 0)
+	{
+		readInterpolate = true;
+		pFileData += cnt;
+	}
+
 	/* This can be either texture or levels */
 	if (sscanf(pFileData, "%255s %d%n", buffer, &nlevels, &cnt) != 2)
 	{
@@ -1007,50 +1249,90 @@ static void iV_ProcessIMD(const WzString &filename, const char **ppFileData, con
 	// load texture page if specified
 	if (bTextured)
 	{
-		int texpage = iV_GetTexture(texfile);
-		int normalpage = iV_TEX_INVALID;
-		int specpage = iV_TEX_INVALID;
+		optional<size_t> texpage = iV_GetTexture(texfile);
+		optional<size_t> normalpage;
+		optional<size_t> specpage;
 
-		ASSERT_OR_RETURN(, texpage >= 0, "%s could not load tex page %s", filename.toUtf8().c_str(), texfile);
+		ASSERT_OR_RETURN(, texpage.has_value(), "%s could not load tex page %s", filename.toUtf8().c_str(), texfile);
 
 		if (normalfile[0] != '\0')
 		{
 			debug(LOG_TEXTURE, "Loading normal map %s for %s", normalfile, filename.toUtf8().c_str());
 			normalpage = iV_GetTexture(normalfile, false);
-			ASSERT_OR_RETURN(, normalpage >= 0, "%s could not load tex page %s", filename.toUtf8().c_str(), normalfile);
+			ASSERT_OR_RETURN(, normalpage.has_value(), "%s could not load tex page %s", filename.toUtf8().c_str(), normalfile);
 		}
 
 		if (specfile[0] != '\0')
 		{
 			debug(LOG_TEXTURE, "Loading specular map %s for %s", specfile, filename.toUtf8().c_str());
-			specpage = iV_GetTexture(specfile, false);
-			ASSERT_OR_RETURN(, specpage >= 0, "%s could not load tex page %s", filename.toUtf8().c_str(), specfile);
+			specpage = iV_GetTransformTexture(specfile, [filename, specfile](iV_Image& sSprite){
+				if (sSprite.depth == 1) { return; }
+				// Otherwise, expecting 3 or 4-channel (RGB/RGBA)
+				ASSERT_OR_RETURN(, sSprite.depth == 3 || sSprite.depth == 4, "(%s): Does not have 1, 3 or 4 channels", specfile);
+				auto originalBmpData = sSprite.bmp;
+				const size_t numPixels = static_cast<size_t>(sSprite.height) * static_cast<size_t>(sSprite.width);
+				sSprite.bmp = (unsigned char *)malloc(numPixels);
+				for (size_t pixelIdx = 0; pixelIdx < numPixels; pixelIdx++)
+				{
+					uint32_t red = originalBmpData[(pixelIdx * sSprite.depth)];
+					uint32_t green = originalBmpData[(pixelIdx * sSprite.depth) + 1];
+					uint32_t blue = originalBmpData[(pixelIdx * sSprite.depth) + 2];
+					if (red == green && red == blue)
+					{
+						// all channels are the same - just use the first channel
+						sSprite.bmp[pixelIdx] = static_cast<unsigned char>(red);
+					}
+					else
+					{
+						// quick approximation of a weighted RGB -> Luma method
+						// (R+R+B+G+G+G)/6
+						uint32_t lum = (red+red+blue+green+green+green) / 6;
+						sSprite.bmp[pixelIdx] = static_cast<unsigned char>(std::min<uint32_t>(lum, 255));
+					}
+				}
+				sSprite.depth = 1;
+				// free the original bitmap data
+				free(originalBmpData);
+			});
+			ASSERT_OR_RETURN(, specpage.has_value(), "%s could not load tex page %s", filename.toUtf8().c_str(), specfile);
 		}
 
 		// assign tex pages and flags to all levels
 		for (iIMDShape *psShape = shape; psShape != nullptr; psShape = psShape->next)
 		{
-			psShape->texpage = texpage;
-			psShape->normalpage = normalpage;
-			psShape->specularpage = specpage;
+			psShape->texpage = texpage.value();
+			psShape->normalpage = (normalpage.has_value()) ? normalpage.value() : iV_TEX_INVALID;
+			psShape->specularpage = (specpage.has_value()) ? specpage.value() : iV_TEX_INVALID;
 			psShape->flags = imd_flags;
+			psShape->interpolate = (readInterpolate) ? interpolate : 1;
 		}
 
 		// check if model should use team colour mask
 		if (imd_flags & iV_IMD_TCMASK)
 		{
-			int texpage_mask;
+			std::string tcmask_name = pie_MakeTexPageTCMaskName(texfile);
+			tcmask_name += ".png";
+			optional<size_t> texpage_mask = iV_GetTransformTexture(tcmask_name.c_str(), [filename, tcmask_name](iV_Image& sSprite){
+				ASSERT_OR_RETURN(, sSprite.depth == 4, "(%s) tcmask png (%s) does not have 4 channels, as expected", filename.toUtf8().c_str(), tcmask_name.c_str());
+				auto originalBmpData = sSprite.bmp;
+				const size_t numPixels = static_cast<size_t>(sSprite.height) * static_cast<size_t>(sSprite.width);
+				// copy just the alpha channel over
+				sSprite.bmp = (unsigned char *)malloc(numPixels);
+				for (size_t pixelIdx = 0; pixelIdx < numPixels; pixelIdx++)
+				{
+					sSprite.bmp[pixelIdx] = originalBmpData[(pixelIdx * 4) + 3];
+				}
+				sSprite.depth = 1;
+				// free the original bitmap data
+				free(originalBmpData);
+			});
 
-			pie_MakeTexPageTCMaskName(texfile);
-			sstrcat(texfile, ".png");
-			texpage_mask = iV_GetTexture(texfile);
-
-			ASSERT_OR_RETURN(, texpage_mask >= 0, "%s could not load tcmask %s", filename.toUtf8().c_str(), texfile);
+			ASSERT_OR_RETURN(, texpage_mask.has_value(), "%s could not load tcmask %s", filename.toUtf8().c_str(), tcmask_name.c_str());
 
 			// Propagate settings through levels
 			for (iIMDShape *psShape = shape; psShape != nullptr; psShape = psShape->next)
 			{
-				psShape->tcmaskpage = texpage_mask;
+				psShape->tcmaskpage = (texpage_mask.has_value()) ? texpage_mask.value() : iV_TEX_INVALID;
 			}
 		}
 	}
