@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -20,10 +20,10 @@
 
 #include "lib/framework/frame.h"
 #include "lib/framework/file.h"
-#include "lib/framework/opengl.h"
 #include <stdlib.h>
 #include <string.h>
 #include "lib/framework/string_ext.h"
+#include "lib/framework/geometry.h"
 #include "lib/ivis_opengl/ivisdef.h"
 #include "lib/ivis_opengl/piestate.h"
 #include "lib/ivis_opengl/pieclip.h"
@@ -49,6 +49,7 @@ static float font_colour[4] = {1.f, 1.f, 1.f, 1.f};
 #include "ft2build.h"
 #include <unordered_map>
 #include <memory>
+#include <limits>
 
 #if defined(HB_VERSION_ATLEAST) && HB_VERSION_ATLEAST(1,0,5)
 //	#define WZ_FT_LOAD_FLAGS (FT_LOAD_DEFAULT | FT_LOAD_TARGET_LCD) // Needs further testing on low-DPI displays
@@ -101,7 +102,7 @@ struct GlyphMetrics
 
 struct FTFace
 {
-	FTFace(FT_Library &lib, const std::string &fileName, int32_t charSize, int32_t horizDPI, int32_t vertDPI)
+	FTFace(FT_Library &lib, const std::string &fileName, int32_t charSize, uint32_t horizDPI, uint32_t vertDPI)
 	{
 		UDWORD pFileSize = 0;
 		if (!loadFile(fileName.c_str(), &pFileData, &pFileSize))
@@ -426,15 +427,16 @@ public:
 	ShapingResult shapeText(const TextRun& text, FTFace &face)
 	{
 		hb_buffer_reset(m_buffer);
-		size_t length = text.text.size();
+		size_t length = std::min(text.text.size(), static_cast<size_t>(std::numeric_limits<int>::max()));
+		int textLength = static_cast<int>(length);
 
-		hb_buffer_add_utf8(m_buffer, text.text.c_str(), length, 0, length);
+		hb_buffer_add_utf8(m_buffer, text.text.c_str(), textLength, 0, textLength);
 		hb_buffer_guess_segment_properties(m_buffer);
 		hb_buffer_set_flags(m_buffer, (hb_buffer_flags_t)(HB_BUFFER_FLAG_BOT | HB_BUFFER_FLAG_EOT));
 
 		// harfbuzz shaping
 		std::array<hb_feature_t, 3> features = { {HBFeature::KerningOn, HBFeature::LigatureOn, HBFeature::CligOn} };
-		hb_shape(face.m_font, m_buffer, features.data(), features.size());
+		hb_shape(face.m_font, m_buffer, features.data(), static_cast<unsigned int>(features.size()));
 
 		unsigned int glyphCount;
 		hb_glyph_info_t *glyphInfo = hb_buffer_get_glyph_infos(m_buffer, &glyphCount);
@@ -498,11 +500,22 @@ inline float iV_GetVertScaleFactor()
 #define DEFAULT_DPI 72.0f
 
 static FTFace *regular = nullptr;
+static FTFace *regularBold = nullptr;
 static FTFace *bold = nullptr;
 static FTFace *medium = nullptr;
 static FTFace *small = nullptr;
 static FTFace *smallBold = nullptr;
 
+struct iVFontsHash
+{
+    std::size_t operator()(iV_fonts FontID) const
+    {
+        return static_cast<std::size_t>(FontID);
+    }
+};
+typedef std::unordered_map<iV_fonts, WzText, iVFontsHash> FontToEllipsisMapType;
+static FontToEllipsisMapType fontToEllipsisMap;
+
 static FTFace &getFTFace(iV_fonts FontID)
 {
 	switch (FontID)
@@ -510,6 +523,8 @@ static FTFace &getFTFace(iV_fonts FontID)
 	default:
 	case font_regular:
 		return *regular;
+	case font_regular_bold:
+		return *regularBold;
 	case font_large:
 		return *bold;
 	case font_medium:
@@ -523,8 +538,6 @@ static FTFace &getFTFace(iV_fonts FontID)
 
 static gfx_api::texture* textureID = nullptr;
 
-const GLint text_filtering = GL_LINEAR;
-
 void iV_TextInit(float horizScaleFactor, float vertScaleFactor)
 {
 	assert(horizScaleFactor >= 1.0f);
@@ -533,11 +546,12 @@ void iV_TextInit(float horizScaleFactor, float vertScaleFactor)
 	// Use the scaling factors to multiply the default DPI (72) to determine the desired internal font rendering DPI.
 	_horizScaleFactor = horizScaleFactor;
 	_vertScaleFactor = vertScaleFactor;
-	float horizDPI = DEFAULT_DPI * horizScaleFactor;
-	float vertDPI = DEFAULT_DPI * vertScaleFactor;
-	debug(LOG_WZ, "Text-Rendering Scaling Factor: %f x %f; Internal Font DPI: %f x %f", _horizScaleFactor, _vertScaleFactor, horizDPI, vertDPI);
+	uint32_t horizDPI = static_cast<uint32_t>(DEFAULT_DPI * horizScaleFactor);
+	uint32_t vertDPI = static_cast<uint32_t>(DEFAULT_DPI * vertScaleFactor);
+	debug(LOG_WZ, "Text-Rendering Scaling Factor: %f x %f; Internal Font DPI: %" PRIu32 " x %" PRIu32 "", _horizScaleFactor, _vertScaleFactor, horizDPI, vertDPI);
 
 	regular = new FTFace(getGlobalFTlib().lib, "fonts/DejaVuSans.ttf", 12 * 64, horizDPI, vertDPI);
+	regularBold = new FTFace(getGlobalFTlib().lib, "fonts/DejaVuSans-Bold.ttf", 12 * 64, horizDPI, vertDPI);
 	bold = new FTFace(getGlobalFTlib().lib, "fonts/DejaVuSans-Bold.ttf", 21 * 64, horizDPI, vertDPI);
 	medium = new FTFace(getGlobalFTlib().lib, "fonts/DejaVuSans.ttf", 16 * 64, horizDPI, vertDPI);
 	small = new FTFace(getGlobalFTlib().lib, "fonts/DejaVuSans.ttf", 9 * 64, horizDPI, vertDPI);
@@ -559,6 +573,7 @@ void iV_TextShutdown()
 	smallBold = nullptr;
 	delete textureID;
 	textureID = nullptr;
+	fontToEllipsisMap.clear();
 }
 
 void iV_TextUpdateScaleFactor(float horizScaleFactor, float vertScaleFactor)
@@ -567,13 +582,34 @@ void iV_TextUpdateScaleFactor(float horizScaleFactor, float vertScaleFactor)
 	iV_TextInit(horizScaleFactor, vertScaleFactor);
 }
 
+static WzText& iV_Internal_GetEllipsis(iV_fonts fontID)
+{
+	auto it = fontToEllipsisMap.find(fontID);
+	if (it == fontToEllipsisMap.end())
+	{
+		// We must create + cache an ellipsis for this fontID
+		it = fontToEllipsisMap.insert(FontToEllipsisMapType::value_type(fontID, WzText("\u2026", fontID))).first;
+	}
+	return it->second;
+}
+
+int iV_GetEllipsisWidth(iV_fonts fontID)
+{
+	return iV_Internal_GetEllipsis(fontID).width();
+}
+
+void iV_DrawEllipsis(iV_fonts fontID, Vector2f position, PIELIGHT colour)
+{
+	iV_Internal_GetEllipsis(fontID).render(position, colour);
+}
+
 unsigned int width_pixelsToPoints(unsigned int widthInPixels)
 {
-	return ceil((float)widthInPixels / _horizScaleFactor);
+	return static_cast<int>(ceil((float)widthInPixels / _horizScaleFactor));
 }
 unsigned int height_pixelsToPoints(unsigned int heightInPixels)
 {
-	return ceil((float)heightInPixels / _vertScaleFactor);
+	return static_cast<int>(ceil((float)heightInPixels / _vertScaleFactor));
 }
 
 // Returns the text width *in points*
@@ -607,7 +643,7 @@ unsigned int iV_GetCharWidth(uint32_t charCode, iV_fonts fontID)
 int metricsHeight_PixelsToPoints(int heightMetric)
 {
 	float ptMetric = (float)heightMetric / _vertScaleFactor;
-	return (ptMetric < 0) ? floor(ptMetric) : ceil(ptMetric);
+	return (ptMetric < 0) ? static_cast<int>(floor(ptMetric)) : static_cast<int>(ceil(ptMetric));
 }
 
 int iV_GetTextLineSize(iV_fonts fontID)
@@ -636,22 +672,24 @@ void iV_SetTextColour(PIELIGHT colour)
 	font_colour[3] = colour.byte.a / 255.0f;
 }
 
-/** Draws formatted text with word wrap, long word splitting, embedded newlines
- *  (uses '@' rather than '\n') and colour toggle mode ('#') which enables or
- *  disables font colouring.
- *
- *  @param String   the string to display.
- *  @param x,y      X and Y coordinates of top left of formatted text.
- *  @param width    the maximum width of the formatted text (beyond which line
- *                  wrapping is used).
- *  @param justify  The alignment style to use, which is one of the following:
- *                  FTEXT_LEFTJUSTIFY, FTEXT_CENTRE or FTEXT_RIGHTJUSTIFY.
- *  @return the Y coordinate for the next text line.
- */
-int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, UDWORD Justify, iV_fonts fontID)
+static bool breaksLine(char const c)
 {
+	return c == ASCII_NEWLINE || c == '\n';
+}
+
+static bool breaksWord(char const c)
+{
+	return c == ASCII_SPACE || breaksLine(c);
+}
+
+std::vector<TextLine> iV_FormatText(const char *String, UDWORD MaxWidth, UDWORD Justify, iV_fonts fontID, bool ignoreNewlines /*= false*/)
+{
+	std::vector<TextLine> lineDrawResults;
+
 	std::string FString;
 	std::string FWord;
+	const int x = 0;
+	const int y = 0;
 	int i;
 	int jx = x;		// Default to left justify.
 	int jy = y;
@@ -669,8 +707,10 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 
 		WWidth = 0;
 
+		auto indexWithinLine = 0;
+
 		// Parse through the string, adding words until width is achieved.
-		while (*curChar != 0 && WWidth < Width && !NewLine)
+		while (*curChar != 0 && (WWidth == 0 || WWidth < MaxWidth) && !NewLine)
 		{
 			const char *startOfWord = curChar;
 			const unsigned int FStringWidth = iV_GetTextWidth(FString.c_str(), fontID);
@@ -678,11 +718,11 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 			// Get the next word.
 			i = 0;
 			FWord.clear();
-			for (; *curChar != 0
-			     && *curChar != ASCII_SPACE
-			     && *curChar != ASCII_NEWLINE
-			     && *curChar != '\n';
-			     ++i, ++curChar)
+			for (
+				;
+				*curChar && ((indexWithinLine == 0 && !breaksLine(*curChar)) || !breaksWord(*curChar));
+				++i, ++curChar, ++indexWithinLine
+			)
 			{
 				if (*curChar == ASCII_COLOURMODE) // If it's a colour mode toggle char then just add it to the word.
 				{
@@ -692,37 +732,37 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 					continue;
 				}
 
+				FWord.push_back(*curChar);
+
 				// Update this line's pixel width.
 				//WWidth = FStringWidth + iV_GetCountedTextWidth(FWord.c_str(), i + 1, fontID);  // This triggers tonnes of valgrind warnings, if the string contains unicode. Adding lots of trailing garbage didn't help... Using iV_GetTextWidth with a null-terminated string, instead.
 				WWidth = FStringWidth + iV_GetTextWidth(FWord.c_str(), fontID);
 
 				// If this word doesn't fit on the current line then break out
-				if (WWidth > Width)
+				if (indexWithinLine != 0 && WWidth > MaxWidth)
 				{
+					FWord.erase(FWord.size() - 1);
 					break;
 				}
-
-				// If width ok then add this character to the current word.
-				FWord.push_back(*curChar);
 			}
 
 			// Don't forget the space.
 			if (*curChar == ASCII_SPACE)
 			{
-				WWidth += iV_GetCharWidth(' ', fontID);
-				if (WWidth <= Width)
+				FWord.push_back(' ');
+				++i;
+				++curChar;
+				GotSpace = true;
+				auto spaceWidth = iV_GetCharWidth(' ', fontID);
+				if (WWidth + spaceWidth <= MaxWidth)
 				{
-					FWord.push_back(' ');
-					++i;
-					++curChar;
-					GotSpace = true;
+					WWidth += spaceWidth;
 				}
 			}
 			// Check for new line character.
-			else if (*curChar == ASCII_NEWLINE
-			         || *curChar == '\n')
+			else if (breaksLine(*curChar))
 			{
-				if (!bMultiPlayer)
+				if (!ignoreNewlines)
 				{
 					NewLine = true;
 				}
@@ -734,7 +774,7 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 			// rewind to the start of this word and finish this line.
 			if (GotSpace
 			    && i != 0
-			    && WWidth > Width
+			    && WWidth > MaxWidth
 			    && FWord[i - 1] != ' ')
 			{
 				// Skip back to the beginning of this
@@ -760,11 +800,11 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 		switch (Justify)
 		{
 		case FTEXT_CENTRE:
-			jx = x + (Width - TWidth) / 2;
+			jx = x + (MaxWidth - TWidth) / 2;
 			break;
 
 		case FTEXT_RIGHTJUSTIFY:
-			jx = x + Width - TWidth;
+			jx = x + MaxWidth - TWidth;
 			break;
 
 		case FTEXT_LEFTJUSTIFY:
@@ -772,87 +812,44 @@ int iV_DrawFormattedText(const char *String, UDWORD x, UDWORD y, UDWORD Width, U
 			break;
 		}
 
-		// draw the text.
-		iV_DrawText(FString.c_str(), jx, jy, fontID);
+		// Store the line of text and its position in the bounding rect
+		lineDrawResults.push_back({FString, Vector2i(TWidth, iV_GetTextLineSize(fontID)), Vector2i(jx, jy)});
 
 		// and move down a line.
 		jy += iV_GetTextLineSize(fontID);
 	}
 
-	return jy;
+	return lineDrawResults;
 }
 
 void iV_DrawTextRotated(const char *string, float XPos, float YPos, float rotation, iV_fonts fontID)
 {
 	ASSERT_OR_RETURN(, string, "Couldn't render string!");
-	pie_SetTexturePage(TEXPAGE_EXTERN);
 
 	if (rotation != 0.f)
 	{
-		rotation = 180. - rotation;
+		rotation = 180.f - rotation;
 	}
 
 	PIELIGHT color;
-	color.vector[0] = font_colour[0] * 255.f;
-	color.vector[1] = font_colour[1] * 255.f;
-	color.vector[2] = font_colour[2] * 255.f;
-	color.vector[3] = font_colour[3] * 255.f;
+	color.vector[0] = static_cast<UBYTE>(font_colour[0] * 255.f);
+	color.vector[1] = static_cast<UBYTE>(font_colour[1] * 255.f);
+	color.vector[2] = static_cast<UBYTE>(font_colour[2] * 255.f);
+	color.vector[3] = static_cast<UBYTE>(font_colour[3] * 255.f);
 
 	TextRun tr(string, "en", HB_SCRIPT_COMMON, HB_DIRECTION_LTR);
 	DrawTextResult drawResult = getShaper().drawText(tr, getFTFace(fontID));
 
 	if (drawResult.text.width > 0 && drawResult.text.height > 0)
 	{
-		pie_SetTexturePage(TEXPAGE_EXTERN);
 		if (textureID)
 			delete textureID;
-		textureID = gfx_api::context::get().create_texture(drawResult.text.width, drawResult.text.height, gfx_api::pixel_format::rgba);
-		textureID->upload(0u, 0u, 0u, drawResult.text.width, drawResult.text.height, gfx_api::pixel_format::rgba, drawResult.text.data.get());
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, text_filtering);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, text_filtering);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-		glDisable(GL_CULL_FACE);
-		iV_DrawImageText(*textureID, Vector2i(XPos, YPos), Vector2f((float)drawResult.text.offset_x / _horizScaleFactor, (float)drawResult.text.offset_y / _vertScaleFactor), Vector2f((float)drawResult.text.width / _horizScaleFactor, (float)drawResult.text.height / _vertScaleFactor), rotation, REND_TEXT, color);
-		glEnable(GL_CULL_FACE);
+		textureID = gfx_api::context::get().create_texture(1, drawResult.text.width, drawResult.text.height, gfx_api::pixel_format::FORMAT_RGBA8_UNORM_PACK8);
+		textureID->upload(0u, 0u, 0u, drawResult.text.width, drawResult.text.height, gfx_api::pixel_format::FORMAT_RGBA8_UNORM_PACK8, drawResult.text.data.get());
+		iV_DrawImageText(*textureID, Vector2f(XPos, YPos), Vector2f((float)drawResult.text.offset_x / _horizScaleFactor, (float)drawResult.text.offset_y / _vertScaleFactor), Vector2f((float)drawResult.text.width / _horizScaleFactor, (float)drawResult.text.height / _vertScaleFactor), rotation, color);
 	}
 }
 
-#if 0
-static void iV_DrawTextRotatedFv(float x, float y, float rotation, const char *format, va_list ap, iV_fonts fontID)
-{
-	va_list aq;
-	size_t size;
-	char *str;
-
-	/* Required because we're using the va_list ap twice otherwise, which
-	 * results in undefined behaviour. See stdarg(3) for details.
-	 */
-	va_copy(aq, ap);
-
-	// Allocate a buffer large enough to hold our string on the stack
-	size = vsnprintf(NULL, 0, format, ap);
-	str = (char *)alloca(size + 1);
-
-	// Print into our newly created string buffer
-	vsprintf(str, format, aq);
-
-	va_end(aq);
-
-	// Draw the produced string to the screen at the given position and rotation
-	iV_DrawTextRotated(str, x, y, rotation, fontID);
-}
-
-void iV_DrawTextF(float x, float y, const char *format, ...)
-{
-	va_list ap;
-
-	va_start(ap, format);
-	iV_DrawTextRotatedFv(x, y, 0.f, format, ap);
-	va_end(ap);
-}
-#endif
-
 int WzText::width()
 {
 	updateCacheIfNecessary();
@@ -916,15 +913,9 @@ void WzText::drawAndCacheText(const std::string &string, iV_fonts fontID)
 
 	if (dimensions.x > 0 && dimensions.y > 0)
 	{
-		pie_SetTexturePage(TEXPAGE_EXTERN);
-		texture = gfx_api::context::get().create_texture(dimensions.x, dimensions.y, gfx_api::pixel_format::rgba);
-		texture->upload(0u, 0u, 0u, dimensions.x , dimensions.y, gfx_api::pixel_format::rgba, drawResult.text.data.get());
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, text_filtering);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, text_filtering);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
+		texture = gfx_api::context::get().create_texture(1, dimensions.x, dimensions.y, gfx_api::pixel_format::FORMAT_RGBA8_UNORM_PACK8);
+		texture->upload(0u, 0u, 0u, dimensions.x , dimensions.y, gfx_api::pixel_format::FORMAT_RGBA8_UNORM_PACK8, drawResult.text.data.get());
 	}
-	glBindTexture(GL_TEXTURE_2D, 0);
 }
 
 void WzText::redrawAndCacheText()
@@ -991,7 +982,7 @@ inline void WzText::updateCacheIfNecessary()
 	}
 }
 
-void WzText::render(Vector2i position, PIELIGHT colour, float rotation)
+void WzText::render(Vector2f position, PIELIGHT colour, float rotation, int maxWidth, int maxHeight)
 {
 	updateCacheIfNecessary();
 
@@ -1004,11 +995,32 @@ void WzText::render(Vector2i position, PIELIGHT colour, float rotation)
 
 	if (rotation != 0.f)
 	{
-		rotation = 180. - rotation;
+		rotation = 180.f - rotation;
 	}
-	glDisable(GL_CULL_FACE);
-	iV_DrawImageText(*texture, position, Vector2f(offsets.x / mRenderingHorizScaleFactor, offsets.y / mRenderingVertScaleFactor), Vector2f(dimensions.x / mRenderingHorizScaleFactor, dimensions.y / mRenderingVertScaleFactor), rotation, REND_TEXT, colour);
-	glEnable(GL_CULL_FACE);
+
+	if (maxWidth <= 0 && maxHeight <= 0)
+	{
+		iV_DrawImageText(*texture, position, Vector2f(offsets.x / mRenderingHorizScaleFactor, offsets.y / mRenderingVertScaleFactor), Vector2f(dimensions.x / mRenderingHorizScaleFactor, dimensions.y / mRenderingVertScaleFactor), rotation, colour);
+	}
+	else
+	{
+		WzRect clippingRectInPixels;
+		clippingRectInPixels.setWidth((maxWidth > 0) ? static_cast<int>((float)maxWidth * mRenderingHorizScaleFactor) : dimensions.x);
+		clippingRectInPixels.setHeight((maxHeight > 0) ? static_cast<int>((float)maxHeight * mRenderingVertScaleFactor) : dimensions.y);
+		iV_DrawImageTextClipped(*texture, dimensions, position, Vector2f(offsets.x / mRenderingHorizScaleFactor, offsets.y / mRenderingVertScaleFactor), Vector2f((maxWidth > 0) ? maxWidth : dimensions.x / mRenderingHorizScaleFactor, (maxHeight > 0) ? maxHeight : dimensions.y / mRenderingVertScaleFactor), rotation, colour, clippingRectInPixels);
+	}
+}
+
+void WzText::renderOutlined(int x, int y, PIELIGHT colour, PIELIGHT outlineColour)
+{
+	for (auto i = -1; i <= 1; i++)
+	{
+		for (auto j = -1; j <= 1; j++)
+		{
+			render(x + i, y + j, outlineColour);
+		}
+	}
+	render(x, y, colour);
 }
 
 // Sets the text, truncating to a desired width limit (in *points*) if needed
