@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -36,6 +36,10 @@
 #endif
 #include <zlib.h>
 
+#if defined(__clang__)
+	#pragma clang diagnostic ignored "-Wshorten-64-to-32" // FIXME!!
+#endif
+
 enum
 {
 	SOCK_CONNECTION,
@@ -436,11 +440,11 @@ static int socketThreadFunction(void *)
 
 				// Write data.
 				// FIXME SOMEHOW AAARGH This send() call can't block, but unless the socket is not set to blocking (setting the socket to nonblocking had better work, or else), does anyway (at least sometimes, when someone quits). Not reproducible except in public releases.
-				ssize_t ret = send(sock->fd[SOCK_CONNECTION], reinterpret_cast<char *>(&writeQueue[0]), writeQueue.size(), MSG_NOSIGNAL);
-				if (ret != SOCKET_ERROR)
+				ssize_t retSent = send(sock->fd[SOCK_CONNECTION], reinterpret_cast<char *>(&writeQueue[0]), writeQueue.size(), MSG_NOSIGNAL);
+				if (retSent != SOCKET_ERROR)
 				{
 					// Erase as much data as written.
-					writeQueue.erase(writeQueue.begin(), writeQueue.begin() + ret);
+					writeQueue.erase(writeQueue.begin(), writeQueue.begin() + retSent);
 					if (writeQueue.empty())
 					{
 						socketThreadWrites.erase(w);  // Nothing left to write, delete from pending list.
@@ -688,7 +692,7 @@ ssize_t writeAll(Socket *sock, const void *buf, size_t size, size_t *rawByteCoun
 	return size;
 }
 
-void socketFlush(Socket *sock, size_t *rawByteCount)
+void socketFlush(Socket *sock, uint8_t player, size_t *rawByteCount)
 {
 	size_t ignored;
 	size_t &rawBytes = rawByteCount != nullptr ? *rawByteCount : ignored;
@@ -699,6 +703,8 @@ void socketFlush(Socket *sock, size_t *rawByteCount)
 		return;  // Not compressed, so don't mess with zlib.
 	}
 
+	ASSERT(!sock->writeError, "Socket write error?? (Player: %" PRIu8 "", player);
+
 	// Flush data out of zlib compression state.
 	do
 	{
@@ -825,6 +831,47 @@ void SocketSet_DelSocket(SocketSet *set, Socket *socket)
 	}
 }
 
+#if !defined(SOCK_CLOEXEC)
+static bool setSocketInheritable(SOCKET fd, bool inheritable)
+{
+#if   defined(WZ_OS_UNIX)
+	int sockopts = fcntl(fd, F_SETFD);
+	if (sockopts == SOCKET_ERROR)
+	{
+		debug(LOG_NET, "Failed to retrieve current socket options: %s", strSockError(getSockErr()));
+		return false;
+	}
+
+	// Set or clear FD_CLOEXEC flag
+	if (inheritable)
+	{
+		sockopts &= ~FD_CLOEXEC;
+	}
+	else
+	{
+		sockopts |= FD_CLOEXEC;
+	}
+
+	if (fcntl(fd, F_SETFD, sockopts) == SOCKET_ERROR)
+	{
+		debug(LOG_NET, "Failed to set socket %sinheritable: %s", (inheritable ? "" : "non-"), strSockError(getSockErr()));
+		return false;
+	}
+#elif defined(WZ_OS_WIN)
+	DWORD dwFlags = (inheritable) ? HANDLE_FLAG_INHERIT : 0;
+	if (::SetHandleInformation((HANDLE)fd, HANDLE_FLAG_INHERIT, dwFlags) == 0)
+	{
+		DWORD dwErr = GetLastError();
+		debug(LOG_NET, "Failed to set socket %sinheritable: %s", (inheritable ? "" : "non-"), std::to_string(dwErr).c_str());
+		return false;
+	}
+#endif
+
+	debug(LOG_NET, "Socket is set to %sinheritable.", (inheritable ? "" : "non-"));
+	return true;
+}
+#endif // !defined(SOCK_CLOEXEC)
+
 static bool setSocketBlocking(const SOCKET fd, bool blocking)
 {
 #if   defined(WZ_OS_UNIX)
@@ -973,7 +1020,7 @@ ssize_t readAll(Socket *sock, void *buf, size_t size, unsigned int timeout)
 
 	if (sock->fd[SOCK_CONNECTION] == INVALID_SOCKET)
 	{
-		debug(LOG_ERROR, "Invalid socket (%p), sock->fd[SOCK_CONNECTION]=%x  (error: EBADF)", static_cast<void *>(sock), sock->fd[SOCK_CONNECTION]);
+		debug(LOG_ERROR, "Invalid socket (%p), sock->fd[SOCK_CONNECTION]=%" PRIuPTR"x  (error: EBADF)", static_cast<void *>(sock), static_cast<uintptr_t>(sock->fd[SOCK_CONNECTION]));
 		setSockErr(EBADF);
 		return SOCKET_ERROR;
 	}
@@ -1003,7 +1050,7 @@ ssize_t readAll(Socket *sock, void *buf, size_t size, unsigned int timeout)
 		sock->ready = false;
 		if (ret == 0)
 		{
-			debug(LOG_NET, "Socket %x disconnected.", sock->fd[SOCK_CONNECTION]);
+			debug(LOG_NET, "Socket %" PRIuPTR"x disconnected.", static_cast<uintptr_t>(sock->fd[SOCK_CONNECTION]));
 			sock->readDisconnected = true;
 			setSockErr(ECONNRESET);
 			return received;
@@ -1090,7 +1137,11 @@ Socket *socketAccept(Socket *sock)
 			Socket *conn;
 			unsigned int j;
 
+#if defined(SOCK_CLOEXEC)
+			const SOCKET newConn = accept4(sock->fd[i], (struct sockaddr *)&addr, &addr_len, SOCK_CLOEXEC);
+#else
 			const SOCKET newConn = accept(sock->fd[i], (struct sockaddr *)&addr, &addr_len);
+#endif
 			if (newConn == INVALID_SOCKET)
 			{
 				// Ignore the case where no connection is pending
@@ -1111,6 +1162,14 @@ Socket *socketAccept(Socket *sock)
 				return nullptr;
 			}
 
+#if !defined(SOCK_CLOEXEC)
+			if (!setSocketInheritable(newConn, false))
+			{
+				debug(LOG_NET, "Couldn't set socket (%p) inheritable status (false). Ignoring...", static_cast<void *>(conn));
+				// ignore and continue
+			}
+#endif
+
 			debug(LOG_NET, "setting socket (%p) blocking status (false).", static_cast<void *>(conn));
 			if (!setSocketBlocking(newConn, false))
 			{
@@ -1172,7 +1231,11 @@ Socket *socketOpen(const SocketAddress *addr, unsigned timeout)
 		conn->fd[i] = INVALID_SOCKET;
 	}
 
-	conn->fd[SOCK_CONNECTION] = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
+	int socket_type = addr->ai_socktype;
+#if defined(SOCK_CLOEXEC)
+	socket_type |= SOCK_CLOEXEC;
+#endif
+	conn->fd[SOCK_CONNECTION] = socket(addr->ai_family, socket_type, addr->ai_protocol);
 
 	if (conn->fd[SOCK_CONNECTION] == INVALID_SOCKET)
 	{
@@ -1181,6 +1244,14 @@ Socket *socketOpen(const SocketAddress *addr, unsigned timeout)
 		return nullptr;
 	}
 
+#if !defined(SOCK_CLOEXEC)
+	if (!setSocketInheritable(conn->fd[SOCK_CONNECTION], false))
+	{
+		debug(LOG_NET, "Couldn't set socket (%p) inheritable status (false). Ignoring...", static_cast<void *>(conn));
+		// ignore and continue
+	}
+#endif
+
 	debug(LOG_NET, "setting socket (%p) blocking status (false).", static_cast<void *>(conn));
 	if (!setSocketBlocking(conn->fd[SOCK_CONNECTION], false))
 	{
@@ -1309,8 +1380,12 @@ Socket *socketListen(unsigned int port)
 	addr6.sin6_flowinfo = 0;
 	addr6.sin6_scope_id = 0;
 
-	conn->fd[SOCK_IPV4_LISTEN] = socket(addr4.sin_family, SOCK_STREAM, 0);
-	conn->fd[SOCK_IPV6_LISTEN] = socket(addr6.sin6_family, SOCK_STREAM, 0);
+	int socket_type = SOCK_STREAM;
+#if defined(SOCK_CLOEXEC)
+	socket_type |= SOCK_CLOEXEC;
+#endif
+	conn->fd[SOCK_IPV4_LISTEN] = socket(addr4.sin_family, socket_type, 0);
+	conn->fd[SOCK_IPV6_LISTEN] = socket(addr6.sin6_family, socket_type, 0);
 
 	if (conn->fd[SOCK_IPV4_LISTEN] == INVALID_SOCKET
 	    && conn->fd[SOCK_IPV6_LISTEN] == INVALID_SOCKET)
@@ -1352,6 +1427,14 @@ Socket *socketListen(unsigned int port)
 
 	if (conn->fd[SOCK_IPV4_LISTEN] != INVALID_SOCKET)
 	{
+#if !defined(SOCK_CLOEXEC)
+		if (!setSocketInheritable(conn->fd[SOCK_IPV4_LISTEN], false))
+		{
+			debug(LOG_NET, "Couldn't set socket (%p) inheritable status (false). Ignoring...", static_cast<void *>(conn));
+			// ignore and continue
+		}
+#endif
+
 		if (setsockopt(conn->fd[SOCK_IPV4_LISTEN], SOL_SOCKET, SO_REUSEADDR, (const char *)&so_reuseaddr, sizeof(so_reuseaddr)) == SOCKET_ERROR)
 		{
 			debug(LOG_WARNING, "Failed to set SO_REUSEADDR on IPv4 socket. Error: %s", strSockError(getSockErr()));
@@ -1374,6 +1457,14 @@ Socket *socketListen(unsigned int port)
 
 	if (conn->fd[SOCK_IPV6_LISTEN] != INVALID_SOCKET)
 	{
+#if !defined(SOCK_CLOEXEC)
+		if (!setSocketInheritable(conn->fd[SOCK_IPV6_LISTEN], false))
+		{
+			debug(LOG_NET, "Couldn't set socket (%p) inheritable status (false). Ignoring...", static_cast<void *>(conn));
+			// ignore and continue
+		}
+#endif
+
 		if (setsockopt(conn->fd[SOCK_IPV6_LISTEN], SOL_SOCKET, SO_REUSEADDR, (const char *)&so_reuseaddr, sizeof(so_reuseaddr)) == SOCKET_ERROR)
 		{
 			debug(LOG_INFO, "Failed to set SO_REUSEADDR on IPv6 socket. Error: %s", strSockError(getSockErr()));
@@ -1442,15 +1533,105 @@ void socketArrayClose(Socket **sockets, size_t maxSockets)
 	std::fill(sockets, sockets + maxSockets, (Socket *)nullptr);      // Set the pointers to NULL.
 }
 
+WZ_DECL_NONNULL(1) bool socketHasIPv4(Socket *sock)
+{
+	if (sock->fd[SOCK_IPV4_LISTEN] != INVALID_SOCKET)
+	{
+		return true;
+	}
+	else
+	{
+#if defined(IPV6_V6ONLY)
+		if (sock->fd[SOCK_IPV6_LISTEN] != INVALID_SOCKET)
+		{
+			int ipv6_v6only = 1;
+			socklen_t len = sizeof(ipv6_v6only);
+			if (getsockopt(sock->fd[SOCK_IPV6_LISTEN], IPPROTO_IPV6, IPV6_V6ONLY, (char *)&ipv6_v6only, &len) == 0)
+			{
+				return ipv6_v6only == 0;
+			}
+		}
+#endif
+		return false;
+	}
+}
+
+WZ_DECL_NONNULL(1) bool socketHasIPv6(Socket *sock)
+{
+	return sock->fd[SOCK_IPV6_LISTEN] != INVALID_SOCKET;
+}
+
 char const *getSocketTextAddress(Socket const *sock)
 {
 	return sock->textAddress;
 }
 
+std::vector<unsigned char> ipv4_AddressString_To_NetBinary(const std::string& ipv4Address)
+{
+	std::vector<unsigned char> binaryForm(sizeof(struct in_addr), 0);
+	if (inet_pton(AF_INET, ipv4Address.c_str(), binaryForm.data()) <= 0)
+	{
+		// inet_pton failed
+		binaryForm.clear();
+	}
+	return binaryForm;
+}
+
+#ifndef INET_ADDRSTRLEN
+# define INET_ADDRSTRLEN 16
+#endif
+
+std::string ipv4_NetBinary_To_AddressString(const std::vector<unsigned char>& ip4NetBinaryForm)
+{
+	if (ip4NetBinaryForm.size() != sizeof(struct in_addr))
+	{
+		return "";
+	}
+
+	char buf[INET_ADDRSTRLEN] = {0};
+	if (inet_ntop(AF_INET, ip4NetBinaryForm.data(), buf, sizeof(buf)) == nullptr)
+	{
+		return "";
+	}
+	std::string ipv4Address = buf;
+	return ipv4Address;
+}
+
+std::vector<unsigned char> ipv6_AddressString_To_NetBinary(const std::string& ipv6Address)
+{
+	std::vector<unsigned char> binaryForm(sizeof(struct in6_addr), 0);
+	if (inet_pton(AF_INET6, ipv6Address.c_str(), binaryForm.data()) <= 0)
+	{
+		// inet_pton failed
+		binaryForm.clear();
+	}
+	return binaryForm;
+}
+
+#ifndef INET6_ADDRSTRLEN
+# define INET6_ADDRSTRLEN 46
+#endif
+
+std::string ipv6_NetBinary_To_AddressString(const std::vector<unsigned char>& ip6NetBinaryForm)
+{
+	if (ip6NetBinaryForm.size() != sizeof(struct in6_addr))
+	{
+		return "";
+	}
+
+	char buf[INET6_ADDRSTRLEN] = {0};
+	if (inet_ntop(AF_INET6, ip6NetBinaryForm.data(), buf, sizeof(buf)) == nullptr)
+	{
+		return "";
+	}
+	std::string ipv6Address = buf;
+	return ipv6Address;
+}
+
 SocketAddress *resolveHost(const char *host, unsigned int port)
 {
 	struct addrinfo *results;
-	char *service;
+	std::string service;
 	struct addrinfo hint;
 	int error, flags = 0;
 
@@ -1469,12 +1650,12 @@ SocketAddress *resolveHost(const char *host, unsigned int port)
 	hint.ai_canonname = nullptr;
 	hint.ai_next      = nullptr;
 
-	sasprintf(&service, "%u", port);
+	service = astringf("%u", port);
 
-	error = getaddrinfo(host, service, &hint, &results);
+	error = getaddrinfo(host, service.c_str(), &hint, &results);
 	if (error != 0)
 	{
-		debug(LOG_NET, "getaddrinfo failed for %s:%s: %s", host, service, gai_strerror(error));
+		debug(LOG_NET, "getaddrinfo failed for %s:%s: %s", host, service.c_str(), gai_strerror(error));
 		return nullptr;
 	}
 
