@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -25,13 +25,14 @@
 #include "lib/ivis_opengl/bitimage.h"
 #include "lib/ivis_opengl/tex.h"
 #include "src/warzoneconfig.h"
-#include "src/frontend.h"
 #include "cursors_sdl.h"
 #include <SDL.h>
 
 static CURSOR currentCursor = CURSOR_MAX;
-static SDL_Cursor *aCursors[CURSOR_MAX];
-static bool monoCursor;
+static CURSOR lastAppliedCursor = CURSOR_MAX;
+static SDL_Cursor *aCursors[CURSOR_MAX] = {};
+static bool monoCursor = false;
+static bool cursorsEnabled = false;
 
 /* TODO: do bridge and attach need swapping? */
 static const char *cursor_arrow[] =
@@ -1290,6 +1291,11 @@ SDL_Cursor *init_system_ColorCursor(CURSOR cur, const char *fileName)
 #endif
 
 	SDL_Surface *surface = SDL_CreateRGBSurfaceFrom(psSprite->bmp, psSprite->width, psSprite->height, psSprite->depth * 8, psSprite->width * 4, rmask, gmask, bmask, amask);
+	if (!surface)
+	{
+		debug(LOG_FATAL, "Failed to create cursor because: %s", SDL_GetError());
+		return nullptr;
+	}
 	SDL_Cursor *pointer = SDL_CreateColorCursor(surface, psSprite->width / 2, psSprite->height / 2);	// We center the hotspot for all (FIXME ?)
 	if (!pointer)
 	{
@@ -1297,10 +1303,10 @@ SDL_Cursor *init_system_ColorCursor(CURSOR cur, const char *fileName)
 		exit(-1);
 	}
 
-	// free up image & surface data
+	// free up surface & image data
+	SDL_FreeSurface(surface);
 	free(psSprite->bmp);
 	free(psSprite);
-	SDL_FreeSurface(surface);
 
 	return pointer;
 }
@@ -1356,24 +1362,37 @@ SDL_Cursor *init_system_cursor32(CURSOR cur)
 }
 
 /**
-	wzSetCursor()-- Set the current cursor
+	wzSetCursor()-- Set the current cursor. Cursor is actually applied in wzApplyCursor() in mainLoop()
  */
 void wzSetCursor(CURSOR cur)
 {
 	ASSERT(cur < CURSOR_MAX, "Specified cursor(%d) is over our limit of (%d)!", (int)cur, (int)CURSOR_MAX);
-	// we reset mouse cursors on the fly...(only in the mouse options screen!)
-	if ((!(war_GetColouredCursor() ^ monoCursor)) && (titleMode == MOUSE_OPTIONS))
+	
+	currentCursor = cur;
+}
+
+void wzApplyCursor()
+{
+	if (!cursorsEnabled) return;
+
+	// If mouse cursor options change, change cursors (used to only work on mouse options screen for some reason)
+	if (!(war_GetColouredCursor() ^ monoCursor))
 	{
 		sdlFreeCursors();
 		war_GetColouredCursor() ? sdlInitColoredCursors() : sdlInitCursors();
-		SDL_SetCursor(aCursors[cur]);
+		SDL_SetCursor(aCursors[currentCursor]);
+		lastAppliedCursor = currentCursor;
+		return;
 	}
-	// If we are already using this cursor then  return
-	if (cur != currentCursor)
+
+	// If we are already using this cursor then return
+	if (currentCursor == lastAppliedCursor)
 	{
-		SDL_SetCursor(aCursors[cur]);
-		currentCursor = cur;
+		return;
 	}
+
+	SDL_SetCursor(aCursors[currentCursor]);
+	lastAppliedCursor = currentCursor;
 }
 
 /**
@@ -1410,6 +1429,7 @@ void sdlInitColoredCursors()
 	aCursors[CURSOR_PICKUP]      = init_system_ColorCursor(CURSOR_PICKUP, "images/intfac/image_cursor_pickup.png");
 	aCursors[CURSOR_SEEKREPAIR]  = init_system_ColorCursor(CURSOR_SEEKREPAIR, "images/intfac/image_cursor_repair.png");
 	aCursors[CURSOR_SELECT]      = init_system_ColorCursor(CURSOR_SELECT, "images/intfac/image_cursor_select.png");
+	cursorsEnabled = true;
 }
 
 /**
@@ -1446,6 +1466,7 @@ void sdlInitCursors()
 	aCursors[CURSOR_PICKUP]      = init_system_cursor32(CURSOR_PICKUP);
 	aCursors[CURSOR_SEEKREPAIR]  = init_system_cursor32(CURSOR_SEEKREPAIR);
 	aCursors[CURSOR_SELECT]      = init_system_cursor32(CURSOR_SELECT);
+	cursorsEnabled = true;
 }
 
 void sdlFreeCursors()
@@ -1453,6 +1474,12 @@ void sdlFreeCursors()
 	unsigned int i;
 	for (i = 0 ; i < ARRAY_SIZE(aCursors); ++i)
 	{
+		if (aCursors[i] == nullptr)
+		{
+			continue;
+		}
 		SDL_FreeCursor(aCursors[i]);
+		aCursors[i] = nullptr;
 	}
+	cursorsEnabled = false;
 }
