@@ -1,6 +1,6 @@
 /*
 	This file is part of Warzone 2100.
-	Copyright (C) 2008-2019  Warzone 2100 Project
+	Copyright (C) 2008-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -56,7 +56,7 @@
  */
 
 #include "lib/framework/frame.h"
-#include "lib/framework/opengl.h"
+#include "lib/ivis_opengl/gfx_api.h"
 #include "sequence.h"
 #include "timer.h"
 #include "lib/framework/math_ext.h"
@@ -76,11 +76,7 @@
 #include <physfs.h>
 
 #include <vorbis/codec.h>
-#if defined(WZ_OS_MAC)
-#include <OpenAL/al.h>
-#else
 #include <AL/al.h>
-#endif
 
 #include "lib/framework/physfs_ext.h"
 
@@ -159,6 +155,7 @@ static gfx_api::gfxFloat vertices[NUM_VERTICES][2];
 static gfx_api::gfxFloat Scrnvidpos[3];
 
 static SCANLINE_MODE use_scanlines;
+static bool scanlinesDisabled = false;
 
 // Helper; just grab some more compressed bitstream and sync it for page extraction
 static int buffer_data(PHYSFS_file *in, ogg_sync_state *oy)
@@ -166,7 +163,7 @@ static int buffer_data(PHYSFS_file *in, ogg_sync_state *oy)
 	// read in 256K chunks
 	const int size = 262144;
 	char *buffer = ogg_sync_buffer(oy, size);
-	int bytes = WZ_PHYSFS_readBytes(in, buffer, size);
+	int bytes = static_cast<int>(WZ_PHYSFS_readBytes(in, buffer, size));
 
 	ogg_sync_wrote(oy, bytes);
 	return (bytes);
@@ -202,7 +199,7 @@ static void open_audio(void)
 {
 	float volume = 1.0;
 
-	audiodata.audiofd_fragsize = (((videodata.vi.channels * 16) / 8) * videodata.vi.rate);
+	audiodata.audiofd_fragsize = static_cast<int>((((videodata.vi.channels * 16) / 8) * videodata.vi.rate));
 	audiobuf = (ogg_int16_t *)malloc(audiodata.audiofd_fragsize);
 
 	// FIX ME:  This call will fail, since we have, most likely, already
@@ -262,15 +259,15 @@ static double getRelativeTime(void)
 	return ((getTimeNow() - basetime) * .001);
 }
 
-const gfx_api::gfxFloat texture_width = 1024.0f;
-const gfx_api::gfxFloat texture_height = 1024.0f;
+const size_t texture_width = 1024;
+const size_t texture_height = 1024;
 
 /** Allocates memory to hold the decoded video frame
  */
 static void Allocate_videoFrame(void)
 {
 	int size = videodata.ti.frame_width * videodata.ti.frame_height * 4;
-	if (use_scanlines)
+	if (!seq_getScanlinesDisabled() && seq_getScanlineMode())
 	{
 		size *= 2;
 	}
@@ -310,8 +307,9 @@ static void video_write(bool update)
 	unsigned int x = 0, y = 0;
 	const int video_width = videodata.ti.frame_width;
 	const int video_height = videodata.ti.frame_height;
+	SCANLINE_MODE scanMode = seq_getScanlinesDisabled() ? SCANLINES_OFF : seq_getScanlineMode();
 	// when using scanlines we need to double the height
-	const int height_factor = (use_scanlines ? 2 : 1);
+	const int height_factor = (scanMode ? 2 : 1);
 	yuv_buffer yuv;
 
 	if (update)
@@ -345,12 +343,12 @@ static void video_write(bool update)
 				uint32_t rgba = (R << Rshift) | (G << Gshift) | (B << Bshift) | (0xFF << Ashift);
 
 				RGBAframe[rgb_offset] = rgba;
-				if (use_scanlines == SCANLINES_50)
+				if (scanMode == SCANLINES_50)
 				{
 					// halve the rgb values for a dimmed scanline
 					RGBAframe[rgb_offset + video_width] = (rgba >> 1 & RGBmask) | Amask;
 				}
-				else if (use_scanlines == SCANLINES_BLACK)
+				else if (scanMode == SCANLINES_BLACK)
 				{
 					RGBAframe[rgb_offset + video_width] = Amask;
 				}
@@ -366,33 +364,32 @@ static void video_write(bool update)
 
 				rgba = (R << Rshift) | (G << Gshift) | (B << Bshift) | (0xFF << Ashift);
 				RGBAframe[rgb_offset] = rgba;
-				if (use_scanlines == SCANLINES_50)
+				if (scanMode == SCANLINES_50)
 				{
 					// halve the rgb values for a dimmed scanline
 					RGBAframe[rgb_offset + video_width] = (rgba >> 1 & RGBmask) | Amask;
 				}
-				else if (use_scanlines == SCANLINES_BLACK)
+				else if (scanMode == SCANLINES_BLACK)
 				{
 					RGBAframe[rgb_offset + video_width] = Amask;
 				}
 				rgb_offset++;
 			}
-			if (use_scanlines)
+			if (scanMode)
 			{
 				rgb_offset += video_width;
 			}
 		}
 
-		videoGfx->updateTexture(RGBAframe, video_width, video_height * height_factor);
+		videoGfx->updateTexture(RGBAframe, static_cast<size_t>(video_width), static_cast<size_t>(video_height) * static_cast<size_t>(height_factor));
 	}
 
-	glDisable(GL_DEPTH_TEST);
-	glDepthMask(GL_FALSE);
+	const auto& modelViewProjectionMatrix = glm::ortho(0.f, static_cast<float>(pie_GetVideoBufferWidth()), static_cast<float>(pie_GetVideoBufferHeight()), 0.f) *
+	glm::translate(glm::vec3(Scrnvidpos[0], Scrnvidpos[1], Scrnvidpos[2]));
 
-	videoGfx->draw(
-		glm::ortho(0.f, static_cast<float>(pie_GetVideoBufferWidth()), static_cast<float>(pie_GetVideoBufferHeight()), 0.f) *
-		glm::translate(glm::vec3(Scrnvidpos[0], Scrnvidpos[1], Scrnvidpos[2]))
-	);
+	gfx_api::VideoPSO::get().bind();
+	gfx_api::VideoPSO::get().bind_constants({ modelViewProjectionMatrix, glm::vec2(0), glm::vec2(0), glm::vec4(1), 0 });
+	videoGfx->draw<gfx_api::VideoPSO>(modelViewProjectionMatrix);
 }
 
 // FIXME: perhaps we should use wz's routine for audio?
@@ -432,7 +429,7 @@ static void audio_write(void)
 		}
 
 		alBufferData(oldbuffer, (videodata.vi.channels == 1 ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16),
-		             audiobuf, audiodata.audiobuf_fill, videodata.vi.rate);
+		             audiobuf, audiodata.audiobuf_fill, static_cast<ALsizei>(videodata.vi.rate));
 
 		alSourceQueueBuffers(audiodata.source, 1, &oldbuffer);
 		audiodata.totbufstarted++;
@@ -461,6 +458,7 @@ static void seq_InitOgg(void)
 	vorbis_p = 0;
 
 	videoplaying = false;
+	seq_setScanlinesDisabled(false);
 
 	/* single frame video buffering */
 	videobuf_ready = false;
@@ -511,7 +509,7 @@ bool seq_Play(const char *filename)
 	fpInfile = PHYSFS_openRead(filename);
 	if (fpInfile == nullptr)
 	{
-		info("unable to open '%s' for playback", filename);
+		wz_info("unable to open '%s' for playback", filename);
 
 		fpInfile = PHYSFS_openRead("novideo.ogg");
 		if (fpInfile == nullptr)
@@ -622,7 +620,7 @@ bool seq_Play(const char *filename)
 		}
 		else
 		{
-			int ret = buffer_data(fpInfile, &videodata.oy);   /* someone needs more data */
+			ret = buffer_data(fpInfile, &videodata.oy);   /* someone needs more data */
 
 			if (ret == 0)
 			{
@@ -678,12 +676,12 @@ bool seq_Play(const char *filename)
 	}
 
 	/* open video */
-	videoGfx = new GFX(GFX_TEXTURE, GL_TRIANGLE_STRIP, 2);
+	videoGfx = new GFX(GFX_TEXTURE, 2);
 	if (theora_p)
 	{
 		if (videodata.ti.frame_width > texture_width || videodata.ti.frame_height > texture_height)
 		{
-			debug(LOG_ERROR, "Video size too large, must be below %.gx%.g!",
+			debug(LOG_ERROR, "Video size too large, must be below %zu x %zu!",
 			      texture_width, texture_height);
 			delete videoGfx;
 			videoGfx = nullptr;
@@ -698,20 +696,20 @@ bool seq_Play(const char *filename)
 		}
 		char *blackframe = (char *)calloc(1, texture_width * texture_height * 4);
 
-		// disable scanlines if the video is too large for the texture or shown too small
+		// disable scanlines temporarily if the video is too large for the texture or shown too small
 		if (videodata.ti.frame_height * 2 > texture_height || vertices[3][1] < videodata.ti.frame_height * 2)
 		{
-			use_scanlines = SCANLINES_OFF;
+			seq_setScanlinesDisabled(true);
 		}
 
 		Allocate_videoFrame();
-		videoGfx->makeTexture(texture_width, texture_height, GL_LINEAR, gfx_api::pixel_format::rgba, blackframe);
+		videoGfx->makeTexture(texture_width, texture_height, gfx_api::pixel_format::FORMAT_RGBA8_UNORM_PACK8, blackframe);
 		free(blackframe);
 
 		// when using scanlines we need to double the height
-		const int height_factor = (use_scanlines ? 2 : 1);
-		const gfx_api::gfxFloat vtwidth = (float)videodata.ti.frame_width / texture_width;
-		const gfx_api::gfxFloat vtheight = (float)videodata.ti.frame_height * height_factor / texture_height;
+		const uint32_t height_factor = ((!seq_getScanlinesDisabled() && seq_getScanlineMode()) ? 2 : 1);
+		const gfx_api::gfxFloat vtwidth = (float)videodata.ti.frame_width / (float)texture_width;
+		const gfx_api::gfxFloat vtheight = (float)videodata.ti.frame_height * height_factor / (float)texture_height;
 		gfx_api::gfxFloat texcoords[NUM_VERTICES * 2] = { 0.0f, 0.0f, vtwidth, 0.0f, 0.0f, vtheight, vtwidth, vtheight };
 		videoGfx->buffers(NUM_VERTICES, vertices, texcoords);
 	}
@@ -766,7 +764,7 @@ bool seq_Update()
 			{
 				for (j = 0; j < videodata.vi.channels; j++)
 				{
-					int val = nearbyint(pcm[j][i] * 32767.f);
+					int val = static_cast<int>(nearbyint(pcm[j][i] * 32767.f));
 
 					if (val > 32767)
 					{
@@ -955,12 +953,13 @@ void seq_Shutdown()
 	}
 
 	videoplaying = false;
+	seq_setScanlinesDisabled(false);
 	Timer_stop();
 
 	audioTime = 0;
 	sampletimeOffset = last_time = timer_expire = timer_started = 0;
 	basetime = -1;
-	pie_SetTexturePage(-1);
+
 	debug(LOG_VIDEO, " **** frames = %d dropped = %d ****", frames, dropped);
 }
 
@@ -992,7 +991,7 @@ void seq_SetDisplaySize(int sizeX, int sizeY, int posX, int posY)
 
 		if (aspect > videoAspect) // x offset
 		{
-			int offset = (screenWidth - screenHeight * videoAspect) / 2;
+			int offset = static_cast<int>((screenWidth - screenHeight * videoAspect) / 2);
 			vertices[1][0] += offset;
 			vertices[3][0] += offset;
 			vertices[0][0] -= offset;
@@ -1000,7 +999,7 @@ void seq_SetDisplaySize(int sizeX, int sizeY, int posX, int posY)
 		}
 		else // y offset
 		{
-			int offset = (screenHeight - screenWidth / videoAspect) / 2;
+			int offset = static_cast<int>((screenHeight - screenWidth / videoAspect) / 2);
 			vertices[0][1] += offset;
 			vertices[1][1] += offset;
 			vertices[2][1] -= offset;
@@ -1013,6 +1012,16 @@ void seq_SetDisplaySize(int sizeX, int sizeY, int posX, int posY)
 	Scrnvidpos[2] = 0.0f;
 }
 
+void seq_setScanlinesDisabled(bool flag)
+{
+	scanlinesDisabled = flag;
+}
+
+bool seq_getScanlinesDisabled()
+{
+	return scanlinesDisabled;
+}
+
 void seq_setScanlineMode(SCANLINE_MODE mode)
 {
 	use_scanlines = mode;
