@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -96,7 +96,7 @@ bool audio_Disabled(void)
 // =======================================================================================================================
 // =======================================================================================================================
 //
-bool audio_Init(AUDIO_CALLBACK pStopTrackCallback, bool really_enable)
+bool audio_Init(AUDIO_CALLBACK pStopTrackCallback, HRTFMode hrtf, bool really_enable)
 {
 	// init audio system
 	g_sPreviousSample.iTrack = NO_SAMPLE;
@@ -106,7 +106,7 @@ bool audio_Init(AUDIO_CALLBACK pStopTrackCallback, bool really_enable)
 	g_bAudioEnabled = really_enable;
 	if (g_bAudioEnabled)
 	{
-		g_bAudioEnabled = sound_Init();
+		g_bAudioEnabled = sound_Init(hrtf);
 	}
 	if (g_bAudioEnabled)
 	{
@@ -722,7 +722,7 @@ static bool audio_Play3DTrack(SDWORD iX, SDWORD iY, SDWORD iZ, int iTrack, SIMPL
 	distance = sqrtf(dX * dX + dY * dY + dZ * dZ); // Pythagorean theorem
 
 	// compute gain
-	gain = (1.0 - (distance * ATTENUATION_FACTOR)) ;//* 1.0f * sfx3d_volume
+	gain = (1.0f - (distance * ATTENUATION_FACTOR)) ;//* 1.0f * sfx3d_volume
 	if (gain > 1.0f)
 	{
 		gain = 1.0f;
@@ -878,7 +878,7 @@ AUDIO_STREAM *audio_PlayStream(const char *fileName, float volume, void (*onFini
 
 	// Open up the file
 	fileHandle = PHYSFS_openRead(fileName);
-	debug(LOG_WZ, "Reading...[directory: %s] %s", PHYSFS_getRealDir(fileName), fileName);
+	debug(LOG_WZ, "Reading...[directory: %s] %s", WZ_PHYSFS_getRealDir_String(fileName).c_str(), fileName);
 	if (fileHandle == nullptr)
 	{
 		debug(LOG_ERROR, "sound_LoadTrackFromFile: PHYSFS_openRead(\"%s\") failed with error: %s\n", fileName, WZ_PHYSFS_getLastError());
@@ -926,6 +926,17 @@ void audio_StopObjTrack(SIMPLE_OBJECT *psObj, int iTrack)
 		psSample = psSample->psNext;
 	}
 }
+static UDWORD lastTimeBuildFailedPlayed = 0;
+#define REPEAT_AUDIO_CUE_THRESHOLD 100
+// Play annoying sound only once per tick
+void audio_PlayBuildFailedOnce()
+{
+	if (realTime - lastTimeBuildFailedPlayed >= REPEAT_AUDIO_CUE_THRESHOLD)
+	{
+		audio_PlayTrack(ID_SOUND_BUILD_FAIL);
+		lastTimeBuildFailedPlayed = realTime;
+	}
+}
 
 //*
 //
