@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -27,18 +27,17 @@
 #include "lib/framework/frameresource.h"
 #include "lib/exceptionhandler/dumpinfo.h"
 
-#ifdef WZ_OS_MAC
-#include <OpenAL/al.h>
-#include <OpenAL/alc.h>
-#else
 #include <AL/al.h>
 #include <AL/alc.h>
+#if defined(HAVE_OPENAL_ALEXT_H)
+# include <AL/alext.h>
 #endif
 
 #include <physfs.h>
 #include "lib/framework/physfs_ext.h"
 #include <string.h>
 #include <math.h>
+#include <limits>
 
 #include "tracklib.h"
 #include "audio.h"
@@ -46,6 +45,7 @@
 #include "oggvorbis.h"
 #include "openal_error.h"
 #include "mixer.h"
+#include "openal_info.h"
 
 static ALuint current_queue_sample = -1;
 
@@ -53,19 +53,19 @@ static bool openal_initialized = false;
 
 struct AUDIO_STREAM
 {
-	ALuint                  source;        // OpenAL name of the sound source
-	struct OggVorbisDecoderState *decoder;
-	PHYSFS_file *fileHandle;
-	float                   volume;
+	ALuint                  source = -1;        // OpenAL name of the sound source
+	struct OggVorbisDecoderState *decoder = nullptr;
+	PHYSFS_file *fileHandle = nullptr;
+	float                   volume = 0.f;
 
 	// Callbacks
-	void (*onFinished)(const void *);
-	const void              *user_data;
+	std::function<void (const void *)> onFinished;
+	const void              *user_data = nullptr;
 
-	size_t                  bufferSize;
+	size_t                  bufferSize = 0;
 
 	// Linked list pointer
-	AUDIO_STREAM           *next;
+	AUDIO_STREAM           *next = nullptr;
 };
 
 struct SAMPLE_LIST
@@ -84,6 +84,11 @@ static ALfloat		sfx3d_volume = 1.0;
 static ALCdevice *device = nullptr;
 static ALCcontext *context = nullptr;
 
+#if defined(ALC_SOFT_HRTF)
+static LPALCGETSTRINGISOFT alcGetStringiSOFT = nullptr;
+static LPALCRESETDEVICESOFT alcResetDeviceSOFT = nullptr;
+#endif
+
 
 /** Removes the given sample from the "active_samples" linked list
  *  \param previous either NULL (if \c to_remove is the first item in the
@@ -110,11 +115,31 @@ static void sound_RemoveSample(SAMPLE_LIST *previous, SAMPLE_LIST *to_remove)
 	}
 }
 
+ALCint HRTFModeToALCint(HRTFMode mode)
+{
+#if defined(ALC_SOFT_HRTF)
+	switch (mode)
+	{
+	case HRTFMode::Unsupported:
+		// should never be called with unsupported, log it and fall through to disabled
+		debug(LOG_ERROR, "HRTFModeToALCint called with HRTFMode::Unsupported");
+		// fallthrough
+	case HRTFMode::Disabled:
+		return ALC_FALSE;
+	case HRTFMode::Enabled:
+		return ALC_TRUE;
+	case HRTFMode::Auto:
+		return ALC_DONT_CARE_SOFT;
+	}
+#endif
+	return ALC_FALSE;
+}
+
 //*
 // =======================================================================================================================
 // =======================================================================================================================
 //
-bool sound_InitLibrary(void)
+bool sound_InitLibrary(HRTFMode hrtf)
 {
 	int err;
 	const ALfloat listenerVel[3] = { 0.0, 0.0, 0.0 };
@@ -137,28 +162,44 @@ bool sound_InitLibrary(void)
 	}
 #endif
 
-#ifdef WZ_OS_WIN
-	/* HACK: Select the "software" OpenAL device on Windows because it
-	 *       provides 256 sound sources (unlike most Creative's default
-	 *       which provides only 16), causing our lack of source-management
-	 *       to be significantly less noticeable.
-	 */
-	device = alcOpenDevice("Generic Software");
-
-	// If the software device isn't available, fall back to default
-	if (!device)
-#endif
+	if (enabled_debug[LOG_SOUND])
 	{
-		// Open default device
-		device = alcOpenDevice(nullptr);
+		OpenALInfo::Output_PlaybackDevices([](const std::string &output) {
+			addDumpInfo(output.c_str());
+			if (enabled_debug[LOG_SOUND])
+			{
+				_debug_multiline(0, LOG_SOUND, "sound", output);
+			}
+		});
 	}
 
+	// Open default device
+	device = alcOpenDevice(nullptr);
+
 	if (!device)
 	{
 		debug(LOG_ERROR, "Couldn't open audio device.");
 		return false;
 	}
 
+	if (enabled_debug[LOG_SOUND])
+	{
+		OpenALInfo::Output_ALCInfo(device, [](const std::string &output) {
+			// addDumpInfo(output.c_str());
+			//if (enabled_debug[LOG_SOUND])
+			//{
+				_debug_multiline(0, LOG_SOUND, "sound", output);
+			//}
+		});
+	}
+
+#if defined(ALC_SOFT_HRTF)
+	// Load some extensions from OpenAL-Soft (if available)
+	alcGetStringiSOFT = (LPALCGETSTRINGISOFT)alcGetProcAddress(device, "alcGetStringiSOFT");
+	alcResetDeviceSOFT = (LPALCRESETDEVICESOFT)alcGetProcAddress(device, "alcResetDeviceSOFT");
+#endif
+
+
 	// Print current device name and add it to dump info
 	deviceName = alcGetString(device, ALC_DEVICE_SPECIFIER);
 	debug(LOG_SOUND, "Current audio device: %s", deviceName);
@@ -201,6 +242,63 @@ bool sound_InitLibrary(void)
 
 	openal_initialized = true;
 
+#if defined(ALC_SOFT_HRTF)
+	if(alcIsExtensionPresent(device, "ALC_SOFT_HRTF"))
+	{
+		// Set desired HRTF mode
+		sound_SetHRTFMode(hrtf);
+
+		// Get current HRTF status
+		ALCint hrtfStatus;
+		alcGetIntegerv(device, ALC_HRTF_STATUS_SOFT, 1, &hrtfStatus);
+
+		const char *hrtfStatusString = nullptr;
+		switch (hrtfStatus)
+		{
+		case ALC_HRTF_DISABLED_SOFT:
+			hrtfStatusString = "ALC_HRTF_DISABLED_SOFT: HRTF is disabled";
+			break;
+		case ALC_HRTF_ENABLED_SOFT:
+			hrtfStatusString = "ALC_HRTF_ENABLED_SOFT: HRTF is enabled";
+			break;
+		case ALC_HRTF_DENIED_SOFT:
+			// This may be caused by invalid resource permissions, or other user configuration that disallows HRTF.
+			hrtfStatusString = "ALC_HRTF_DENIED_SOFT: HRTF is disabled because it's not allowed on the device.";
+			break;
+		case ALC_HRTF_REQUIRED_SOFT:
+			// This may be caused by a device that can only use HRTF, or other user configuration that forces HRTF to be used.
+			hrtfStatusString = "ALC_HRTF_REQUIRED_SOFT: HRTF is enabled because it must be used on the device.";
+			break;
+		case ALC_HRTF_HEADPHONES_DETECTED_SOFT:
+			hrtfStatusString = "ALC_HRTF_HEADPHONES_DETECTED_SOFT: HRTF is enabled automatically because the device reported itself as headphones.";
+			break;
+		case ALC_HRTF_UNSUPPORTED_FORMAT_SOFT:
+			// HRTF is disabled because the device does not support it with the current format.
+			// Typically this is caused by non-stereo output or an incompatible output frequency.
+			hrtfStatusString = "ALC_HRTF_UNSUPPORTED_FORMAT_SOFT: HRTF is disabled because the device does not support it with the current format.";
+			break;
+		default:
+			hrtfStatusString = nullptr;
+			break;
+		}
+
+		if (hrtfStatusString)
+		{
+			debug(LOG_SOUND, "%s", hrtfStatusString);
+		}
+		else
+		{
+			debug(LOG_SOUND, "OpenAL-Soft returned an unknown ALC_HRTF_STATUS_SOFT result: %d", hrtfStatus);
+		}
+	}
+	else
+	{
+		debug(LOG_SOUND, "alcIsExtensionPresent(..., \"ALC_SOFT_HRTF\") returned false");
+	}
+#else
+	debug(LOG_SOUND, "ALC_SOFT_HRTF not defined");
+#endif
+
 	// Clear Error Codes
 	alGetError();
 	alcGetError(device);
@@ -214,6 +312,57 @@ bool sound_InitLibrary(void)
 	return true;
 }
 
+HRTFMode sound_GetHRTFMode()
+{
+#if defined(ALC_SOFT_HRTF)
+	if(alcIsExtensionPresent(device, "ALC_SOFT_HRTF"))
+	{
+		ALCint hrtfStatus;
+		alcGetIntegerv(device, ALC_HRTF_SOFT, 1, &hrtfStatus);
+		switch (hrtfStatus)
+		{
+		case ALC_TRUE:
+			return HRTFMode::Enabled;
+		case ALC_FALSE:
+			return HRTFMode::Disabled;
+		default:
+			debug(LOG_SOUND, "OpenAL-Soft returned an unexpected ALC_HRTF_SOFT result: %d", hrtfStatus);
+		}
+	}
+#endif
+	return HRTFMode::Unsupported;
+}
+
+bool sound_SetHRTFMode(HRTFMode mode)
+{
+#if defined(ALC_SOFT_HRTF)
+	if(alcIsExtensionPresent(device, "ALC_SOFT_HRTF"))
+	{
+		ALCint hrtfSetting = HRTFModeToALCint(mode);
+
+		ALCint attrs[] = {
+			ALC_HRTF_SOFT, hrtfSetting, /* configure HRTF */
+			0 /* end of list */
+		};
+		if (alcResetDeviceSOFT)
+		{
+			ASSERT(device, "device is null");
+			if (!alcResetDeviceSOFT(device, attrs))
+			{
+				debug(LOG_ERROR, "Failed to reset device: %s\n", alcGetString(device, alcGetError(device)));
+				return false;
+			}
+			return true;
+		}
+		else
+		{
+			debug(LOG_ERROR, "ALC_SOFT_HRTF extension is available, but alcResetDeviceSOFT is null");
+		}
+	}
+#endif
+	return false;
+}
+
 static void sound_UpdateStreams(void);
 
 void sound_ShutdownLibrary(void)
@@ -250,6 +399,7 @@ void sound_ShutdownLibrary(void)
 	{
 		debug(LOG_SOUND, "OpenAl could not close the audio device.");
 	}
+	device = nullptr;
 
 	while (aSample)
 	{
@@ -305,7 +455,6 @@ void sound_Update()
 {
 	SAMPLE_LIST *node = active_samples;
 	SAMPLE_LIST *previous = nullptr;
-	ALCenum err;
 	ALfloat gain;
 
 	if (!openal_initialized)
@@ -376,7 +525,7 @@ void sound_Update()
 
 	alcProcessContext(context);
 
-	err = sound_GetContextError(device);
+	ALCenum err = sound_GetContextError(device);
 	if (err != ALC_NO_ERROR)
 	{
 		debug(LOG_ERROR, "Error while processing audio context: %s", alGetString(err));
@@ -494,7 +643,8 @@ static inline TRACK *sound_DecodeOggVorbisTrack(TRACK *psTrack, PHYSFS_file *PHY
 	// Create an OpenAL buffer and fill it with the decoded data
 	alGenBuffers(1, &buffer);
 	sound_GetError();
-	alBufferData(buffer, format, soundBuffer->data, soundBuffer->size, soundBuffer->frequency);
+	ASSERT(soundBuffer->size <= static_cast<size_t>(std::numeric_limits<ALsizei>::max()), "soundBuffer->size (%zu) exceeds ALsizei::max", soundBuffer->size);
+	alBufferData(buffer, format, soundBuffer->data, static_cast<ALsizei>(soundBuffer->size), soundBuffer->frequency);
 	sound_GetError();
 
 	free(soundBuffer);
@@ -518,7 +668,7 @@ TRACK *sound_LoadTrackFromFile(const char *fileName)
 
 	// Use PhysicsFS to open the file
 	fileHandle = PHYSFS_openRead(fileName);
-	debug(LOG_NEVER, "Reading...[directory: %s] %s", PHYSFS_getRealDir(fileName), fileName);
+	debug(LOG_NEVER, "Reading...[directory: %s] %s", WZ_PHYSFS_getRealDir_String(fileName).c_str(), fileName);
 	if (fileHandle == nullptr)
 	{
 		debug(LOG_ERROR, "sound_LoadTrackFromFile: PHYSFS_openRead(\"%s\") failed with error: %s\n", fileName, WZ_PHYSFS_getLastError());
@@ -730,9 +880,13 @@ bool sound_Play3DSample(TRACK *psTrack, AUDIO_SAMPLE *psSample)
 		 */
 	}
 
+#if defined(WZ_OS_UNIX) && !defined(WZ_OS_MAC)
 	// HACK: this is a workaround for a bug in the 64bit implementation of OpenAL on GNU/Linux
 	// The AL_PITCH value really should be 1.0.
 	alSourcef(psSample->iSample, AL_PITCH, 1.001f);
+#else
+	alSourcef(psSample->iSample, AL_PITCH, 1.0f);
+#endif
 
 	sound_SetObjectPosition(psSample);
 	alSourcefv(psSample->iSample, AL_VELOCITY, zero);
@@ -788,10 +942,11 @@ AUDIO_STREAM *sound_PlayStream(PHYSFS_file *fileHandle, float volume, void (*onF
  *  \see sound_PlayStream() for details about the rest of the function
  *       parameters and other details.
  */
-AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, void (*onFinished)(const void *), const void *user_data, size_t streamBufferSize, unsigned int buffer_count)
+AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, const std::function<void (const void *)>& onFinished, const void *user_data, size_t streamBufferSize, unsigned int buffer_count, bool allowSeeking)
 {
 	AUDIO_STREAM *stream;
-	ALuint       *buffers = (ALuint *)alloca(sizeof(ALuint) * buffer_count);
+	ALuint       *buffers = nullptr;
+	bool freeBuffers = false;
 	ALint error;
 	unsigned int i;
 
@@ -801,7 +956,7 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 		return nullptr;
 	}
 
-	stream = (AUDIO_STREAM *)malloc(sizeof(AUDIO_STREAM));
+	stream = new AUDIO_STREAM();
 	if (stream == nullptr)
 	{
 		debug(LOG_FATAL, "sound_PlayStream: Out of memory");
@@ -820,17 +975,17 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 	{
 		// Failed to create OpenAL sound source, so bail out...
 		debug(LOG_SOUND, "alGenSources failed, most likely out of sound sources");
-		free(stream);
+		delete stream;
 		return nullptr;
 	}
 
 	stream->fileHandle = fileHandle;
 
-	stream->decoder = sound_CreateOggVorbisDecoder(stream->fileHandle, false);
+	stream->decoder = sound_CreateOggVorbisDecoder(stream->fileHandle, allowSeeking);
 	if (stream->decoder == nullptr)
 	{
 		debug(LOG_ERROR, "sound_PlayStream: Failed to open audio file for decoding");
-		free(stream);
+		delete stream;
 		return nullptr;
 	}
 
@@ -839,11 +994,25 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 
 	alSourcef(stream->source, AL_GAIN, stream->volume);
 
+#if defined(WZ_OS_UNIX) && !defined(WZ_OS_MAC)
 	// HACK: this is a workaround for a bug in the 64bit implementation of OpenAL on GNU/Linux
 	// The AL_PITCH value really should be 1.0.
 	alSourcef(stream->source, AL_PITCH, 1.001f);
+#else
+	alSourcef(stream->source, AL_PITCH, 1.0f);
+#endif
 
 	// Create some OpenAL buffers to store the decoded data in
+	if (buffer_count <= (1024 / sizeof(ALuint))) // See CMakeLists.txt for value of -Walloca-larger-than=<N>
+	{
+		buffers = (ALuint *)alloca(buffer_count * sizeof(ALuint));
+	}
+	else
+	{
+		// Too many buffers - don't allocate on the stack!
+		buffers = (ALuint *)malloc(buffer_count * sizeof(ALuint));
+		freeBuffers = true;
+	}
 	alGenBuffers(buffer_count, buffers);
 	sound_GetError();
 
@@ -860,7 +1029,8 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 			ALenum format = (soundBuffer->channelCount == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
 
 			// Copy the audio data into one of OpenAL's own buffers
-			alBufferData(buffers[i], format, soundBuffer->data, soundBuffer->size, soundBuffer->frequency);
+			ASSERT(soundBuffer->size <= static_cast<size_t>(std::numeric_limits<ALsizei>::max()), "soundBuffer->size (%zu) exceeds ALsizei::max", soundBuffer->size);
+			alBufferData(buffers[i], format, soundBuffer->data, static_cast<ALint>(soundBuffer->size), soundBuffer->frequency);
 			sound_GetError();
 
 			// Clean up our memory
@@ -894,7 +1064,13 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 		alDeleteSources(1, &stream->source);
 
 		// Free allocated memory
-		free(stream);
+		delete stream;
+
+		if(freeBuffers)
+		{
+			free(buffers);
+		}
+		buffers = nullptr;
 
 		return nullptr;
 	}
@@ -917,6 +1093,12 @@ AUDIO_STREAM *sound_PlayStreamWithBuf(PHYSFS_file *fileHandle, float volume, voi
 	stream->next = active_streams;
 	active_streams = stream;
 
+	if(freeBuffers)
+	{
+		free(buffers);
+	}
+	buffers = nullptr;
+
 	return stream;
 }
 
@@ -1032,6 +1214,11 @@ void sound_SetStreamVolume(AUDIO_STREAM *stream, float volume)
 	sound_GetError();
 }
 
+double sound_GetStreamTotalTime(AUDIO_STREAM *stream)
+{
+	return sound_GetOggVorbisTotalTime(stream->decoder);
+}
+
 /** Update the given stream by making sure its buffers remain full
  *  \param stream the stream to update
  *  \return true when the stream is still playing, false when it has stopped
@@ -1072,7 +1259,8 @@ static bool sound_UpdateStream(AUDIO_STREAM *stream)
 			ALenum format = (soundBuffer->channelCount == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
 
 			// Insert the data into the buffer
-			alBufferData(buffer, format, soundBuffer->data, soundBuffer->size, soundBuffer->frequency);
+			ASSERT(soundBuffer->size <= static_cast<size_t>(std::numeric_limits<ALsizei>::max()), "soundBuffer->size (%zu) exceeds ALsizei::max", soundBuffer->size);
+			alBufferData(buffer, format, soundBuffer->data, static_cast<ALsizei>(soundBuffer->size), soundBuffer->frequency);
 			sound_GetError();
 
 			// Reattach the buffer to the source
@@ -1134,16 +1322,24 @@ static void sound_DestroyStream(AUDIO_STREAM *stream)
 	// Detach all buffers and retrieve their ID numbers
 	if (buffer_count > 0)
 	{
-		if (buffer_count <= (1024 / sizeof(ALuint))) // See CMakeLists.txt for value of -Walloca-larger-than=<N>
+#if !defined(__clang__) && defined(__GNUC__) && __GNUC__ >= 7
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Walloca-larger-than="
+#endif
+		const size_t buffer_count_sizet = static_cast<size_t>(buffer_count);
+		if (buffer_count_sizet <= (1024 / sizeof(ALuint))) // See CMakeLists.txt for value of -Walloca-larger-than=<N>
 		{
-			buffers = (ALuint *)alloca(buffer_count * sizeof(ALuint));
+			buffers = (ALuint *)alloca(buffer_count_sizet * sizeof(ALuint));
 		}
 		else
 		{
 			// Too many buffers - don't allocate on the stack!
-			buffers = (ALuint *)malloc(buffer_count * sizeof(ALuint));
+			buffers = (ALuint *)malloc(buffer_count_sizet * sizeof(ALuint));
 			freeBuffers = true;
 		}
+#if !defined(__clang__) && defined(__GNUC__) && __GNUC__ >= 7
+#pragma GCC diagnostic pop
+#endif
 		alSourceUnqueueBuffers(stream->source, buffer_count, buffers);
 		sound_GetError();
 
@@ -1179,7 +1375,7 @@ static void sound_DestroyStream(AUDIO_STREAM *stream)
 	}
 
 	// Free the memory used by this stream
-	free(stream);
+	delete stream;
 }
 
 /** Update all currently running streams and destroy them when they're finished.
