@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -27,7 +27,6 @@
 #include "lib/framework/frame.h"
 #include "lib/framework/math_ext.h"
 #include "lib/framework/fixedpoint.h"
-#include "lib/script/script.h"
 #include "lib/sound/audio.h"
 #include "lib/sound/audio_id.h"
 #include "lib/netplay/netplay.h"
@@ -39,7 +38,6 @@
 #include "projectile.h"
 #include "qtscript.h"
 #include "random.h"
-#include "scriptcb.h"
 #include "transporter.h"
 #include "mapgrid.h"
 #include "hci.h"
@@ -414,6 +412,7 @@ bool actionVisibleTarget(DROID *psDroid, BASE_OBJECT *psTarget, int weapon_slot)
 {
 	CHECK_DROID(psDroid);
 	ASSERT_OR_RETURN(false, psTarget != nullptr, "Target is NULL");
+	ASSERT_OR_RETURN(false, psDroid->player < MAX_PLAYERS, "psDroid->player (%" PRIu8 ") must be < MAX_PLAYERS", psDroid->player);
 	if (!psTarget->visible[psDroid->player])
 	{
 		return false;
@@ -529,6 +528,17 @@ static void actionCalcPullBackPoint(BASE_OBJECT *psObj, BASE_OBJECT *psTarget, i
 	clip_world_offmap(px, py);
 }
 
+// check whether a droid is in the neighboring tile of another droid
+bool actionReachedDroid(DROID const *psDroid, DROID const *psOther)
+{
+	ASSERT_OR_RETURN(false, psDroid != nullptr && psOther != nullptr, "Bad droids");
+	CHECK_DROID(psDroid);
+	Vector2i xy = map_coord(psDroid->pos.xy());
+	Vector2i otherxy = map_coord(psOther->pos.xy());
+	Vector2i delta = xy - otherxy;
+	return delta.x >=-1 && delta.x <=1 &&  delta.y >=-1 && delta.y <=1 ;
+}
+
 
 // check whether a droid is in the neighboring tile to a build position
 bool actionReachedBuildPos(DROID const *psDroid, int x, int y, uint16_t dir, BASE_STATS const *psStats)
@@ -655,6 +665,7 @@ void actionUpdateDroid(DROID *psDroid)
 	bool (*actionUpdateFunc)(DROID * psDroid) = nullptr;
 	bool nonNullWeapon[MAX_WEAPONS] = { false };
 	BASE_OBJECT *psTargets[MAX_WEAPONS] = { nullptr };
+	bool hasValidWeapon = false;
 	bool hasVisibleTarget = false;
 	bool targetVisibile[MAX_WEAPONS] = { false };
 	bool bHasTarget = false;
@@ -729,7 +740,7 @@ void actionUpdateDroid(DROID *psDroid)
 						if (secondaryGetState(psDroid, DSO_ATTACK_LEVEL) == DSS_ALEV_ALWAYS)
 						{
 							psDroid->action = DACTION_ATTACK;
-							setDroidActionTarget(psDroid, psTemp, 0);
+							setDroidActionTarget(psDroid, psTemp, i);
 						}
 					}
 				}
@@ -743,20 +754,32 @@ void actionUpdateDroid(DROID *psDroid)
 			psDroid->action = DACTION_NONE;
 			break;
 		}
-		// move back to the repair facility if necessary
-		if (DROID_STOPPED(psDroid) &&
-		    !actionReachedBuildPos(psDroid,
-		                           order->psObj->pos.x, order->psObj->pos.y, ((STRUCTURE *)order->psObj)->rot.direction,
-		                           ((STRUCTURE *)order->psObj)->pStructureType))
+		if (order->type == DORDER_RTR && order->rtrType == RTR_TYPE_REPAIR_FACILITY)
 		{
-			moveDroidToNoFormation(psDroid, order->psObj->pos.x, order->psObj->pos.y);
+			// move back to the repair facility if necessary
+			if (DROID_STOPPED(psDroid) &&
+				!actionReachedBuildPos(psDroid,
+									order->psObj->pos.x, order->psObj->pos.y, ((STRUCTURE *)order->psObj)->rot.direction,
+									((STRUCTURE *)order->psObj)->pStructureType))
+			{
+				moveDroidToNoFormation(psDroid, order->psObj->pos.x, order->psObj->pos.y);
+			}
+		}
+		else if (order->type == DORDER_RTR && order->rtrType == RTR_TYPE_DROID && DROID_STOPPED(psDroid)) {
+			if (!actionReachedDroid(psDroid, static_cast<DROID *> (order->psObj)))
+			{
+				moveDroidToNoFormation(psDroid, order->psObj->pos.x, order->psObj->pos.y);
+			} else {
+				moveStopDroid(psDroid);
+			}
 		}
 		break;
 	case DACTION_TRANSPORTWAITTOFLYIN:
 		//if we're moving droids to safety and currently waiting to fly back in, see if time is up
 		if (psDroid->player == selectedPlayer && getDroidsToSafetyFlag())
 		{
-			if ((SDWORD)(mission.ETA - (gameTime - missionGetReinforcementTime())) <= 0)
+			bool enoughTimeRemaining = (mission.time - (gameTime - mission.startTime)) >= (60 * GAME_TICKS_PER_SEC);
+			if (((SDWORD)(mission.ETA - (gameTime - missionGetReinforcementTime())) <= 0) && enoughTimeRemaining)
 			{
 				UDWORD droidX, droidY;
 
@@ -791,11 +814,6 @@ void actionUpdateDroid(DROID *psDroid)
 				/* notify scripts we have reached the destination
 				*  also triggers when patrolling and reached a waypoint
 				*/
-				psScrCBOrder = order->type;
-				psScrCBOrderDroid = psDroid;
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_DROID_REACH_LOCATION);
-				psScrCBOrderDroid = nullptr;
-				psScrCBOrder = DORDER_NONE;
 
 				triggerEventDroidIdle(psDroid);
 			}
@@ -960,6 +978,10 @@ void actionUpdateDroid(DROID *psDroid)
 		break;
 	case DACTION_ATTACK:
 	case DACTION_ROTATETOATTACK:
+		if (psDroid->psActionTarget[0] == nullptr &&  psDroid->psActionTarget[1] != nullptr)
+		{
+			break;
+		}
 		ASSERT_OR_RETURN(, psDroid->psActionTarget[0] != nullptr, "target is NULL while attacking");
 
 		if (psDroid->action == DACTION_ROTATETOATTACK)
@@ -1139,10 +1161,15 @@ void actionUpdateDroid(DROID *psDroid)
 					psDroid->action = DACTION_NONE;
 				}
 			}
-			else
+			else if (secondaryGetState(psDroid, DSO_HALTTYPE) != DSS_HALT_HOLD)
 			{
 				psDroid->action = DACTION_MOVETOATTACK;	// out of range - chase it
 			}
+			else
+			{
+				psDroid->order.psObj = nullptr;
+				psDroid->action = DACTION_NONE;
+			}
 		}
 
 		break;
@@ -1150,17 +1177,18 @@ void actionUpdateDroid(DROID *psDroid)
 	case DACTION_VTOLATTACK:
 		{
 			WEAPON_STATS *psWeapStats = nullptr;
-
+			const bool targetIsValid = validTarget(psDroid, psDroid->psActionTarget[0], 0);
 			//uses vtResult
 			if (psDroid->psActionTarget[0] != nullptr &&
-			    validTarget(psDroid, psDroid->psActionTarget[0], 0))
+			    targetIsValid)
 			{
 				//check if vtol that its armed
 				if ((vtolEmpty(psDroid)) ||
 				    (psDroid->psActionTarget[0] == nullptr) ||
 				    //check the target hasn't become one the same player ID - Electronic Warfare
 				    (electronicDroid(psDroid) && (psDroid->player == psDroid->psActionTarget[0]->player)) ||
-				    !validTarget(psDroid, psDroid->psActionTarget[0], 0))
+					// Huh? !targetIsValid can't be true, we just checked for it
+				    !targetIsValid)
 				{
 					moveToRearm(psDroid);
 					break;
@@ -1205,15 +1233,16 @@ void actionUpdateDroid(DROID *psDroid)
 			{
 				actionAddVtolAttackRun(psDroid);
 			}
-			else
+			else if(psDroid->psActionTarget[0] != nullptr &&
+			    targetIsValid)
 			{
 				// if the vtol is close to the target, go around again
-				const Vector2i diff = (psDroid->pos - psDroid->psActionTarget[0]->pos).xy();
+				Vector2i diff = (psDroid->pos - psDroid->psActionTarget[0]->pos).xy();
 				const unsigned rangeSq = dot(diff, diff);
 				if (rangeSq < VTOL_ATTACK_TARDIST * VTOL_ATTACK_TARDIST)
 				{
 					// don't do another attack run if already moving away from the target
-					const Vector2i diff = psDroid->sMove.destination - psDroid->psActionTarget[0]->pos.xy();
+					diff = psDroid->sMove.destination - psDroid->psActionTarget[0]->pos.xy();
 					if (dot(diff, diff) < VTOL_ATTACK_TARDIST * VTOL_ATTACK_TARDIST)
 					{
 						actionAddVtolAttackRun(psDroid);
@@ -1227,7 +1256,7 @@ void actionUpdateDroid(DROID *psDroid)
 					if (rangeSq > maxRange * maxRange)
 					{
 						// don't do another attack run if already heading for the target
-						const Vector2i diff = psDroid->sMove.destination - psDroid->psActionTarget[0]->pos.xy();
+						diff = psDroid->sMove.destination - psDroid->psActionTarget[0]->pos.xy();
 						if (dot(diff, diff) > VTOL_ATTACK_TARDIST * VTOL_ATTACK_TARDIST)
 						{
 							moveDroidToDirect(psDroid, psDroid->psActionTarget[0]->pos.x, psDroid->psActionTarget[0]->pos.y);
@@ -1246,10 +1275,12 @@ void actionUpdateDroid(DROID *psDroid)
 		}
 
 		ASSERT_OR_RETURN(, psDroid->psActionTarget[0] != nullptr, "action update move to attack target is NULL");
-
+		for (unsigned i = 0; i < psDroid->numWeaps; ++i)
+		{
+			hasValidWeapon |= validTarget(psDroid, psDroid->psActionTarget[0], i);
+		}
 		//check the target hasn't become one the same player ID - Electronic Warfare, and that the target is still valid.
-		if ((electronicDroid(psDroid) && psDroid->player == psDroid->psActionTarget[0]->player) ||
-		    !validTarget(psDroid, psDroid->psActionTarget[0], 0))
+		if ((electronicDroid(psDroid) && psDroid->player == psDroid->psActionTarget[0]->player) || !hasValidWeapon)
 		{
 			for (unsigned i = 0; i < psDroid->numWeaps; ++i)
 			{
@@ -1403,9 +1434,69 @@ void actionUpdateDroid(DROID *psDroid)
 			psDroid->action = DACTION_NONE;
 			break;
 		}
-		// moving to a location to build a structure
+		else
+		{
+			// Determine if the droid can still build or help to build the ordered structure at the specified location
+			const STRUCTURE_STATS* const desiredStructure = order->psStats;
+			const STRUCTURE* const structureAtBuildPosition = getTileStructure(map_coord(psDroid->actionPos.x), map_coord(psDroid->actionPos.y));
+
+			if (nullptr != structureAtBuildPosition)
+			{
+				bool droidCannotBuild = false;
+
+				if (!aiCheckAlliances(structureAtBuildPosition->player, psDroid->player))
+				{
+					// Not our structure
+					droidCannotBuild = true;
+				}
+				else
+				// There's an allied structure already there.  Is it a wall, and can the droid upgrade it to a defence or gate?
+				if (isWall(structureAtBuildPosition->pStructureType->type) &&
+					(desiredStructure->type == REF_DEFENSE || desiredStructure->type == REF_GATE))
+				{
+					// It's always valid to upgrade a wall to a defence or gate
+					droidCannotBuild = false; // Just to avoid an empty branch
+				}
+				else
+				if ((structureAtBuildPosition->pStructureType != desiredStructure) && // ... it's not the exact same type as the droid was ordered to build
+					(structureAtBuildPosition->pStructureType->type == REF_WALLCORNER && desiredStructure->type != REF_WALL)) // and not a wall corner when the droid wants to build a wall
+				{
+					// And so the droid can't build or help with building this structure
+					droidCannotBuild = true;
+				}
+				else
+				// So it's a structure that the droid could help to build, but is it already complete?
+				if (structureAtBuildPosition->status == SS_BUILT &&
+					(!IsStatExpansionModule(desiredStructure) || !canStructureHaveAModuleAdded(structureAtBuildPosition)))
+				{
+					// The building is complete and the droid hasn't been told to add a module, or can't add one, so can't help with that.
+					droidCannotBuild = true;
+				}
+
+				if (droidCannotBuild)
+				{
+					if (order->type == DORDER_LINEBUILD && map_coord(psDroid->order.pos) != map_coord(psDroid->order.pos2))
+					{
+						// The droid is doing a line build, and there's more to build. This will force the droid to move to the next structure in the line build
+						objTrace(psDroid->id, "DACTION_MOVETOBUILD: line target is already built, or can't be built - moving to next structure in line");
+						psDroid->action = DACTION_NONE;
+					}
+					else
+					{
+						// Cancel the current build order. This will move the truck onto the next order, if it has one, or halt in place.
+						objTrace(psDroid->id, "DACTION_MOVETOBUILD: target is already built, or can't be built - executing next order or halting");
+						cancelBuild(psDroid);
+					}
+
+					break;
+				}
+			}
+		} // End of check for whether the droid can still succesfully build the ordered structure
+
+		// The droid can still build or help with a build, and is moving to a location to do so - are we there yet, are we there yet, are we there yet?
 		if (actionReachedBuildPos(psDroid, psDroid->actionPos.x, psDroid->actionPos.y, order->direction, order->psStats))
 		{
+			// We're there, go ahead and build or help to build the structure
 			bool buildPosEmpty = actionRemoveDroidsFromBuildPos(psDroid->player, psDroid->actionPos, order->direction, order->psStats);
 			if (!buildPosEmpty)
 			{
@@ -1476,9 +1567,7 @@ void actionUpdateDroid(DROID *psDroid)
 					psDroid->actionPoints = 0;
 				}
 			}
-			else if ((order->type == DORDER_LINEBUILD || order->type == DORDER_BUILD)
-			         && (psStructStats->type == REF_WALL || psStructStats->type == REF_WALLCORNER || psStructStats->type == REF_GATE ||
-			             psStructStats->type == REF_DEFENSE || psStructStats->type == REF_REARM_PAD))
+			else if (order->type == DORDER_LINEBUILD || order->type == DORDER_BUILD)
 			{
 				// building a wall.
 				MAPTILE *const psTile = worldTile(psDroid->actionPos);
@@ -1508,7 +1597,19 @@ void actionUpdateDroid(DROID *psDroid)
 							// building a gun tower over a wall - OK
 							if (droidStartBuild(psDroid))
 							{
-								objTrace(psDroid->id, "DACTION_MOVETOBUILD: start building");
+								objTrace(psDroid->id, "DACTION_MOVETOBUILD: start building defense");
+								psDroid->action = DACTION_BUILD;
+							}
+						}
+						else if ((psStruct->pStructureType->type == REF_FACTORY && order->psStats->type == REF_FACTORY_MODULE) ||
+								(psStruct->pStructureType->type == REF_RESEARCH && order->psStats->type == REF_RESEARCH_MODULE) ||
+								(psStruct->pStructureType->type == REF_POWER_GEN && order->psStats->type == REF_POWER_MODULE) ||
+								(psStruct->pStructureType->type == REF_VTOL_FACTORY && order->psStats->type == REF_FACTORY_MODULE))
+							{
+							// upgrade current structure in a row
+							if (droidStartBuild(psDroid))
+							{
+								objTrace(psDroid->id, "DACTION_MOVETOBUILD: start building module");
 								psDroid->action = DACTION_BUILD;
 							}
 						}
@@ -1521,6 +1622,19 @@ void actionUpdateDroid(DROID *psDroid)
 # pragma GCC diagnostic pop
 #endif
 					}
+					else if (TileHasFeature(psTile))
+					{
+						FEATURE *feature = getTileFeature(map_coord(psDroid->actionPos.x), map_coord(psDroid->actionPos.y));
+						objTrace(psDroid->id, "DACTION_MOVETOBUILD: tile has feature %d", feature->psStats->subType);
+						if (feature->psStats->subType == FEAT_OIL_RESOURCE && order->psStats->type == REF_RESOURCE_EXTRACTOR)
+						{
+							if (droidStartBuild(psDroid))
+							{
+								objTrace(psDroid->id, "DACTION_MOVETOBUILD: start building oil derrick");
+								psDroid->action = DACTION_BUILD;
+							}
+						}
+					}
 					else
 					{
 						objTrace(psDroid->id, "DACTION_MOVETOBUILD: blocked line build");
@@ -1530,7 +1644,6 @@ void actionUpdateDroid(DROID *psDroid)
 				else if (droidStartBuild(psDroid))
 				{
 					psDroid->action = DACTION_BUILD;
-					intBuildStarted(psDroid);
 				}
 			}
 			else
@@ -1547,7 +1660,6 @@ void actionUpdateDroid(DROID *psDroid)
 				{
 					objTrace(psDroid->id, "DACTION_MOVETOBUILD: starting help build");
 					psDroid->action = DACTION_BUILD;
-					intBuildStarted(psDroid);
 				}
 			}
 		}
@@ -1597,6 +1709,41 @@ void actionUpdateDroid(DROID *psDroid)
 			psDroid->action = DACTION_NONE;
 			break;
 		}
+		else
+		{
+			const STRUCTURE* structureAtPos = getTileStructure(map_coord(psDroid->actionPos.x), map_coord(psDroid->actionPos.y));
+
+			if (structureAtPos == nullptr)
+			{
+				//No structure located at desired position. Move on.
+				psDroid->action = DACTION_NONE;
+				break;
+			}
+			else if (order->type != DORDER_RESTORE)
+			{
+				bool cantDoRepairLikeAction = false;
+
+				if (!aiCheckAlliances(structureAtPos->player, psDroid->player))
+				{
+					cantDoRepairLikeAction = true;
+				}
+				else if (order->type != DORDER_DEMOLISH && structureAtPos->body == structureBody(structureAtPos))
+				{
+					cantDoRepairLikeAction = true;
+				}
+				else if (order->type == DORDER_DEMOLISH && structureAtPos->player != psDroid->player)
+				{
+					cantDoRepairLikeAction = true;
+				}
+
+				if (cantDoRepairLikeAction)
+				{
+					psDroid->action = DACTION_NONE;
+					moveStopDroid(psDroid);
+					break;
+				}
+			}
+		}
 		// see if the droid is at the edge of what it is moving to
 		if (actionReachedBuildPos(psDroid, psDroid->actionPos.x, psDroid->actionPos.y, ((STRUCTURE *)psDroid->psActionTarget[0])->rot.direction, order->psStats))
 		{
@@ -1692,17 +1839,47 @@ void actionUpdateDroid(DROID *psDroid)
 		}
 		break;
 	case DACTION_MOVETOREPAIRPOINT:
-		/* moving from front to rear of repair facility or rearm pad */
-		if (actionReachedBuildPos(psDroid, psDroid->psActionTarget[0]->pos.x, psDroid->psActionTarget[0]->pos.y, ((STRUCTURE *)psDroid->psActionTarget[0])->rot.direction, ((STRUCTURE *)psDroid->psActionTarget[0])->pStructureType))
+		if (psDroid->order.rtrType == RTR_TYPE_REPAIR_FACILITY)
 		{
-			objTrace(psDroid->id, "Arrived at repair point - waiting for our turn");
-			moveStopDroid(psDroid);
-			psDroid->action = DACTION_WAITDURINGREPAIR;
-		}
-		else if (DROID_STOPPED(psDroid))
+			/* moving from front to rear of repair facility or rearm pad */
+			if (actionReachedBuildPos(psDroid, psDroid->psActionTarget[0]->pos.x, psDroid->psActionTarget[0]->pos.y, ((STRUCTURE *)psDroid->psActionTarget[0])->rot.direction, ((STRUCTURE *)psDroid->psActionTarget[0])->pStructureType))
+			{
+				objTrace(psDroid->id, "Arrived at repair point - waiting for our turn");
+				moveStopDroid(psDroid);
+				psDroid->action = DACTION_WAITDURINGREPAIR;
+			}
+			else if (DROID_STOPPED(psDroid))
+			{
+				moveDroidToNoFormation(psDroid, psDroid->psActionTarget[0]->pos.x,
+									psDroid->psActionTarget[0]->pos.y);
+			}
+		} else if (psDroid->order.rtrType == RTR_TYPE_DROID)
 		{
-			moveDroidToNoFormation(psDroid, psDroid->psActionTarget[0]->pos.x,
-			                       psDroid->psActionTarget[0]->pos.y);
+			bool reached = actionReachedDroid(psDroid, (DROID*) psDroid->order.psObj);
+			if (reached)
+			{
+				if (psDroid->body >= psDroid->originalBody)
+				{
+					objTrace(psDroid->id, "Repair not needed of droid %d", (int)psDroid->id);
+					/* set droid points to max */
+					psDroid->body = psDroid->originalBody;
+					// if completely repaired then reset order
+					secondarySetState(psDroid, DSO_RETURN_TO_LOC, DSS_NONE);
+					orderDroidObj(psDroid, DORDER_GUARD, psDroid->order.psObj, ModeImmediate);
+				}
+				else
+				{
+					objTrace(psDroid->id, "Stopping and waiting for repairs %d", (int)psDroid->id);
+					moveStopDroid(psDroid);
+					psDroid->action = DACTION_WAITDURINGREPAIR;
+				}
+
+			}
+			else if (DROID_STOPPED(psDroid))
+			{
+				//objTrace(psDroid->id, "Droid was stopped, but havent reach the target, moving now");
+				//moveDroidToNoFormation(psDroid, psDroid->order.psObj->pos.x, psDroid->order.psObj->pos.y);
+			}
 		}
 		break;
 	case DACTION_OBSERVE:
@@ -1803,6 +1980,16 @@ void actionUpdateDroid(DROID *psDroid)
 		break;
 	case DACTION_MOVETODROIDREPAIR:
 		{
+			BASE_OBJECT* actionTargetObj = psDroid->psActionTarget[0];
+			ASSERT_OR_RETURN(, actionTargetObj != nullptr && actionTargetObj->type == OBJ_DROID, "unexpected repair target");
+			const DROID* actionTarget = (const DROID *) actionTargetObj;
+			if (actionTarget->body == actionTarget->originalBody)
+			{
+				// target is healthy: nothing to do
+				psDroid->action = DACTION_NONE;
+				moveStopDroid(psDroid);
+				break;
+			}
 			Vector2i diff = (psDroid->pos - psDroid->psActionTarget[0]->pos).xy();
 			// moving to repair a droid
 			if (!psDroid->psActionTarget[0] ||  // Target missing.
@@ -1887,6 +2074,7 @@ void actionUpdateDroid(DROID *psDroid)
 				if (!droidUpdateDroidRepair(psDroid))
 				{
 					psDroid->action = DACTION_NONE;
+					moveStopDroid(psDroid);
 					//if the order is RTR then resubmit order so that the unit will go to repair facility point
 					if (orderState(psDroid, DORDER_RTR))
 					{
@@ -2026,7 +2214,11 @@ static void actionDroidBase(DROID *psDroid, DROID_ACTION_DATA *psAction)
 	objTrace(psDroid->id, "base set action to %s (was %s)", getDroidActionName(psAction->action), getDroidActionName(psDroid->action));
 
 	DROID_ORDER_DATA *order = &psDroid->order;
-
+	bool hasValidWeapon = false;
+	for (int i = 0; i < MAX_WEAPONS; i++)
+	{
+		hasValidWeapon |= validTarget(psDroid, psAction->psObj, i);
+	}
 	switch (psAction->action)
 	{
 	case DACTION_NONE:
@@ -2061,7 +2253,12 @@ static void actionDroidBase(DROID *psDroid, DROID_ACTION_DATA *psAction)
 		{
 			break;
 		}
-
+		if (!hasValidWeapon)
+		{
+			// continuing is pointless, we were given an invalid target
+			// for ex. AA gun can't attack ground unit
+			break;
+		}
 		if (electronicDroid(psDroid))
 		{
 			//check for low or zero resistance - just zero resistance!
@@ -2113,11 +2310,10 @@ static void actionDroidBase(DROID *psDroid, DROID_ACTION_DATA *psAction)
 					moveTurnDroid(psDroid, psDroid->psActionTarget[0]->pos.x, psDroid->psActionTarget[0]->pos.y);
 				}
 			}
-			else if (order->type != DORDER_HOLD)
+			else if (order->type != DORDER_HOLD && secondaryGetState(psDroid, DSO_HALTTYPE) != DSS_HALT_HOLD)
 			{
 				int pbx = 0;
 				int pby = 0;
-
 				/* direct fire - try and extend the range */
 				psDroid->action = DACTION_MOVETOATTACK;
 				actionCalcPullBackPoint(psDroid, psAction->psObj, &pbx, &pby);
@@ -2127,13 +2323,17 @@ static void actionDroidBase(DROID *psDroid, DROID_ACTION_DATA *psAction)
 				turnOffMultiMsg(false);
 			}
 		}
-		else if (order->type != DORDER_HOLD) // approach closer?
+		else if (order->type != DORDER_HOLD && secondaryGetState(psDroid, DSO_HALTTYPE) != DSS_HALT_HOLD) // approach closer?
 		{
 			psDroid->action = DACTION_MOVETOATTACK;
 			turnOffMultiMsg(true);
 			moveDroidTo(psDroid, psAction->psObj->pos.x, psAction->psObj->pos.y);
 			turnOffMultiMsg(false);
 		}
+		else if (order->type != DORDER_HOLD && secondaryGetState(psDroid, DSO_HALTTYPE) == DSS_HALT_HOLD)
+		{
+			psDroid->action = DACTION_ATTACK;
+		}
 		break;
 
 	case DACTION_MOVETOREARM:
@@ -2278,23 +2478,29 @@ static void actionDroidBase(DROID *psDroid, DROID_ACTION_DATA *psAction)
 		ensureRearmPadClear((STRUCTURE *)psAction->psObj, psDroid);
 		break;
 	case DACTION_DROIDREPAIR:
-		psDroid->action = psAction->action;
-		psDroid->actionPos.x = psAction->x;
-		psDroid->actionPos.y = psAction->y;
-		setDroidActionTarget(psDroid, psAction->psObj, 0);
-		//initialise the action points
-		psDroid->actionPoints  = 0;
-		psDroid->actionStarted = gameTime;
-		if (secHoldActive && (order->type == DORDER_NONE || order->type == DORDER_HOLD))
 		{
-			psDroid->action = DACTION_DROIDREPAIR;
+			psDroid->action = psAction->action;
+			psDroid->actionPos.x = psAction->x;
+			psDroid->actionPos.y = psAction->y;
+			setDroidActionTarget(psDroid, psAction->psObj, 0);
+			//initialise the action points
+			psDroid->actionPoints  = 0;
+			psDroid->actionStarted = gameTime;
+			const auto xdiff = (SDWORD)psDroid->pos.x - (SDWORD) psAction->x;
+			const auto ydiff = (SDWORD)psDroid->pos.y - (SDWORD) psAction->y;
+			if (secHoldActive && (order->type == DORDER_NONE || order->type == DORDER_HOLD))
+			{
+				psDroid->action = DACTION_DROIDREPAIR;
+			}
+			else if (((!secHoldActive && order->type != DORDER_HOLD) || (secHoldActive && order->type == DORDER_DROIDREPAIR))
+					// check that we actually need to move closer
+					&& ((xdiff * xdiff + ydiff * ydiff) > REPAIR_RANGE * REPAIR_RANGE))
+			{
+				psDroid->action = DACTION_MOVETODROIDREPAIR;
+				moveDroidTo(psDroid, psAction->x, psAction->y);
+			}
+			break;
 		}
-		else if ((!secHoldActive && order->type != DORDER_HOLD) || (secHoldActive && order->type == DORDER_DROIDREPAIR))
-		{
-			psDroid->action = DACTION_MOVETODROIDREPAIR;
-			moveDroidTo(psDroid, psAction->x, psAction->y);
-		}
-		break;
 	case DACTION_RESTORE:
 		ASSERT_OR_RETURN(, order->type == DORDER_RESTORE, "cannot start restore action without a restore order");
 		psDroid->action = psAction->action;
