@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -57,7 +57,7 @@
 /// A coordinate.
 struct PathCoord
 {
-	PathCoord() {}
+	PathCoord(): x(0), y(0) {}
 	PathCoord(int16_t x_, int16_t y_) : x(x_), y(y_) {}
 	bool operator ==(PathCoord const &z) const
 	{
@@ -190,7 +190,7 @@ struct PathfindContext
 			map.clear();  // There are no values of iteration guaranteed not to exist in map, so clear the map.
 			iteration = 0;
 		}
-		map.resize(mapWidth * mapHeight);  // Allocate space for map, if needed.
+		map.resize(static_cast<size_t>(mapWidth) * static_cast<size_t>(mapHeight));  // Allocate space for map, if needed.
 	}
 
 	PathCoord       tileS;                // Start tile for pathfinding. (May be either source or target tile.)
@@ -400,18 +400,18 @@ static PathCoord fpathAStarExplore(PathfindContext &context, PathCoord tileF)
 			*/
 			if (dir % 2 != 0 && !context.dstIgnore.isNonblocking(node.p.x, node.p.y) && !context.dstIgnore.isNonblocking(x, y))
 			{
-				int x, y;
+				int x2, y2;
 
 				// We cannot cut corners
-				x = node.p.x + aDirOffset[(dir + 1) % 8].x;
-				y = node.p.y + aDirOffset[(dir + 1) % 8].y;
-				if (context.isBlocked(x, y))
+				x2 = node.p.x + aDirOffset[(dir + 1) % 8].x;
+				y2 = node.p.y + aDirOffset[(dir + 1) % 8].y;
+				if (context.isBlocked(x2, y2))
 				{
 					continue;
 				}
-				x = node.p.x + aDirOffset[(dir + 7) % 8].x;
-				y = node.p.y + aDirOffset[(dir + 7) % 8].y;
-				if (context.isBlocked(x, y))
+				x2 = node.p.x + aDirOffset[(dir + 7) % 8].x;
+				y2 = node.p.y + aDirOffset[(dir + 7) % 8].y;
+				if (context.isBlocked(x2, y2))
 				{
 					continue;
 				}
@@ -520,7 +520,7 @@ ASR_RETVAL fpathAStarRoute(MOVE_CONTROL *psMove, PATHJOB *psJob)
 	for (Vector2i p(world_coord(endCoord.x) + TILE_UNITS / 2, world_coord(endCoord.y) + TILE_UNITS / 2); true; p = newP)
 	{
 		ASSERT_OR_RETURN(ASR_FAILED, worldOnMap(p.x, p.y), "Assigned XY coordinates (%d, %d) not on map!", (int)p.x, (int)p.y);
-		ASSERT_OR_RETURN(ASR_FAILED, path.size() < (unsigned)mapWidth * mapHeight, "Pathfinding got in a loop.");
+		ASSERT_OR_RETURN(ASR_FAILED, path.size() < (static_cast<size_t>(mapWidth) * static_cast<size_t>(mapHeight)), "Pathfinding got in a loop.");
 
 		path.push_back(p);
 
@@ -627,7 +627,7 @@ void fpathSetBlockingMap(PATHJOB *psJob)
 		// blockMap now points to an empty map with no data. Fill the map.
 		blockMap->type = type;
 		std::vector<bool> &map = blockMap->map;
-		map.resize(mapWidth * mapHeight);
+		map.resize(static_cast<size_t>(mapWidth) * static_cast<size_t>(mapHeight));
 		uint32_t checksumMap = 0, checksumDangerMap = 0, factor = 0;
 		for (int y = 0; y < mapHeight; ++y)
 			for (int x = 0; x < mapWidth; ++x)
@@ -638,7 +638,7 @@ void fpathSetBlockingMap(PATHJOB *psJob)
 		if (!isHumanPlayer(type.owner) && type.moveType == FMT_MOVE)
 		{
 			std::vector<bool> &dangerMap = blockMap->dangerMap;
-			dangerMap.resize(mapWidth * mapHeight);
+			dangerMap.resize(static_cast<size_t>(mapWidth) * static_cast<size_t>(mapHeight));
 			for (int y = 0; y < mapHeight; ++y)
 				for (int x = 0; x < mapWidth; ++x)
 				{
