@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -26,13 +26,16 @@
 #include "lib/framework/frame.h"
 #include "lib/exceptionhandler/dumpinfo.h"
 #include "lib/netplay/netplay.h"
+#include "lib/framework/string_ext.h"
 
+#include "input/debugmappings.h"
+#include "display.h"
 #include "cheat.h"
 #include "keybind.h"
-#include "keymap.h"
 #include "multiplay.h"
 #include "qtscript.h"
 #include "template.h"
+#include "activity.h"
 
 struct CHEAT_ENTRY
 {
@@ -44,9 +47,9 @@ bool Cheated = false;
 static CHEAT_ENTRY cheatCodes[] =
 {
 	{"templates", listTemplates}, // print templates
-	{"jsload", jsAutogame}, // load an AI script for selectedPlayer
 	{"jsdebug", jsShowDebug}, // show scripting states
 	{"teach us", kf_TeachSelected}, // give experience to selected units
+	{"makemehero", kf_MakeMeHero}, // make selected units Heros
 	{"untouchable", kf_Unselectable}, // make selected droids unselectable
 	{"clone wars", []{ kf_CloneSelected(10); }}, // clone selected units
 	{"clone wars!", []{ kf_CloneSelected(40); }}, // clone selected units
@@ -60,7 +63,6 @@ static CHEAT_ENTRY cheatCodes[] =
 	{"deity", kf_ToggleGodMode},	//from above
 	{"droidinfo", kf_DebugDroidInfo},	//show unit stats
 	{"sensors", kf_ToggleSensorDisplay},	//show sensor ranges
-	{"let me win", kf_AddMissionOffWorld},	//let me win
 	{"timedemo", kf_FrameRate},	 //timedemo
 	{"kill", kf_KillSelected},	//kill selected
 	{"john kettley", kf_ToggleWeather},	//john kettley
@@ -77,6 +79,7 @@ static CHEAT_ENTRY cheatCodes[] =
 	{"work harder", kf_FinishResearch},
 	{"tileinfo", kf_TileInfo}, // output debug info about a tile
 	{"showfps", kf_ToggleFPS},	//displays your average FPS
+	{"showunits", kf_ToggleUnitCount},	//displays unit count information
 	{"showsamples", kf_ToggleSamples}, //displays the # of Sound samples in Queue & List
 	{"showorders", kf_ToggleOrders}, //displays unit order/action state.
 	{"pause", kf_TogglePauseMode}, // Pause the game.
@@ -86,10 +89,11 @@ static CHEAT_ENTRY cheatCodes[] =
 	{"damage me", kf_DamageMe},
 	{"autogame on", kf_AutoGame},
 	{"autogame off", kf_AutoGame},
+	{"shakey", kf_ToggleShakeStatus}, //shakey
 
 };
 
-bool attemptCheatCode(const char *cheat_name)
+bool _attemptCheatCode(const char *cheat_name)
 {
 	const CHEAT_ENTRY *curCheat;
 	static const CHEAT_ENTRY *const EndCheat = &cheatCodes[ARRAY_SIZE(cheatCodes)];
@@ -100,21 +104,32 @@ bool attemptCheatCode(const char *cheat_name)
 		kf_ToggleFPS();
 		return true;
 	}
+	if (!strcasecmp("showunits", cheat_name))
+	{
+		kf_ToggleUnitCount();
+		return true;
+	}
+	if (!strcasecmp("specstats", cheat_name))
+	{
+		kf_ToggleSpecOverlays();
+		return true;
+	}
 
+	const DebugInputManager& dbgInputManager = gInputManager.debugManager();
 	if (strcmp(cheat_name, "cheat on") == 0 || strcmp(cheat_name, "debug") == 0)
 	{
-		if (!getDebugMappingStatus())
+		if (!dbgInputManager.debugMappingsAllowed())
 		{
 			kf_ToggleDebugMappings();
 		}
 		return true;
 	}
-	if (strcmp(cheat_name, "cheat off") == 0 && getDebugMappingStatus())
+	if (strcmp(cheat_name, "cheat off") == 0 && dbgInputManager.debugMappingsAllowed())
 	{
 		kf_ToggleDebugMappings();
 		return true;
 	}
-	if (!getDebugMappingStatus())
+	if (!dbgInputManager.debugMappingsAllowed())
 	{
 		return false;
 	}
@@ -141,13 +156,47 @@ bool attemptCheatCode(const char *cheat_name)
 	return false;
 }
 
+bool attemptCheatCode(const char *cheat_name)
+{
+	bool result = _attemptCheatCode(cheat_name);
+	if (result)
+	{
+		ActivityManager::instance().cheatUsed(cheat_name);
+	}
+	return result;
+}
+
 void sendProcessDebugMappings(bool val)
 {
+	if (NETisReplay())
+	{
+		return;
+	}
+	if (selectedPlayer >= MAX_PLAYERS)
+	{
+		return;
+	}
 	NETbeginEncode(NETgameQueue(selectedPlayer), GAME_DEBUG_MODE);
 	NETbool(&val);
 	NETend();
 }
 
+static std::string getWantedDebugMappingStatuses(const DebugInputManager& dbgInputManager, bool bStatus)
+{
+	char ret[MAX_PLAYERS + 1] = "\0";
+	char* p = ret;
+	for (unsigned n = 0; n < MAX_PLAYERS; ++n)
+	{
+		if (NetPlay.players[n].allocated && !NetPlay.players[n].isSpectator && (dbgInputManager.getPlayerWantsDebugMappings(n) == bStatus))
+		{
+			*p++ = '0' + NetPlay.players[n].position;
+		}
+	}
+	std::sort(ret, p);
+	*p++ = '\0';
+	return ret;
+}
+
 void recvProcessDebugMappings(NETQUEUE queue)
 {
 	bool val = false;
@@ -155,30 +204,46 @@ void recvProcessDebugMappings(NETQUEUE queue)
 	NETbool(&val);
 	NETend();
 
-	bool oldDebugMode = getDebugMappingStatus();
-	processDebugMappings(queue.index, val);
-	bool newDebugMode = getDebugMappingStatus();
+	DebugInputManager& dbgInputManager = gInputManager.debugManager();
+	bool oldDebugMode = dbgInputManager.debugMappingsAllowed();
+	dbgInputManager.setPlayerWantsDebugMappings(queue.index, val);
+	bool newDebugMode = dbgInputManager.debugMappingsAllowed();
 
-	char *cmsg;
+	std::string cmsg;
 	if (val)
 	{
-		sasprintf(&cmsg, _("%s wants to enable debug mode. Enabled: %s, Disabled: %s."), getPlayerName(queue.index), getWantedDebugMappingStatuses(true).c_str(), getWantedDebugMappingStatuses(false).c_str());
+		cmsg = astringf(
+			_("%s wants to enable debug mode. Enabled: %s, Disabled: %s."),
+			getPlayerName(queue.index),
+			getWantedDebugMappingStatuses(dbgInputManager, true).c_str(),
+			getWantedDebugMappingStatuses(dbgInputManager, false).c_str()
+		);
 	}
 	else
 	{
-		sasprintf(&cmsg, _("%s wants to disable debug mode. Enabled: %s, Disabled: %s."), getPlayerName(queue.index), getWantedDebugMappingStatuses(true).c_str(), getWantedDebugMappingStatuses(false).c_str());
+		cmsg = astringf(
+			_("%s wants to disable debug mode. Enabled: %s, Disabled: %s."),
+			getPlayerName(queue.index),
+			getWantedDebugMappingStatuses(dbgInputManager, true).c_str(),
+			getWantedDebugMappingStatuses(dbgInputManager, false).c_str()
+		);
 	}
-	addConsoleMessage(cmsg, DEFAULT_JUSTIFY,  SYSTEM_MESSAGE);
+	addConsoleMessage(cmsg.c_str(), DEFAULT_JUSTIFY,  SYSTEM_MESSAGE);
 
 	if (!oldDebugMode && newDebugMode)
 	{
 		addConsoleMessage(_("Debug mode now enabled!"), DEFAULT_JUSTIFY,  SYSTEM_MESSAGE);
 		Cheated = true;
+		gInputManager.contexts().set(InputContext::DEBUG_MISC, InputContext::State::ACTIVE);
 		triggerEventCheatMode(true);
 	}
 	else if (oldDebugMode && !newDebugMode)
 	{
 		addConsoleMessage(_("Debug mode now disabled!"), DEFAULT_JUSTIFY,  SYSTEM_MESSAGE);
+		if (!NETisReplay())
+		{
+			gInputManager.contexts().set(InputContext::DEBUG_MISC, InputContext::State::INACTIVE);
+		}
 		triggerEventCheatMode(false);
 	}
 }
