@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 #include "lib/framework/frame.h"
-#include "lib/framework/opengl.h"
+#include "lib/framework/string_ext.h"
 #include "lib/ivis_opengl/screen.h"
 #include "lib/netplay/netplay.h"
 #include "lib/ivis_opengl/pieclip.h"
@@ -42,6 +42,9 @@
 #include "version.h"
 #include "warzoneconfig.h"
 #include "wrappers.h"
+#include "multilobbycommands.h"
+
+#include <cwchar>
 
 //////
 // Our fine replacement for the popt abomination follows
@@ -79,6 +82,80 @@ typedef struct _poptContext
 static bool wz_autogame = false;
 static std::string wz_saveandquit;
 static std::string wz_test;
+static std::string wz_autoratingUrl;
+static bool wz_cli_headless = false;
+static bool wz_streamer_spectator_mode = false;
+static bool wz_lobby_slashcommands = false;
+static WZ_Command_Interface wz_cmd_interface = WZ_Command_Interface::None;
+static int wz_min_autostart_players = -1;
+
+#if defined(WZ_OS_WIN)
+
+#include <ntverp.h>				// Windows SDK - include for access to VER_PRODUCTBUILD
+#if VER_PRODUCTBUILD >= 9200
+	// 9200 is the Windows SDK 8.0 (which introduced family support)
+	#include <winapifamily.h>	// Windows SDK
+#else
+	// Earlier SDKs don't have the concept of families - provide simple implementation
+	// that treats everything as "desktop"
+	#if !defined(WINAPI_PARTITION_DESKTOP)
+		#define WINAPI_PARTITION_DESKTOP			0x00000001
+	#endif
+	#if !defined(WINAPI_FAMILY_PARTITION)
+		#define WINAPI_FAMILY_PARTITION(Partition)	((WINAPI_PARTITION_DESKTOP & Partition) == Partition)
+	#endif
+#endif
+
+#include <fcntl.h>
+#include <io.h>
+#include <iostream>
+
+void SetStdOutToConsole_Win()
+{
+#if WINAPI_FAMILY_PARTITION(WINAPI_PARTITION_DESKTOP)
+	static bool has_setup_console = false;
+	const WORD MAX_CONSOLE_LINES = 9999;
+
+	if (has_setup_console) { return; }
+
+	if (AttachConsole(ATTACH_PARENT_PROCESS) == 0)
+	{
+		// failed to attach to parent process console
+		// allocate a console for this app
+		if (AllocConsole() == 0)
+		{
+			// failed to allocate a console
+			return;
+		}
+
+		// give the new console window a nicer title
+		SetConsoleTitleW(L"Warzone 2100");
+	}
+
+	SetConsoleOutputCP(CP_UTF8);
+
+	HANDLE consoleHandle = GetStdHandle(STD_OUTPUT_HANDLE);
+
+	// give the console window a bigger buffer size / scroll-back
+	CONSOLE_SCREEN_BUFFER_INFO csbi;
+	if (GetConsoleScreenBufferInfo(consoleHandle, &csbi))
+	{
+		csbi.dwSize.Y = MAX_CONSOLE_LINES;
+		SetConsoleScreenBufferSize(consoleHandle, csbi.dwSize);
+	}
+
+	FILE* fi = 0;
+	freopen_s(&fi, "CONOUT$", "w", stdout);
+	fi = 0;
+	freopen_s(&fi, "CONOUT$", "w", stderr);
+
+	std::ios::sync_with_stdio();
+
+	has_setup_console = true;
+#endif /* WINAPI_FAMILY_PARTITION(WINAPI_PARTITION_DESKTOP) */
+}
+
+#endif /* defined(WZ_OS_WIN) */
 
 static void poptPrintHelp(poptContext ctx, FILE *output)
 {
@@ -92,13 +169,18 @@ static void poptPrintHelp(poptContext ctx, FILE *output)
 		if (ctx->table[i].argument)
 		{
 			sstrcat(txt, "=");
-			sstrcat(txt, ctx->table[i].argDescrip);
+			if (ctx->table[i].argDescrip)
+			{
+				sstrcat(txt, ctx->table[i].argDescrip);
+			}
 		}
 
 		// calculate number of terminal columns required to print
 		// for languages with multibyte characters
 		const size_t txtSize = strlen(txt) + 1;
-		int txtOffset = (int) mbstowcs(nullptr, txt, txtSize) + 1 - txtSize;
+		std::mbstate_t state = std::mbstate_t();
+		const char *pTxt = txt;
+		size_t txtOffset = std::mbsrtowcs(nullptr, &pTxt, 0, &state) + 1 - txtSize;
 		// CJK characters take up two columns
 		char language[3]; // stores ISO 639-1 code
 		strlcpy(language, setlocale(LC_MESSAGES, nullptr), sizeof(language));
@@ -107,7 +189,7 @@ static void poptPrintHelp(poptContext ctx, FILE *output)
 			txtOffset /= 2;
 		}
 
-		fprintf(output, "%-*s", 40 - txtOffset, txt);
+		fprintf(output, "%-*s", static_cast<int>(40 - txtOffset), txt);
 
 		if (ctx->table[i].descrip)
 		{
@@ -152,7 +234,7 @@ static int poptGetNextOpt(poptContext ctx)
 
 	sstrcpy(match, ctx->argv[ctx->current]);
 	ctx->current++;
-	pparam = strrchr(match, '=');
+	pparam = strchr(match, '=');
 	if (pparam)									// option's got a parameter
 	{
 		*pparam++ = '\0';							// split option from parameter and increment past '='
@@ -231,6 +313,7 @@ typedef enum
 	CLI_MOD_MP,
 	CLI_LOADSKIRMISH,
 	CLI_LOADCAMPAIGN,
+	CLI_LOADREPLAY,
 	CLI_WINDOW,
 	CLI_VERSION,
 	CLI_RESOLUTION,
@@ -239,14 +322,33 @@ typedef enum
 	CLI_SOUND,
 	CLI_NOSOUND,
 	CLI_CONNECTTOIP,
+	CLI_CONNECTTOIP_SPECTATE,
 	CLI_HOSTLAUNCH,
 	CLI_NOASSERT,
 	CLI_CRASH,
 	CLI_TEXTURECOMPRESSION,
 	CLI_NOTEXTURECOMPRESSION,
+	CLI_GFXBACKEND,
+	CLI_GFXDEBUG,
+	CLI_JSBACKEND,
 	CLI_AUTOGAME,
 	CLI_SAVEANDQUIT,
 	CLI_SKIRMISH,
+	CLI_CONTINUE,
+	CLI_AUTOHOST,
+	CLI_AUTORATING,
+	CLI_AUTOHEADLESS,
+#if defined(WZ_OS_WIN)
+	CLI_WIN_ENABLE_CONSOLE,
+#endif
+	CLI_GAMEPORT,
+	CLI_WZ_CRASH_RPT,
+	CLI_STREAMER_SPECTATOR,
+	CLI_LOBBY_SLASHCOMMANDS,
+	CLI_ADD_LOBBY_ADMINHASH,
+	CLI_ADD_LOBBY_ADMINPUBLICKEY,
+	CLI_COMMAND_INTERFACE,
+	CLI_STARTPLAYERS,
 } CLI_OPTIONS;
 
 static const struct poptOption *getOptionsTable()
@@ -268,6 +370,7 @@ static const struct poptOption *getOptionsTable()
 		{ "crash", POPT_ARG_NONE, CLI_CRASH,      N_("Causes a crash to test the crash handler"), nullptr },
 		{ "loadskirmish", POPT_ARG_STRING, CLI_LOADSKIRMISH, N_("Load a saved skirmish game"),     N_("savegame") },
 		{ "loadcampaign", POPT_ARG_STRING, CLI_LOADCAMPAIGN, N_("Load a saved campaign game"),     N_("savegame") },
+		{ "loadreplay", POPT_ARG_STRING, CLI_LOADREPLAY, N_("Load a replay"),     N_("replay file") },
 		{ "window", POPT_ARG_NONE, CLI_WINDOW,     N_("Play in windowed mode"),             nullptr },
 		{ "version", POPT_ARG_NONE, CLI_VERSION,    N_("Show version information and exit"), nullptr },
 		{ "resolution", POPT_ARG_STRING, CLI_RESOLUTION, N_("Set the resolution to use"),         N_("WIDTHxHEIGHT") },
@@ -276,12 +379,44 @@ static const struct poptOption *getOptionsTable()
 		{ "sound", POPT_ARG_NONE, CLI_SOUND,      N_("Enable sound"),                      nullptr },
 		{ "nosound", POPT_ARG_NONE, CLI_NOSOUND,    N_("Disable sound"),                     nullptr },
 		{ "join", POPT_ARG_STRING, CLI_CONNECTTOIP, N_("Connect directly to IP/hostname"),  N_("host") },
+		{ "spectate", POPT_ARG_STRING, CLI_CONNECTTOIP_SPECTATE, N_("Connect directly to IP/hostname as a spectator"),  N_("host") },
 		{ "host", POPT_ARG_NONE, CLI_HOSTLAUNCH, N_("Go directly to host screen"),        nullptr },
 		{ "texturecompression", POPT_ARG_NONE, CLI_TEXTURECOMPRESSION, N_("Enable texture compression"), nullptr },
 		{ "notexturecompression", POPT_ARG_NONE, CLI_NOTEXTURECOMPRESSION, N_("Disable texture compression"), nullptr },
+		{ "gfxbackend", POPT_ARG_STRING, CLI_GFXBACKEND, N_("Set gfx backend"),
+			"(opengl, opengles"
+#if defined(WZ_VULKAN_ENABLED)
+			", vulkan"
+#endif
+#if defined(WZ_BACKEND_DIRECTX)
+			", directx"
+#endif
+			")"
+		},
+		{ "gfxdebug", POPT_ARG_NONE, CLI_GFXDEBUG, N_("Use gfx backend debug"), nullptr },
+		{ "jsbackend", POPT_ARG_STRING, CLI_JSBACKEND, N_("Set JS backend"),
+					"("
+					"quickjs"
+					")"
+		},
 		{ "autogame", POPT_ARG_NONE, CLI_AUTOGAME,   N_("Run games automatically for testing"), nullptr },
+		{ "headless", POPT_ARG_NONE, CLI_AUTOHEADLESS,   N_("Headless mode (only supported when also specifying --autogame, --autohost, --skirmish)"), nullptr },
 		{ "saveandquit", POPT_ARG_STRING, CLI_SAVEANDQUIT, N_("Immediately save game and quit"), N_("save name") },
 		{ "skirmish", POPT_ARG_STRING, CLI_SKIRMISH,   N_("Start skirmish game with given settings file"), N_("test") },
+		{ "continue", POPT_ARG_NONE, CLI_CONTINUE,   N_("Continue the last saved game"), nullptr },
+		{ "autohost", POPT_ARG_STRING, CLI_AUTOHOST,   N_("Start host game with given settings file"), N_("autohost") },
+		{ "autorating", POPT_ARG_STRING, CLI_AUTORATING,   N_("Query ratings from given server url (containing \"{HASH}\"), when hosting"), N_("autorating") },
+#if defined(WZ_OS_WIN)
+		{ "enableconsole", POPT_ARG_NONE, CLI_WIN_ENABLE_CONSOLE,   N_("Attach or create a console window and display console output (Windows only)"), nullptr },
+#endif
+		{ "gameport", POPT_ARG_STRING, CLI_GAMEPORT,   N_("Set game server port"), N_("port") },
+		{ "wz-crash-rpt", POPT_ARG_NONE, CLI_WZ_CRASH_RPT, nullptr, nullptr },
+		{ "spectator-min-ui", POPT_ARG_NONE, CLI_STREAMER_SPECTATOR, nullptr, nullptr},
+		{ "enablelobbyslashcmd", POPT_ARG_NONE, CLI_LOBBY_SLASHCOMMANDS, N_("Enable lobby slash commands (for connecting clients)"), nullptr},
+		{ "addlobbyadminhash", POPT_ARG_STRING, CLI_ADD_LOBBY_ADMINHASH, N_("Add a lobby admin identity hash (for slash commands)"), _("hash string")},
+		{ "addlobbyadminpublickey", POPT_ARG_STRING, CLI_ADD_LOBBY_ADMINPUBLICKEY, N_("Add a lobby admin public key (for slash commands)"), N_("b64-pub-key")},
+		{ "enablecmdinterface", POPT_ARG_STRING, CLI_COMMAND_INTERFACE, N_("Enable command interface"), N_("(stdin)")},
+		{ "startplayers", POPT_ARG_STRING, CLI_STARTPLAYERS, N_("Minimum required players to auto-start game"), N_("startplayers")},
 		// Terminating entry
 		{ nullptr, 0, 0,              nullptr,                                    nullptr },
 	};
@@ -403,6 +538,14 @@ bool ParseCommandLineEarly(int argc, const char * const *argv)
 			printf("Warzone 2100 - %s\n", version_getFormattedVersionString());
 			return false;
 
+#if defined(WZ_OS_WIN)
+		case CLI_WIN_ENABLE_CONSOLE:
+			SetStdOutToConsole_Win();
+			break;
+#endif
+		case CLI_WZ_CRASH_RPT:
+			// this is currently a no-op because it must be parsed even earlier than ParseCommandLineEarly
+			break;
 		default:
 			break;
 		};
@@ -437,6 +580,10 @@ bool ParseCommandLine(int argc, const char * const *argv)
 		case CLI_CONFIGDIR:
 		case CLI_HELP:
 		case CLI_VERSION:
+#if defined(WZ_OS_WIN)
+		case CLI_WIN_ENABLE_CONSOLE:
+#endif
+		case CLI_WZ_CRASH_RPT:
 			// These options are parsed in ParseCommandLineEarly() already, so ignore them
 			break;
 
@@ -463,9 +610,10 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			break;
 
 		case CLI_FULLSCREEN:
-			war_setFullscreen(true);
+			war_setWindowMode(WINDOW_MODE::fullscreen);
 			break;
 		case CLI_CONNECTTOIP:
+		case CLI_CONNECTTOIP_SPECTATE:
 			//get the ip we want to connect with, and go directly to join screen.
 			token = poptGetOptArg(poptCon);
 			if (token == nullptr)
@@ -473,10 +621,12 @@ bool ParseCommandLine(int argc, const char * const *argv)
 				qFatal("No IP/hostname given");
 			}
 			sstrcpy(iptoconnect, token);
+			// also set spectate flag
+			cliConnectToIpAsSpectator = (option == CLI_CONNECTTOIP_SPECTATE);
 			break;
 		case CLI_HOSTLAUNCH:
 			// go directly to host screen, bypass all others.
-			hostlaunch = 1;
+			setHostLaunch(HostLaunch::Host);
 			break;
 		case CLI_GAME:
 			// retrieve the game name
@@ -491,9 +641,9 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			NetPlay.bComms = false;
 			bMultiPlayer = false;
 			bMultiMessages = false;
-			for (int i = 0; i < MAX_PLAYERS; i++)
+			for (unsigned int i = 0; i < MAX_PLAYERS; i++)
 			{
-				NET_InitPlayer(i, true, false);
+				NET_InitPlayer(i, true);
 			}
 
 			//NET_InitPlayer deallocates Player 0, who must be allocated so that a later invocation of processDebugMappings does not trigger DEBUG mode
@@ -501,11 +651,11 @@ bool ParseCommandLine(int argc, const char * const *argv)
 
 			if (!strcmp(token, "CAM_1A") || !strcmp(token, "CAM_2A") || !strcmp(token, "CAM_3A"))
 			{
-				game.type = CAMPAIGN;
+				game.type = LEVEL_TYPE::CAMPAIGN;
 			}
 			else
 			{
-				game.type = SKIRMISH; // tutorial is skirmish for some reason
+				game.type = LEVEL_TYPE::SKIRMISH; // tutorial is skirmish for some reason
 			}
 			sstrcpy(aLevelName, token);
 			SetGameMode(GS_NORMAL);
@@ -582,9 +732,8 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			}
 			snprintf(saveGameName, sizeof(saveGameName), "%s/skirmish/%s.gam", SaveGamePath, token);
 			sstrcpy(sRequestResult, saveGameName); // hack to avoid crashes
-			SPinit();
+			SPinit(LEVEL_TYPE::SKIRMISH);
 			bMultiPlayer = true;
-			game.type = SKIRMISH; // tutorial is skirmish for some reason
 			SetGameMode(GS_SAVEGAMELOAD);
 			break;
 		case CLI_LOADCAMPAIGN:
@@ -595,12 +744,60 @@ bool ParseCommandLine(int argc, const char * const *argv)
 				qFatal("Unrecognised campaign savegame name");
 			}
 			snprintf(saveGameName, sizeof(saveGameName), "%s/campaign/%s.gam", SaveGamePath, token);
-			SPinit();
+			SPinit(LEVEL_TYPE::CAMPAIGN);
 			SetGameMode(GS_SAVEGAMELOAD);
 			break;
+		case CLI_LOADREPLAY:
+		{
+			// retrieve the replay name
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr)
+			{
+				qFatal("Unrecognised replay name");
+			}
+			std::string extension;
+			if (!strEndsWith(token, ".wzrp"))
+			{
+				extension = ".wzrp";
+			}
+			// check if we have a full path (relative to the replay dir)
+			snprintf(saveGameName, sizeof(saveGameName), "%s/%s%s", ReplayPath, token, extension.c_str());
+			bool foundReplayFile = PHYSFS_exists(saveGameName) != 0;
+			if (!foundReplayFile)
+			{
+				// look in all possible replay subdirs (maybe we just have a filename)
+				std::vector<std::string> replaySubdirs = {"skirmish", "multiplay"};
+				for (auto& replaySubdir : replaySubdirs)
+				{
+					snprintf(saveGameName, sizeof(saveGameName), "%s/%s/%s%s", ReplayPath, replaySubdir.c_str(), token, extension.c_str());
+					if (PHYSFS_exists(saveGameName))
+					{
+						foundReplayFile = true;
+						break;
+					}
+				}
+			}
+			if (!foundReplayFile)
+			{
+				qFatal("Unable to find specified replay");
+			}
+			sstrcpy(sRequestResult, saveGameName); // hack to avoid crashes
+			SPinit(LEVEL_TYPE::SKIRMISH);
+			bMultiPlayer = true;
+			game.maxPlayers = 4; //DEFAULTSKIRMISHMAPMAXPLAYERS;
+			SetGameMode(GS_SAVEGAMELOAD);
+			break;
+		}
+		case CLI_CONTINUE:
+			if (findLastSave())
+			{
+				runContinue();
+				SetGameMode(GS_SAVEGAMELOAD);
+			}
+			break;
 
 		case CLI_WINDOW:
-			war_setFullscreen(false);
+			war_setWindowMode(WINDOW_MODE::windowed);
 			break;
 
 		case CLI_SHADOWS:
@@ -627,8 +824,54 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			wz_texture_compression = false;
 			break;
 
+		case CLI_GFXBACKEND:
+			{
+				// retrieve the backend
+				token = poptGetOptArg(poptCon);
+				if (token == nullptr)
+				{
+					qFatal("Unrecognised backend");
+				}
+				video_backend gfxBackend;
+				if (video_backend_from_str(token, gfxBackend))
+				{
+					war_setGfxBackend(gfxBackend);
+				}
+				else
+				{
+					qFatal("Unsupported / invalid gfxbackend value");
+				}
+				break;
+			}
+
+		case CLI_GFXDEBUG:
+			uses_gfx_debug = true;
+			break;
+
+		case CLI_JSBACKEND:
+			{
+				// retrieve the backend
+				token = poptGetOptArg(poptCon);
+				if (token == nullptr)
+				{
+					qFatal("Unrecognised backend");
+				}
+				JS_BACKEND jsBackend;
+				if (js_backend_from_str(token, jsBackend))
+				{
+					war_setJSBackend(jsBackend);
+				}
+				else
+				{
+					qFatal("Unsupported / invalid jsbackend value");
+				}
+				break;
+			}
+
 		case CLI_AUTOGAME:
 			wz_autogame = true;
+			// need to cause wrappers.cpp to update calculated effective headless mode
+			setHeadlessGameMode(wz_cli_headless);
 			break;
 
 		case CLI_SAVEANDQUIT:
@@ -641,7 +884,7 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			break;
 
 		case CLI_SKIRMISH:
-			hostlaunch = 2;
+			setHostLaunch(HostLaunch::Skirmish);
 			token = poptGetOptArg(poptCon);
 			if (token == nullptr)
 			{
@@ -649,6 +892,100 @@ bool ParseCommandLine(int argc, const char * const *argv)
 			}
 			wz_test = token;
 			break;
+
+		case CLI_AUTOHOST:
+			setHostLaunch(HostLaunch::Autohost);
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr)
+			{
+				qFatal("Bad test key");
+			}
+			wz_test = token;
+			break;
+
+		case CLI_AUTORATING:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr)
+			{
+				qFatal("Bad autorating server");
+			}
+			wz_autoratingUrl = token;
+			debug(LOG_INFO, "Using \"%s\" for ratings.", wz_autoratingUrl.c_str());
+			break;
+
+		case CLI_AUTOHEADLESS:
+			wz_cli_headless = true;
+			setHeadlessGameMode(true);
+			break;
+
+		case CLI_GAMEPORT:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr)
+			{
+				qFatal("Bad game server port");
+			}
+			NETsetGameserverPort(atoi(token));
+			netGameserverPortOverride = true;
+			debug(LOG_INFO, "Games will be hosted on port [%d]", NETgetGameserverPort());
+			break;
+
+		case CLI_STREAMER_SPECTATOR:
+			wz_streamer_spectator_mode = true;
+			break;
+
+		case CLI_LOBBY_SLASHCOMMANDS:
+			wz_lobby_slashcommands = true;
+			break;
+
+		case CLI_ADD_LOBBY_ADMINHASH:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr || strlen(token) == 0)
+			{
+				qFatal("Bad admin hash");
+			}
+			addLobbyAdminIdentityHash(token);
+			break;
+
+		case CLI_ADD_LOBBY_ADMINPUBLICKEY:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr || strlen(token) == 0)
+			{
+				qFatal("Bad admin public key");
+			}
+			addLobbyAdminPublicKey(token);
+			break;
+
+		case CLI_COMMAND_INTERFACE:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr || strlen(token) == 0)
+			{
+				// use default, which is currently "stdin"
+				token = "stdin";
+			}
+			if (strcmp(token, "stdin") == 0)
+			{
+				// enable stdin
+				wz_cmd_interface = WZ_Command_Interface::StdIn_Interface;
+			}
+			else
+			{
+				qFatal("Unsupported / invalid enablecmdinterface value");
+			}
+			break;
+		case CLI_STARTPLAYERS:
+			token = poptGetOptArg(poptCon);
+			if (token == nullptr)
+			{
+				qFatal("Bad start players count");
+			}
+			wz_min_autostart_players = atoi(token);
+			if (wz_min_autostart_players < 0)
+			{
+				qFatal("Invalid start players count");
+			}
+			debug(LOG_INFO, "Games will automatically start with [%d] players (when ready)", wz_min_autostart_players);
+			break;
+
 		};
 	}
 
@@ -669,3 +1006,41 @@ const std::string &wz_skirmish_test()
 {
 	return wz_test;
 }
+
+std::string autoratingUrl(std::string const &hash) {
+	auto url = wz_autoratingUrl;
+	auto h = wz_autoratingUrl.find_first_of("{HASH}");
+	if (h != std::string::npos)
+	{
+		url.replace(h, 6, hash);
+	}
+	return url;
+}
+
+void setAutoratingUrl(std::string url) {
+	wz_autoratingUrl = url;
+}
+
+std::string getAutoratingUrl() {
+	return wz_autoratingUrl;
+}
+
+bool streamer_spectator_mode()
+{
+	return wz_streamer_spectator_mode;
+}
+
+bool lobby_slashcommands_enabled()
+{
+	return wz_lobby_slashcommands;
+}
+
+WZ_Command_Interface wz_command_interface()
+{
+	return wz_cmd_interface;
+}
+
+int min_autostart_player_count()
+{
+	return wz_min_autostart_players;
+}
