@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -31,5 +31,19 @@ bool ParseCommandLineEarly(int argc, const char * const *argv);
 bool autogame_enabled();
 const std::string &saveandquit_enabled();
 const std::string &wz_skirmish_test();
+std::string autoratingUrl(std::string const &hash);
+void setAutoratingUrl(std::string url);
+std::string getAutoratingUrl();
+bool streamer_spectator_mode();
+bool lobby_slashcommands_enabled();
+
+enum class WZ_Command_Interface
+{
+	None,
+	StdIn_Interface,
+};
+WZ_Command_Interface wz_command_interface();
+
+int min_autostart_player_count();
 
 #endif // __INCLUDED_SRC_CLPARSE_H__
