@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -36,14 +36,18 @@
 #include "main.h"
 #include "radar.h"
 #include "hci.h"
+#include "mission.h" //for TIMER_Y
+#include "challenge.h" //for challengeActive
 #include <string>
-#include <istream>
 #include <sstream>
 #include <deque>
+#include <chrono>
+#include <set>
 
 // FIXME: When we switch over to full JS, use class version of this file
 
-#define	DEFAULT_MESSAGE_DURATION	GAME_TICKS_PER_SEC * 4
+#define	DEFAULT_MESSAGE_DURATION			GAME_TICKS_PER_SEC * 5
+#define	DEFAULT_MESSAGE_DURATION_CAMPAIGN	GAME_TICKS_PER_SEC * 12
 // Chat/history "window"
 #define CON_BORDER_WIDTH			4
 #define CON_BORDER_HEIGHT			4
@@ -64,15 +68,17 @@ struct CONSOLE_MESSAGE
 {
 	WzText display;
 	UDWORD	timeAdded;		// When was it added to our list?
+	UDWORD	duration;
 	CONSOLE_TEXT_JUSTIFICATION JustifyType;	// text justification
 	int		player;			// Player who sent this message or SYSTEM_MESSAGE
-	CONSOLE_MESSAGE(const std::string &text, iV_fonts fontID, UDWORD time, CONSOLE_TEXT_JUSTIFICATION justify, int plr)
-	                : display(text, fontID), timeAdded(time), JustifyType(justify), player(plr) {}
+	CONSOLE_MESSAGE(const std::string &text, iV_fonts fontID, UDWORD time, UDWORD duration, CONSOLE_TEXT_JUSTIFICATION justify, int plr)
+	                : display(text, fontID), timeAdded(time), duration(duration), JustifyType(justify), player(plr) {}
 
-	CONSOLE_MESSAGE& operator =(CONSOLE_MESSAGE&& input)
+	CONSOLE_MESSAGE& operator =(CONSOLE_MESSAGE&& input) noexcept
 	{
 		display = std::move(input.display);
 		timeAdded = input.timeAdded;
+		duration = input.duration;
 		JustifyType = input.JustifyType;
 		player = input.player;
 		return *this;
@@ -96,9 +102,20 @@ static bool	bTextBoxActive = false;				/** Is there a box under the console text
 static bool	bConsoleDisplayEnabled = false;		/** Is the console being displayed? */
 static UDWORD	consoleVisibleLines;			/** How many lines are displayed? */
 static int allowNewMessages;					/** Whether new messages are allowed to be added */
+static std::set<std::shared_ptr<CONSOLE_MESSAGE_LISTENER>> messageListeners;
 
 char ConsoleString[MAX_CONSOLE_TMP_STRING_LENGTH];	/// Globally available string for new console messages.
 
+void consoleAddMessageListener(const std::shared_ptr<CONSOLE_MESSAGE_LISTENER>& listener)
+{
+	messageListeners.insert(listener);
+}
+
+void consoleRemoveMessageListener(const std::shared_ptr<CONSOLE_MESSAGE_LISTENER>& listener)
+{
+	messageListeners.erase(listener);
+}
+
 static CONSOLE_CALC_LAYOUT_FUNC calcLayoutFunc;
 
 /**
@@ -121,15 +138,16 @@ void setConsoleCalcLayout(const CONSOLE_CALC_LAYOUT_FUNC& layoutFunc)
 /** Sets the system up */
 void	initConsoleMessages()
 {
+	unsigned int duration = (game.type == LEVEL_TYPE::SKIRMISH) ? DEFAULT_MESSAGE_DURATION : DEFAULT_MESSAGE_DURATION_CAMPAIGN;
 	linePitch = iV_GetTextLineSize(font_regular);
 	bConsoleDropped = false;
-	setConsoleMessageDuration(DEFAULT_MESSAGE_DURATION);	// Setup how long messages are displayed for
+	setConsoleMessageDuration(duration);					// Setup how long messages are displayed for
 	setConsoleBackdropStatus(true);							// No box under the text
 	enableConsoleDisplay(true);								// Turn on the console display
 
 	//	Set up the main console size and position x,y,width
 	setConsoleCalcLayout([]() {
-		setConsoleSizePos(16, 32, pie_GetVideoBufferWidth() - 32);
+		setConsoleSizePos(16, (!challengeActive && (game.type == LEVEL_TYPE::SKIRMISH)) ? 32 : (32 + TIMER_Y), pie_GetVideoBufferWidth() - 32);
 	});
 	historyConsole.topX = HISTORYBOX_X;
 	historyConsole.topY = HISTORYBOX_Y;
@@ -139,6 +157,13 @@ void	initConsoleMessages()
 	permitNewConsoleMessages(true);							// Allow new messages
 }
 
+void shutdownConsoleMessages()
+{
+	permitNewConsoleMessages(false);
+	flushConsoleMessages();
+	clearInfoMessages();
+}
+
 void consoleScreenDidChangeSize(unsigned int oldWidth, unsigned int oldHeight, unsigned int newWidth, unsigned int newHeight)
 {
 	if (calcLayoutFunc == nullptr) return;
@@ -167,9 +192,36 @@ void	toggleConsoleDrop()
 	}
 }
 
-/** Add a string to the console. */
-bool addConsoleMessage(const char *Text, CONSOLE_TEXT_JUSTIFICATION jusType, SDWORD player, bool team)
+bool addConsoleMessageDebounced(const char * Text, CONSOLE_TEXT_JUSTIFICATION jusType, SDWORD player, const DEBOUNCED_MESSAGE & message, bool team, UDWORD duration)
 {
+	// Messages are debounced individually - one debounced message won't stop a different one from appearing
+	static std::map<const DEBOUNCED_MESSAGE *, std::chrono::steady_clock::time_point> lastAllowedMessageTimes;
+
+	const std::chrono::milliseconds DEBOUNCE_TIME(message.debounceTime);
+	const std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();
+
+	const auto lastAllowedMessageTime = lastAllowedMessageTimes.find(&message);
+
+	if (lastAllowedMessageTime == lastAllowedMessageTimes.end() || std::chrono::duration_cast<std::chrono::milliseconds>(now - lastAllowedMessageTime->second) >= DEBOUNCE_TIME)
+	{
+		lastAllowedMessageTimes[&message] = now;
+		return addConsoleMessage(Text, jusType, player, team, duration);
+	}
+	else
+	{
+		return false;
+	}
+}
+
+/** Add a string to the console. */
+bool addConsoleMessage(const char *Text, CONSOLE_TEXT_JUSTIFICATION jusType, SDWORD player, bool team, UDWORD duration)
+{
+	ConsoleMessage message = { Text, jusType, player, team, duration };
+	for (const auto &listener: messageListeners)
+	{
+		(*listener)(message);
+	}
+
 	if (!allowNewMessages)
 	{
 		return false;	// Don't allow it to be added if we've disabled adding of new messages
@@ -186,7 +238,7 @@ bool addConsoleMessage(const char *Text, CONSOLE_TEXT_JUSTIFICATION jusType, SDW
 		std::string FitText(lines);
 		while (!FitText.empty())
 		{
-			int pixelWidth = iV_GetTextWidth(FitText.c_str(), font_regular);
+			unsigned pixelWidth = iV_GetTextWidth(FitText.c_str(), font_regular);
 			if (pixelWidth <= mainConsole.width)
 			{
 				break;
@@ -196,20 +248,23 @@ bool addConsoleMessage(const char *Text, CONSOLE_TEXT_JUSTIFICATION jusType, SDW
 
 		debug(LOG_CONSOLE, "(to player %d): %s", (int)player, FitText.c_str());
 
+		UDWORD newMsgDuration = (duration != DEFAULT_CONSOLE_MESSAGE_DURATION) ? duration : messageDuration;
+
 		if (player == INFO_MESSAGE)
 		{
-			InfoMessages.emplace_back(FitText, font_regular, realTime, jusType, player);
+			InfoMessages.emplace_back(FitText, font_regular, realTime, newMsgDuration, jusType, player);
 		}
 		else
 		{
-			ActiveMessages.emplace_back(FitText, font_regular, realTime, jusType, player);	// everything gets logged here for a specific period of time
+			ActiveMessages.emplace_back(FitText, font_regular, realTime, newMsgDuration, jusType, player);	// everything gets logged here for a specific period of time
 			if (team)
 			{
-				TeamMessages.emplace_back(FitText, font_regular, realTime, jusType, player);	// persistent team specific logs
+				TeamMessages.emplace_back(FitText, font_regular, realTime, newMsgDuration, jusType, player);	// persistent team specific logs
 			}
-			HistoryMessages.emplace_back(FitText, font_regular, realTime, jusType, player);	// persistent messages (all types)
+			HistoryMessages.emplace_back(FitText, font_regular, realTime, newMsgDuration, jusType, player);	// persistent messages (all types)
 		}
 	}
+
 	return true;
 }
 
@@ -231,7 +286,7 @@ void	updateConsoleMessages()
 	}
 	for (auto i = InfoMessages.begin(); i != InfoMessages.end();)
 	{
-		if (realTime - i->timeAdded > messageDuration)
+		if ((i->duration != MAX_CONSOLE_MESSAGE_DURATION) && (realTime - i->timeAdded > i->duration))
 		{
 			i = InfoMessages.erase(i);
 		}
@@ -243,7 +298,7 @@ void	updateConsoleMessages()
 	// Time to kill all expired ones
 	for (auto i = ActiveMessages.begin(); i != ActiveMessages.end();)
 	{
-		if (realTime - i->timeAdded > messageDuration)
+		if ((i->duration != MAX_CONSOLE_MESSAGE_DURATION) && (realTime - i->timeAdded > i->duration))
 		{
 			i = ActiveMessages.erase(i);
 		}
@@ -298,21 +353,33 @@ static PIELIGHT getConsoleTextColor(int player)
 	{
 		return WZCOL_CONS_TEXT_INFO;
 	}
+	else if (player == SPECTATOR_MESSAGE)
+	{
+		return WZCOL_TEXT_MEDIUM;
+	}
 	else
 	{
-		// Don't use friend-foe colors in the lobby
-		if (bEnemyAllyRadarColor && (GetGameMode() == GS_NORMAL))
+		// Only use friend-foe colors if we are (potentially) a player
+		if (selectedPlayer < MAX_PLAYERS)
 		{
-			if (aiCheckAlliances(player, selectedPlayer))
+			// Don't use friend-foe colors in the lobby
+			if (bEnemyAllyRadarColor && (GetGameMode() == GS_NORMAL))
 			{
-				return WZCOL_CONS_TEXT_USER_ALLY;
-			}
-			else
-			{
-				return WZCOL_CONS_TEXT_USER_ENEMY;
+				if (aiCheckAlliances(player, selectedPlayer))
+				{
+					if (selectedPlayer == player)
+					{
+						return WZCOL_TEXT_BRIGHT;
+					}
+					return WZCOL_CONS_TEXT_USER_ALLY;
+				}
+				else
+				{
+					return WZCOL_CONS_TEXT_USER_ENEMY;
+				}
 			}
 		}
-		// Friend-foe is off
+
 		return WZCOL_TEXT_BRIGHT;
 	}
 }
@@ -333,7 +400,7 @@ static void console_drawtext(WzText &display, PIELIGHT colour, int x, int y, CON
 	display.render(x, y, colour);
 }
 
-// Show global (mode=true) or team (mode=false) history messages
+// Show global (mode=false) or team (mode=true) history messages
 void displayOldMessages(bool mode)
 {
 	int startpos = 0;
@@ -399,7 +466,7 @@ void displayOldMessages(bool mode)
 void	displayConsoleMessages()
 {
 	// Check if we have any messages we want to show
-	if (!getNumberConsoleMessages() && !bConsoleDropped && InfoMessages.empty())
+	if (ActiveMessages.empty() && !bConsoleDropped && InfoMessages.empty())
 	{
 		return;
 	}
@@ -410,7 +477,6 @@ void	displayConsoleMessages()
 		return;
 	}
 
-	pie_SetDepthBufferStatus(DEPTH_CMP_ALWAYS_WRT_ON);
 	pie_SetFogStatus(false);
 
 	if (bConsoleDropped)
@@ -426,17 +492,24 @@ void	displayConsoleMessages()
 		tmp -= i->display.width();
 		console_drawtext(i->display, getConsoleTextColor(i->player), tmp - 6, linePitch - 2, i->JustifyType, i->display.width());
 	}
-	int TextYpos = mainConsole.topY;
-	// Draw the blue background for the text (only in game, not lobby)
-	if (bTextBoxActive && GetGameMode() == GS_NORMAL)
+
+	if (!ActiveMessages.empty())
 	{
-		iV_TransBoxFill(mainConsole.topX - CON_BORDER_WIDTH, mainConsole.topY - mainConsole.textDepth - CON_BORDER_HEIGHT,
-						mainConsole.topX + mainConsole.width, mainConsole.topY + (getNumberConsoleMessages() * linePitch) + CON_BORDER_HEIGHT - linePitch);
-	}
-	for (auto & ActiveMessage : ActiveMessages)
-	{
-		console_drawtext(ActiveMessage.display, getConsoleTextColor(ActiveMessage.player), mainConsole.topX, TextYpos, ActiveMessage.JustifyType, mainConsole.width);
-		TextYpos += ActiveMessage.display.lineSize();
+		int TextYpos = mainConsole.topY;
+		// Draw the blue background for the text (only in game, not lobby)
+		if (bTextBoxActive && GetGameMode() == GS_NORMAL)
+		{
+			iV_TransBoxFill(mainConsole.topX - CON_BORDER_WIDTH,
+							mainConsole.topY - mainConsole.textDepth - CON_BORDER_HEIGHT,
+							mainConsole.topX + mainConsole.width,
+							mainConsole.topY + (getNumberConsoleMessages() * linePitch) + CON_BORDER_HEIGHT - linePitch);
+		}
+		for (auto &ActiveMessage : ActiveMessages)
+		{
+			console_drawtext(ActiveMessage.display, getConsoleTextColor(ActiveMessage.player), mainConsole.topX,
+							 TextYpos, ActiveMessage.JustifyType, mainConsole.width);
+			TextYpos += ActiveMessage.display.lineSize();
+		}
 	}
 }
 
