@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #include "objectdef.h"
 #include "stats.h"
 #include "visibility.h"
+#include "selection.h"
 
 #include <queue>
 
@@ -77,36 +78,36 @@ struct INITIAL_DROID_ORDERS
 };
 /*Builds an instance of a Structure - the x/y passed in are in world coords.*/
 /// Sends a GAME_DROID message if bMultiMessages is true, or actually creates it if false. Only uses initialOrders if sending a GAME_DROID message.
-DROID *buildDroid(DROID_TEMPLATE *pTemplate, UDWORD x, UDWORD y, UDWORD player, bool onMission, const INITIAL_DROID_ORDERS *initialOrders);
+DROID *buildDroid(DROID_TEMPLATE *pTemplate, UDWORD x, UDWORD y, UDWORD player, bool onMission, const INITIAL_DROID_ORDERS *initialOrders, Rotation rot = Rotation());
 /// Creates a droid locally, instead of sending a message, even if the bMultiMessages HACK is set to true.
-DROID *reallyBuildDroid(DROID_TEMPLATE *pTemplate, Position pos, UDWORD player, bool onMission, Rotation rot = Rotation());
+DROID *reallyBuildDroid(const DROID_TEMPLATE *pTemplate, Position pos, UDWORD player, bool onMission, Rotation rot = Rotation());
 
 /* Set the asBits in a DROID structure given it's template. */
-void droidSetBits(DROID_TEMPLATE *pTemplate, DROID *psDroid);
+void droidSetBits(const DROID_TEMPLATE *pTemplate, DROID *psDroid);
 
 /* Calculate the weight of a droid from it's template */
-UDWORD calcDroidWeight(DROID_TEMPLATE *psTemplate);
+UDWORD calcDroidWeight(const DROID_TEMPLATE *psTemplate);
 
 /* Calculate the power points required to build/maintain a droid */
-UDWORD calcDroidPower(DROID *psDroid);
+UDWORD calcDroidPower(const DROID *psDroid);
 
 // Calculate the number of points required to build a droid
 UDWORD calcDroidPoints(DROID *psDroid);
 
 /* Calculate the body points of a droid from it's template */
-UDWORD calcTemplateBody(DROID_TEMPLATE *psTemplate, UBYTE player);
+UDWORD calcTemplateBody(const DROID_TEMPLATE *psTemplate, UBYTE player);
 
 /* Calculate the base speed of a droid from it's template */
-UDWORD calcDroidBaseSpeed(DROID_TEMPLATE *psTemplate, UDWORD weight, UBYTE player);
+UDWORD calcDroidBaseSpeed(const DROID_TEMPLATE *psTemplate, UDWORD weight, UBYTE player);
 
 /* Calculate the speed of a droid over a terrain */
 UDWORD calcDroidSpeed(UDWORD baseSpeed, UDWORD terrainType, UDWORD propIndex, UDWORD level);
 
 /* Calculate the points required to build the template */
-UDWORD calcTemplateBuild(DROID_TEMPLATE *psTemplate);
+UDWORD calcTemplateBuild(const DROID_TEMPLATE *psTemplate);
 
 /* Calculate the power points required to build/maintain the droid */
-UDWORD calcTemplatePower(DROID_TEMPLATE *psTemplate);
+UDWORD calcTemplatePower(const DROID_TEMPLATE *psTemplate);
 
 // return whether a droid is IDF
 bool idfDroid(DROID *psDroid);
@@ -162,13 +163,16 @@ bool droidTemplateShutDown();
 DROID_TYPE droidType(DROID *psDroid);
 
 /* Return the type of a droid from it's template */
-DROID_TYPE droidTemplateType(DROID_TEMPLATE *psTemplate);
+DROID_TYPE droidTemplateType(const DROID_TEMPLATE *psTemplate);
 
-void assignDroidsToGroup(UDWORD	playerNumber, UDWORD groupNumber);
+void assignDroidsToGroup(UDWORD	playerNumber, UDWORD groupNumber, bool clearGroup);
+void removeDroidsFromGroup(UDWORD playerNumber);
+
+bool activateNoGroup(UDWORD playerNumber, const SELECTIONTYPE selectionType, const SELECTION_CLASS selectionClass, const bool bOnScreen);
 
 bool activateGroup(UDWORD playerNumber, UDWORD groupNumber);
 
-UDWORD getNumDroidsForLevel(UDWORD level);
+UDWORD getNumDroidsForLevel(uint32_t player, UDWORD level);
 
 bool activateGroupAndMove(UDWORD playerNumber, UDWORD groupNumber);
 /* calculate muzzle tip location in 3d world added int weapon_slot to fix the always slot 0 hack*/
@@ -176,9 +180,6 @@ bool calcDroidMuzzleLocation(const DROID *psDroid, Vector3i *muzzle, int weapon_
 /* calculate muzzle base location in 3d world added int weapon_slot to fix the always slot 0 hack*/
 bool calcDroidMuzzleBaseLocation(const DROID *psDroid, Vector3i *muzzle, int weapon_slot);
 
-// finds a droid for the player and sets it to be the current selected droid
-bool selectDroidByID(UDWORD id, UDWORD player);
-
 /* Droid experience stuff */
 unsigned int getDroidLevel(const DROID *psDroid);
 UDWORD getDroidEffectiveLevel(const DROID *psDroid);
@@ -273,7 +274,7 @@ bool cbSensorDroid(const DROID *psDroid);
 bool standardSensorDroid(const DROID *psDroid);
 
 // give a droid from one player to another - used in Electronic Warfare and multiplayer
- DROID *giftSingleDroid(DROID *psD, UDWORD to);
+ DROID *giftSingleDroid(DROID *psD, UDWORD to, bool electronic = false);
 
 /// Calculates the electronic resistance of a droid based on its experience level
 SWORD droidResistance(const DROID *psDroid);
@@ -286,6 +287,9 @@ const char *getDroidNameForRank(UDWORD rank);
 /*called when a Template is deleted in the Design screen*/
 void deleteTemplateFromProduction(DROID_TEMPLATE *psTemplate, unsigned player, QUEUE_MODE mode);  // ModeQueue deletes from production queues, which are not yet synchronised. ModeImmediate deletes from current production which is synchronised.
 
+// Check if a droid can be selected.
+bool isSelectable(DROID const *psDroid);
+
 // Select a droid and do any necessary housekeeping.
 void SelectDroid(DROID *psDroid);
 
@@ -307,7 +311,7 @@ bool droidOnMap(const DROID *psDroid);
 void droidSetPosition(DROID *psDroid, int x, int y);
 
 /// Return a percentage of how fully armed the object is, or -1 if N/A.
-int droidReloadBar(BASE_OBJECT *psObj, WEAPON *psWeap, int weapon_slot);
+int droidReloadBar(const BASE_OBJECT *psObj, const WEAPON *psWeap, int weapon_slot);
 
 static inline int droidSensorRange(const DROID *psDroid)
 {
