@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #include "lib/framework/frameresource.h"
 #include "lib/framework/file.h"
 #include "lib/framework/physfs_ext.h"
+#include "3rdparty/physfs_memoryio.h"
 #include "lib/framework/wzapp.h"
 #include "lib/ivis_opengl/piemode.h"
 #include "lib/ivis_opengl/piestate.h"
@@ -38,13 +39,13 @@
 #include "lib/ivis_opengl/tex.h"
 #include "lib/ivis_opengl/imd.h"
 #include "lib/netplay/netplay.h"
-#include "lib/script/script.h"
 #include "lib/sound/audio_id.h"
 #include "lib/sound/cdaudio.h"
 #include "lib/sound/mixer.h"
 
 #include "init.h"
 
+#include "input/manager.h"
 #include "advvis.h"
 #include "atmos.h"
 #include "challenge.h"
@@ -64,9 +65,9 @@
 #include "gateway.h"
 #include "hci.h"
 #include "intdisplay.h"
-#include "keymap.h"
 #include "levels.h"
 #include "lighting.h"
+#include "loadsave.h"
 #include "loop.h"
 #include "mapgrid.h"
 #include "mechanics.h"
@@ -76,15 +77,13 @@
 #include "multiint.h"
 #include "multigifts.h"
 #include "multiplay.h"
+#include "multistat.h"
+#include "notifications.h"
 #include "projectile.h"
 #include "order.h"
 #include "radar.h"
 #include "research.h"
 #include "lib/framework/cursors.h"
-#include "scriptextern.h"
-#include "scriptfuncs.h"
-#include "scripttabs.h"
-#include "scriptvals.h"
 #include "text.h"
 #include "transporter.h"
 #include "warzoneconfig.h"
@@ -94,8 +93,14 @@
 #include "ingameop.h"
 #include "qtscript.h"
 #include "template.h"
+#include "activity.h"
+#include "spectatorwidgets.h"
+#include "seqdisp.h"
+#include "version.h"
 
 #include <algorithm>
+#include <unordered_map>
+#include <array>
 
 static void initMiscVars();
 
@@ -108,6 +113,72 @@ IMAGEFILE *FrontImages;
 
 static wzSearchPath *searchPathRegistry = nullptr;
 
+static std::string inMemoryMapVirtualFilenameUID;
+static std::vector<uint8_t> inMemoryMapArchiveData;
+static size_t inMemoryMapArchiveMounted = 0;
+
+struct WZmapInfo
+{
+public:
+	WZmapInfo(bool isMapMod, bool isRandom)
+	: isMapMod(isMapMod)
+	, isRandom(isRandom)
+	{ }
+public:
+	bool isMapMod;
+	bool isRandom;
+};
+
+typedef std::string MapName;
+typedef std::unordered_map<MapName, WZmapInfo> WZMapInfo_Map;
+WZMapInfo_Map WZ_Maps;
+
+enum MODS_PATHS: size_t
+{
+	MODS_MUSIC,
+	MODS_GLOBAL,
+	MODS_AUTOLOAD,
+	MODS_CAMPAIGN,
+	MODS_MULTIPLAY,
+	MODS_PATHS_MAX
+};
+
+static std::string getFullModPath(MODS_PATHS type)
+{
+	switch (type)
+	{
+		case MODS_MUSIC:
+			return version_getVersionedModsFolderPath("music");
+		case MODS_GLOBAL:
+			return version_getVersionedModsFolderPath("global");
+		case MODS_AUTOLOAD:
+			return version_getVersionedModsFolderPath("autoload");
+		case MODS_CAMPAIGN:
+			return version_getVersionedModsFolderPath("campaign");
+		case MODS_MULTIPLAY:
+			return version_getVersionedModsFolderPath("multiplay");
+		case MODS_PATHS_MAX:
+			break;
+	}
+	return version_getVersionedModsFolderPath();
+}
+
+static std::array<std::string, MODS_PATHS_MAX> buildFullModsPaths()
+{
+	std::array<std::string, MODS_PATHS_MAX> result;
+	for (size_t i = 0; i < MODS_PATHS_MAX; ++i)
+	{
+		result[i] = getFullModPath(static_cast<MODS_PATHS>(i));
+	}
+	return result;
+}
+
+static const char* versionedModsPath(MODS_PATHS type)
+{
+	static std::array<std::string, MODS_PATHS_MAX> cachedFullModsPaths = buildFullModsPaths();
+	return cachedFullModsPaths[type].c_str();
+}
+
 // Each module in the game should have a call from here to initialise
 // any globals and statics to there default values each time the game
 // or frontend restarts.
@@ -133,28 +204,28 @@ static bool InitialiseGlobals()
 }
 
 
-bool loadLevFile(const char *filename, searchPathMode datadir, bool ignoreWrf, char const *realFileName)
+bool loadLevFile(const std::string& filename, searchPathMode pathMode, bool ignoreWrf, char const *realFileName)
 {
 	char *pBuffer;
 	UDWORD size;
 
 	if (realFileName == nullptr)
 	{
-		debug(LOG_WZ, "Loading lev file: \"%s\", builtin\n", filename);
+		debug(LOG_WZ, "Loading lev file: \"%s\", builtin\n", filename.c_str());
 	}
 	else
 	{
-		debug(LOG_WZ, "Loading lev file: \"%s\" from \"%s\"\n", filename, realFileName);
+		debug(LOG_WZ, "Loading lev file: \"%s\" from \"%s\"\n", filename.c_str(), realFileName);
 	}
 
-	if (!PHYSFS_exists(filename) || !loadFile(filename, &pBuffer, &size))
+	if (!PHYSFS_exists(filename.c_str()) || !loadFile(filename.c_str(), &pBuffer, &size))
 	{
-		debug(LOG_ERROR, "File not found: %s\n", filename);
+		debug(LOG_ERROR, "File not found: %s\n", filename.c_str());
 		return false; // only in NDEBUG case
 	}
-	if (!levParse(pBuffer, size, datadir, ignoreWrf, realFileName))
+	if (!levParse(pBuffer, size, pathMode, ignoreWrf, realFileName))
 	{
-		debug(LOG_ERROR, "Parse error in %s\n", filename);
+		debug(LOG_ERROR, "Parse error in %s\n", filename.c_str());
 		free(pBuffer);
 		return false;
 	}
@@ -239,6 +310,76 @@ void registerSearchPath(const char path[], unsigned int priority)
 	}
 }
 
+std::list<std::string> getPhysFSSearchPathsAsStr()
+{
+	std::list<std::string> results;
+	char **list = PHYSFS_getSearchPath();
+	if (list == NULL)
+	{
+		return {};
+	}
+	for (char **i = list; *i != NULL; i++)
+	{
+		results.push_back(*i);
+	}
+	PHYSFS_freeList(list);
+	return results;
+}
+
+static void clearAllPhysFSSearchPaths()
+{
+	auto searchPaths = getPhysFSSearchPathsAsStr();
+	if (searchPaths.empty())
+	{
+		return;
+	}
+
+	std::unordered_map<std::string, std::string> mountPathToErrorUnmounting;
+	size_t searchPathsRemoved = 0;
+	do
+	{
+		searchPathsRemoved = 0;
+		mountPathToErrorUnmounting.clear();
+		for (auto i = searchPaths.begin(); i != searchPaths.end();)
+		{
+			if (WZ_PHYSFS_unmount(i->c_str()) != 0)
+			{
+				++searchPathsRemoved;
+				i = searchPaths.erase(i);
+			}
+			else
+			{
+				const char* pErrorStr = WZ_PHYSFS_getLastError();
+				mountPathToErrorUnmounting[*i] = (pErrorStr) ? pErrorStr : "<unknown>";
+				++i;
+			}
+		}
+	} while (!searchPaths.empty() && searchPathsRemoved > 0);
+
+
+	for (auto i = searchPaths.begin(); i != searchPaths.end(); i++)
+	{
+		auto it_error = mountPathToErrorUnmounting.find(*i);
+		debug(LOG_WZ, "Unable to unmount search path: %s; (Reason: %s)", i->c_str(), (it_error != mountPathToErrorUnmounting.end()) ? it_error->second.c_str() : "");
+	}
+}
+
+static void clearInMemoryMapFile(void *pData)
+{
+	ASSERT_OR_RETURN(, !inMemoryMapArchiveData.empty(), "Already freed??");
+	ASSERT_OR_RETURN(, inMemoryMapArchiveData.data() == pData, "Unexpected pointer received?");
+	if (!inMemoryMapVirtualFilenameUID.empty())
+	{
+		levRemoveDataSetByRealFileName(inMemoryMapVirtualFilenameUID.c_str(), nullptr);
+		WZ_Maps.erase(inMemoryMapVirtualFilenameUID);
+	}
+	inMemoryMapVirtualFilenameUID.clear();
+	inMemoryMapArchiveData.clear();
+	if (inMemoryMapArchiveMounted > 0)
+	{
+		--inMemoryMapArchiveMounted;
+	}
+}
 
 /*!
  * \brief Rebuilds the PHYSFS searchPath with mode specific subdirs
@@ -283,13 +424,12 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 #endif // DEBUG
 				// Remove maps and mods
 				removeSubdirs(curSearchPath->path, "maps");
-				removeSubdirs(curSearchPath->path, "mods/music");
-				removeSubdirs(curSearchPath->path, "mods/global");
-				removeSubdirs(curSearchPath->path, "mods");
-				removeSubdirs(curSearchPath->path, "mods/autoload");
-				removeSubdirs(curSearchPath->path, "mods/campaign");
-				removeSubdirs(curSearchPath->path, "mods/multiplay");
-				removeSubdirs(curSearchPath->path, "mods/downloads");
+				removeSubdirs(curSearchPath->path, versionedModsPath(MODS_MUSIC));
+				removeSubdirs(curSearchPath->path, versionedModsPath(MODS_GLOBAL));
+				removeSubdirs(curSearchPath->path, versionedModsPath(MODS_AUTOLOAD));
+				removeSubdirs(curSearchPath->path, versionedModsPath(MODS_CAMPAIGN));
+				removeSubdirs(curSearchPath->path, versionedModsPath(MODS_MULTIPLAY));
+				removeSubdirs(curSearchPath->path, "mods/downloads"); // not versioned
 
 				// Remove multiplay patches
 				sstrcpy(tmpstr, curSearchPath->path);
@@ -316,6 +456,10 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 				WZ_PHYSFS_unmount(tmpstr);
 				curSearchPath = curSearchPath->higherPriority;
 			}
+
+			// This should properly remove all paths, but testing is needed to ensure that all supported versions of PhysFS behave as expected
+			// For now, keep the old code above as well as this new method
+			clearAllPhysFSSearchPaths();
 			break;
 		case mod_campaign:
 			debug(LOG_WZ, "*** Switching to campaign mods ***");
@@ -342,11 +486,10 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 				// Add global and campaign mods
 				PHYSFS_mount(curSearchPath->path, NULL, PHYSFS_APPEND);
 
-				addSubdirs(curSearchPath->path, "mods/music", PHYSFS_APPEND, nullptr, false);
-				addSubdirs(curSearchPath->path, "mods/global", PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
-				addSubdirs(curSearchPath->path, "mods", PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
-				addSubdirs(curSearchPath->path, "mods/autoload", PHYSFS_APPEND, use_override_mods ? &override_mods : nullptr, true);
-				addSubdirs(curSearchPath->path, "mods/campaign", PHYSFS_APPEND, use_override_mods ? &override_mods : &campaign_mods, true);
+				addSubdirs(curSearchPath->path, versionedModsPath(MODS_MUSIC), PHYSFS_APPEND, nullptr, false);
+				addSubdirs(curSearchPath->path, versionedModsPath(MODS_GLOBAL), PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
+				addSubdirs(curSearchPath->path, versionedModsPath(MODS_AUTOLOAD), PHYSFS_APPEND, use_override_mods ? &override_mods : nullptr, true);
+				addSubdirs(curSearchPath->path, versionedModsPath(MODS_CAMPAIGN), PHYSFS_APPEND, use_override_mods ? &override_mods : &campaign_mods, true);
 				if (!WZ_PHYSFS_unmount(curSearchPath->path))
 				{
 					debug(LOG_WZ, "* Failed to remove path %s again", curSearchPath->path);
@@ -381,9 +524,25 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 			// Add the selected map first, for mapmod support
 			if (current_map != nullptr)
 			{
-				WzString realPathAndDir = WzString::fromUtf8(PHYSFS_getRealDir(current_map)) + current_map;
-				realPathAndDir.replace("/", PHYSFS_getDirSeparator()); // Windows fix
-				PHYSFS_mount(realPathAndDir.toUtf8().c_str(), NULL, PHYSFS_APPEND);
+				if (inMemoryMapVirtualFilenameUID.empty() || current_map != inMemoryMapVirtualFilenameUID)
+				{
+					// mount it as a normal physical map path
+					WzString realPathAndDir = WzString::fromUtf8(PHYSFS_getRealDir(current_map)) + current_map;
+					realPathAndDir.replace("/", PHYSFS_getDirSeparator()); // Windows fix
+					PHYSFS_mount(realPathAndDir.toUtf8().c_str(), NULL, PHYSFS_APPEND);
+				}
+				else if (!inMemoryMapArchiveData.empty())
+				{
+					// mount the in-memory map archive as a virtual file
+					if (PHYSFS_mountMemory_fixed(inMemoryMapArchiveData.data(), inMemoryMapArchiveData.size(), clearInMemoryMapFile, inMemoryMapVirtualFilenameUID.c_str(), NULL, PHYSFS_APPEND) != 0)
+					{
+						inMemoryMapArchiveMounted++;
+					}
+				}
+				else
+				{
+					debug(LOG_ERROR, "Specified virtual map file, but no data?");
+				}
 			}
 			curSearchPath = searchPathRegistry;
 			while (curSearchPath->lowerPriority)
@@ -397,13 +556,14 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 #endif // DEBUG
 				// Add global and multiplay mods
 				PHYSFS_mount(curSearchPath->path, NULL, PHYSFS_APPEND);
-				addSubdirs(curSearchPath->path, "mods/music", PHYSFS_APPEND, nullptr, false);
-				if (NetPlay.isHost || !NetPlay.bComms || ingame.bHostSetup)
+				addSubdirs(curSearchPath->path, versionedModsPath(MODS_MUSIC), PHYSFS_APPEND, nullptr, false);
+
+				// Only load if we are host or singleplayer (Initial mod load relies on this, too)
+				if (ingame.side == InGameSide::HOST_OR_SINGLEPLAYER || !NetPlay.bComms)
 				{
-					addSubdirs(curSearchPath->path, "mods/global", PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
-					addSubdirs(curSearchPath->path, "mods", PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
-					addSubdirs(curSearchPath->path, "mods/autoload", PHYSFS_APPEND, use_override_mods ? &override_mods : nullptr, true);
-					addSubdirs(curSearchPath->path, "mods/multiplay", PHYSFS_APPEND, use_override_mods ? &override_mods : &multiplay_mods, true);
+					addSubdirs(curSearchPath->path, versionedModsPath(MODS_GLOBAL), PHYSFS_APPEND, use_override_mods ? &override_mods : &global_mods, true);
+					addSubdirs(curSearchPath->path, versionedModsPath(MODS_AUTOLOAD), PHYSFS_APPEND, use_override_mods ? &override_mods : nullptr, true);
+					addSubdirs(curSearchPath->path, versionedModsPath(MODS_MULTIPLAY), PHYSFS_APPEND, use_override_mods ? &override_mods : &multiplay_mods, true);
 				}
 				else
 				{
@@ -411,7 +571,7 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 					for (Sha256 &hash : game.modHashes)
 					{
 						hashList = {hash.toString()};
-						addSubdirs(curSearchPath->path, "mods/downloads", PHYSFS_APPEND, &hashList, true);
+						addSubdirs(curSearchPath->path, "mods/downloads", PHYSFS_APPEND, &hashList, true); // not versioned
 					}
 				}
 				WZ_PHYSFS_unmount(curSearchPath->path);
@@ -459,6 +619,8 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 #ifdef DEBUG
 		printSearchPath();
 #endif // DEBUG
+
+		ActivityManager::instance().rebuiltSearchPath();
 	}
 	else if (use_override_mods)
 	{
@@ -468,25 +630,35 @@ bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map)
 	return true;
 }
 
-typedef std::vector<std::string> MapFileList;
+struct MapFileListPath
+{
+public:
+	MapFileListPath(const std::string& platformIndependentNotation, const std::string& platformDependentNotation)
+	: platformIndependent(platformIndependentNotation)
+	, platformDependent(platformDependentNotation)
+	{ }
+	std::string platformIndependent;
+	std::string platformDependent;
+};
+typedef std::vector<MapFileListPath> MapFileList;
 static MapFileList listMapFiles()
 {
-	MapFileList ret, filtered, oldSearchPath;
+	MapFileList ret, filtered;
+	std::vector<std::string> oldSearchPath;
 
-	char **subdirlist = PHYSFS_enumerateFiles("maps");
-
-	for (char **i = subdirlist; *i != nullptr; ++i)
-	{
-		std::string wzfile = *i;
-		if (*i[0] == '.' || wzfile.substr(wzfile.find_last_of('.') + 1) != "wz")
+	WZ_PHYSFS_enumerateFiles("maps", [&](const char *i) -> bool {
+		std::string wzfile = i;
+		if (i[0] == '.' || wzfile.substr(wzfile.find_last_of('.') + 1) != "wz")
 		{
-			continue;
+			return true; // continue;
 		}
 
-		std::string realFileName = std::string("maps/") + *i;
-		ret.push_back(realFileName);
-	}
-	PHYSFS_freeList(subdirlist);
+		std::string realFileName_platformIndependent = std::string("maps") + "/" + i;
+		std::string realFileName_platformDependent = std::string("maps") + PHYSFS_getDirSeparator() + i;
+		ret.push_back(MapFileListPath(realFileName_platformIndependent, realFileName_platformDependent));
+		return true; // continue
+	});
+
 	// save our current search path(s)
 	debug(LOG_WZ, "Map search paths:");
 	char **searchPath = PHYSFS_getSearchPath();
@@ -500,31 +672,34 @@ static MapFileList listMapFiles()
 
 	for (const auto &realFileName : ret)
 	{
-		std::string realFilePathAndName = PHYSFS_getWriteDir() + realFileName;
+		std::string realFilePathAndName = PHYSFS_getWriteDir() + realFileName.platformDependent;
 		if (PHYSFS_mount(realFilePathAndName.c_str(), NULL, PHYSFS_APPEND))
 		{
 			int unsafe = 0;
-			char **filelist = PHYSFS_enumerateFiles("multiplay/maps");
-
-			for (char **file = filelist; *file != nullptr; ++file)
-			{
-				std::string isDir = std::string("multiplay/maps/") + *file;
+			bool enumSuccess = WZ_PHYSFS_enumerateFiles("multiplay/maps", [&unsafe, &realFilePathAndName](const char *file) -> bool {
+				std::string isDir = std::string("multiplay/maps/") + file;
 				if (WZ_PHYSFS_isDirectory(isDir.c_str()))
 				{
-					continue;
+					return true; // continue;
 				}
-				std::string checkfile = *file;
-				debug(LOG_WZ, "checking ... %s", *file);
+				std::string checkfile = file;
+				debug(LOG_WZ, "checking ... %s", file);
 				if (checkfile.substr(checkfile.find_last_of('.') + 1) == "gam")
 				{
 					if (unsafe++ > 1)
 					{
 						debug(LOG_ERROR, "Map packs are not supported! %s NOT added.", realFilePathAndName.c_str());
-						break;
+						return false; // break;
 					}
 				}
+				return true; // continue
+			});
+			if (!enumSuccess)
+			{
+				// Failed to enumerate contents - corrupt map archive (ignore it)
+				debug(LOG_ERROR, "Failed to enumerate - ignoring corrupt map file: %s", realFilePathAndName.c_str());
+				unsafe = std::numeric_limits<int>::max() - 1;
 			}
-			PHYSFS_freeList(filelist);
 			if (unsafe < 2)
 			{
 				filtered.push_back(realFileName);
@@ -549,83 +724,220 @@ static MapFileList listMapFiles()
 }
 
 // Map processing
-struct WZmaps
-{
-	std::string MapName;
-	bool isMapMod;
-};
 
-std::vector<struct WZmaps> WZ_Maps;
-
-struct FindMap
+static inline WZMapInfo_Map::iterator findMap(char const *name)
 {
-	const std::string map;
-	FindMap(const std::string &name) : map(name) {};
-	bool operator()(const WZmaps &wzm) const
-	{
-		return wzm.MapName == map;
-	};
-};
+	return name != nullptr? WZ_Maps.find(name) : WZ_Maps.end();
+}
 
-bool CheckForMod(char *theMap)
+bool CheckForMod(char const *mapFile)
 {
-	std::vector<struct WZmaps>::iterator it;
-	if (theMap == nullptr)
-	{
-		return false;
-	}
-	it = std::find_if(WZ_Maps.begin(), WZ_Maps.end(), FindMap(theMap));
+	auto it = findMap(mapFile);
 	if (it != WZ_Maps.end())
 	{
-		return it->isMapMod;
+		return it->second.isMapMod;
+	}
+	if (mapFile != nullptr)
+	{
+		debug(LOG_ERROR, "Couldn't find map %s", mapFile);
+	}
+
+	return false;
+}
+
+bool CheckForRandom(char const *mapFile, char const *mapDataFile0)
+{
+	auto it = findMap(mapFile);
+	if (it != WZ_Maps.end())
+	{
+		return it->second.isRandom;
+	}
+
+	if (mapFile != nullptr) {
+		debug(LOG_ERROR, "Couldn't find map %s", mapFile);
+	}
+	else if (mapDataFile0 != nullptr && strlen(mapDataFile0) > 4)
+	{
+		std::string fn = mapDataFile0;
+		fn.resize(fn.size() - 4);
+		fn += "/game.js";
+		return PHYSFS_exists(fn.c_str());
 	}
-	debug(LOG_ERROR, "Couldn't find map %s", theMap);
 
 	return false;
 }
 
 // Mount the archive under the mountpoint, and enumerate the archive according to lookin
-static bool CheckInMap(const char *archive, const char *mountpoint, const char *lookin)
+static inline optional<WZmapInfo> CheckInMap(const char *archive, const std::string& mountpoint, const std::vector<std::string>& lookin_list)
 {
 	bool mapmod = false;
+	bool isRandom = false;
 
-	if (!PHYSFS_mount(archive, mountpoint, PHYSFS_APPEND))
+	for (auto lookin_subdir : lookin_list)
 	{
-		// We already checked to see if this was valid before, and now, something went seriously wrong.
-		debug(LOG_FATAL, "Could not mount %s, because: %s. Please delete the file, and run the game again. Game will now exit.", archive, WZ_PHYSFS_getLastError());
-		exit(-1);
-	}
-
-	std::string checkpath = lookin;
-	checkpath.append("/");
-	char **filelist = PHYSFS_enumerateFiles(lookin);
-	for (char **file = filelist; *file != nullptr; ++file)
-	{
-		std::string checkfile = *file;
-		if (WZ_PHYSFS_isDirectory((checkpath + checkfile).c_str()))
+		std::string lookin = mountpoint;
+		if (!lookin_subdir.empty())
 		{
-			if (checkfile.compare("wrf") == 0 || checkfile.compare("stats") == 0 || checkfile.compare("components") == 0
-			    || checkfile.compare("effects") == 0 || checkfile.compare("messages") == 0
-			    || checkfile.compare("audio") == 0 || checkfile.compare("sequenceaudio") == 0 || checkfile.compare("misc") == 0
-			    || checkfile.compare("features") == 0 || checkfile.compare("script") == 0 || checkfile.compare("structs") == 0
-			    || checkfile.compare("tileset") == 0 || checkfile.compare("images") == 0 || checkfile.compare("texpages") == 0
-			    || checkfile.compare("skirmish") == 0)
+			lookin.append("/");
+			lookin.append(lookin_subdir);
+		}
+		bool enumResult = WZ_PHYSFS_enumerateFiles(lookin.c_str(), [&](const char *file) -> bool {
+			std::string checkfile = file;
+			if (WZ_PHYSFS_isDirectory((lookin + "/" + checkfile).c_str()))
 			{
-				debug(LOG_WZ, "Detected: %s %s" , archive, checkfile.c_str());
-				mapmod = true;
-				break;
+				if (checkfile.compare("wrf") == 0 || checkfile.compare("stats") == 0 || checkfile.compare("components") == 0
+					|| checkfile.compare("effects") == 0 || checkfile.compare("messages") == 0
+					|| checkfile.compare("audio") == 0 || checkfile.compare("sequenceaudio") == 0 || checkfile.compare("misc") == 0
+					|| checkfile.compare("features") == 0 || checkfile.compare("script") == 0 || checkfile.compare("structs") == 0
+					|| checkfile.compare("tileset") == 0 || checkfile.compare("images") == 0 || checkfile.compare("texpages") == 0
+					|| checkfile.compare("skirmish") == 0 || checkfile.compare("shaders") == 0 || checkfile.compare("fonts") == 0
+					|| checkfile.compare("icons") == 0)
+				{
+					debug(LOG_WZ, "Detected: %s %s" , archive, checkfile.c_str());
+					mapmod = true;
+					return false; // break;
+				}
 			}
+			return true; // continue
+		});
+		if (!enumResult)
+		{
+			// failed to enumerate - just exit out
+			return nullopt;
+		}
+
+		std::string maps = lookin + "/multiplay/maps";
+		enumResult = WZ_PHYSFS_enumerateFiles(maps.c_str(), [&](const char *file) -> bool {
+			if (WZ_PHYSFS_isDirectory((maps + "/" + file).c_str()) && PHYSFS_exists((maps + "/" + file + "/game.js").c_str()))
+			{
+				isRandom = true;
+				return false; // break;
+			}
+			return true; // continue
+		});
+		if (!enumResult)
+		{
+			// failed to enumerate - just exit out
+			return nullopt;
 		}
 	}
-	PHYSFS_freeList(filelist);
 
-	if (!WZ_PHYSFS_unmount(archive))
-	{
-		debug(LOG_ERROR, "Could not unmount %s, %s", archive, WZ_PHYSFS_getLastError());
-	}
-	return mapmod;
+	return WZmapInfo(mapmod, isRandom);
 }
 
+static std::vector<std::string> map_lookin_list = { "", "multiplay" };
+
+// Process a map that has been mounted in the PhysFS virtual filesystem
+//
+// Verifies the index data, determines attributes, and adds to the level loading system so it can be loaded
+//
+// - archive: Directory or archive added to the path, in platform-dependent notation
+// - realFileName_platformIndependent:
+//     For actual map archives, this is the platform independent unique filename + parent path for the map
+//     For "virtual" map archives this is basically a lookup key that should be unique
+// - mountPoint: Location in the interpolated PhysFS tree that this archive was "mounted" (in platform-independent notation)
+bool processMap(const char* archive, const char* realFileName_platformIndependent, const std::string& mountPoint, bool rejectMapMods = false)
+{
+	auto WZmapInfoResult = CheckInMap(archive, mountPoint, map_lookin_list);
+	if (!WZmapInfoResult.has_value())
+	{
+		// failed to enumerate contents
+		return false;
+	}
+	if (rejectMapMods && WZmapInfoResult.value().isMapMod)
+	{
+		debug(LOG_WZ, "Rejecting map mod: %s", archive);
+		return false;
+	}
+
+	bool containsMap = false;
+	bool enumSuccess = WZ_PHYSFS_enumerateFiles(mountPoint.c_str(), [&](const char *file) -> bool {
+		size_t len = strlen(file);
+		if ((len > 10 && !strcasecmp(file + (len - 10), ".addon.lev"))  // Do not add addon.lev again // <--- Err, what? The code has loaded .addon.lev for a while...
+			|| (len > 13 && !strcasecmp(file + (len - 13), ".xplayers.lev"))) // add support for X player maps using a new name to prevent conflicts.
+		{
+			std::string fullPath = mountPoint + "/" + file;
+			if (loadLevFile(mountPoint + "/" + file, mod_multiplay, true, realFileName_platformIndependent))
+			{
+				containsMap = true;
+				return false; // stop enumerating
+			}
+		}
+		return true; // continue
+	});
+
+	if (!containsMap)
+	{
+		// not sure what this is, but it doesn't seem to contain a map
+		return false;
+	}
+
+	if (!enumSuccess)
+	{
+		// Failed to enumerate contents - corrupt map archive
+		return false;
+	}
+
+	std::string MapName = realFileName_platformIndependent;
+	WZ_Maps.insert(WZMapInfo_Map::value_type(MapName, WZmapInfoResult.value()));
+
+	return true;
+}
+
+#if defined(HAS_PHYSFS_IO_SUPPORT)
+bool setSpecialInMemoryMap(std::vector<uint8_t>&& mapArchiveData)
+{
+	ASSERT_OR_RETURN(false, !mapArchiveData.empty(), "Null map archive data passed?");
+	ASSERT_OR_RETURN(false, inMemoryMapArchiveMounted == 0, "In-memory map archive already mounted");
+
+	// calculate a hash for the map data
+	Sha256 mapHash = sha256Sum(mapArchiveData.data(), mapArchiveData.size());
+
+	// generate a new unique filename UID for the in-memory map
+	inMemoryMapVirtualFilenameUID = "<in-memory>::mapArchive::" + mapHash.toString();
+
+	// store the map archive data
+	inMemoryMapArchiveData = std::move(mapArchiveData);
+
+	// try to mount the in-memory archive
+	if (PHYSFS_mountMemory_fixed(inMemoryMapArchiveData.data(), inMemoryMapArchiveData.size(), nullptr, inMemoryMapVirtualFilenameUID.c_str(), "WZMap", PHYSFS_APPEND) == 0)
+	{
+		// Failed to mount data - corrupt map archive
+		debug(LOG_ERROR, "Failed to mount - corrupt / invalid map file: %s", inMemoryMapVirtualFilenameUID.c_str());
+		inMemoryMapVirtualFilenameUID.clear();
+		inMemoryMapArchiveData.clear();
+		return false;
+	}
+
+	// load it into the level-loading system
+	if (!processMap(inMemoryMapVirtualFilenameUID.c_str(), inMemoryMapVirtualFilenameUID.c_str(), "WZMap", true))
+	{
+		// Failed to enumerate contents - corrupt map archive
+		debug(LOG_ERROR, "Failed to enumerate - corrupt / invalid map file: %s", inMemoryMapVirtualFilenameUID.c_str());
+		inMemoryMapVirtualFilenameUID.clear();
+		inMemoryMapArchiveData.clear();
+		return false;
+	}
+
+	if (WZ_PHYSFS_unmount(inMemoryMapVirtualFilenameUID.c_str()) == 0)
+	{
+		debug(LOG_ERROR, "Could not unmount %s, %s", inMemoryMapVirtualFilenameUID.c_str(), WZ_PHYSFS_getLastError());
+	}
+
+	// fix-up level hash
+	levSetFileHashByRealFileName(inMemoryMapVirtualFilenameUID.c_str(), mapHash);
+
+	return true;
+}
+#else
+bool setSpecialInMemoryMap(std::vector<uint8_t>&& mapArchiveData)
+{
+	// Sadly, the version of PhysFS used for compilation is too old
+	debug(LOG_INFO, "The version of PhysFS used for compilation is too old, and does not support PHYSFS_Io");
+	return false;
+}
+#endif
+
 bool buildMapList()
 {
 	if (!loadLevFile("gamedesc.lev", mod_campaign, false, nullptr))
@@ -634,46 +946,38 @@ bool buildMapList()
 	}
 	loadLevFile("addon.lev", mod_multiplay, false, nullptr);
 	WZ_Maps.clear();
+	if (!inMemoryMapArchiveMounted && !inMemoryMapArchiveData.empty())
+	{
+		debug(LOG_INFO, "Clearing in-memory map archive (since it isn't currently loaded)");
+		clearInMemoryMapFile(inMemoryMapArchiveData.data());
+	}
 	MapFileList realFileNames = listMapFiles();
 	for (auto &realFileName : realFileNames)
 	{
-		bool mapmod = false;
-		struct WZmaps CurrentMap;
-		std::string realFilePathAndName = PHYSFS_getRealDir(realFileName.c_str()) + realFileName;
-
-		PHYSFS_mount(realFilePathAndName.c_str(), NULL, PHYSFS_APPEND);
-
-		char **filelist = PHYSFS_enumerateFiles("");
-		for (char **file = filelist; *file != nullptr; ++file)
+		const char * pRealDirStr = PHYSFS_getRealDir(realFileName.platformIndependent.c_str());
+		if (!pRealDirStr)
 		{
-			std::string checkfile = *file;
-			size_t len = strlen(*file);
-			if (len > 10 && !strcasecmp(*file + (len - 10), ".addon.lev"))  // Do not add addon.lev again
-			{
-				loadLevFile(*file, mod_multiplay, true, realFileName.c_str());
-			}
-			// add support for X player maps using a new name to prevent conflicts.
-			if (len > 13 && !strcasecmp(*file + (len - 13), ".xplayers.lev"))
-			{
-				loadLevFile(*file, mod_multiplay, true, realFileName.c_str());
-			}
+			debug(LOG_ERROR, "Failed to find realdir for: %s", realFileName.platformIndependent.c_str());
+			continue; // skip
+		}
+		std::string realFilePathAndName = pRealDirStr + realFileName.platformDependent;
+
+		if (PHYSFS_mount(realFilePathAndName.c_str(), "WZMap", PHYSFS_APPEND) == 0)
+		{
+			debug(LOG_POPUP, "Could not mount %s, because: %s.\nPlease delete or move the file specified.", realFilePathAndName.c_str(), WZ_PHYSFS_getLastError());
+			continue; // skip
+		}
+
+		if (!processMap(realFilePathAndName.c_str(), realFileName.platformIndependent.c_str(), "WZMap"))
+		{
+			// Failed to enumerate contents - corrupt map archive
+			debug(LOG_ERROR, "Failed to enumerate - corrupt / invalid map file: %s", realFilePathAndName.c_str());
 		}
-		PHYSFS_freeList(filelist);
 
 		if (WZ_PHYSFS_unmount(realFilePathAndName.c_str()) == 0)
 		{
 			debug(LOG_ERROR, "Could not unmount %s, %s", realFilePathAndName.c_str(), WZ_PHYSFS_getLastError());
 		}
-
-		mapmod = CheckInMap(realFilePathAndName.c_str(), "WZMap", "WZMap");
-		if (!mapmod)
-		{
-			mapmod = CheckInMap(realFilePathAndName.c_str(), "WZMap", "WZMap/multiplay");
-		}
-
-		CurrentMap.MapName = realFileName;
-		CurrentMap.isMapMod = mapmod;
-		WZ_Maps.push_back(CurrentMap);
 	}
 
 	return true;
@@ -690,6 +994,11 @@ bool systemInitialise(float horizScaleFactor, float vertScaleFactor)
 		return false;
 	}
 
+	if (!notificationsInitialize())
+	{
+		return false;
+	}
+
 	buildMapList();
 
 	// Initialize render engine
@@ -699,7 +1008,7 @@ bool systemInitialise(float horizScaleFactor, float vertScaleFactor)
 		return false;
 	}
 
-	if (!audio_Init(droidAudioTrackStopped, war_getSoundEnabled()))
+	if (!audio_Init(droidAudioTrackStopped, war_GetHRTFMode(), war_getSoundEnabled()))
 	{
 		debug(LOG_SOUND, "Continuing without audio");
 	}
@@ -734,12 +1043,20 @@ bool systemInitialise(float horizScaleFactor, float vertScaleFactor)
 //
 void systemShutdown()
 {
+	if (bLoadSaveUp)
+	{
+		closeLoadSaveOnShutdown(); // TODO: Ideally this would not be required here (refactor loadsave.cpp / frontend.cpp?)
+	}
+
+	seqReleaseAll();
+
 	pie_ShutdownRadar();
 	clearLoadedMods();
+	shutdownConsoleMessages();
 
 	shutdownEffectsSystem();
 	wzSceneEnd(nullptr);  // Might want to end the "Main menu loop" or "Main game loop".
-	keyMappings.clear();
+	gInputManager.shutdown();
 
 	// free up all the load functions (all the data should already have been freed)
 	resReleaseAll();
@@ -750,6 +1067,9 @@ void systemShutdown()
 		abort();
 	}
 
+	// shut down various databases
+	shutdownKnownPlayers();
+
 	debug(LOG_MAIN, "shutting down audio subsystems");
 
 	debug(LOG_MAIN, "shutting down CD audio");
@@ -763,6 +1083,7 @@ void systemShutdown()
 
 	debug(LOG_MAIN, "shutting down graphics subsystem");
 	levShutDown();
+	notificationsShutDown();
 	widgShutDown();
 	fpathShutdown();
 	mapShutdown();
@@ -770,6 +1091,7 @@ void systemShutdown()
 	pal_ShutDown();		// currently unused stub
 	frameShutDown();	// close screen / SDL / resources / cursors / trig
 	screenShutDown();
+	gfx_api::context::get().shutdown();
 	cleanSearchPath();	// clean PHYSFS search paths
 	debug_exit();		// cleanup debug routines
 	PHYSFS_deinit();	// cleanup PHYSFS (If failure, state of PhysFS is undefined, and probably badly screwed up.)
@@ -791,11 +1113,6 @@ bool frontendInitialise(const char *ResourceFile)
 		return false;
 	}
 
-	if (!scrTabInitialise())				// Initialise the script system
-	{
-		return false;
-	}
-
 	if (!stringsInitialise())				// Initialise the string system
 	{
 		return false;
@@ -824,6 +1141,18 @@ bool frontendInitialise(const char *ResourceFile)
 	}
 
 	FrontImages = (IMAGEFILE *)resGetData("IMG", "frontend.img");
+	if (FrontImages == nullptr)
+	{
+		std::string errorMessage = astringf(_("Unable to load: %s."), "frontend.img");
+		if (!getLoadedMods().empty())
+		{
+			errorMessage += " ";
+			errorMessage += _("Please remove all incompatible mods.");
+		}
+		debug(LOG_FATAL, "%s", errorMessage.c_str());
+		return false;
+	}
+
 	/* Shift the interface initialisation here temporarily so that it
 		can pick up the stats after they have been loaded */
 	if (!intInitialise())
@@ -832,7 +1161,7 @@ bool frontendInitialise(const char *ResourceFile)
 	}
 
 	// reinitialise key mappings
-	keyInitMappings(false);
+	gInputManager.resetMappings(false, gKeyFuncConfig);
 
 	// Set the default uncoloured cursor here, since it looks slightly
 	// better for menus and such.
@@ -862,8 +1191,16 @@ bool frontendShutdown()
 		return false;
 	}
 
+	changeTitleUI(nullptr);
+	if (challengesUp)
+	{
+		closeChallenges(); // TODO: Ideally this would not be required here (refactor challenge.cpp / frontend.cpp?)
+	}
+	if (bLoadSaveUp)
+	{
+		closeLoadSaveOnShutdown(); // TODO: Ideally this would not be required here (refactor loadsave.cpp / frontend.cpp?)
+	}
 	interfaceShutDown();
-	scrShutDown();
 
 	//do this before shutting down the iV library
 	resReleaseAllData();
@@ -949,11 +1286,6 @@ bool stageOneInitialise()
 		return false;
 	}
 
-	if (!scrTabInitialise())	// Initialise the old wzscript system
-	{
-		return false;
-	}
-
 	if (!gridInitialise())
 	{
 		return false;
@@ -963,11 +1295,9 @@ bool stageOneInitialise()
 	initTransporters();
 	scriptInit();
 
-	// do this here so that the very first mission has it initialised
-	initRunData();
-
 	gameTimeInit();
-	eventTimeReset(gameTime / SCR_TICKRATE);
+	transitionInit();
+	resetScroll();
 
 	return true;
 }
@@ -1018,7 +1348,6 @@ bool stageOneShutDown()
 		return false;
 	}
 
-	scrShutDown();
 	gridShutDown();
 
 	debug(LOG_TEXTURE, "== stageOneShutDown ==");
@@ -1097,7 +1426,13 @@ bool stageTwoInitialise()
 	}
 
 	// reinitialise key mappings
-	keyInitMappings(false);
+	gInputManager.resetMappings(false, gKeyFuncConfig);
+
+	if (NETisReplay())
+	{
+		// Debug keybinds start enabled for replays
+		gInputManager.contexts().set(InputContext::DEBUG_MISC, InputContext::State::ACTIVE);
+	}
 
 	// Set the default uncoloured cursor here, since it looks slightly
 	// better for menus and such.
@@ -1109,13 +1444,16 @@ bool stageTwoInitialise()
 	// Don't ask why this doesn't go in stage three. In fact, don't even ask me what stage one/two/three is supposed to mean, it seems about as descriptive as stage doStuff, stage doMoreStuff and stage doEvenMoreStuff...
 	debug(LOG_MAIN, "Init game queues, I am %d.", selectedPlayer);
 	sendQueuedDroidInfo();  // Discard any pending orders which could later get flushed into the game queue.
-	for (i = 0; i < MAX_PLAYERS; ++i)
+	if (!NETisReplay())
 	{
-		NETinitQueue(NETgameQueue(i));
-
-		if (!myResponsibility(i))
+		for (i = 0; i < MAX_CONNECTED_PLAYERS; ++i)
 		{
-			NETsetNoSendOverNetwork(NETgameQueue(i));
+			NETinitQueue(NETgameQueue(i));
+
+			if (!myResponsibility(i))
+			{
+				NETsetNoSendOverNetwork(NETgameQueue(i));
+			}
 		}
 	}
 
@@ -1178,6 +1516,8 @@ bool stageTwoShutDown()
 		return false;
 	}
 
+	shutdown3DView();
+
 	return true;
 }
 
@@ -1220,10 +1560,7 @@ bool stageThreeInitialise()
 
 	if (bMultiPlayer)
 	{
-		if (!fromSave)
-		{
-			multiGameInit();
-		}
+		multiGameInit();
 		initTemplates();
 	}
 
@@ -1247,7 +1584,13 @@ bool stageThreeInitialise()
 
 	// Re-inititialise some static variables.
 
-	playerBuiltHQ = false;
+	bInTutorial = false;
+	rangeOnScreen = false;
+
+	if (fromSave && ActivityManager::instance().getCurrentGameMode() == ActivitySink::GameMode::CHALLENGE)
+	{
+		challengeActive = true;
+	}
 
 	resizeRadar();
 
@@ -1266,7 +1609,7 @@ bool stageThreeInitialise()
 				/* Structures */
 				for (psStr = apsStructLists[i]; psStr; psStr = psStr->psNext)
 				{
-					if (aiCheckAlliances(psStr->player, selectedPlayer))
+					if (selectedPlayer < MAX_PLAYERS && aiCheckAlliances(psStr->player, selectedPlayer))
 					{
 						visTilesUpdate((BASE_OBJECT *)psStr);
 					}
@@ -1275,7 +1618,7 @@ bool stageThreeInitialise()
 				/* Droids */
 				for (psDroid = apsDroidLists[i]; psDroid; psDroid = psDroid->psNext)
 				{
-					if (aiCheckAlliances(psDroid->player, selectedPlayer))
+					if (selectedPlayer < MAX_PLAYERS && aiCheckAlliances(psDroid->player, selectedPlayer))
 					{
 						visTilesUpdate((BASE_OBJECT *)psDroid);
 					}
@@ -1285,16 +1628,25 @@ bool stageThreeInitialise()
 	}
 
 	countUpdate();
-	if (getLevelLoadType() != GTYPE_SAVE_MIDMISSION)
+
+	if (getLevelLoadType() == GTYPE_SAVE_MIDMISSION || getLevelLoadType() == GTYPE_SAVE_START)
 	{
-		if (getDebugMappingStatus())
+		triggerEvent(TRIGGER_GAME_LOADED);
+	}
+	else
+	{
+		const DebugInputManager& dbgInputManager = gInputManager.debugManager();
+		if (dbgInputManager.debugMappingsAllowed())
 		{
 			triggerEventCheatMode(true);
 		}
-		eventFireCallbackTrigger((TRIGGER_TYPE)CALL_GAMEINIT);
 		triggerEvent(TRIGGER_GAME_INIT);
+		playerBuiltHQ = structureExists(selectedPlayer, REF_HQ, true, false);
 	}
 
+	// Start / randomize in-game music
+	cdAudio_PlayTrack(SONG_INGAME);
+
 	return true;
 }
 
@@ -1305,7 +1657,7 @@ bool stageThreeShutDown()
 {
 	debug(LOG_WZ, "== stageThreeShutDown ==");
 
-	hostlaunch = 0;
+	setHostLaunch(HostLaunch::Normal);
 
 	removeSpotters();
 
@@ -1316,9 +1668,13 @@ bool stageThreeShutDown()
 		return false;
 	}
 
+	specStatsViewShutdown();
+
 	challengesUp = false;
 	challengeActive = false;
 	isInGamePopupUp = false;
+	InGameOpUp = false;
+	bInTutorial = false;
 
 	shutdownTemplates();
 
@@ -1332,24 +1688,12 @@ bool stageThreeShutDown()
 		multiGameShutdown();
 	}
 
-	eventReset();
-
-	// reset the script values system
-	scrvReset();
-
 	//call this here before mission data is released
 	if (!missionShutDown())
 	{
 		return false;
 	}
 
-	scrExternReset();
-
-	// reset the run data so that doesn't need to be initialised in the scripts
-	initRunData();
-
-	resetVTOLLandingPos();
-
 	setScriptWinLoseVideo(PLAY_NONE);
 
 	return true;
@@ -1386,7 +1730,6 @@ bool saveGameReset()
 	//free up the gateway stuff?
 	gwShutDown();
 	intResetScreen(true);
-	intResetPreviousObj();
 
 	if (!mapShutdown())
 	{
@@ -1411,8 +1754,9 @@ static void initMiscVars()
 	includeRedundantDesigns = false;
 	enableConsoleDisplay(true);
 
+	DebugInputManager& dbgInputManager = gInputManager.debugManager();
 	for (unsigned n = 0; n < MAX_PLAYERS; ++n)
 	{
-		processDebugMappings(n, false);
+		dbgInputManager.setPlayerWantsDebugMappings(n, false);
 	}
 }
