@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -24,6 +24,8 @@
 #ifndef __INCLUDED_SRC_INIT_H__
 #define __INCLUDED_SRC_INIT_H__
 
+#include <vector>
+
 struct IMAGEFILE;
 
 // the size of the file loading buffer
@@ -60,9 +62,11 @@ void registerSearchPath(const char path[], unsigned int priority);
 bool rebuildSearchPath(searchPathMode mode, bool force, const char *current_map = NULL);
 
 bool buildMapList();
-bool CheckForMod(char *theMap);
+bool CheckForMod(char const *mapFile);
+bool CheckForRandom(char const *mapFile, char const *mapDataFile0);
+bool setSpecialInMemoryMap(std::vector<uint8_t>&& mapArchiveData);
 
-bool loadLevFile(const char *filename, searchPathMode datadir, bool ignoreWrf, char const *realFileName);
+bool loadLevFile(const std::string& filename, searchPathMode datadir, bool ignoreWrf, char const *realFileName);
 
 extern IMAGEFILE	*FrontImages;
 
