@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -70,6 +70,7 @@
 #include "mission.h"
 
 #include "multiplay.h"
+#include "qtscript.h"
 
 // Is a button widget highlighted, either because the cursor is over it or it is flashing.
 //
@@ -92,9 +93,6 @@ static int FormCloseCount;	// Count used to ensure only one sfx played when two
 
 static UDWORD ManuPower = 0;	// Power required to manufacture the current item.
 
-// Get the first factory assigned to a command droid
-static STRUCTURE *droidGetCommandFactory(DROID *psDroid);
-
 // Set audio IDs for form opening/closing anims.
 // Use -1 to dissable audio.
 //
@@ -108,6 +106,7 @@ void SetFormAudioIDs(int OpenID, int CloseID)
 
 static void setBarGraphValue(W_BARGRAPH *barGraph, PIELIGHT colour, int value, int range)
 {
+	ASSERT_OR_RETURN(, range != 0, "range is 0");
 	barGraph->majorCol = colour;
 	barGraph->majorSize = PERNUM(WBAR_SCALE, clip(value, 0, range), range);
 	barGraph->show();
@@ -127,7 +126,7 @@ static void formatTimeText(W_BARGRAPH *barGraph, int time)
 	barGraph->textCol = WZCOL_CONSTRUCTION_BARTEXT;
 }
 
-static void formatTime(W_BARGRAPH *barGraph, int buildPointsDone, int buildPointsTotal, int buildRate, char const *toolTip)
+void formatTime(W_BARGRAPH *barGraph, int buildPointsDone, int buildPointsTotal, int buildRate, char const *toolTip)
 {
 	barGraph->setTip(toolTip);
 
@@ -153,8 +152,9 @@ static void formatPowerText(W_BARGRAPH *barGraph, int neededPower)
 	barGraph->textCol = WZCOL_POWERQUEUE_BARTEXT;
 }
 
-static void formatPower(W_BARGRAPH *barGraph, int neededPower, int powerToBuild)
+void formatPower(W_BARGRAPH *barGraph, int neededPower, int powerToBuild)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(, barGraph);
 	if (neededPower == -1 || powerToBuild == 0)
 	{
 		formatEmpty(barGraph);
@@ -166,396 +166,15 @@ static void formatPower(W_BARGRAPH *barGraph, int neededPower, int powerToBuild)
 	setBarGraphValue(barGraph, WZCOL_GREEN, powerToBuild - neededPower, powerToBuild);
 }
 
-// Widget callback to update the progress bar in the object stats screen.
-//
-void intUpdateProgressBar(WIDGET *psWidget, W_CONTEXT *psContext)
+std::string PowerBar::getTip()
 {
-	BASE_OBJECT			*psObj;
-	DROID				*Droid;
-	STRUCTURE			*Structure;
-	FACTORY				*Manufacture;
-	RESEARCH_FACILITY	*Research;
-	W_BARGRAPH			*BarGraph = (W_BARGRAPH *)psWidget;
-
-	psObj = (BASE_OBJECT *)BarGraph->pUserData;	// Get the object associated with this widget.
-
-	if (psObj == nullptr)
-	{
-		BarGraph->hide();
-		return;
-	}
-
-	if (isDead((BASE_OBJECT *)psObj))
-	{
-		return;
-	}
-
-	BarGraph->majorSize = 0;
-	BarGraph->hide();
-
-	switch (psObj->type)
-	{
-	case OBJ_DROID:						// If it's a droid and...
-		Droid = (DROID *)psObj;
-
-		if (DroidIsBuilding(Droid)) // Is it a building.
-		{
-			ASSERT(Droid->asBits[COMP_CONSTRUCT], "Invalid droid type");
-
-			Structure = DroidGetBuildStructure(Droid);  // Get the structure's building.
-
-			if (Structure)
-			{
-				//show progress of build
-				if (Structure->currentBuildPts != 0)
-				{
-					formatTime(BarGraph, Structure->currentBuildPts, Structure->pStructureType->buildPoints, Structure->lastBuildRate, _("Build Progress"));
-				}
-				else
-				{
-					formatPower(BarGraph, checkPowerRequest(Structure), Structure->pStructureType->powerToBuild);
-				}
-			}
-		}
-		break;
-
-	case OBJ_STRUCTURE:					// If it's a structure and...
-		Structure = (STRUCTURE *)psObj;
-
-		if (StructIsFactory(Structure) && StructureIsManufacturingPending(Structure))  // Is it manufacturing.
-		{
-			Manufacture = StructureGetFactory(Structure);
-
-			if (Manufacture->psSubject != nullptr && Manufacture->buildPointsRemaining < calcTemplateBuild(Manufacture->psSubject))
-			{
-				// Started production. Set the colour of the bar to yellow.
-				int buildPointsTotal = calcTemplateBuild(FactoryGetTemplate(Manufacture));
-				int buildRate = Manufacture->timeStartHold == 0 ? getBuildingProductionPoints(Structure) : 0;
-				formatTime(BarGraph, buildPointsTotal - Manufacture->buildPointsRemaining, buildPointsTotal, buildRate, _("Construction Progress"));
-			}
-			else
-			{
-				// Not yet started production.
-				int neededPower = checkPowerRequest(Structure);
-				int powerToBuild = Manufacture->psSubject ? calcTemplatePower(Manufacture->psSubject) : 0;
-				formatPower(BarGraph, neededPower, powerToBuild);
-			}
-		}
-		else if (structureIsResearchingPending(Structure)) // Is it researching.
-		{
-			Research = StructureGetResearch(Structure);
-			unsigned currentPoints = 0;
-			if (Research->psSubject != nullptr)
-			{
-				currentPoints = asPlayerResList[selectedPlayer][Research->psSubject->index].currentPoints;
-			}
-			if (currentPoints != 0)
-			{
-				int researchRate = Research->timeStartHold == 0 ? getBuildingResearchPoints(Structure) : 0;
-				formatTime(BarGraph, currentPoints, Research->psSubject->researchPoints, researchRate, _("Research Progress"));
-			}
-			else
-			{
-				// Not yet started production.
-				int neededPower = checkPowerRequest(Structure);
-				int powerToBuild = Research->psSubject != nullptr ? Research->psSubject->researchPower : 0;
-				formatPower(BarGraph, neededPower, powerToBuild);
-			}
-		}
-
-		break;
-
-	default:
-		ASSERT(false, "Invalid object type");
-	}
+	auto income = getApproxPowerGeneratedPerSecForDisplay(selectedPlayer);
+	return astringf("%s\n%s: %s", _("Power"), _("Power Per Second"), income.c_str());
 }
 
-
-void intUpdateQuantity(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL *Label = (W_LABEL *)psWidget;
-	BASE_OBJECT *psObj = (BASE_OBJECT *)Label->pUserData; // Get the object associated with this widget.
-
-	int remaining = -1;
-
-	if (STRUCTURE *psStruct = castStructure(psObj))
-	{
-		if (StructIsFactory(psStruct) && StructureIsManufacturingPending(psStruct))
-		{
-			ASSERT(!isDead(psObj), "Object is dead");
-
-			DROID_TEMPLATE *psTemplate = FactoryGetTemplate(StructureGetFactory(psStruct));
-			remaining = getProduction(psStruct, psTemplate).numRemaining();
-		}
-	}
-	else if (DROID *psDroid = castDroid(psObj))
-	{
-		STRUCTURE_STATS const *stats = nullptr;
-		int count = 0;
-		auto processOrder = [&](DroidOrder const &order) {
-			STRUCTURE_STATS *newStats = nullptr;
-			int deltaCount = 0;
-			switch (order.type)
-			{
-				case DORDER_BUILD:
-				case DORDER_LINEBUILD:
-					newStats = order.psStats;
-					deltaCount = order.type == DORDER_LINEBUILD? 1 + (abs(order.pos.x - order.pos2.x) + abs(order.pos.y - order.pos2.y))/TILE_UNITS : 1;
-					break;
-				case DORDER_HELPBUILD:
-					if (STRUCTURE *target = castStructure(order.psObj))
-					{
-						newStats = target->pStructureType;
-						deltaCount = 1;
-					}
-					break;
-				default:
-					return false;
-			}
-			if (newStats != nullptr && (stats == nullptr || stats == newStats))
-			{
-				stats = newStats;
-				count += deltaCount;
-				return true;
-			}
-			return false;
-		};
-		if (processOrder(psDroid->order))
-		{
-			for (auto const &order : psDroid->asOrderList)
-			{
-				if (!processOrder(order))
-				{
-					break;
-				}
-			}
-		}
-		if (count > 1)
-		{
-			remaining = count;
-		}
-	}
-
-	if (remaining != -1)
-	{
-		char tmp[20];
-		ssprintf(tmp, "%d", remaining);
-		Label->aText = WzString::fromUtf8(tmp);
-		Label->show();
-	}
-	else
-	{
-		Label->hide();
-	}
-}
-
-//callback to display the factory number
-void intAddFactoryInc(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL		*Label = (W_LABEL *)psWidget;
-	BASE_OBJECT	*psObj = (BASE_OBJECT *)Label->pUserData;
-
-	// Get the object associated with this widget.
-	if (psObj != nullptr && !isDead(psObj))
-	{
-		STRUCTURE	*Structure = (STRUCTURE *)psObj;
-		if (StructIsFactory(Structure))
-		{
-			FACTORY		*Factory = &Structure->pFunctionality->factory;
-			char tmp[20];
-			ssprintf(tmp, "%u", Factory->psAssemblyPoint->factoryInc + 1);
-			Label->aText = WzString::fromUtf8(tmp);
-			Label->show();
-			return;
-		}
-	}
-	Label->aText.clear();
-	Label->hide();
-}
-
-//callback to display the production quantity number for a template
-void intAddProdQuantity(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL				*Label = (W_LABEL *)psWidget;
-	DROID_TEMPLATE                 *psTemplate = (DROID_TEMPLATE *)Label->pUserData;
-
-	// Get the object associated with this widget.
-	if (psTemplate != nullptr)
-	{
-		STRUCTURE	*psStructure = nullptr;
-		BASE_OBJECT	*psObj = getCurrentSelected();
-
-		if (psObj != nullptr && psObj->type == OBJ_STRUCTURE && !isDead(psObj))
-		{
-			psStructure = (STRUCTURE *)psObj;
-		}
-
-		ProductionRunEntry entry;
-		if (psStructure != nullptr && StructIsFactory(psStructure))
-		{
-			entry = getProduction(psStructure, psTemplate);
-		}
-
-		// now find out how many we have built
-		if (entry.isValid())
-		{
-			char tmp[40];
-			if (psStructure->pFunctionality->factory.productionLoops != 0)
-			{
-				ssprintf(tmp, "%u/%u", entry.numRemaining(), entry.quantity);
-			}
-			else
-			{
-				ssprintf(tmp, "%u", entry.numRemaining());
-			}
-			Label->aText = WzString::fromUtf8(tmp);
-			Label->show();
-		}
-		else
-		{
-			Label->aText.clear();
-			Label->hide();
-		}
-	}
-}
-
-//callback to display the production loop quantity number for a factory
-void intAddLoopQuantity(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL		*Label = (W_LABEL *)psWidget;
-	STRUCTURE	*psStruct = (STRUCTURE *)Label->pUserData;
-
-	//loop depends on the factory
-	if (psStruct && psStruct->pFunctionality && !psStruct->died)
-	{
-		FACTORY *psFactory = &psStruct->pFunctionality->factory;
-
-		if (psFactory->productionLoops == INFINITE_PRODUCTION)
-		{
-			Label->aText = WzString::fromUtf8("∞");
-		}
-		else if (psFactory->productionLoops != 0)
-		{
-			char tmp[20];
-			ssprintf(tmp, "%u", psFactory->productionLoops + DEFAULT_LOOP);
-			Label->aText = WzString::fromUtf8(tmp);
-		}
-		else
-		{
-			Label->aText.clear();  // Don't show "1" loop.
-		}
-		Label->show();
-	}
-	else
-	{
-		//hide the label if no factory
-		Label->aText.clear();
-		Label->hide();
-	}
-}
-
-// callback to update the command droid size label
-void intUpdateCommandSize(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL				*Label = (W_LABEL *)psWidget;
-	BASE_OBJECT			*psObj = (BASE_OBJECT *)Label->pUserData;
-
-	// Get the object associated with this widget.
-	if (psObj != nullptr && !isDead(psObj))
-	{
-		DROID	*psDroid = (DROID *)psObj;
-
-		ASSERT(psDroid->droidType == DROID_COMMAND,
-		       "Droid is not a command droid");
-
-		char tmp[40];
-		ssprintf(tmp, "%u/%u", psDroid->psGroup ? psDroid->psGroup->getNumMembers() : 0, cmdDroidMaxGroup(psDroid));
-		Label->aText = WzString::fromUtf8(tmp);
-		Label->show();
-	}
-	else
-	{
-		Label->aText.clear();
-		Label->hide();
-	}
-}
-
-// callback to update the command droid experience
-void intUpdateCommandExp(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL				*Label = (W_LABEL *)psWidget;
-	BASE_OBJECT			*psObj = (BASE_OBJECT *)Label->pUserData;
-
-	// Get the object associated with this widget.
-	if (psObj != nullptr && !isDead(psObj))
-	{
-		DROID	*psDroid = (DROID *)psObj;
-
-		ASSERT(psObj->type == OBJ_DROID, "Invalid droid pointer");
-		ASSERT(psDroid->droidType == DROID_COMMAND, "Droid is not a command droid");
-
-		int numStars = std::max((int)getDroidLevel(psDroid) - 1, 0);
-		Label->aText = WzString(numStars, WzUniCodepoint::fromASCII('*'));
-		Label->show();
-	}
-	else
-	{
-		Label->aText.clear();
-		Label->hide();
-	}
-}
-
-// callback to update the command droid factories
-void intUpdateCommandFact(WIDGET *psWidget, W_CONTEXT *psContext)
-{
-	W_LABEL				*Label = (W_LABEL *)psWidget;
-	BASE_OBJECT			*psObj = (BASE_OBJECT *)Label->pUserData;
-	SDWORD                          i, start;
-
-	// Get the object associated with this widget.
-	if (psObj != nullptr && !isDead(psObj))
-	{
-		DROID		*psDroid = (DROID *)psObj;
-
-		ASSERT(psObj->type == OBJ_DROID, "Invalid droid pointer");
-		ASSERT(psDroid->droidType == DROID_COMMAND, "Droid is not a command droid");
-
-		// see which type of factory this is for
-		if (Label->id >= IDOBJ_COUNTSTART && Label->id < IDOBJ_COUNTEND)
-		{
-			start = DSS_ASSPROD_SHIFT;
-		}
-		else if (Label->id >= IDOBJ_CMDFACSTART && Label->id < IDOBJ_CMDFACEND)
-		{
-			start = DSS_ASSPROD_CYBORG_SHIFT;
-		}
-		else
-		{
-			start = DSS_ASSPROD_VTOL_SHIFT;
-		}
-
-		Label->aText.clear();
-		for (i = 0; i < 5; ++i)  // TODO Support up to MAX_FACTORY (which won't fit in the ugly secondaryOrder bitmask hack).
-		{
-			if (psDroid->secondaryOrder & (1 << (i + start)))
-			{
-				Label->aText.append(WzUniCodepoint::fromASCII((char)('0' + i + 1)));
-			}
-		}
-		Label->show();
-	}
-	else
-	{
-		Label->aText.clear();
-		Label->hide();
-	}
-}
-
-// Widget callback to update and display the power bar.
 // !!!!!!!!!!!!!!!!!!!!!!ONLY WORKS ON A SIDEWAYS POWERBAR!!!!!!!!!!!!!!!!!
-void intDisplayPowerBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
+void PowerBar::display(int xOffset, int yOffset)
 {
-	W_BARGRAPH *BarGraph = (W_BARGRAPH *)psWidget;
 	SDWORD		Avail, ManPow, realPower;
 	SDWORD		Empty;
 	SDWORD		BarWidth, textWidth = 0;
@@ -569,16 +188,13 @@ void intDisplayPowerBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 	lastRealTime = realTime;
 
 	ManPow = ManuPower / POWERBAR_SCALE;
-	Avail = (displayPower + 1e-8) / POWERBAR_SCALE;
-	realPower = (displayPower + 1e-8) - ManuPower;
+	Avail = static_cast<SDWORD>((displayPower + 1e-8) / POWERBAR_SCALE);
+	realPower = static_cast<SDWORD>((displayPower + 1e-8) - ManuPower);
+	ManuPower = 0;
 
-	BarWidth = BarGraph->width();
+	BarWidth = this->width();
 	sprintf(szVal, "%d", realPower);
 
-	// Any widget using intDisplayPowerBar must have its pUserData initialized to a (DisplayPowerBarCache*)
-	assert(psWidget->pUserData != nullptr);
-	DisplayPowerBarCache& cache = *static_cast<DisplayPowerBarCache*>(psWidget->pUserData);
-
 	cache.wzText.setText(szVal, font_regular);
 
 	textWidth = cache.wzText.width();
@@ -607,10 +223,9 @@ void intDisplayPowerBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 		Empty = 0;
 	}
 
-	int x0 = xOffset + BarGraph->x();
-	int y0 = yOffset + BarGraph->y();
+	int x0 = xOffset + this->x();
+	int y0 = yOffset + this->y();
 
-	pie_SetDepthBufferStatus(DEPTH_CMP_ALWAYS_WRT_ON);
 	pie_SetFogStatus(false);
 
 	int top_x0 = x0;
@@ -664,31 +279,38 @@ void intDisplayPowerBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 
 	imageDrawBatch.draw(true);
 
-	PIELIGHT colour;
-	if (Avail < 0)
+	auto unusedDerricks = countPlayerUnusedDerricks();
+
+	auto showNeedMessage = true;
+	if (unusedDerricks > 0)
 	{
-		const char *need = _("Need more resources!");
-		cache.wzNeedText.setText(need, font_small);
-		if ((realTime / 1250) % 5 == 0)
-		{
-			cache.wzNeedText.render(iX + 102, iY - 1, WZCOL_BLACK);
-		}
-		else
-		{
-			cache.wzNeedText.render(iX + 102, iY - 1, WZCOL_RED);
-		}
-		colour = WZCOL_RED;
+		char unusedText[50];
+		ssprintf(unusedText, _("%d derrick(s) inactive"), unusedDerricks);
+		cache.wzNeedText.setText(unusedText, font_small);
+	}
+	else if (Avail < 0)
+	{
+		cache.wzNeedText.setText(_("Need more resources!"), font_small);
 	}
 	else
 	{
-		colour = WZCOL_TEXT_BRIGHT;
+		showNeedMessage = false;
 	}
+
+	if (showNeedMessage)
+	{
+		auto needTextWidth = cache.wzNeedText.width();
+		auto textX = iX + (this->width() - needTextWidth) / 2;
+		pie_UniTransBoxFill(textX - 3, y0 + 1, textX + needTextWidth + 3, y0 + this->height() - 1, WZCOL_TRANSPARENT_BOX);
+		cache.wzNeedText.render(textX, iY - 1, (realTime / 1250) % 5 ? WZCOL_WHITE: WZCOL_RED);
+	}
+
 	// draw text value
-	cache.wzText.render(iX, iY, colour);
+	cache.wzText.render(iX, iY, Avail < 0 ? WZCOL_RED : WZCOL_TEXT_BRIGHT);
 }
 
-IntFancyButton::IntFancyButton(WIDGET *parent)
-	: W_CLICKFORM(parent)
+IntFancyButton::IntFancyButton()
+	: W_CLICKFORM()
 	, buttonType(TOPBUTTON)
 {
 	model.position.x = 0;
@@ -702,13 +324,42 @@ IntFancyButton::IntFancyButton(WIDGET *parent)
 
 void IntFancyButton::initDisplay()
 {
-	model.rate += realTimeAdjustedAverage(isHighlighted() ? 2 * BUTTONOBJ_ROTSPEED : -4 * BUTTONOBJ_ROTSPEED);
-	model.rate = clip(model.rate, 0, BUTTONOBJ_ROTSPEED);
-	model.rotation.y += realTimeAdjustedAverage(model.rate);
-}
-
-void IntFancyButton::doneDisplay()
-{
+	if (isHighlighted())
+	{
+		model.rate += realTimeAdjustedAverage(isHighlighted() ? 2 * BUTTONOBJ_ROTSPEED : -4 * BUTTONOBJ_ROTSPEED);
+		model.rate = clip(model.rate, 0, BUTTONOBJ_ROTSPEED);
+		model.rotation.y += realTimeAdjustedAverage(model.rate);
+		model.rotation.y = model.rotation.y % 360;
+	}
+	else if (model.rotation.y != DEFAULT_BUTTON_ROTATION)
+	{
+		// return to default rotation using the nearest direction
+		if (model.rotation.y > (180 + DEFAULT_BUTTON_ROTATION))
+		{
+			model.rate += realTimeAdjustedAverage(5 * BUTTONOBJ_ROTSPEED);
+			model.rate = clip(model.rate, 0, 5 * BUTTONOBJ_ROTSPEED);
+			model.rotation.y += realTimeAdjustedAverage(model.rate);
+			if (model.rotation.y > 360)
+			{
+				model.rotation.y = model.rotation.y % 360;
+				model.rotation.y = std::min(model.rotation.y, DEFAULT_BUTTON_ROTATION);
+			}
+		}
+		else if (model.rotation.y < DEFAULT_BUTTON_ROTATION)
+		{
+			model.rate += realTimeAdjustedAverage(5 * BUTTONOBJ_ROTSPEED);
+			model.rate = clip(model.rate, 0, 5 * BUTTONOBJ_ROTSPEED);
+			model.rotation.y += realTimeAdjustedAverage(model.rate);
+			model.rotation.y = std::min(model.rotation.y, DEFAULT_BUTTON_ROTATION);
+		}
+		else
+		{
+			model.rate -= realTimeAdjustedAverage(7 * BUTTONOBJ_ROTSPEED);
+			model.rate = clip(model.rate, -5 * BUTTONOBJ_ROTSPEED, BUTTONOBJ_ROTSPEED);
+			model.rotation.y += realTimeAdjustedAverage(model.rate);
+			model.rotation.y = std::max(model.rotation.y, DEFAULT_BUTTON_ROTATION);
+		}
+	}
 }
 
 void IntFancyButton::displayIfHighlight(int xOffset, int yOffset)
@@ -719,8 +370,8 @@ void IntFancyButton::displayIfHighlight(int xOffset, int yOffset)
 	}
 }
 
-IntStatusButton::IntStatusButton(WIDGET *parent)
-	: IntObjectButton(parent)
+IntStatusButton::IntStatusButton()
+	: IntObjectButton()
 	, theStats(nullptr)
 {
 	buttonType = TOPBUTTON;
@@ -869,8 +520,8 @@ void IntStatusButton::display(int xOffset, int yOffset)
 	}
 }
 
-IntObjectButton::IntObjectButton(WIDGET *parent)
-	: IntFancyButton(parent)
+IntObjectButton::IntObjectButton()
+	: IntFancyButton()
 	, psObj(nullptr)
 {
 	buttonType = BTMBUTTON;
@@ -908,12 +559,10 @@ void IntObjectButton::display(int xOffset, int yOffset)
 
 	displayIMD(Image(), object, xOffset, yOffset);
 	displayIfHighlight(xOffset, yOffset);
-
-	doneDisplay();
 }
 
-IntStatsButton::IntStatsButton(WIDGET *parent)
-	: IntFancyButton(parent)
+IntStatsButton::IntStatsButton()
+	: IntFancyButton()
 	, Stat(nullptr)
 {}
 
@@ -997,12 +646,19 @@ void IntStatsButton::display(int xOffset, int yOffset)
 
 	displayIMD(image, object, xOffset, yOffset);
 	displayIfHighlight(xOffset, yOffset);
-
-	doneDisplay();
 }
 
-IntFormAnimated::IntFormAnimated(WIDGET *parent, bool openAnimate)
-	: W_FORM(parent)
+IntFormTransparent::IntFormTransparent()
+	: W_FORM()
+{
+}
+
+void IntFormTransparent::display(int xOffset, int yOffset)
+{
+}
+
+IntFormAnimated::IntFormAnimated(bool openAnimate)
+	: W_FORM()
 	, startTime(0)
 	, currentAction(openAnimate ? 0 : 2)
 {
@@ -1015,6 +671,11 @@ void IntFormAnimated::closeAnimateDelete()
 	disableChildren = true;
 }
 
+bool IntFormAnimated::isClosing() const
+{
+	return currentAction >= 3;
+}
+
 void IntFormAnimated::display(int xOffset, int yOffset)
 {
 	WzRect aOpen(xOffset + x(), yOffset + y(), width(), height());
@@ -1135,7 +796,7 @@ void intDisplayImageHilight(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 		break;
 
 	case WIDG_SLIDER:
-		if (((W_SLIDER *)psWidget)->state & SLD_HILITE)
+		if (((W_SLIDER *)psWidget)->isHighlighted())
 		{
 			Hilight = true;
 		}
@@ -1265,10 +926,15 @@ void intDisplayEditBox(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 
 // Initialise all the surfaces,graphics etc. used by the interface.
 //
-void intInitialiseGraphics()
+bool intInitialiseGraphics()
 {
 	// Initialise any bitmaps used by the interface.
-	imageInitBitmaps();
+	if (!imageInitBitmaps())
+	{
+		return false;
+	}
+
+	return true;
 }
 
 // Clear a button bitmap. ( copy the button background ).
@@ -1442,11 +1108,11 @@ void IntFancyButton::displayIMD(Image image, ImdObject imdObject, int xOffset, i
 			BASE_STATS *psStats = (BASE_STATS *)Object;
 			if (psStats->id.compare("SuperTransportBody") == 0)
 			{
-				model.scale *= .4;
+				model.scale = static_cast<int>(model.scale * .4f);
 			}
 			else if (psStats->id.compare("TransporterBody") == 0)
 			{
-				model.scale *= .6;
+				model.scale = static_cast<int>(model.scale * .6f);
 			}
 		}
 		else if (IMDType == IMDTYPE_RESEARCH)
@@ -1548,11 +1214,9 @@ void IntFancyButton::displayIMD(Image image, ImdObject imdObject, int xOffset, i
 
 		if (!image.isNull())
 		{
-			iV_DrawImage(image, ButXPos + ox, ButYPos + oy);
+			iV_DrawImageImage(image, ButXPos + ox, ButYPos + oy);
 		}
 
-		pie_SetDepthBufferStatus(DEPTH_CMP_LEQ_WRT_ON);
-
 		/* all non droid buttons */
 		if (IMDType == IMDTYPE_COMPONENT)
 		{
@@ -1578,8 +1242,6 @@ void IntFancyButton::displayIMD(Image image, ImdObject imdObject, int xOffset, i
 		{
 			displayIMDButton((iIMDShape *)Object, &model.rotation, &model.position, model.scale);
 		}
-
-		pie_SetDepthBufferStatus(DEPTH_CMP_ALWAYS_WRT_ON);
 	}
 }
 
@@ -1593,7 +1255,7 @@ void IntFancyButton::displayImage(Image image, int xOffset, int yOffset)
 	}
 
 	displayClear(xOffset, yOffset);
-	iV_DrawImage(image, xOffset + x(), yOffset + y());
+	iV_DrawImageImage(image, xOffset + x(), yOffset + y());
 }
 
 // Create a blank button.
@@ -1620,8 +1282,8 @@ void IntFancyButton::displayBlank(int xOffset, int yOffset)
 //
 bool DroidIsBuilding(DROID *Droid)
 {
-	BASE_STATS	*Stats;
-	UDWORD x, y;
+	STRUCTURE_STATS	*Stats;
+	ASSERT_NOT_NULLPTR_OR_RETURN(false, Droid);
 
 	if (!(droidType(Droid) == DROID_CONSTRUCT ||
 	      droidType(Droid) == DROID_CYBORG_CONSTRUCT))
@@ -1629,7 +1291,7 @@ bool DroidIsBuilding(DROID *Droid)
 		return false;
 	}
 
-	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats, &x, &y))
+	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats))
 	{
 		// Moving to build location?
 		return false;
@@ -1648,8 +1310,8 @@ bool DroidIsBuilding(DROID *Droid)
 //
 bool DroidGoingToBuild(DROID *Droid)
 {
-	BASE_STATS	*Stats;
-	UDWORD x, y;
+	STRUCTURE_STATS	*Stats;
+	ASSERT_NOT_NULLPTR_OR_RETURN(false, Droid);
 
 	if (!(droidType(Droid) == DROID_CONSTRUCT ||
 	      droidType(Droid) == DROID_CYBORG_CONSTRUCT))
@@ -1657,7 +1319,7 @@ bool DroidGoingToBuild(DROID *Droid)
 		return false;
 	}
 
-	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats, &x, &y))  	// Moving to build location?
+	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats)) // Moving to build location?
 	{
 		return true;
 	}
@@ -1681,7 +1343,7 @@ STRUCTURE *DroidGetBuildStructure(DROID *Droid)
 }
 
 // Get the first factory assigned to a command droid
-static STRUCTURE *droidGetCommandFactory(DROID *psDroid)
+STRUCTURE *droidGetCommandFactory(DROID *psDroid)
 {
 	SDWORD		inc;
 	STRUCTURE	*psCurr;
@@ -1736,10 +1398,9 @@ static STRUCTURE *droidGetCommandFactory(DROID *psDroid)
 //
 BASE_STATS *DroidGetBuildStats(DROID *Droid)
 {
-	BASE_STATS *Stats;
-	UDWORD x, y;
+	STRUCTURE_STATS *Stats;
 
-	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats, &x, &y))  	// Moving to build location?
+	if (orderStateStatsLoc(Droid, DORDER_BUILD, &Stats))  	// Moving to build location?
 	{
 		return Stats;
 	}
@@ -1764,6 +1425,7 @@ static inline bool _structureIsManufacturingPending(Functionality const &functio
 
 bool StructureIsManufacturingPending(STRUCTURE *structure)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(false, structure);
 	switch (structure->pStructureType->type)
 	{
 	case REF_FACTORY:
@@ -1777,11 +1439,13 @@ bool StructureIsManufacturingPending(STRUCTURE *structure)
 
 FACTORY *StructureGetFactory(STRUCTURE *Structure)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(nullptr, Structure);
 	return &Structure->pFunctionality->factory;
 }
 
 bool structureIsResearchingPending(STRUCTURE *structure)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(false, structure);
 	return structure->pStructureType->type == REF_RESEARCH && _structureIsManufacturingPending(structure->pFunctionality->researchFacility);
 }
 
@@ -1797,6 +1461,7 @@ static inline bool structureIsOnHoldPending(Functionality const &functionality)
 
 bool StructureIsOnHoldPending(STRUCTURE *structure)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(false, structure);
 	switch (structure->pStructureType->type)
 	{
 	case REF_FACTORY:
@@ -1813,12 +1478,15 @@ bool StructureIsOnHoldPending(STRUCTURE *structure)
 
 RESEARCH_FACILITY *StructureGetResearch(STRUCTURE *Structure)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(nullptr, Structure);
+	ASSERT_NOT_NULLPTR_OR_RETURN(nullptr, Structure->pFunctionality);
 	return &Structure->pFunctionality->researchFacility;
 }
 
 
 DROID_TEMPLATE *FactoryGetTemplate(FACTORY *Factory)
 {
+	ASSERT_NOT_NULLPTR_OR_RETURN(nullptr, Factory);
 	if (Factory->psSubjectPending != nullptr)
 	{
 		return (DROID_TEMPLATE *)Factory->psSubjectPending;
@@ -1829,14 +1497,12 @@ DROID_TEMPLATE *FactoryGetTemplate(FACTORY *Factory)
 
 bool StatIsStructure(BASE_STATS const *Stat)
 {
-	return (Stat->ref >= REF_STRUCTURE_START && Stat->ref <
-	        REF_STRUCTURE_START + REF_RANGE);
+	return Stat->hasType(STAT_STRUCTURE);
 }
 
 bool StatIsFeature(BASE_STATS const *Stat)
 {
-	return (Stat->ref >= REF_FEATURE_START && Stat->ref <
-	        REF_FEATURE_START + REF_RANGE);
+	return Stat->hasType(STAT_FEATURE);
 }
 
 iIMDShape *StatGetStructureIMD(BASE_STATS *Stat, UDWORD Player)
@@ -1847,61 +1513,23 @@ iIMDShape *StatGetStructureIMD(BASE_STATS *Stat, UDWORD Player)
 
 bool StatIsTemplate(BASE_STATS *Stat)
 {
-	return (Stat->ref >= REF_TEMPLATE_START &&
-	        Stat->ref < REF_TEMPLATE_START + REF_RANGE);
+	return Stat->hasType(STAT_TEMPLATE);
 }
 
-SDWORD StatIsComponent(BASE_STATS *Stat)
+COMPONENT_TYPE StatIsComponent(BASE_STATS *Stat)
 {
-	if (Stat->ref >= REF_BODY_START &&
-	    Stat->ref < REF_BODY_START + REF_RANGE)
+	switch (StatType(Stat->ref & STAT_MASK))
 	{
-		return COMP_BODY;
+		case STAT_BODY: return COMP_BODY;
+		case STAT_BRAIN: return COMP_BRAIN;
+		case STAT_PROPULSION: return COMP_PROPULSION;
+		case STAT_WEAPON: return COMP_WEAPON;
+		case STAT_SENSOR: return COMP_SENSOR;
+		case STAT_ECM: return COMP_ECM;
+		case STAT_CONSTRUCT: return COMP_CONSTRUCT;
+		case STAT_REPAIR: return COMP_REPAIRUNIT;
+		default: return COMP_NUMCOMPONENTS;
 	}
-
-	if (Stat->ref >= REF_BRAIN_START &&
-	    Stat->ref < REF_BRAIN_START + REF_RANGE)
-	{
-		return COMP_BRAIN;
-	}
-
-	if (Stat->ref >= REF_PROPULSION_START &&
-	    Stat->ref < REF_PROPULSION_START + REF_RANGE)
-	{
-		return COMP_PROPULSION;
-	}
-
-	if (Stat->ref >= REF_WEAPON_START &&
-	    Stat->ref < REF_WEAPON_START + REF_RANGE)
-	{
-		return COMP_WEAPON;
-	}
-
-	if (Stat->ref >= REF_SENSOR_START &&
-	    Stat->ref < REF_SENSOR_START + REF_RANGE)
-	{
-		return COMP_SENSOR;
-	}
-
-	if (Stat->ref >= REF_ECM_START &&
-	    Stat->ref < REF_ECM_START + REF_RANGE)
-	{
-		return COMP_ECM;
-	}
-
-	if (Stat->ref >= REF_CONSTRUCT_START &&
-	    Stat->ref < REF_CONSTRUCT_START + REF_RANGE)
-	{
-		return COMP_CONSTRUCT;
-	}
-
-	if (Stat->ref >= REF_REPAIR_START &&
-	    Stat->ref < REF_REPAIR_START + REF_RANGE)
-	{
-		return COMP_REPAIRUNIT;
-	}
-
-	return COMP_NUMCOMPONENTS;
 }
 
 bool StatGetComponentIMD(BASE_STATS *Stat, SDWORD compID, iIMDShape **CompIMD, iIMDShape **MountIMD)
@@ -1962,8 +1590,7 @@ bool StatGetComponentIMD(BASE_STATS *Stat, SDWORD compID, iIMDShape **CompIMD, i
 
 bool StatIsResearch(BASE_STATS *Stat)
 {
-	return (Stat->ref >= REF_RESEARCH_START && Stat->ref <
-	        REF_RESEARCH_START + REF_RANGE);
+	return Stat->hasType(STAT_RESEARCH);
 }
 
 static void StatGetResearchImage(BASE_STATS *psStat, Image *image, iIMDShape **Shape, BASE_STATS **ppGraphicData, bool drawTechIcon)
@@ -1988,83 +1615,6 @@ static void StatGetResearchImage(BASE_STATS *psStat, Image *image, iIMDShape **S
 	}
 }
 
-static void intDisplayBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset, bool isPowerBar, DisplayBarCache& cache)
-{
-	W_BARGRAPH *BarGraph = (W_BARGRAPH *)psWidget;
-	char szVal[30];
-	char const *szCheckWidth = "00000";
-	int x0 = xOffset + BarGraph->x();
-	int y0 = yOffset + BarGraph->y();
-	int arbitraryOffset = 3;
-	int iX, iY;
-	int barWidth = 100, width;
-	int i, precisionFactor = 1, value;
-
-	cache.wzCheckWidthText.setText(szCheckWidth, font_regular);
-
-	if (isPowerBar)
-	{
-		//draw the background image
-		iV_DrawImage(IntImages, IMAGE_DES_POWERBAR_LEFT, x0, y0);
-		iV_DrawImage(IntImages, IMAGE_DES_POWERBAR_RIGHT, x0 + psWidget->width() - iV_GetImageWidth(IntImages, IMAGE_DES_POWERBAR_RIGHT), y0);
-	}
-
-	// Arbitrary increment for the position of the bars
-	x0 += arbitraryOffset;
-	y0 += arbitraryOffset;
-
-	/* indent to allow text value */
-	iX = x0 + cache.wzCheckWidthText.width();
-	iY = y0 + (iV_GetImageHeight(IntImages, IMAGE_DES_STATSCURR) - iV_GetTextLineSize(font_regular)) / 2 - iV_GetTextAboveBase(font_regular);
-
-	if (isPowerBar)
-	{
-		// Adjust the width based on the text drawn
-		barWidth = BarGraph->width() - (iX - x0 + arbitraryOffset);
-	}
-
-	//draw current value section
-	width = MIN(BarGraph->majorSize * barWidth / 100, barWidth);
-	iV_DrawImageRepeatX(IntImages, IMAGE_DES_STATSCURR, iX, y0, width, defaultProjectionMatrix(), true);
-
-	/* draw text value */
-	for (i = 0; i < BarGraph->precision; ++i)
-	{
-		precisionFactor *= 10;
-	}
-	value = (BarGraph->iOriginal * precisionFactor + BarGraph->denominator / 2) / BarGraph->denominator;
-	sprintf(szVal, "%d%s%.*d", value / precisionFactor, precisionFactor == 1 ? "" : ".", BarGraph->precision, value % precisionFactor);
-	cache.wzText.setText(szVal, font_regular);
-	cache.wzText.render(x0, iY, WZCOL_TEXT_BRIGHT);
-
-	//draw the comparison value - only if not zero
-	if (BarGraph->minorSize != 0)
-	{
-		width = MIN(BarGraph->minorSize * barWidth / 100, barWidth);
-		iV_DrawImage(IntImages, IMAGE_DES_STATSCOMP, iX + width, y0 - 1);
-	}
-}
-
-/* Draws a stats bar for the design screen */
-void intDisplayStatsBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
-{
-	// Any widget using intDisplayStatsBar must have its pUserData initialized to a (DisplayBarCache*)
-	assert(psWidget->pUserData != nullptr);
-	DisplayBarCache& cache = *static_cast<DisplayBarCache*>(psWidget->pUserData);
-
-	intDisplayBar(psWidget, xOffset, yOffset, false, cache);
-}
-
-/* Draws a Template Power Bar for the Design Screen */
-void intDisplayDesignPowerBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
-{
-	// Any widget using intDisplayDesignPowerBar must have its pUserData initialized to a (DisplayBarCache*)
-	assert(psWidget->pUserData != nullptr);
-	DisplayBarCache& cache = *static_cast<DisplayBarCache*>(psWidget->pUserData);
-
-	intDisplayBar(psWidget, xOffset, yOffset, true, cache);
-}
-
 // Widget callback function to play an audio track.
 //
 #define WIDGETBEEPGAP (200)	// 200 milliseconds between each beep please
@@ -2086,8 +1636,8 @@ void WidgetAudioCallback(int AudioID)
 	}
 }
 
-IntTransportButton::IntTransportButton(WIDGET *parent)
-	: IntFancyButton(parent)
+IntTransportButton::IntTransportButton()
+	: IntFancyButton()
 	, psDroid(nullptr)
 {}
 
@@ -2111,7 +1661,6 @@ void IntTransportButton::display(int xOffset, int yOffset)
 			iV_DrawImage(IntImages, gfxId, xOffset + x() + 50, yOffset + y() + 10);
 		}
 	}
-	doneDisplay();
 }
 
 /* Draws blips on radar to represent Proximity Display and damaged structures */
@@ -2134,6 +1683,7 @@ void drawRadarBlips(int radarX, int radarY, float pixSizeH, float pixSizeV, cons
 	height = scrollMaxY - scrollMinY;
 
 	// Check if it's time to remove beacons
+	bool removedAMessage = false;
 	for (i = 0; i < MAX_PLAYERS; i++)
 	{
 		/* Go through all the proximity Displays*/
@@ -2153,15 +1703,20 @@ void drawRadarBlips(int radarX, int radarY, float pixSizeH, float pixSizeV, cons
 					{
 						debug(LOG_MSG, "blip timeout for %d, from %d", i, (((VIEW_PROXIMITY *)pViewData->pData)->sender));
 						removeMessage(psCurrMsg, i);	//remove beacon
+						removedAMessage = true;
 						break;	//there can only be 1 beacon per player
 					}
 				}
 			}
 		}
 	}
+	if (removedAMessage)
+	{
+		jsDebugMessageUpdate();
+	}
 
 	/* Go through all the proximity Displays */
-	for (psProxDisp = apsProxDisp[selectedPlayer]; psProxDisp != nullptr; psProxDisp = psProxDisp->psNext)
+	for (psProxDisp = (selectedPlayer < MAX_PLAYERS) ? apsProxDisp[selectedPlayer] : nullptr; psProxDisp != nullptr; psProxDisp = psProxDisp->psNext)
 	{
 		unsigned        animationLength = ARRAY_SIZE(imagesEnemy) - 1;  // Same size as imagesResource and imagesArtifact.
 		const uint16_t *images;
@@ -2217,13 +1772,13 @@ void drawRadarBlips(int radarX, int radarY, float pixSizeH, float pixSizeV, cons
 		{
 			const VIEW_PROXIMITY *psViewProx = (VIEW_PROXIMITY *)psProxDisp->psMessage->pViewData->pData;
 
-			x = (psViewProx->x / TILE_UNITS - scrollMinX) * pixSizeH;
-			y = (psViewProx->y / TILE_UNITS - scrollMinY) * pixSizeV;
+			x = static_cast<int>((psViewProx->x / TILE_UNITS - scrollMinX) * pixSizeH);
+			y = static_cast<int>((psViewProx->y / TILE_UNITS - scrollMinY) * pixSizeV);
 		}
 		else if (psProxDisp->type == POS_PROXOBJ)
 		{
-			x = (psProxDisp->psMessage->psObj->pos.x / TILE_UNITS - scrollMinX) * pixSizeH;
-			y = (psProxDisp->psMessage->psObj->pos.y / TILE_UNITS - scrollMinY) * pixSizeV;
+			x = static_cast<int>((psProxDisp->psMessage->psObj->pos.x / TILE_UNITS - scrollMinX) * pixSizeH);
+			y = static_cast<int>((psProxDisp->psMessage->psObj->pos.y / TILE_UNITS - scrollMinY) * pixSizeV);
 		}
 		else
 		{
@@ -2245,8 +1800,8 @@ void drawRadarBlips(int radarX, int radarY, float pixSizeH, float pixSizeV, cons
 	{
 		unsigned        animationLength = ARRAY_SIZE(imagesEnemy) - 1;
 		int             strobe = (realTime / delay) % animationLength;
-		x = (x / TILE_UNITS - scrollMinX) * pixSizeH;
-		y = (y / TILE_UNITS - scrollMinY) * pixSizeV;
+		x = static_cast<int>((x / TILE_UNITS - scrollMinX) * pixSizeH);
+		y = static_cast<int>((y / TILE_UNITS - scrollMinY) * pixSizeV);
 		imageID = imagesEnemy[strobe];
 
 		// NOTE:  On certain missions (limbo & expand), there is still valid data that is stored outside the
@@ -2327,7 +1882,7 @@ void intUpdateQuantitySlider(WIDGET *psWidget, W_CONTEXT *psContext)
 {
 	W_SLIDER *Slider = (W_SLIDER *)psWidget;
 
-	if (Slider->state & SLD_HILITE)
+	if (Slider->isHighlighted())
 	{
 		if (keyDown(KEY_LEFTARROW))
 		{
@@ -2346,48 +1901,8 @@ void intUpdateQuantitySlider(WIDGET *psWidget, W_CONTEXT *psContext)
 	}
 }
 
-void intDisplayResSubGroup(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
+void intDisplayUpdateAllyBar(W_BARGRAPH *psBar, const RESEARCH &research, const std::vector<AllyResearch> &researches)
 {
-	W_LABEL		*Label = (W_LABEL *)psWidget;
-	UDWORD		x = Label->x() + xOffset;
-	UDWORD		y = Label->y() + yOffset;
-	RESEARCH    *psResearch = (RESEARCH *)Label->pUserData;
-
-	if (psResearch->subGroup != NO_RESEARCH_ICON)
-	{
-		iV_DrawImage(IntImages, psResearch->subGroup, x, y);
-	}
-}
-
-void intDisplayAllyIcon(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
-{
-	W_LABEL		*Label = (W_LABEL *)psWidget;
-	UDWORD		x = Label->x() + xOffset;
-	UDWORD		y = Label->y() + yOffset;
-
-	unsigned ref = UNPACKDWORD_HI(psWidget->UserData) + REF_RESEARCH_START;
-	unsigned num = UNPACKDWORD_LOW(psWidget->UserData);
-
-	std::vector<AllyResearch> const &researches = listAllyResearch(ref);
-	if (num >= researches.size())
-	{
-		return;  // No icon to display. (Shouldn't really get here...)
-	}
-
-	if (!researches[num].active && realTime % 500 >= 250)
-	{
-		return;  // If inactive, blink the icon. (Alternatively, we could use a different icon instead.)
-	}
-	iV_DrawImageTc(IntImages, IMAGE_ALLY_RESEARCH, IMAGE_ALLY_RESEARCH_TC, x, y, pal_GetTeamColour(getPlayerColour(researches[num].player)));
-}
-
-void intDisplayAllyBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
-{
-	W_BARGRAPH *psBar = (W_BARGRAPH *)psWidget;
-
-	RESEARCH const &research = asResearch[psWidget->UserData];
-	std::vector<AllyResearch> const &researches = listAllyResearch(research.ref);
-
 	unsigned bestCompletion = 0;
 	const int researchNotStarted = 3600000;
 	int bestPowerNeeded = researchNotStarted;
@@ -2432,7 +1947,6 @@ void intDisplayAllyBar(WIDGET *psWidget, UDWORD xOffset, UDWORD yOffset)
 		formatPowerText(psBar, bestPowerNeeded);
 		setBarGraphValue(psBar, psBar->majorCol, researchPowerCost - bestPowerNeeded, researchPowerCost);
 	}
-	barGraphDisplayTrough(psWidget, xOffset, yOffset);
 }
 
 /* Set the shadow power for the selected player */
