@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -32,7 +32,6 @@
 #include "lib/framework/file.h"
 #include "lib/framework/crc.h"
 #include "lib/framework/physfs_ext.h"
-#include "lib/framework/rational.h"
 #include "lib/gamelib/gtime.h"
 #include "lib/exceptionhandler/dumpinfo.h"
 #include "clparse.h"
@@ -45,8 +44,6 @@
 #include "game.h"
 #include "lib/ivis_opengl/piestate.h"
 #include "data.h"
-#include "lib/script/script.h"
-#include "scripttabs.h"
 #include "research.h"
 #include "lib/framework/lexer_input.h"
 #include "effects.h"
@@ -54,6 +51,11 @@
 #include "multiint.h"
 #include "qtscript.h"
 #include "wrappers.h"
+#include "activity.h"
+
+#include <unordered_set>
+
+#include "3rdparty/gsl_finally.h"
 
 extern int lev_get_lineno();
 extern char *lev_get_text();
@@ -74,7 +76,7 @@ static LEVEL_DATASET	*psBaseData = nullptr;
 static LEVEL_DATASET	*psCurrLevel = nullptr;
 
 // dummy level data for single WRF loads
-static LEVEL_DATASET	sSingleWRF = {LDS_COMPLETE, 0, 0, nullptr, mod_clean, {nullptr}, nullptr, nullptr, nullptr, {{0}}};
+static LEVEL_DATASET	sSingleWRF = { LEVEL_TYPE::LDS_COMPLETE, 0, 0, nullptr, mod_clean, {nullptr}, nullptr, nullptr, nullptr, {{0}}};
 
 // return values from the lexer
 char *pLevToken;
@@ -111,22 +113,27 @@ SDWORD getLevelLoadType()
 	return levelLoadType;
 }
 
+static inline void freeLevel(LEVEL_DATASET* toDelete)
+{
+	for (auto &apDataFile : toDelete->apDataFiles)
+	{
+		if (apDataFile != nullptr)
+		{
+			free(apDataFile);
+		}
+	}
+
+	free(toDelete->pName);
+	free(toDelete->realFileName);
+	free(toDelete);
+}
+
 // shutdown the level system
 void levShutDown()
 {
 	for (auto toDelete : psLevels)
 	{
-		for (auto &apDataFile : toDelete->apDataFiles)
-		{
-			if (apDataFile != nullptr)
-			{
-				free(apDataFile);
-			}
-		}
-
-		free(toDelete->pName);
-		free(toDelete->realFileName);
-		free(toDelete);
+		freeLevel(toDelete);
 	}
 	psLevels.clear();
 }
@@ -163,6 +170,61 @@ LEVEL_DATASET *levFindDataSet(char const *name, Sha256 const *hash)
 	return nullptr;
 }
 
+LEVEL_DATASET *levFindDataSetByRealFileName(char const *realFileName, Sha256 const *hash)
+{
+	if (hash != nullptr && hash->isZero())
+	{
+		hash = nullptr;  // Don't check hash if it's just 0000000000000000000000000000000000000000000000000000000000000000. Assuming real map files probably won't have that particular SHA-256 hash.
+	}
+
+	for (auto psNewLevel : psLevels)
+	{
+		if (psNewLevel->realFileName && strcmp(psNewLevel->realFileName, realFileName) == 0)
+		{
+			if (hash == nullptr || levGetFileHash(psNewLevel) == *hash)
+			{
+				return psNewLevel;
+			}
+		}
+	}
+
+	return nullptr;
+}
+
+bool levRemoveDataSetByRealFileName(char const *realFileName, Sha256 const *hash)
+{
+	if (hash != nullptr && hash->isZero())
+	{
+		hash = nullptr;  // Don't check hash if it's just 0000000000000000000000000000000000000000000000000000000000000000. Assuming real map files probably won't have that particular SHA-256 hash.
+	}
+
+	size_t numRemoved = 0;
+	for (auto it = psLevels.begin(); it != psLevels.end();)
+	{
+		LEVEL_DATASET *level = *it;
+		if (level && level->realFileName && strcmp(level->realFileName, realFileName) == 0)
+		{
+			if (hash == nullptr || levGetFileHash(level) == *hash)
+			{
+				if (psCurrLevel == *it)
+				{
+					ASSERT(false, "Trying to remove what is still the current level");
+					continue;
+				}
+
+				LEVEL_DATASET* toDelete = *it;
+				it = psLevels.erase(it);
+				freeLevel(toDelete);
+				++numRemoved;
+				continue;
+			}
+		}
+		++it;
+	}
+
+	return numRemoved > 0;
+}
+
 Sha256 levGetFileHash(LEVEL_DATASET *level)
 {
 	if (level->realFileName != nullptr && level->realFileHash.isZero())
@@ -173,6 +235,21 @@ Sha256 levGetFileHash(LEVEL_DATASET *level)
 	return level->realFileHash;
 }
 
+bool levSetFileHashByRealFileName(char const *realFileName, Sha256 const &hash)
+{
+	size_t numAffected = 0;
+	for (auto psNewLevel : psLevels)
+	{
+		if (psNewLevel && psNewLevel->realFileName && strcmp(psNewLevel->realFileName, realFileName) == 0)
+		{
+			ASSERT(psNewLevel->realFileHash.isZero(), "Level already has a hash??");
+			psNewLevel->realFileHash = hash;
+			++numAffected;
+		}
+	}
+	return numAffected > 0;
+}
+
 Sha256 levGetMapNameHash(char const *mapName)
 {
 	LEVEL_DATASET *level = levFindDataSet(mapName, nullptr);
@@ -189,7 +266,7 @@ Sha256 levGetMapNameHash(char const *mapName)
 // parse a level description data file
 // the ignoreWrf hack is for compatibility with old maps that try to link in various
 // data files that we have removed
-bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool ignoreWrf, char const *realFileName)
+bool levParse(const char *buffer, size_t size, searchPathMode pathMode, bool ignoreWrf, char const *realFileName)
 {
 	lexerinput_t input;
 	LEVELPARSER_STATE state;
@@ -203,6 +280,8 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 	lev_set_extra(&input);
 
 	state = LP_START;
+	auto always_destroy_on_return = gsl::finally([] { lev_lex_destroy(); });
+
 	for (token = lev_lex(); token != 0; token = lev_lex())
 	{
 		switch (token)
@@ -230,7 +309,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 				memset(psDataSet, 0, sizeof(LEVEL_DATASET));
 				psDataSet->players = 1;
 				psDataSet->game = -1;
-				psDataSet->dataDir = datadir;
+				psDataSet->dataDir = pathMode;
 				psDataSet->realFileName = realFileName != nullptr ? strdup(realFileName) : nullptr;
 				psDataSet->realFileHash.setZero();  // The hash is only calculated on demand; for example, if the map name matches.
 				psLevels.push_back(psDataSet);
@@ -240,34 +319,34 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 				switch (token)
 				{
 				case LTK_LEVEL:
-					psDataSet->type = LDS_COMPLETE;
+					psDataSet->type = LEVEL_TYPE::LDS_COMPLETE;
 					break;
 				case LTK_CAMPAIGN:
-					psDataSet->type = LDS_CAMPAIGN;
+					psDataSet->type = LEVEL_TYPE::LDS_CAMPAIGN;
 					break;
 				case LTK_CAMSTART:
-					psDataSet->type = LDS_CAMSTART;
+					psDataSet->type = LEVEL_TYPE::LDS_CAMSTART;
 					break;
 				case LTK_BETWEEN:
-					psDataSet->type = LDS_BETWEEN;
+					psDataSet->type = LEVEL_TYPE::LDS_BETWEEN;
 					break;
 				case LTK_MKEEP:
-					psDataSet->type = LDS_MKEEP;
+					psDataSet->type = LEVEL_TYPE::LDS_MKEEP;
 					break;
 				case LTK_CAMCHANGE:
-					psDataSet->type = LDS_CAMCHANGE;
+					psDataSet->type = LEVEL_TYPE::LDS_CAMCHANGE;
 					break;
 				case LTK_EXPAND:
-					psDataSet->type = LDS_EXPAND;
+					psDataSet->type = LEVEL_TYPE::LDS_EXPAND;
 					break;
 				case LTK_MCLEAR:
-					psDataSet->type = LDS_MCLEAR;
+					psDataSet->type = LEVEL_TYPE::LDS_MCLEAR;
 					break;
 				case LTK_EXPAND_LIMBO:
-					psDataSet->type = LDS_EXPAND_LIMBO;
+					psDataSet->type = LEVEL_TYPE::LDS_EXPAND_LIMBO;
 					break;
 				case LTK_MKEEP_LIMBO:
-					psDataSet->type = LDS_MKEEP_LIMBO;
+					psDataSet->type = LEVEL_TYPE::LDS_MKEEP_LIMBO;
 					break;
 				default:
 					ASSERT(false, "eh?");
@@ -283,7 +362,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			break;
 		case LTK_PLAYERS:
 			if (state == LP_LEVELDONE &&
-			    (psDataSet->type == LDS_COMPLETE || psDataSet->type >= LDS_MULTI_TYPE_START))
+			    (psDataSet->type == LEVEL_TYPE::LDS_COMPLETE || psDataSet->type >= LEVEL_TYPE::LDS_MULTI_TYPE_START))
 			{
 				state = LP_PLAYERS;
 			}
@@ -294,7 +373,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			}
 			break;
 		case LTK_TYPE:
-			if (state == LP_LEVELDONE && psDataSet->type == LDS_COMPLETE)
+			if (state == LP_LEVELDONE && psDataSet->type == LEVEL_TYPE::LDS_COMPLETE)
 			{
 				state = LP_TYPE;
 			}
@@ -311,7 +390,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			}
 			else if (state == LP_TYPE)
 			{
-				if (levVal < LDS_MULTI_TYPE_START)
+				if (levVal < LEVEL_TYPE::LDS_MULTI_TYPE_START)
 				{
 					lev_error("invalid type number");
 					return false;
@@ -327,7 +406,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			state = LP_LEVELDONE;
 			break;
 		case LTK_DATASET:
-			if (state == LP_LEVELDONE && psDataSet->type != LDS_COMPLETE)
+			if (state == LP_LEVELDONE && psDataSet->type != LEVEL_TYPE::LDS_COMPLETE)
 			{
 				state = LP_DATASET;
 			}
@@ -344,13 +423,13 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			}
 			else if (state == LP_LEVELDONE)
 			{
-				if (psDataSet->type == LDS_CAMSTART ||
-				    psDataSet->type == LDS_MKEEP
-				    || psDataSet->type == LDS_CAMCHANGE ||
-				    psDataSet->type == LDS_EXPAND ||
-				    psDataSet->type == LDS_MCLEAR ||
-				    psDataSet->type == LDS_EXPAND_LIMBO ||
-				    psDataSet->type == LDS_MKEEP_LIMBO
+				if (psDataSet->type == LEVEL_TYPE::LDS_CAMSTART ||
+				    psDataSet->type == LEVEL_TYPE::LDS_MKEEP
+				    || psDataSet->type == LEVEL_TYPE::LDS_CAMCHANGE ||
+				    psDataSet->type == LEVEL_TYPE::LDS_EXPAND ||
+				    psDataSet->type == LEVEL_TYPE::LDS_MCLEAR ||
+				    psDataSet->type == LEVEL_TYPE::LDS_EXPAND_LIMBO ||
+				    psDataSet->type == LEVEL_TYPE::LDS_MKEEP_LIMBO
 				   )
 				{
 					lev_error("Missing dataset command");
@@ -366,7 +445,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 			break;
 		case LTK_GAME:
 			if ((state == LP_WAITDATA || state == LP_LEVELDONE) &&
-			    psDataSet->game == -1 && psDataSet->type != LDS_CAMPAIGN)
+			    psDataSet->game == -1 && psDataSet->type != LEVEL_TYPE::LDS_CAMPAIGN)
 			{
 				state = LP_GAME;
 			}
@@ -379,7 +458,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 		case LTK_IDENT:
 			if (state == LP_LEVEL)
 			{
-				if (psDataSet->type == LDS_CAMCHANGE)
+				if (psDataSet->type == LEVEL_TYPE::LDS_CAMCHANGE)
 				{
 					// This is a campaign change dataset, we need to find the full data set.
 					LEVEL_DATASET *const psFoundData = levFindDataSet(pLevToken);
@@ -390,7 +469,7 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 						return false;
 					}
 
-					if (psFoundData->type != LDS_CAMSTART)
+					if (psFoundData->type != LEVEL_TYPE::LDS_CAMSTART)
 					{
 						lev_error("Invalid data set name for cam change");
 						return false;
@@ -466,8 +545,6 @@ bool levParse(const char *buffer, size_t size, searchPathMode datadir, bool igno
 		}
 	}
 
-	lev_lex_destroy();
-
 	// Accept empty files when parsing (indicated by currData < 0)
 	if (currData >= 0
 	    && (state != LP_WAITDATA
@@ -511,6 +588,7 @@ bool levReleaseMissionData()
 			}
 		}
 	}
+	releaseObjectives = true; // allow releasing mission objectives after quitting / saveload
 	return true;
 }
 
@@ -552,6 +630,8 @@ bool levReleaseAll()
 			}
 		}
 
+		psCurrLevel = nullptr;
+
 		if (!stageOneShutDown())
 		{
 			debug(LOG_ERROR, "Failed stage one shutdown");
@@ -559,8 +639,6 @@ bool levReleaseAll()
 		}
 	}
 
-	psCurrLevel = nullptr;
-
 	return true;
 }
 
@@ -613,9 +691,6 @@ const char *getLevelName()
 // load up the data for a level
 bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYPE saveType)
 {
-	LEVEL_DATASET	*psNewLevel, *psBaseData, *psChangeLevel;
-	bool            bCamChangeSaveGame;
-
 	debug(LOG_WZ, "Loading level %s hash %s (%s, type %d)", name, hash == nullptr ? "builtin" : hash->toString().c_str(), pSaveName, (int)saveType);
 	if (saveType == GTYPE_SAVE_START || saveType == GTYPE_SAVE_MIDMISSION)
 	{
@@ -632,7 +707,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	levelLoadType = saveType;
 
 	// find the level dataset
-	psNewLevel = levFindDataSet(name, hash);
+	LEVEL_DATASET* psNewLevel = levFindDataSet(name, hash);
 	if (psNewLevel == nullptr)
 	{
 		debug(LOG_INFO, "Dataset %s not found - trying to load as WRF", name);
@@ -643,18 +718,14 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	/* Keep a copy of the present level name */
 	sstrcpy(currentLevelName, name);
 
-	bCamChangeSaveGame = false;
-	if (pSaveName && saveType == GTYPE_SAVE_START)
+	const bool bCamChangeSaveGame = pSaveName && saveType == GTYPE_SAVE_START && psNewLevel->psChange != nullptr;
+	if (bCamChangeSaveGame)
 	{
-		if (psNewLevel->psChange != nullptr)
-		{
-			bCamChangeSaveGame = true;
-			debug(LOG_WZ, "** CAMCHANGE FOUND");
-		}
+		debug(LOG_WZ, "** CAMCHANGE FOUND");
 	}
 
 	// select the change dataset if there is one
-	psChangeLevel = nullptr;
+	LEVEL_DATASET* psChangeLevel = nullptr;
 	if (((psNewLevel->psChange != nullptr) && (psCurrLevel != nullptr)) || bCamChangeSaveGame)
 	{
 		//store the level name
@@ -664,7 +735,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	}
 
 	// ensure the correct dataset is loaded
-	if (psNewLevel->type == LDS_CAMPAIGN)
+	if (psNewLevel->type == LEVEL_TYPE::LDS_CAMPAIGN)
 	{
 		debug(LOG_ERROR, "Cannot load a campaign dataset (%s)", psNewLevel->pName);
 		return false;
@@ -674,8 +745,8 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 		if (psCurrLevel != nullptr)
 		{
 			if ((psCurrLevel->psBaseData != psNewLevel->psBaseData) ||
-			    (psCurrLevel->type < LDS_NONE && psNewLevel->type  >= LDS_NONE) ||
-			    (psCurrLevel->type >= LDS_NONE && psNewLevel->type  < LDS_NONE))
+			    (psCurrLevel->type < LEVEL_TYPE::LDS_NONE && psNewLevel->type  >= LEVEL_TYPE::LDS_NONE) ||
+			    (psCurrLevel->type >= LEVEL_TYPE::LDS_NONE && psNewLevel->type  < LEVEL_TYPE::LDS_NONE))
 			{
 				// there is a dataset loaded but it isn't the correct one
 				debug(LOG_WZ, "Incorrect base dataset loaded (%p != %p, %d - %d)",
@@ -736,7 +807,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	}
 
 	// initialise if necessary
-	if (psNewLevel->type == LDS_COMPLETE || psBaseData != nullptr)
+	if (psNewLevel->type == LEVEL_TYPE::LDS_COMPLETE || psBaseData != nullptr)
 	{
 		debug(LOG_WZ, "Calling stageOneInitialise!");
 		if (!stageOneInitialise())
@@ -755,7 +826,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 			if (psBaseData->apDataFiles[i])
 			{
 				// load the data
-				debug(LOG_WZ, "Loading [directory: %s] %s ...", PHYSFS_getRealDir(psBaseData->apDataFiles[i]), psBaseData->apDataFiles[i]);
+				debug(LOG_WZ, "Loading [directory: %s] %s ...", WZ_PHYSFS_getRealDir_String(psBaseData->apDataFiles[i]).c_str(), psBaseData->apDataFiles[i]);
 				if (!resLoad(psBaseData->apDataFiles[i], i))
 				{
 					debug(LOG_ERROR, "Failed resLoad(%s)!", psBaseData->apDataFiles[i]);
@@ -764,7 +835,26 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 			}
 		}
 	}
-	if (psNewLevel->type == LDS_CAMCHANGE)
+	// preload faction IMDs
+	std::unordered_set<FactionID> enabledNonNormalFactions = getEnabledFactions(true);
+	if (!enabledNonNormalFactions.empty())
+	{
+		enumerateLoadedModels([enabledNonNormalFactions](const std::string &modelName, iIMDShape &){
+			for (const auto& faction : enabledNonNormalFactions)
+			{
+				auto factionModel = getFactionModelName(faction, WzString::fromUtf8(modelName));
+				if (factionModel.has_value())
+				{
+					iIMDShape *retval = modelGet(factionModel.value());
+					ASSERT(retval != nullptr, "Cannot find the faction PIE model %s (for normal model: %s)",
+						   factionModel.value().toUtf8().c_str(), modelName.c_str());
+				}
+			}
+		});
+		resDoResLoadCallback();		// do callback.
+	}
+
+	if (psNewLevel->type == LEVEL_TYPE::LDS_CAMCHANGE)
 	{
 		if (!campaignReset())
 		{
@@ -774,7 +864,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	}
 	if (psNewLevel->game == -1)  //no .gam file to load - BETWEEN missions (for Editor games only)
 	{
-		ASSERT(psNewLevel->type == LDS_BETWEEN, "Only BETWEEN missions do not need a .gam file");
+		ASSERT(psNewLevel->type == LEVEL_TYPE::LDS_BETWEEN, "Only BETWEEN missions do not need a .gam file");
 		debug(LOG_WZ, "No .gam file for level: BETWEEN mission");
 		if (pSaveName != nullptr)
 		{
@@ -793,7 +883,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 				debug(LOG_WZ, "Init mission stuff");
 				if (!startMissionSave(psNewLevel->type))
 				{
-					debug(LOG_ERROR, "Failed startMissionSave(%d)!", psNewLevel->type);
+					debug(LOG_ERROR, "Failed startMissionSave(%d)!", static_cast<int8_t>(psNewLevel->type));
 					return false;
 				}
 
@@ -814,7 +904,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 			debug(LOG_NEVER, "Start mission - no .gam");
 			if (!startMission((LEVEL_TYPE)psNewLevel->type, nullptr))
 			{
-				debug(LOG_ERROR, "Failed startMission(%d)!", psNewLevel->type);
+				debug(LOG_ERROR, "Failed startMission(%d)!", static_cast<int8_t>(psNewLevel->type));
 				return false;
 			}
 		}
@@ -853,7 +943,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 		if (psNewLevel->game == i)
 		{
 			// do some more initialising if necessary
-			if (psNewLevel->type == LDS_COMPLETE || psNewLevel->type >= LDS_MULTI_TYPE_START || (psBaseData != nullptr && !bCamChangeSaveGame))
+			if (psNewLevel->type == LEVEL_TYPE::LDS_COMPLETE || psNewLevel->type >= LEVEL_TYPE::LDS_MULTI_TYPE_START || (psBaseData != nullptr && !bCamChangeSaveGame))
 			{
 				if (!stageTwoInitialise())
 				{
@@ -871,7 +961,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 					debug(LOG_WZ, "Init mission stuff");
 					if (!startMissionSave(psNewLevel->type))
 					{
-						debug(LOG_ERROR, "Failed startMissionSave(%d)!", psNewLevel->type);
+						debug(LOG_ERROR, "Failed startMissionSave(%d)!", static_cast<uint8_t>(psNewLevel->type));
 						return false;
 					}
 
@@ -893,80 +983,80 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 				debug(LOG_WZ, "Loading scenario file %s", psNewLevel->apDataFiles[i]);
 				switch (psNewLevel->type)
 				{
-				case LDS_COMPLETE:
-				case LDS_CAMSTART:
+				case LEVEL_TYPE::LDS_COMPLETE:
+				case LEVEL_TYPE::LDS_CAMSTART:
 					debug(LOG_WZ, "LDS_COMPLETE / LDS_CAMSTART");
-					if (!startMission(LDS_CAMSTART, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_CAMSTART, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_CAMSTART, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_CAMSTART), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
-				case LDS_BETWEEN:
+				case LEVEL_TYPE::LDS_BETWEEN:
 					debug(LOG_WZ, "LDS_BETWEEN");
-					if (!startMission(LDS_BETWEEN, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_BETWEEN, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_BETWEEN, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_BETWEEN), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
 
-				case LDS_MKEEP:
+				case LEVEL_TYPE::LDS_MKEEP:
 					debug(LOG_WZ, "LDS_MKEEP");
-					if (!startMission(LDS_MKEEP, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_MKEEP, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_MKEEP, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_MKEEP), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
-				case LDS_CAMCHANGE:
+				case LEVEL_TYPE::LDS_CAMCHANGE:
 					debug(LOG_WZ, "LDS_CAMCHANGE");
-					if (!startMission(LDS_CAMCHANGE, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_CAMCHANGE, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_CAMCHANGE, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_CAMCHANGE), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
 
-				case LDS_EXPAND:
+				case LEVEL_TYPE::LDS_EXPAND:
 					debug(LOG_WZ, "LDS_EXPAND");
-					if (!startMission(LDS_EXPAND, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_EXPAND, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_EXPAND, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_EXPAND), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
-				case LDS_EXPAND_LIMBO:
+				case LEVEL_TYPE::LDS_EXPAND_LIMBO:
 					debug(LOG_WZ, "LDS_LIMBO");
-					if (!startMission(LDS_EXPAND_LIMBO, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_EXPAND_LIMBO, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_EXPAND_LIMBO, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_EXPAND_LIMBO), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
 
-				case LDS_MCLEAR:
+				case LEVEL_TYPE::LDS_MCLEAR:
 					debug(LOG_WZ, "LDS_MCLEAR");
-					if (!startMission(LDS_MCLEAR, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_MCLEAR, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_MCLEAR, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_MCLEAR), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
-				case LDS_MKEEP_LIMBO:
+				case LEVEL_TYPE::LDS_MKEEP_LIMBO:
 					debug(LOG_WZ, "LDS_MKEEP_LIMBO");
-					if (!startMission(LDS_MKEEP_LIMBO, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_MKEEP_LIMBO, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s)!", LDS_MKEEP_LIMBO, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s)!", static_cast<int8_t>(LEVEL_TYPE::LDS_MKEEP_LIMBO), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
 				default:
-					ASSERT(psNewLevel->type >= LDS_MULTI_TYPE_START, "Unexpected mission type");
+					ASSERT(psNewLevel->type >= LEVEL_TYPE::LDS_MULTI_TYPE_START, "Unexpected mission type");
 					debug(LOG_WZ, "default (MULTIPLAYER)");
-					if (!startMission(LDS_CAMSTART, psNewLevel->apDataFiles[i]))
+					if (!startMission(LEVEL_TYPE::LDS_CAMSTART, psNewLevel->apDataFiles[i]))
 					{
-						debug(LOG_ERROR, "Failed startMission(%d, %s) (default)!", LDS_CAMSTART, psNewLevel->apDataFiles[i]);
+						debug(LOG_ERROR, "Failed startMission(%d, %s) (default)!", static_cast<int8_t>(LEVEL_TYPE::LDS_CAMSTART), psNewLevel->apDataFiles[i]);
 						return false;
 					}
 					break;
@@ -996,7 +1086,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 		//load MidMission Extras
 		if (!loadMissionExtras(pSaveName, psNewLevel->type))
 		{
-			debug(LOG_ERROR, "Failed loadMissionExtras(%s, %d)!", pSaveName, psNewLevel->type);
+			debug(LOG_ERROR, "Failed loadMissionExtras(%s, %d)!", pSaveName, static_cast<int8_t>(psNewLevel->type));
 			return false;
 		}
 	}
@@ -1012,7 +1102,7 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 			return false;
 		}
 	}
-
+	// this will trigger upgrades
 	if (!stageThreeInitialise())
 	{
 		debug(LOG_ERROR, "Failed stageThreeInitialise()!");
@@ -1021,13 +1111,6 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 
 	dataClearSaveFlag();
 
-	//this enables us to to start cam2/cam3 without going via a save game and get the extra droids
-	//in from the script-controlled Transporters
-	if (!pSaveName && psNewLevel->type == LDS_CAMSTART)
-	{
-		eventFireCallbackTrigger((TRIGGER_TYPE)CALL_NO_REINFORCEMENTS_LEFT);
-	}
-
 	//restore the level name for comparisons on next mission load up
 	if (psChangeLevel == nullptr)
 	{
@@ -1044,46 +1127,98 @@ bool levLoadData(char const *name, Sha256 const *hash, char *pSaveName, GAME_TYP
 	ssprintf(buf, "Current Level/map is %s", psCurrLevel->pName);
 	addDumpInfo(buf);
 
-	triggerEvent(TRIGGER_GAME_LOADED);
-
 	if (autogame_enabled())
 	{
 		gameTimeSetMod(Rational(500));
-		if (hostlaunch != 2) // tests will specify the AI manually
+		if (getHostLaunch() != HostLaunch::Skirmish) // tests will specify the AI manually
 		{
-			jsAutogameSpecific("multiplay/skirmish/semperfi.js", selectedPlayer);
+			if (selectedPlayer < MAX_PLAYERS && !NetPlay.players[selectedPlayer].isSpectator)
+			{
+				jsAutogameSpecific("multiplay/skirmish/semperfi.js", selectedPlayer);
+			}
+			else
+			{
+				debug(LOG_INFO, "Skipping autogame auto-AI for selectedPlayer %" PRIu32 "", selectedPlayer);
+			}
 		}
 	}
 
+	ActivityManager::instance().loadedLevel(psCurrLevel->type, mapNameWithoutTechlevel(getLevelName()));
+
 	return true;
 }
 
+std::string mapNameWithoutTechlevel(const char *mapName)
+{
+	ASSERT_OR_RETURN("", mapName != nullptr, "null mapName provided");
+	std::string result(mapName);
+	size_t len = result.length();
+	if (len > 2 && result[len - 3] == '-' && result[len - 2] == 'T' && isdigit(result[len - 1]))
+	{
+		result.resize(len - 3);
+	}
+	return result;
+}
+
 /// returns maps of the right 'type'
 LEVEL_LIST enumerateMultiMaps(int camToUse, int numPlayers)
 {
 	LEVEL_LIST list;
 
-	for (auto lev : psLevels)
+	if (game.type == LEVEL_TYPE::SKIRMISH)
 	{
-		if (game.type == SKIRMISH)
+		// Add maps with exact match to type
+		for (auto lev : psLevels)
 		{
 			int cam = 1;
-			if (lev->type == MULTI_SKIRMISH2)
+			if (lev->type == LEVEL_TYPE::MULTI_SKIRMISH2)
 			{
 				cam = 2;
 			}
-			else if (lev->type == MULTI_SKIRMISH3)
+			else if (lev->type == LEVEL_TYPE::MULTI_SKIRMISH3)
 			{
 				cam = 3;
 			}
+			else if (lev->type == LEVEL_TYPE::MULTI_SKIRMISH4)
+			{
+				cam = 4;
+			}
 
-			if ((lev->type == SKIRMISH || lev->type == MULTI_SKIRMISH2 || lev->type == MULTI_SKIRMISH3)
+			if ((lev->type == LEVEL_TYPE::SKIRMISH || lev->type == LEVEL_TYPE::MULTI_SKIRMISH2 || lev->type == LEVEL_TYPE::MULTI_SKIRMISH3 || lev->type == LEVEL_TYPE::MULTI_SKIRMISH4)
 			    && (numPlayers == 0 || numPlayers == lev->players)
 			    && cam == camToUse)
 			{
 				list.push_back(lev);
 			}
 		}
+		// Also add maps where only the tech level is different, if a more specific map has not been added
+		for (auto lev : psLevels)
+		{
+			if ((lev->type == LEVEL_TYPE::SKIRMISH || lev->type == LEVEL_TYPE::MULTI_SKIRMISH2 || lev->type == LEVEL_TYPE::MULTI_SKIRMISH3 || lev->type == LEVEL_TYPE::MULTI_SKIRMISH4)
+			    && (numPlayers == 0 || numPlayers == lev->players)
+			    && lev->pName)
+			{
+				bool already_added = false;
+				for (auto map : list)
+				{
+					if (!map->pName)
+					{
+						continue;
+					}
+					std::string levelBaseName = mapNameWithoutTechlevel(lev->pName);
+					std::string mapBaseName = mapNameWithoutTechlevel(map->pName);
+					if (strcmp(levelBaseName.c_str(), mapBaseName.c_str()) == 0)
+					{
+						already_added = true;
+						break;
+					}
+				}
+				if (!already_added)
+				{
+					list.push_back(lev);
+				}
+			}
+		}
 	}
 
 	return list;
