@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -35,9 +35,8 @@ Vector3f getTheSun();
 
 void processLight(LIGHT *psLight);
 void initLighting(UDWORD x1, UDWORD y1, UDWORD x2, UDWORD y2);
-void lightValueForTile(UDWORD tileX, UDWORD tileY);
 void doBuildingLights();
-void UpdateFogDistance(float distance);
+void updateFogDistance(float distance);
 void calcDroidIllumination(DROID *psDroid);
 
 #endif // __INCLUDED_SRC_LIGHTNING_H__
