@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -27,7 +27,6 @@
 #include "lib/framework/input.h"
 #include "lib/framework/strres.h"
 #include "lib/framework/wzapp.h"
-#include "lib/framework/rational.h"
 
 #include "lib/ivis_opengl/pieblitfunc.h"
 #include "lib/ivis_opengl/piestate.h" //ivis render code
@@ -36,7 +35,6 @@
 #include "lib/ivis_opengl/screen.h"
 
 #include "lib/gamelib/gtime.h"
-#include "lib/script/script.h"
 #include "lib/sound/audio.h"
 #include "lib/sound/cdaudio.h"
 #include "lib/sound/mixer.h"
@@ -60,7 +58,6 @@
 #include "warzoneconfig.h"
 
 #include "multiplay.h" //ajl
-#include "scripttabs.h"
 #include "levels.h"
 #include "visibility.h"
 #include "multimenu.h"
@@ -78,13 +75,15 @@
 #include "mapgrid.h"
 #include "edit3d.h"
 #include "fpath.h"
-#include "scriptextern.h"
 #include "cmddroid.h"
 #include "keybind.h"
 #include "wrappers.h"
 #include "random.h"
 #include "qtscript.h"
 #include "version.h"
+#include "notifications.h"
+#include "scores.h"
+#include "clparse.h"
 
 #include "warzoneconfig.h"
 
@@ -95,19 +94,18 @@
 #include <numeric>
 
 
-static void fireWaitingCallbacks();
-
 /*
  * Global variables
  */
-unsigned int loopPieCount;
-unsigned int loopPolyCount;
+size_t loopPieCount;
+size_t loopPolyCount;
 
 /*
  * local variables
  */
 static bool paused = false;
 static bool video = false;
+static unsigned short int skipCounter = 0;
 
 //holds which pause is valid at any one time
 struct PAUSE_STATE
@@ -119,6 +117,8 @@ struct PAUSE_STATE
 	bool consolePause;
 };
 static PAUSE_STATE pauseState;
+static size_t maxFastForwardTicks = WZ_DEFAULT_MAX_FASTFORWARD_TICKS;
+static bool fastForwardTicksFixedToNormalTickRate = true; // can be set to false to "catch-up" as quickly as possible (but this may result in more jerky behavior)
 
 static unsigned numDroids[MAX_PLAYERS];
 static unsigned numMissionDroids[MAX_PLAYERS];
@@ -131,7 +131,7 @@ static SDWORD videoMode = 0;
 LOOP_MISSION_STATE		loopMissionState = LMS_NORMAL;
 
 // this is set by scrStartMission to say what type of new level is to be started
-LEVEL_TYPE nextMissionType = LDS_NONE;
+LEVEL_TYPE nextMissionType = LEVEL_TYPE::LDS_NONE;
 
 static GAMECODE renderLoop()
 {
@@ -145,13 +145,17 @@ static GAMECODE renderLoop()
 	wzShowMouse(true);
 
 	INT_RETVAL intRetVal = INT_NONE;
-	CURSOR cursor = CURSOR_DEFAULT;
 	if (!paused)
 	{
+		/* Always refresh the widgets' backing stores if needed, even if we don't process clicks below */
+		intDoScreenRefresh();
+
 		/* Run the in game interface and see if it grabbed any mouse clicks */
-		if (!rotActive && getWidgetsStatus() && dragBox3D.status != DRAG_DRAGGING && wallDrag.status != DRAG_DRAGGING)
+		if (!getRotActive() && getWidgetsStatus() && dragBox3D.status != DRAG_DRAGGING && wallDrag.status != DRAG_DRAGGING)
 		{
 			intRetVal = intRunWidgets();
+			screen_FlipIfBackDropTransition();
+
 			// Send droid orders, if any. (Should do between intRunWidgets() calls, to avoid droid orders getting mixed up, in the case of multiple orders given while the game freezes due to net lag.)
 			sendQueuedDroidInfo();
 		}
@@ -199,8 +203,7 @@ static GAMECODE renderLoop()
 		}
 		if (!scrollPaused() && dragBox3D.status != DRAG_DRAGGING && intMode != INT_INGAMEOP)
 		{
-			cursor = scroll();
-			zoom();
+			displayRenderLoop();
 		}
 	}
 	else  // paused
@@ -210,8 +213,7 @@ static GAMECODE renderLoop()
 
 		if (dragBox3D.status != DRAG_DRAGGING)
 		{
-			cursor = scroll();
-			zoom();
+			displayRenderLoop();
 		}
 
 		if (InGameOpUp || isInGamePopupUp)		// ingame options menu up, run it!
@@ -256,7 +258,7 @@ static GAMECODE renderLoop()
 						ASSERT(false, "Mission Results: saveGame Failed");
 						sstrcpy(msgbuffer, _("Could not save game!"));
 						addConsoleMessage(msgbuffer, LEFT_JUSTIFY, NOTIFY_MESSAGE);
-						deleteSaveGame(sRequestResult);
+						deleteSaveGame_classic(sRequestResult);
 					}
 				}
 				else if (bMultiPlayer || saveMidMission())
@@ -272,7 +274,7 @@ static GAMECODE renderLoop()
 						ASSERT(!"saveGame(sRequestResult, GTYPE_SAVE_MIDMISSION) failed", "Mid Mission: saveGame Failed");
 						sstrcpy(msgbuffer, _("Could not save game!"));
 						addConsoleMessage(msgbuffer, LEFT_JUSTIFY, NOTIFY_MESSAGE);
-						deleteSaveGame(sRequestResult);
+						deleteSaveGame_classic(sRequestResult);
 					}
 				}
 				else
@@ -296,7 +298,7 @@ static GAMECODE renderLoop()
 			pie_LoadBackDrop(SCREEN_RANDOMBDROP);
 		}
 	}
-	if (!loop_GetVideoStatus() && !quitting)
+	if (!loop_GetVideoStatus() && !quitting && !headlessGameMode())
 	{
 		if (!gameUpdatePaused())
 		{
@@ -309,17 +311,15 @@ static GAMECODE renderLoop()
 			processInput();
 
 			//no key clicks or in Intelligence Screen
-			if (!isMouseOverRadar() && intRetVal == INT_NONE && !InGameOpUp && !isInGamePopupUp)
+			if (!isMouseOverRadar() && !isDraggingInGameNotification() && !isMouseClickDownOnScreenOverlayChild() && intRetVal == INT_NONE && !InGameOpUp && !isInGamePopupUp)
 			{
-				CURSOR cursor2 = processMouseClickInput();
-				cursor = cursor2 == CURSOR_DEFAULT? cursor : cursor2;
+				processMouseClickInput();
 			}
 			bRender3DOnly = false;
 			displayWorld();
 		}
 		wzPerfBegin(PERF_GUI, "User interface");
 		/* Display the in game interface */
-		pie_SetDepthBufferStatus(DEPTH_CMP_ALWAYS_WRT_ON);
 		pie_SetFogStatus(false);
 
 		if (bMultiPlayer && bDisplayMultiJoiningStatus)
@@ -332,25 +332,12 @@ static GAMECODE renderLoop()
 		{
 			intDisplayWidgets();
 		}
-		pie_SetDepthBufferStatus(DEPTH_CMP_LEQ_WRT_ON);
 		pie_SetFogStatus(true);
 		wzPerfEnd(PERF_GUI);
 	}
 
-	wzSetCursor(cursor);
-
 	pie_GetResetCounts(&loopPieCount, &loopPolyCount);
 
-	if (!quitting)
-	{
-		/* Check for toggling display mode */
-		if ((keyDown(KEY_LALT) || keyDown(KEY_RALT)) && keyPressed(KEY_RETURN))
-		{
-			war_setFullscreen(!war_getFullscreen());
-			wzToggleFullscreen();
-		}
-	}
-
 	// deal with the mission state
 	switch (loopMissionState)
 	{
@@ -374,8 +361,7 @@ static GAMECODE renderLoop()
 		// just wait for this to be changed when the new mission starts
 		break;
 	case LMS_NEWLEVEL:
-		//nextMissionType = MISSION_NONE;
-		nextMissionType = LDS_NONE;
+		nextMissionType = LEVEL_TYPE::LDS_NONE;
 		return GAMECODE_NEWLEVEL;
 		break;
 	case LMS_LOADGAME:
@@ -400,12 +386,6 @@ static GAMECODE renderLoop()
 
 	if (quitting)
 	{
-		/* Check for toggling display mode */
-		if ((keyDown(KEY_LALT) || keyDown(KEY_RALT)) && keyPressed(KEY_RETURN))
-		{
-			war_setFullscreen(!war_getFullscreen());
-			wzToggleFullscreen();
-		}
 		return GAMECODE_QUITGAME;
 	}
 	else if (loop_GetVideoStatus())
@@ -544,15 +524,6 @@ static void gameStateUpdate()
 
 	if (!paused && !scriptPaused())
 	{
-		/* Update the event system */
-		if (!bInTutorial)
-		{
-			eventProcessTriggers(gameTime / SCR_TICKRATE);
-		}
-		else
-		{
-			eventProcessTriggers(realTime / SCR_TICKRATE);
-		}
 		updateScripts();
 	}
 
@@ -577,8 +548,6 @@ static void gameStateUpdate()
 	// update the command droids
 	cmdDroidUpdate();
 
-	fireWaitingCallbacks(); //Now is the good time to fire waiting callbacks (since interpreter is off now)
-
 	for (unsigned i = 0; i < MAX_PLAYERS; i++)
 	{
 		//update the current power available for a player
@@ -627,9 +596,6 @@ static void gameStateUpdate()
 		featureUpdate(psCFeat);
 	}
 
-	// Clean up dead droid pointers in UI.
-	hciUpdate();
-
 	// Free dead droid memory.
 	objmemUpdate();
 
@@ -638,12 +604,17 @@ static void gameStateUpdate()
 
 	// Must be at the end of gameStateUpdate, since countUpdate is also called randomly (unsynchronised) between gameStateUpdate calls, but should have no effect if we already called it, and recvMessage requires consistent counts on all clients.
 	countUpdate(true);
+}
 
-	static int i = 0;
-	if (i++ % 10 == 0) // trigger every second
-	{
-		jsDebugUpdate();
-	}
+size_t getMaxFastForwardTicks()
+{
+	return maxFastForwardTicks;
+}
+
+void setMaxFastForwardTicks(optional<size_t> value, bool fixedToNormalTickRate)
+{
+	maxFastForwardTicks = value.value_or(WZ_DEFAULT_MAX_FASTFORWARD_TICKS);
+	fastForwardTicksFixedToNormalTickRate = fixedToNormalTickRate;
 }
 
 /* The main game loop */
@@ -651,6 +622,7 @@ GAMECODE gameLoop()
 {
 	static uint32_t lastFlushTime = 0;
 
+	static size_t numForcedUpdatesLastCall = 0;
 	static int renderBudget = 0;  // Scaled time spent rendering minus scaled time spent updating.
 	static bool previousUpdateWasRender = false;
 	const Rational renderFraction(2, 5);  // Minimum fraction of time spent rendering.
@@ -659,14 +631,38 @@ GAMECODE gameLoop()
 	// Shouldn't this be when initialising the game, rather than randomly called between ticks?
 	countUpdate(false); // kick off with correct counts
 
+	size_t numRegularUpdatesTicks = 0;
+	size_t numFastForwardTicks = 0;
+	gameTimeUpdateBegin();
 	while (true)
 	{
 		// Receive NET_BLAH messages.
 		// Receive GAME_BLAH messages, and if it's time, process exactly as many GAME_BLAH messages as required to be able to tick the gameTime.
 		recvMessage();
 
+		bool selectedPlayerIsSpectator = bMultiPlayer && NetPlay.players[selectedPlayer].isSpectator;
+		bool multiplayerHostDisconnected = bMultiPlayer && !NetPlay.isHostAlive && NetPlay.bComms && !NetPlay.isHost; // do not fast-forward after the host has disconnected
+		bool canFastForwardGameTime =
+			selectedPlayerIsSpectator 			// current player must be a spectator
+			&& !NetPlay.isHost					// AND NOT THE HOST (!)
+			&& !multiplayerHostDisconnected		// and the multiplayer host must not be disconnected ("host quit")
+			&& numFastForwardTicks < maxFastForwardTicks // and the number of forced updates this call of gameLoop must not exceed the max allowed
+			&& checkPlayerGameTime(NET_ALL_PLAYERS);	// and there must be a new game tick available to process from all players
+
+		bool forceTryGameTickUpdate = canFastForwardGameTime && ((!fastForwardTicksFixedToNormalTickRate && numForcedUpdatesLastCall > 0) || numRegularUpdatesTicks > 0) && NETgameIsBehindPlayersByAtLeast(4);
+
 		// Update gameTime and graphicsTime, and corresponding deltas. Note that gameTime and graphicsTime pause, if we aren't getting our GAME_GAME_TIME messages.
-		gameTimeUpdate(renderBudget > 0 || previousUpdateWasRender);
+		auto timeUpdateResult = gameTimeUpdate(renderBudget > 0 || previousUpdateWasRender, forceTryGameTickUpdate);
+
+		if (timeUpdateResult == GameTimeUpdateResult::GAME_TIME_UPDATED_FORCED)
+		{
+			numFastForwardTicks++;
+			// FUTURE TODO: Could trigger the display of a UI icon here (if (numFastForwardTicks == maxFastForwardTicks)?) to indicate that the game is fast-forwarding substantially
+		}
+		else if (timeUpdateResult == GameTimeUpdateResult::GAME_TIME_UPDATED)
+		{
+			numRegularUpdatesTicks++;
+		}
 
 		if (deltaGameTime == 0)
 		{
@@ -687,6 +683,7 @@ GAMECODE gameLoop()
 
 		ASSERT(deltaGraphicsTime == 0, "Shouldn't update graphics and game state at once.");
 	}
+	numForcedUpdatesLastCall = numFastForwardTicks;
 
 	if (realTime - lastFlushTime >= 400u)
 	{
@@ -702,6 +699,12 @@ GAMECODE gameLoop()
 	renderBudget = std::min(renderBudget, (renderFraction * 500).floor());
 	previousUpdateWasRender = true;
 
+	if (headlessGameMode() && autogame_enabled())
+	{
+		// Output occasional stats to stdout
+		stdOutGameSummary();
+	}
+
 	return renderReturn;
 }
 
@@ -716,8 +719,13 @@ void videoLoop()
 	videoFinished = !seq_UpdateFullScreenVideo(nullptr);
 	pie_ScreenFlip(CLEAR_BLACK);
 
+	if (skipCounter <= SEQUENCE_MIN_SKIP_DELAY)
+	{
+		skipCounter += 1; // "time" is stopped so we will count via loop iterations.
+	}
+
 	// should we stop playing?
-	if (videoFinished || keyPressed(KEY_ESC) || mouseReleased(MOUSE_LMB))
+	if (videoFinished || (skipCounter > SEQUENCE_MIN_SKIP_DELAY && (keyPressed(KEY_ESC) || mouseReleased(MOUSE_LMB))))
 	{
 		seq_StopFullScreenVideo();
 
@@ -734,12 +742,7 @@ void videoLoop()
 				intResetScreen(true);
 				setMessageImmediate(false);
 			}
-			//don't do the callback if we're playing the win/lose video
-			if (!getScriptWinLoseVideo())
-			{
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_VIDEO_QUIT);
-			}
-			else if (!bMultiPlayer)
+			if (!bMultiPlayer && getScriptWinLoseVideo())
 			{
 				displayGameOver(getScriptWinLoseVideo() == PLAY_WIN, false);
 			}
@@ -751,6 +754,7 @@ void videoLoop()
 
 void loop_SetVideoPlaybackMode()
 {
+	skipCounter = 0;
 	videoMode += 1;
 	paused = true;
 	video = true;
@@ -765,6 +769,7 @@ void loop_SetVideoPlaybackMode()
 
 void loop_ClearVideoPlaybackMode()
 {
+	skipCounter = 0;
 	videoMode -= 1;
 	paused = false;
 	video = false;
@@ -917,18 +922,3 @@ void droidCountsInTransporter(DROID *droid, int player)
 		}
 	}
 }
-
-/* Fire waiting beacon messages which we couldn't run before */
-static void fireWaitingCallbacks()
-{
-	bool bOK = true;
-
-	while (!isMsgStackEmpty() && bOK)
-	{
-		bOK = msgStackFireTop();
-		if (!bOK)
-		{
-			ASSERT(false, "fireWaitingCallbacks: msgStackFireTop() failed (stack count: %d)", msgStackGetCount());
-		}
-	}
-}
