@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
  *
  */
 #include <time.h>
+#include <algorithm>
 
 #include "lib/framework/frame.h"
 #include "lib/framework/endian_hack.h"
@@ -31,6 +32,7 @@
 #include "lib/framework/physfs_ext.h"
 #include "lib/ivis_opengl/tex.h"
 #include "lib/netplay/netplay.h"  // For syncDebug
+#include <wzmaplib/map.h>
 
 #include "map.h"
 #include "hci.h"
@@ -45,10 +47,10 @@
 #include "gateway.h"
 #include "wrappers.h"
 #include "mapgrid.h"
+#include "qtscript.h"
 #include "astar.h"
 #include "fpath.h"
 #include "levels.h"
-#include "scriptfuncs.h"
 #include "lib/framework/wzapp.h"
 
 #define GAME_TICKS_FOR_DANGER (GAME_TICKS_PER_SEC * 2)
@@ -69,6 +71,9 @@ static int lastDangerPlayer = -1;
 //scroll min and max values
 SDWORD		scrollMinX, scrollMaxX, scrollMinY, scrollMaxY;
 
+//For saves to determine if loading the terrain type override should occur
+bool builtInMap;
+
 /* Structure definitions for loading and saving map data */
 struct MAP_SAVEHEADER  // : public GAME_SAVEHEADER
 {
@@ -104,9 +109,9 @@ struct GATEWAY_SAVE
 
 /* The size and contents of the map */
 SDWORD	mapWidth = 0, mapHeight = 0;
-MAPTILE	*psMapTiles = nullptr;
-uint8_t *psBlockMap[AUX_MAX];
-uint8_t *psAuxMap[MAX_PLAYERS + AUX_MAX];        // yes, we waste one element... eyes wide open... makes API nicer
+std::unique_ptr<MAPTILE[]> psMapTiles;
+std::unique_ptr<uint8_t[]> psBlockMap[AUX_MAX];
+std::unique_ptr<uint8_t[]> psAuxMap[MAX_PLAYERS + AUX_MAX];        // yes, we waste one element... eyes wide open... makes API nicer
 
 #define WATER_MIN_DEPTH 500
 #define WATER_MAX_DEPTH (WATER_MIN_DEPTH + 400)
@@ -118,17 +123,14 @@ static void SetDecals(const char *filename, const char *decal_type);
 static void init_tileNames(int type);
 
 /// The different ground types
-GROUND_TYPE *psGroundTypes;
+std::unique_ptr<GROUND_TYPE[]> psGroundTypes;
 int numGroundTypes;
 char *tilesetDir = nullptr;
 static int numTile_names;
-static char *Tile_names = nullptr;
-#define ARIZONA 1
-#define URBAN 2
-#define ROCKIE 3
-
-static int *map;			// 3D array pointer that holds the texturetype
-static bool *mapDecals;           // array that tells us what tile is a decal
+static std::unique_ptr<char[]> Tile_names = nullptr;
+	
+static std::unique_ptr<int[]> map; // 3D array pointer that holds the texturetype
+static std::unique_ptr<bool[]> mapDecals;           // array that tells us what tile is a decal
 #define MAX_TERRAIN_TILES 0x0200  // max that we support (for now), see TILE_NUMMASK
 
 /* Look up table that returns the terrain type of a given tile texture */
@@ -214,9 +216,7 @@ static void init_tileNames(int type)
 	numTile_names = numlines;
 	//increment the pointer to the start of the next record
 	pFileData = strchr(pFileData, '\n') + 1;
-
-	Tile_names = (char *)malloc(numlines * sizeof(char[MAX_STR_LENGTH]));
-	memset(Tile_names, 0x0, (numlines * sizeof(char[MAX_STR_LENGTH])));
+	Tile_names = std::unique_ptr<char[]>(new char[numlines * MAX_STR_LENGTH]());
 
 	for (i = 0; i < numlines; i++)
 	{
@@ -230,7 +230,7 @@ static void init_tileNames(int type)
 // This is the main loading routine to get all the map's parameters set.
 // Once it figures out what tileset we need, we then parse the files for that tileset.
 // Currently, we only support 3 tilesets.  Arizona, Urban, and Rockie
-static bool mapLoadGroundTypes()
+static bool mapLoadGroundTypes(bool preview)
 {
 	char	*pFileData = nullptr;
 	char	tilename[MAX_STR_LENGTH] = {'\0'};
@@ -248,6 +248,11 @@ static bool mapLoadGroundTypes()
 	if (strcmp(tilesetDir, "texpages/tertilesc1hw") == 0)
 	{
 fallback:
+		// load the override terrain types
+		if (!preview && builtInMap && !loadTerrainTypeMapOverride(ARIZONA))
+		{
+			debug(LOG_POPUP, "Failed to load terrain type override");
+		}
 		init_tileNames(ARIZONA);
 		if (!loadFileToBuffer("tileset/tertilesc1hwGtype.txt", pFileData, FILE_LOAD_BUFFER_SIZE, &fileSize))
 		{
@@ -268,7 +273,7 @@ fallback:
 		//increment the pointer to the start of the next record
 		pFileData = strchr(pFileData, '\n') + 1;
 		numGroundTypes = numlines;
-		psGroundTypes = (GROUND_TYPE *)malloc(sizeof(GROUND_TYPE) * numlines);
+		psGroundTypes = std::unique_ptr<GROUND_TYPE[]> (new GROUND_TYPE[numlines]());
 
 		for (i = 0; i < numlines; i++)
 		{
@@ -278,7 +283,7 @@ fallback:
 			pFileData = strchr(pFileData, '\n') + 1;
 
 			psGroundTypes[getTextureType(textureType)].textureName = strdup(textureName);
-			psGroundTypes[getTextureType(textureType)].textureSize = textureSize ;
+			psGroundTypes[getTextureType(textureType)].textureSize = static_cast<float>(textureSize);
 		}
 
 		SetGroundForTile("tileset/arizonaground.txt", "arizona_ground");
@@ -287,6 +292,11 @@ fallback:
 	// for Urban
 	else if (strcmp(tilesetDir, "texpages/tertilesc2hw") == 0)
 	{
+		// load the override terrain types
+		if (!preview && builtInMap && !loadTerrainTypeMapOverride(URBAN))
+		{
+			debug(LOG_POPUP, "Failed to load terrain type override");
+		}
 		init_tileNames(URBAN);
 		if (!loadFileToBuffer("tileset/tertilesc2hwGtype.txt", pFileData, FILE_LOAD_BUFFER_SIZE, &fileSize))
 		{
@@ -307,7 +317,7 @@ fallback:
 		//increment the pointer to the start of the next record
 		pFileData = strchr(pFileData, '\n') + 1;
 		numGroundTypes = numlines;
-		psGroundTypes = (GROUND_TYPE *)malloc(sizeof(GROUND_TYPE) * numlines);
+		psGroundTypes = std::unique_ptr<GROUND_TYPE[]> (new GROUND_TYPE[numlines]());
 
 		for (i = 0; i < numlines; i++)
 		{
@@ -317,7 +327,7 @@ fallback:
 			pFileData = strchr(pFileData, '\n') + 1;
 
 			psGroundTypes[getTextureType(textureType)].textureName = strdup(textureName);
-			psGroundTypes[getTextureType(textureType)].textureSize = textureSize;
+			psGroundTypes[getTextureType(textureType)].textureSize = static_cast<float>(textureSize);
 		}
 
 		SetGroundForTile("tileset/urbanground.txt", "urban_ground");
@@ -326,6 +336,11 @@ fallback:
 	// for Rockie
 	else if (strcmp(tilesetDir, "texpages/tertilesc3hw") == 0)
 	{
+		// load the override terrain types
+		if (!preview && builtInMap && !loadTerrainTypeMapOverride(ROCKIE))
+		{
+			debug(LOG_POPUP, "Failed to load terrain type override");
+		}
 		init_tileNames(ROCKIE);
 		if (!loadFileToBuffer("tileset/tertilesc3hwGtype.txt", pFileData, FILE_LOAD_BUFFER_SIZE, &fileSize))
 		{
@@ -346,7 +361,7 @@ fallback:
 		//increment the pointer to the start of the next record
 		pFileData = strchr(pFileData, '\n') + 1;
 		numGroundTypes = numlines;
-		psGroundTypes = (GROUND_TYPE *)malloc(sizeof(GROUND_TYPE) * numlines);
+		psGroundTypes = std::unique_ptr<GROUND_TYPE[]> (new GROUND_TYPE[numlines]());
 
 		for (i = 0; i < numlines; i++)
 		{
@@ -356,7 +371,7 @@ fallback:
 			pFileData = strchr(pFileData, '\n') + 1;
 
 			psGroundTypes[getTextureType(textureType)].textureName = strdup(textureName);
-			psGroundTypes[getTextureType(textureType)].textureSize = textureSize;
+			psGroundTypes[getTextureType(textureType)].textureSize = static_cast<float>(textureSize);
 		}
 
 		SetGroundForTile("tileset/rockieground.txt", "rockie_ground");
@@ -403,7 +418,7 @@ static void SetGroundForTile(const char *filename, const char *nametype)
 	//increment the pointer to the start of the next record
 	pFileData = strchr(pFileData, '\n') + 1;
 
-	map = (int *)malloc(sizeof(int) * numlines * 2 * 2);	// this is a 3D array map[numlines][2][2]
+	map = std::unique_ptr<int[]> (new int[numlines * 2 * 2]());
 
 	for (i = 0; i < numlines; i++)
 	{
@@ -450,7 +465,7 @@ static void rotFlip(int tile, int *i, int *j)
 {
 	int texture = TileNumber_texture(tile);
 	int rot;
-	int map[2][2], invmap[4][2];
+	int tmpMap[2][2], invmap[4][2];
 
 	if (texture & TILE_XFLIP)
 	{
@@ -461,11 +476,11 @@ static void rotFlip(int tile, int *i, int *j)
 		*j = 1 - *j;
 	}
 
-	map[0][0] = 0; invmap[0][0] = 0; invmap[0][1] = 0;
-	map[1][0] = 1; invmap[1][0] = 1; invmap[1][1] = 0;
-	map[1][1] = 2; invmap[2][0] = 1; invmap[2][1] = 1;
-	map[0][1] = 3; invmap[3][0] = 0; invmap[3][1] = 1;
-	rot = map[*i][*j];
+	tmpMap[0][0] = 0; invmap[0][0] = 0; invmap[0][1] = 0;
+	tmpMap[1][0] = 1; invmap[1][0] = 1; invmap[1][1] = 0;
+	tmpMap[1][1] = 2; invmap[2][0] = 1; invmap[2][1] = 1;
+	tmpMap[0][1] = 3; invmap[3][0] = 0; invmap[3][1] = 1;
+	rot = tmpMap[*i][*j];
 	rot -= (texture & TILE_ROTMASK) >> TILE_ROTSHIFT;
 	while (rot < 0)
 	{
@@ -595,9 +610,9 @@ static void SetDecals(const char *filename, const char *decal_type)
 	debug(LOG_TERRAIN, "reading: %s, with %d entries", filename, numlines);
 	//increment the pointer to the start of the next record
 	pFileData = strchr(pFileData, '\n') + 1;
-	mapDecals = new bool[MAX_TERRAIN_TILES];
-	std::fill_n(mapDecals, MAX_TERRAIN_TILES, false);  // set everything to false.
-
+	// value initialization sets everything to false.
+	mapDecals = std::unique_ptr<bool[]>(new bool[MAX_TERRAIN_TILES]());
+	
 	for (i = 0; i < numlines; i++)
 	{
 		tiledecal = -1;
@@ -732,59 +747,153 @@ static void generateRiverbed()
 
 }
 
-/* Initialise the map structure */
-bool mapLoad(char *filename, bool preview)
-{
-	UDWORD		numGw, width, height;
-	char		aFileType[4];
-	UDWORD		version;
-	UDWORD		i, x, y;
-	PHYSFS_file	*fp = PHYSFS_openRead(filename);
-	MersenneTwister mt(12345);  // 12345 = random seed.
+static bool afterMapLoad();
 
-	if (!fp)
+class WzMapBinaryPhysFSStream : public WzMap::BinaryIOStream
+{
+public:
+	WzMapBinaryPhysFSStream(const char* pFilename, WzMap::BinaryIOStream::OpenMode mode)
+	{
+		switch (mode)
+		{
+			case WzMap::BinaryIOStream::OpenMode::READ:
+				pFile = PHYSFS_openRead(pFilename);
+				break;
+			case WzMap::BinaryIOStream::OpenMode::WRITE:
+				pFile = PHYSFS_openWrite(pFilename);
+				break;
+		}
+		if (pFile)
+		{
+			WZ_PHYSFS_SETBUFFER(pFile, 4096)//;
+		}
+	}
+	virtual ~WzMapBinaryPhysFSStream()
+	{
+		if (pFile)
+		{
+			PHYSFS_close(pFile);
+			pFile = nullptr;
+		}
+	};
+
+	bool openedFile() const { return pFile != nullptr; }
+
+	virtual optional<size_t> readBytes(void *buffer, size_t len) override
+	{
+		if (!pFile) { return nullopt; }
+		PHYSFS_sint64 result = WZ_PHYSFS_readBytes(pFile, buffer, static_cast<uint32_t>(len));
+		if (result < 0)
+		{
+			// failed
+			return nullopt;
+		}
+		return static_cast<size_t>(result);
+	}
+
+	virtual optional<size_t> writeBytes(const void *buffer, size_t len) override
+	{
+		if (!pFile) { return nullopt; }
+		PHYSFS_sint64 result = WZ_PHYSFS_writeBytes(pFile, buffer, static_cast<uint32_t>(len));
+		if (result < 0)
+		{
+			// failed
+			return nullopt;
+		}
+		return static_cast<size_t>(result);
+	}
+
+	virtual bool endOfStream() override
+	{
+		if (!pFile) { return false; }
+		return PHYSFS_eof(pFile);
+	}
+private:
+	PHYSFS_File *pFile = nullptr;
+};
+
+std::unique_ptr<WzMap::BinaryIOStream> WzMapPhysFSIO::openBinaryStream(const std::string& filename, WzMap::BinaryIOStream::OpenMode mode)
+{
+	WzMapBinaryPhysFSStream* pStream = new WzMapBinaryPhysFSStream(filename.c_str(), mode);
+	if (!pStream->openedFile())
+	{
+		delete pStream;
+		return nullptr;
+	}
+	return std::unique_ptr<WzMap::BinaryIOStream>(pStream);
+}
+
+bool WzMapPhysFSIO::loadFullFile(const std::string& filename, std::vector<char>& fileData)
+{
+	if (!PHYSFS_exists(filename.c_str()))
 	{
-		debug(LOG_ERROR, "%s not found", filename);
 		return false;
 	}
-	else if (WZ_PHYSFS_readBytes(fp, aFileType, 4) != 4
-	         || !PHYSFS_readULE32(fp, &version)
-	         || !PHYSFS_readULE32(fp, &width)
-	         || !PHYSFS_readULE32(fp, &height)
-	         || aFileType[0] != 'm'
-	         || aFileType[1] != 'a'
-	         || aFileType[2] != 'p')
-	{
-		debug(LOG_ERROR, "Bad header in %s", filename);
-		goto failure;
-	}
-	else if (version <= VERSION_9)
-	{
-		debug(LOG_ERROR, "%s: Unsupported save format version %u", filename, version);
-		goto failure;
-	}
-	else if (version > CURRENT_VERSION_NUM)
-	{
-		debug(LOG_ERROR, "%s: Undefined save format version %u", filename, version);
-		goto failure;
-	}
-	else if (width * height > MAP_MAXAREA)
-	{
-		debug(LOG_ERROR, "Map %s too large : %d %d", filename, width, height);
-		goto failure;
-	}
+	return loadFileToBufferVector(filename.c_str(), fileData, true, true);
+}
 
-	if (width <= 1 || height <= 1)
+bool WzMapPhysFSIO::writeFullFile(const std::string& filename, const char *ppFileData, uint32_t fileSize)
+{
+	return saveFile(filename.c_str(), ppFileData, fileSize);
+}
+
+WzMapDebugLogger::~WzMapDebugLogger()
+{ }
+
+void WzMapDebugLogger::printLog(WzMap::LoggingProtocol::LogLevel level, const char *function, int line, const char *str)
+{
+	code_part logPart = LOG_INFO;
+	switch (level)
 	{
-		debug(LOG_ERROR, "Map is too small : %u, %u", width, height);
-		goto failure;
+		case WzMap::LoggingProtocol::LogLevel::Info_Verbose:
+			logPart = LOG_NEVER;
+			break;
+		case WzMap::LoggingProtocol::LogLevel::Info:
+			logPart = LOG_MAP;
+			break;
+		case WzMap::LoggingProtocol::LogLevel::Warning:
+			logPart = LOG_WARNING;
+			break;
+		case WzMap::LoggingProtocol::LogLevel::Error:
+			logPart = LOG_ERROR;
+			break;
 	}
+	if (enabled_debug[logPart])
+	{
+		_debug(line, logPart, function, "%s", str);
+	}
+}
+
+/* Initialise the map structure */
+bool mapLoad(char const *filename)
+{
+	WzMapPhysFSIO mapIO;
+	WzMapDebugLogger debugLoggerInstance;
+
+	std::shared_ptr<WzMap::MapData> loadedMap = loadMapData(filename, mapIO, &debugLoggerInstance);
+	if (!loadedMap)
+	{
+		// loadMapData call handles logging errors
+		return false;
+	}
+	return mapLoadFromWzMapData(loadedMap);
+
+}
+
+///* Initialise the map structure */
+bool mapLoadFromWzMapData(std::shared_ptr<WzMap::MapData> loadedMap)
+{
+	uint32_t		width, height;
+	const bool		preview = false;
+
+	width = loadedMap->width;
+	height = loadedMap->height;
 
 	/* See if this is the first time a map has been loaded */
 	ASSERT(psMapTiles == nullptr, "Map has not been cleared before calling mapLoad()!");
 
 	/* Allocate the memory for the map */
-	psMapTiles = (MAPTILE *)calloc(width * height, sizeof(MAPTILE));
+	psMapTiles = std::unique_ptr<MAPTILE[]>(new MAPTILE[width * height]());
 	ASSERT(psMapTiles != nullptr, "Out of memory");
 
 	mapWidth = width;
@@ -797,27 +906,25 @@ bool mapLoad(char *filename, bool preview)
 	}
 
 	// load the ground types
-	if (!mapLoadGroundTypes())
+	if (!mapLoadGroundTypes(preview))
 	{
-		goto failure;
+		return false;
+	}
+
+	if (!preview)
+	{
+		//preload the terrain textures
+		loadTerrainTextures();
 	}
 
 	//load in the map data itself
 
 	/* Load in the map data */
-	for (i = 0; i < mapWidth * mapHeight; i++)
+	for (int i = 0; i < mapWidth * mapHeight; ++i)
 	{
-		UWORD	texture;
-		UBYTE	height;
-
-		if (!PHYSFS_readULE16(fp, &texture) || !PHYSFS_readULE8(fp, &height))
-		{
-			debug(LOG_ERROR, "%s: Error during savegame load", filename);
-			goto failure;
-		}
-
-		psMapTiles[i].texture = texture;
-		psMapTiles[i].height = height * ELEVATION_SCALE;
+		ASSERT(loadedMap->mMapTiles[i].height <= TILE_MAX_HEIGHT, "Tile height (%" PRIu16 ") exceeds TILE_MAX_HEIGHT (%zu)", loadedMap->mMapTiles[i].height, static_cast<size_t>(TILE_MAX_HEIGHT));
+		psMapTiles[i].texture = loadedMap->mMapTiles[i].texture;
+		psMapTiles[i].height = loadedMap->mMapTiles[i].height;
 
 		// Visibility stuff
 		memset(psMapTiles[i].watchers, 0, sizeof(psMapTiles[i].watchers));
@@ -831,38 +938,37 @@ bool mapLoad(char *filename, bool preview)
 	if (preview)
 	{
 		// no need to do anything else for the map preview
-		goto ok;
+		return true;
 	}
 
-	if (!PHYSFS_readULE32(fp, &version) || !PHYSFS_readULE32(fp, &numGw) || version != 1)
+	size_t gwIdx = 0;
+	for (const auto gateway : loadedMap->mGateways)
 	{
-		debug(LOG_ERROR, "Bad gateway in %s", filename);
-		goto failure;
+		if (!gwNewGateway(gateway.x1, gateway.y1, gateway.x2, gateway.y2))
+		{
+			debug(LOG_ERROR, "Unable to add gateway %zu - dropping it", gwIdx);
+		}
+		gwIdx++;
 	}
 
-	for (i = 0; i < numGw; i++)
+	if (!afterMapLoad())
 	{
-		UBYTE	x0, y0, x1, y1;
-
-		if (!PHYSFS_readULE8(fp, &x0) || !PHYSFS_readULE8(fp, &y0) || !PHYSFS_readULE8(fp, &x1) || !PHYSFS_readULE8(fp, &y1))
-		{
-			debug(LOG_ERROR, "%s: Failed to read gateway info", filename);
-			goto failure;
-		}
-		if (!gwNewGateway(x0, y0, x1, y1))
-		{
-			debug(LOG_ERROR, "%s: Unable to add gateway %d - dropping it", filename, i);
-		}
+		return false;
 	}
 
+	return true;
+}
+
+static bool afterMapLoad()
+{
 	if (!mapSetGroundTypes())
 	{
-		goto failure;
+		return false;
 	}
 
-	for (y = 0; y < mapHeight; y++)
+	for (int y = 0; y < mapHeight; ++y)
 	{
-		for (x = 0; x < mapWidth; x++)
+		for (int x = 0; x < mapWidth; ++x)
 		{
 			// FIXME: magic number
 			mapTile(x, y)->waterLevel = mapTile(x, y)->height - world_coord(1) / 3;
@@ -876,18 +982,20 @@ bool mapLoad(char *filename, bool preview)
 	scrollMaxY = mapHeight;
 
 	/* Allocate aux maps */
-	psBlockMap[AUX_MAP] = (uint8_t *)malloc(mapWidth * mapHeight * sizeof(*psBlockMap[0]));
-	psBlockMap[AUX_ASTARMAP] = (uint8_t *)malloc(mapWidth * mapHeight * sizeof(*psBlockMap[0]));
-	psBlockMap[AUX_DANGERMAP] = (uint8_t *)malloc(mapWidth * mapHeight * sizeof(*psBlockMap[0]));
-	for (x = 0; x < MAX_PLAYERS + AUX_MAX; x++)
+	ASSERT(mapWidth >= 0 && mapHeight >= 0, "Invalid mapWidth or mapHeight (%d x %d)", mapWidth, mapHeight);
+	const size_t mapSize = static_cast<size_t>(mapWidth) * static_cast<size_t>(mapHeight);
+	psBlockMap[AUX_MAP] = std::unique_ptr<uint8_t[]>(new uint8_t[mapSize]());
+	psBlockMap[AUX_ASTARMAP] =  std::unique_ptr<uint8_t[]>(new uint8_t[mapSize]());
+	psBlockMap[AUX_DANGERMAP] = std::unique_ptr<uint8_t[]>(new uint8_t[mapSize]());
+	for (int x = 0; x < MAX_PLAYERS + AUX_MAX; ++x)
 	{
-		psAuxMap[x] = (uint8_t *)malloc(mapWidth * mapHeight * sizeof(*psAuxMap[0]));
+		psAuxMap[x] = std::unique_ptr<uint8_t[]> (new uint8_t[mapSize]());
 	}
 
 	// Set our blocking bits
-	for (y = 0; y < mapHeight; y++)
+	for (int y = 0; y < mapHeight; ++y)
 	{
-		for (x = 0; x < mapWidth; x++)
+		for (int x = 0; x < mapWidth; ++x)
 		{
 			MAPTILE *psTile = mapTile(x, y);
 
@@ -916,106 +1024,50 @@ bool mapLoad(char *filename, bool preview)
 
 	/* Set continents. This should ideally be done in advance by the map editor. */
 	mapFloodFillContinents();
-ok:
-	PHYSFS_close(fp);
-	return true;
 
-failure:
-	PHYSFS_close(fp);
-	return false;
+	return true;
 }
 
 /* Save the map data */
-bool mapSave(char **ppFileData, UDWORD *pFileSize)
+bool mapSaveToWzMapData(WzMap::MapData& output)
 {
-	UDWORD	numGateways = gwNumGateways();
+	output.width = mapWidth;
+	output.height = mapHeight;
 
-	/* Allocate the data buffer */
-	static_assert(SAVE_HEADER_SIZE == (sizeof(char) * 4) + (sizeof(UDWORD) * 3), "SAVE_HEADER_SIZE doesn't match?");
-	static_assert(SAVE_TILE_SIZE == sizeof(UWORD) + sizeof(UBYTE), "SAVE_TILE_SIZE doesn't match?");
-	*pFileSize = SAVE_HEADER_SIZE + mapWidth * mapHeight * SAVE_TILE_SIZE;
-	// Add on the size of the gateway data.
-	*pFileSize += (sizeof(UDWORD) * 2) + ((sizeof(UBYTE) * 4) * numGateways);
-
-	*ppFileData = (char *)malloc(*pFileSize);
-	if (*ppFileData == nullptr)
+	// Write out the map tile data
+	uint32_t numMapTiles = output.width * output.height;
+	output.mMapTiles.clear();
+	output.mMapTiles.reserve(numMapTiles);
+	for (uint32_t i = 0; i < numMapTiles; i++)
 	{
-		debug(LOG_FATAL, "Out of memory");
-		abort();
-		return false;
-	}
-	char *pCurrFileData = *ppFileData;
-
-	auto push_uword = [&](UWORD value) {
-		endian_uword(&value);
-		memcpy(pCurrFileData, &value, sizeof(value));
-		pCurrFileData += sizeof(value);
-		ASSERT(pCurrFileData - *ppFileData <= *pFileSize, "Buffer overrun (buffer size: %" PRIu32") (writing to: %" PRIu32")", *pFileSize, static_cast<uint32_t>(pCurrFileData - *ppFileData));
-	};
-	auto push_udword = [&](UDWORD value) {
-		endian_udword(&value);
-		memcpy(pCurrFileData, &value, sizeof(value));
-		pCurrFileData += sizeof(value);
-		ASSERT(pCurrFileData - *ppFileData <= *pFileSize, "Buffer overrun (buffer size: %" PRIu32") (writing to: %" PRIu32")", *pFileSize, static_cast<uint32_t>(pCurrFileData - *ppFileData));
-	};
-
-	/* Put the file header on the file */
-	char aFileType[4];
-	aFileType[0] = 'm';
-	aFileType[1] = 'a';
-	aFileType[2] = 'p';
-	aFileType[3] = ' ';
-	memcpy(pCurrFileData, aFileType, sizeof(aFileType));
-	pCurrFileData += sizeof(aFileType);
-
-	push_udword(CURRENT_VERSION_NUM);
-	push_udword(mapWidth);
-	push_udword(mapHeight);
-
-	/* Put the map data into the buffer */
-	MAPTILE	*psTile = psMapTiles;
-	for (int i = 0; i < mapWidth * mapHeight; i++)
-	{
-		UWORD	texture;
-		UBYTE	height;
-
-		texture = psTile->texture;
-		if (terrainType(psTile) == TER_WATER)
+		WzMap::MapData::MapTile mapDataTile = {};
+		mapDataTile.texture = psMapTiles[i].texture;
+		if (terrainType(&psMapTiles[i]) == TER_WATER)
 		{
-			height = (psTile->waterLevel + world_coord(1) / 3) / ELEVATION_SCALE;
+			mapDataTile.height = (psMapTiles[i].waterLevel + world_coord(1) / 3); // this magic number stuff should match afterMapLoad()'s handling of water tiles (??)
 		}
 		else
 		{
-			height = psTile->height / ELEVATION_SCALE;
+			mapDataTile.height = psMapTiles[i].height;
 		}
-
-		push_uword(texture);
-		memcpy(pCurrFileData, &height, sizeof(height)); pCurrFileData += sizeof(height);
-
-		psTile++;
+		output.mMapTiles.push_back(std::move(mapDataTile));
 	}
 
-	// Put the gateway header.
-	UDWORD version = 1;
-	push_udword(version);
-	push_udword(numGateways);
-
-	// Put the gateway data.
-	UBYTE	x0, y0, x1, y1;
+	// Write out the gateway data
+	output.mGateways.clear();
+	output.mGateways.reserve(gwNumGateways());
 	for (auto psCurrGate : gwGetGateways())
 	{
-		x0 = psCurrGate->x1;
-		y0 = psCurrGate->y1;
-		x1 = psCurrGate->x2;
-		y1 = psCurrGate->y2;
-		ASSERT(x0 == x1 || y0 == y1, "Invalid gateway coordinates (%d, %d, %d, %d)",
-		       x0, y0, x1, y1);
-		ASSERT(x0 < mapWidth && y0 < mapHeight && x1 < mapWidth && y1 < mapHeight,
-		       "Bad gateway dimensions for savegame");
-		memcpy(pCurrFileData, &x0, sizeof(x0)); pCurrFileData += sizeof(x0);
-		memcpy(pCurrFileData, &y0, sizeof(y0)); pCurrFileData += sizeof(y0);
-		memcpy(pCurrFileData, &x1, sizeof(x1)); pCurrFileData += sizeof(x1);
-		memcpy(pCurrFileData, &y1, sizeof(y1)); pCurrFileData += sizeof(y1);
+		WzMap::MapData::Gateway gw = {};
+		gw.x1 = psCurrGate->x1;
+		gw.y1 = psCurrGate->y1;
+		gw.x2 = psCurrGate->x2;
+		gw.y2 = psCurrGate->y2;
+		ASSERT(gw.x1 == gw.x2 || gw.y1 == gw.y2, "Invalid gateway coordinates (%d, %d, %d, %d)",
+			   gw.x1, gw.y1, gw.x2, gw.y2);
+		ASSERT(gw.x1 < mapWidth && gw.y1 < mapHeight && gw.x2 < mapWidth && gw.y2 < mapHeight,
+			   "Bad gateway dimensions for savegame");
+		output.mGateways.push_back(std::move(gw));
 	}
 
 	return true;
@@ -1039,22 +1091,14 @@ bool mapShutdown()
 		dangerDoneSemaphore = nullptr;
 	}
 
-	free(psMapTiles);
-	delete[] mapDecals;
-	free(psGroundTypes);
-	free(map);
-	free(Tile_names);
-	free(psBlockMap[AUX_MAP]);
+	mapDecals = nullptr;
 	psBlockMap[AUX_MAP] = nullptr;
-	free(psBlockMap[AUX_ASTARMAP]);
 	psBlockMap[AUX_ASTARMAP] = nullptr;
-	free(psBlockMap[AUX_DANGERMAP]);
 	free(floodbucket);
 	psBlockMap[AUX_DANGERMAP] = nullptr;
 	for (x = 0; x < MAX_PLAYERS + AUX_MAX; x++)
 	{
-		free(psAuxMap[x]);
-		psAuxMap[x] = nullptr;
+		psAuxMap[x].reset();
 	}
 
 	map = nullptr;
@@ -1541,6 +1585,8 @@ bool writeVisibilityData(const char *fileName)
 		return false;
 	}
 
+	WZ_PHYSFS_SETBUFFER(fileHandle, 4096)//;
+
 	fileHeader.aFileType[0] = 'v';
 	fileHeader.aFileType[1] = 'i';
 	fileHeader.aFileType[2] = 's';
@@ -1871,14 +1917,10 @@ static int dangerThreadFunc(WZ_DECL_UNUSED void *data)
 
 static inline void threatUpdateTarget(int player, BASE_OBJECT *psObj, bool ground, bool air)
 {
-	int i;
-
 	if (psObj->visible[player] || psObj->born == 2)
 	{
-		for (i = 0; i < psObj->numWatchedTiles; i++)
+		for (TILEPOS pos : psObj->watchedTiles)
 		{
-			const TILEPOS pos = psObj->watchedTiles[i];
-
 			if (ground)
 			{
 				auxSet(pos.x, pos.y, MAX_PLAYERS + AUX_DANGERMAP, AUXBITS_THREAT);	// set ground threat for this tile
@@ -1971,7 +2013,7 @@ void mapInit()
 
 	// Start danger thread (not used for campaign for now - mission map swaps too icky)
 	ASSERT(dangerSemaphore == nullptr && dangerThread == nullptr, "Map data not cleaned up before starting!");
-	if (game.type == SKIRMISH)
+	if (game.type == LEVEL_TYPE::SKIRMISH)
 	{
 		for (player = 0; player < MAX_PLAYERS; player++)
 		{
@@ -2007,7 +2049,7 @@ void mapUpdate()
 			}
 		}
 
-	if (gameTime > lastDangerUpdate + GAME_TICKS_FOR_DANGER && game.type == SKIRMISH)
+	if (gameTime > lastDangerUpdate + GAME_TICKS_FOR_DANGER && game.type == LEVEL_TYPE::SKIRMISH)
 	{
 		syncDebug("Do danger maps.");
 		lastDangerUpdate = gameTime;
