@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -29,11 +29,11 @@
 #include "lib/framework/frame.h"
 #include "lib/framework/wzapp.h"
 #include "lib/framework/math_ext.h"
+#include "lib/framework/physfs_ext.h"
 #include "lib/ivis_opengl/bitimage.h"
 #include "lib/ivis_opengl/pieblitfunc.h"
 #include "lib/ivis_opengl/screen.h"
 #include "lib/gamelib/gtime.h"
-#include "lib/script/script.h"
 #include "lib/sound/audio.h"
 #include "lib/sound/audio_id.h"
 #include "lib/sound/cdaudio.h"
@@ -61,7 +61,6 @@
 #include "main.h"
 #include "display.h"
 #include "loadsave.h"
-#include "scripttabs.h"
 #include "cmddroid.h"
 #include "warcam.h"
 #include "wrappers.h"
@@ -74,11 +73,13 @@
 #include "mapgrid.h"
 #include "selection.h"
 #include "scores.h"
-#include "keymap.h"
 #include "texture.h"
 #include "warzoneconfig.h"
 #include "combat.h"
 #include "qtscript.h"
+#include "activity.h"
+#include "lib/framework/wztime.h"
+#include "keybind.h"
 
 #define		IDMISSIONRES_TXT		11004
 #define		IDMISSIONRES_LOAD		11005
@@ -139,9 +140,6 @@ static LANDING_ZONE		sLandingZone[MAX_NOGO_AREAS];
 //flag to indicate when the droids in a Transporter are flown to safety and not the next mission
 static bool             bDroidsToSafety;
 
-// return positions for vtols
-Vector2i asVTOLReturnPos[MAX_PLAYERS];
-
 static UBYTE   missionCountDown;
 //flag to indicate whether the coded mission countdown is played
 static UBYTE   bPlayCountDown;
@@ -192,16 +190,16 @@ static	UDWORD	camNumber = 1;
 //returns true if on an off world mission
 bool missionIsOffworld()
 {
-	return ((mission.type == LDS_MKEEP)
-	        || (mission.type == LDS_MCLEAR)
-	        || (mission.type == LDS_MKEEP_LIMBO)
+	return ((mission.type == LEVEL_TYPE::LDS_MKEEP)
+	        || (mission.type == LEVEL_TYPE::LDS_MCLEAR)
+	        || (mission.type == LEVEL_TYPE::LDS_MKEEP_LIMBO)
 	       );
 }
 
 //returns true if the correct type of mission for reinforcements
 bool missionForReInforcements()
 {
-	if (mission.type == LDS_CAMSTART || missionIsOffworld() || mission.type == LDS_CAMCHANGE)
+	if (mission.type == LEVEL_TYPE::LDS_CAMSTART || missionIsOffworld() || mission.type == LEVEL_TYPE::LDS_CAMCHANGE)
 	{
 		return true;
 	}
@@ -227,14 +225,14 @@ bool missionCanReEnforce()
 //returns true if the mission is a Limbo Expand mission
 bool missionLimboExpand()
 {
-	return (mission.type == LDS_EXPAND_LIMBO);
+	return (mission.type == LEVEL_TYPE::LDS_EXPAND_LIMBO);
 }
 
 // mission initialisation game code
 void initMission()
 {
 	debug(LOG_SAVE, "*** Init Mission ***");
-	mission.type = LDS_NONE;
+	mission.type = LEVEL_TYPE::LDS_NONE;
 	for (int inc = 0; inc < MAX_PLAYERS; inc++)
 	{
 		mission.apsStructLists[inc] = nullptr;
@@ -257,11 +255,11 @@ void initMission()
 	mission.mapWidth = 0;
 	for (auto &i : mission.psBlockMap)
 	{
-		i = nullptr;
+		i.reset();
 	}
 	for (auto &i : mission.psAuxMap)
 	{
-		i = nullptr;
+		i.reset();
 	}
 
 	//init all the landing zones
@@ -270,12 +268,6 @@ void initMission()
 		inc.x1 = inc.y1 = inc.x2 = inc.y2 = 0;
 	}
 
-	// init the vtol return pos
-	for (size_t i = 0; i < MAX_PLAYERS; ++i)
-	{
-		asVTOLReturnPos[i] = Vector2i(0, 0);
-	}
-
 	bDroidsToSafety = false;
 	setPlayCountDown(true);
 
@@ -283,15 +275,6 @@ void initMission()
 	mission.cheatTime = 0;
 }
 
-// reset the vtol landing pos
-void resetVTOLLandingPos()
-{
-	for (size_t i = 0; i < MAX_PLAYERS; ++i)
-	{
-		asVTOLReturnPos[i] = Vector2i(0, 0);
-	}
-}
-
 //this is called everytime the game is quit
 void releaseMission()
 {
@@ -335,26 +318,22 @@ bool missionShutDown()
 		mission.apsSensorList[0] = nullptr;
 		mission.apsOilList[0] = nullptr;
 
-		psMapTiles = mission.psMapTiles;
+		psMapTiles = std::move(mission.psMapTiles);
 		mapWidth = mission.mapWidth;
 		mapHeight = mission.mapHeight;
 		for (int i = 0; i < ARRAY_SIZE(mission.psBlockMap); ++i)
 		{
-			free(psBlockMap[i]);
-			psBlockMap[i] = mission.psBlockMap[i];
-			mission.psBlockMap[i] = nullptr;
+			psBlockMap[i] = std::move(mission.psBlockMap[i]);
 		}
 		for (int i = 0; i < ARRAY_SIZE(mission.psAuxMap); ++i)
 		{
-			free(psAuxMap[i]);
-			psAuxMap[i] = mission.psAuxMap[i];
-			mission.psAuxMap[i] = nullptr;
+			psAuxMap[i] = std::move(mission.psAuxMap[i]);
 		}
 		std::swap(mission.psGateways, gwGetGateways());
 	}
-
+	keybindShutdown();
 	// sorry if this breaks something - but it looks like it's what should happen - John
-	mission.type = LDS_NONE;
+	mission.type = LEVEL_TYPE::LDS_NONE;
 
 	return true;
 }
@@ -413,7 +392,7 @@ bool startMission(LEVEL_TYPE missionType, char *pGame)
 	// this inits the flag so that 'reinforcements have arrived' message isn't played for the first transporter load
 	initFirstTransporterFlag();
 
-	if (mission.type != LDS_NONE)
+	if (mission.type != LEVEL_TYPE::LDS_NONE)
 	{
 		/*mission type gets set to none when you have returned from a mission
 		so don't want to go another mission when already on one! - so ignore*/
@@ -424,7 +403,7 @@ bool startMission(LEVEL_TYPE missionType, char *pGame)
 	initEffectsSystem();
 
 	//load the game file for all types of mission except a Between Mission
-	if (missionType != LDS_BETWEEN)
+	if (missionType != LEVEL_TYPE::LDS_BETWEEN)
 	{
 		loadGameInit(pGame);
 	}
@@ -434,45 +413,45 @@ bool startMission(LEVEL_TYPE missionType, char *pGame)
 
 	switch (missionType)
 	{
-	case LDS_CAMSTART:
+	case LEVEL_TYPE::LDS_CAMSTART:
 		if (!startMissionCampaignStart(pGame))
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_MKEEP:
-	case LDS_MKEEP_LIMBO:
+	case LEVEL_TYPE::LDS_MKEEP:
+	case LEVEL_TYPE::LDS_MKEEP_LIMBO:
 		if (!startMissionOffKeep(pGame))
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_BETWEEN:
+	case LEVEL_TYPE::LDS_BETWEEN:
 		//do anything?
 		if (!startMissionBetween())
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_CAMCHANGE:
+	case LEVEL_TYPE::LDS_CAMCHANGE:
 		if (!startMissionCampaignChange(pGame))
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_EXPAND:
+	case LEVEL_TYPE::LDS_EXPAND:
 		if (!startMissionCampaignExpand(pGame))
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_EXPAND_LIMBO:
+	case LEVEL_TYPE::LDS_EXPAND_LIMBO:
 		if (!startMissionCampaignExpandLimbo(pGame))
 		{
 			loaded = false;
 		}
 		break;
-	case LDS_MCLEAR:
+	case LEVEL_TYPE::LDS_MCLEAR:
 		if (!startMissionOffClear(pGame))
 		{
 			loaded = false;
@@ -515,7 +494,7 @@ bool startMission(LEVEL_TYPE missionType, char *pGame)
 
 
 // initialise the mission stuff for a save game
-bool startMissionSave(SDWORD missionType)
+bool startMissionSave(LEVEL_TYPE missionType)
 {
 	mission.type = missionType;
 
@@ -549,7 +528,7 @@ void addTransporterTimerInterface()
 	W_CLICKFORM     *psForm;
 
 	//check if reinforcements are allowed
-	if (mission.ETA >= 0)
+	if (mission.ETA >= 0 && selectedPlayer < MAX_PLAYERS)
 	{
 		//check the player has at least one Transporter back at base
 		for (DROID *psDroid = mission.apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psDroid->psNext)
@@ -596,31 +575,6 @@ void addTransporterTimerInterface()
 #define OFFSCREEN_HEIGHT	600
 #define	EDGE_SIZE	1
 
-/* pick nearest map edge to point */
-void missionGetNearestCorner(UWORD iX, UWORD iY, UWORD *piOffX, UWORD *piOffY)
-{
-	UDWORD	iMidX = (scrollMinX + scrollMaxX) / 2,
-	        iMidY = (scrollMinY + scrollMaxY) / 2;
-
-	if (map_coord(iX) < iMidX)
-	{
-		*piOffX = (UWORD)(world_coord(scrollMinX) + (EDGE_SIZE * TILE_UNITS));
-	}
-	else
-	{
-		*piOffX = (UWORD)(world_coord(scrollMaxX) - (EDGE_SIZE * TILE_UNITS));
-	}
-
-	if (map_coord(iY) < iMidY)
-	{
-		*piOffY = (UWORD)(world_coord(scrollMinY) + (EDGE_SIZE * TILE_UNITS));
-	}
-	else
-	{
-		*piOffY = (UWORD)(world_coord(scrollMaxY) - (EDGE_SIZE * TILE_UNITS));
-	}
-}
-
 /* fly in transporters at start of level */
 void missionFlyTransportersIn(SDWORD iPlayer, bool bTrackTransporter)
 {
@@ -640,16 +594,14 @@ void missionFlyTransportersIn(SDWORD iPlayer, bool bTrackTransporter)
 	for (psTransporter = mission.apsDroidLists[iPlayer]; psTransporter != nullptr; psTransporter = psNext)
 	{
 		psNext = psTransporter->psNext;
-		// FIXME: When we convert campaign scripts to use DROID_SUPERTRANSPORTER
-		if (psTransporter->droidType == DROID_TRANSPORTER)
+
+		if (psTransporter->droidType == DROID_SUPERTRANSPORTER)
 		{
 			// Check that this transporter actually contains some droids
 			if (psTransporter->psGroup && psTransporter->psGroup->refCount > 1)
 			{
 				// Remove map information from previous map
-				free(psTransporter->watchedTiles);
-				psTransporter->watchedTiles = nullptr;
-				psTransporter->numWatchedTiles = 0;
+				psTransporter->watchedTiles.clear();
 
 				// Remove out of stored list and add to current Droid list
 				if (droidRemove(psTransporter, mission.apsDroidLists))
@@ -710,30 +662,11 @@ static void saveMissionData()
 
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	//clear out the audio
 	audio_StopAll();
 
-	//save the mission data
-	mission.psMapTiles = psMapTiles;
-	mission.mapWidth = mapWidth;
-	mission.mapHeight = mapHeight;
-	for (int i = 0; i < ARRAY_SIZE(mission.psBlockMap); ++i)
-	{
-		mission.psBlockMap[i] = psBlockMap[i];
-	}
-	for (int i = 0; i < ARRAY_SIZE(mission.psAuxMap); ++i)
-	{
-		mission.psAuxMap[i] = psAuxMap[i];
-	}
-	mission.scrollMinX = scrollMinX;
-	mission.scrollMinY = scrollMinY;
-	mission.scrollMaxX = scrollMaxX;
-	mission.scrollMaxY = scrollMaxY;
-	std::swap(mission.psGateways, gwGetGateways());
-	// save the selectedPlayer's LZ
-	mission.homeLZ_X = getLandingX(selectedPlayer);
-	mission.homeLZ_Y = getLandingY(selectedPlayer);
-
 	bRepairExists = false;
 	//set any structures currently being built to completed for the selected player
 	for (psStruct = apsStructLists[selectedPlayer]; psStruct; psStruct = psStruct->psNext)
@@ -747,7 +680,7 @@ static void saveMissionData()
 				    && psStructBeingBuilt == psStruct)
 				{
 					// just give it all its build points
-					structureBuild(psStruct, nullptr, psStruct->pStructureType->buildPoints);
+					structureBuild(psStruct, nullptr, structureBuildPointsToCompletion(*psStruct));
 					//don't bother looking for any other droids working on it
 					break;
 				}
@@ -788,6 +721,27 @@ static void saveMissionData()
 		}
 	}
 
+	//save the mission data
+	mission.psMapTiles = std::move(psMapTiles);
+	mission.mapWidth = mapWidth;
+	mission.mapHeight = mapHeight;
+	for (int i = 0; i < ARRAY_SIZE(mission.psBlockMap); ++i)
+	{
+		mission.psBlockMap[i] = std::move(psBlockMap[i]);
+	}
+	for (int i = 0; i < ARRAY_SIZE(mission.psAuxMap); ++i)
+	{
+		mission.psAuxMap[i] = std::move(psAuxMap[i]);
+	}
+	mission.scrollMinX = scrollMinX;
+	mission.scrollMinY = scrollMinY;
+	mission.scrollMaxX = scrollMaxX;
+	mission.scrollMaxY = scrollMaxY;
+	std::swap(mission.psGateways, gwGetGateways());
+	// save the selectedPlayer's LZ
+	mission.homeLZ_X = getLandingX(selectedPlayer);
+	mission.homeLZ_Y = getLandingY(selectedPlayer);
+
 	for (inc = 0; inc < MAX_PLAYERS; inc++)
 	{
 		mission.apsStructLists[inc] = apsStructLists[inc];
@@ -799,8 +753,8 @@ static void saveMissionData()
 	mission.apsSensorList[0] = apsSensorList[0];
 	mission.apsOilList[0] = apsOilList[0];
 
-	mission.playerX = player.p.x;
-	mission.playerY = player.p.z;
+	mission.playerX = playerPos.p.x;
+	mission.playerY = playerPos.p.z;
 
 	//save the power settings
 	saveMissionPower();
@@ -837,10 +791,10 @@ void restoreMissionData()
 	freeAllStructs();
 	freeAllFeatures();
 	gwShutDown();
-	if (game.type != CAMPAIGN)
+	if (game.type != LEVEL_TYPE::CAMPAIGN)
 	{
 		ASSERT(false, "game type isn't campaign, but we are in a campaign game!");
-		game.type = CAMPAIGN;	// fix the issue, since it is obviously a bug
+		game.type = LEVEL_TYPE::CAMPAIGN;	// fix the issue, since it is obviously a bug
 	}
 	//restore the game pointers
 	for (inc = 0; inc < MAX_PLAYERS; inc++)
@@ -869,19 +823,17 @@ void restoreMissionData()
 	mission.apsSensorList[0] = nullptr;
 	//swap mission data over
 
-	psMapTiles = mission.psMapTiles;
+	psMapTiles = std::move(mission.psMapTiles);
 
 	mapWidth = mission.mapWidth;
 	mapHeight = mission.mapHeight;
 	for (int i = 0; i < ARRAY_SIZE(mission.psBlockMap); ++i)
 	{
-		psBlockMap[i] = mission.psBlockMap[i];
-		mission.psBlockMap[i] = nullptr;
+		psBlockMap[i] = std::move(mission.psBlockMap[i]);
 	}
 	for (int i = 0; i < ARRAY_SIZE(mission.psAuxMap); ++i)
 	{
-		psAuxMap[i] = mission.psAuxMap[i];
-		mission.psAuxMap[i] = nullptr;
+		psAuxMap[i] = std::move(mission.psAuxMap[i]);
 	}
 	scrollMinX = mission.scrollMinX;
 	scrollMinY = mission.scrollMinY;
@@ -917,6 +869,8 @@ void saveMissionLimboData()
 
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	//clear out the audio
 	audio_StopAll();
 
@@ -958,6 +912,8 @@ void placeLimboDroids()
 
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	// Copy the droids across for the selected Player
 	for (psDroid = apsLimboDroids[selectedPlayer]; psDroid != nullptr; psDroid = psNext)
 	{
@@ -1006,6 +962,8 @@ void restoreMissionLimboData()
 
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	/*the droids stored in the mission droid list need to be added back
 	into the current droid list*/
 	for (psDroid = mission.apsDroidLists[selectedPlayer]; psDroid; psDroid = psNext)
@@ -1031,6 +989,8 @@ void saveCampaignData()
 
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	// If the droids have been moved to safety then get any Transporters that exist
 	if (getDroidsToSafetyFlag())
 	{
@@ -1156,7 +1116,6 @@ bool startMissionOffClear(char *pGame)
 	}
 
 	offWorldKeepLists = false;
-	intResetPreviousObj();
 
 	// The message should have been played at the between stage
 	missionCountDown &= ~NOT_PLAYED_ACTIVATED;
@@ -1177,7 +1136,6 @@ bool startMissionOffKeep(char *pGame)
 	}
 
 	offWorldKeepLists = true;
-	intResetPreviousObj();
 
 	// The message should have been played at the between stage
 	missionCountDown &= ~NOT_PLAYED_ACTIVATED;
@@ -1226,7 +1184,6 @@ bool startMissionCampaignChange(char *pGame)
 	}
 
 	offWorldKeepLists = false;
-	intResetPreviousObj();
 
 	return true;
 }
@@ -1268,6 +1225,8 @@ static bool startMissionBetween()
 //check no units left with any settings that are invalid
 static void clearCampaignUnits()
 {
+	if (selectedPlayer >= MAX_PLAYERS) { return; }
+
 	for (DROID *psDroid = apsDroidLists[selectedPlayer]; psDroid; psDroid = psDroid->psNext)
 	{
 		orderDroid(psDroid, DORDER_STOP, ModeImmediate);
@@ -1285,6 +1244,8 @@ static void processMission()
 	UDWORD			droidX, droidY;
 	PICKTILE		pickRes;
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	//and the rest on the mission map  - for now?
 	for (psDroid = apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psNext)
 	{
@@ -1329,6 +1290,8 @@ void processMissionLimbo()
 	DROID			*psNext, *psDroid;
 	UDWORD	numDroidsAddedToLimboList = 0;
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	//all droids (for selectedPlayer only) are placed into the limbo list
 	for (psDroid = apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psNext)
 	{
@@ -1402,7 +1365,17 @@ void swapMissionPointers()
 
 void endMission()
 {
-	if (mission.type == LDS_NONE)
+	if (mission.type != LEVEL_TYPE::LDS_BETWEEN)
+	{
+		releaseAllFlicMessages(apsMessages); //Needed to remove mission objectives from offworld missions
+		releaseObjectives = true;
+	}
+	else
+	{
+		releaseObjectives = false;
+	}
+
+	if (mission.type == LEVEL_TYPE::LDS_NONE)
 	{
 		//can't go back any further!!
 		debug(LOG_SAVE, "Already returned from mission");
@@ -1411,7 +1384,7 @@ void endMission()
 
 	switch (mission.type)
 	{
-	case LDS_CAMSTART:
+	case LEVEL_TYPE::LDS_CAMSTART:
 		//any transporters that are flying in need to be emptied
 		emptyTransporters(false);
 		//when loading in a save game mid cam2a or cam3a it is loaded as a camstart
@@ -1420,34 +1393,34 @@ void endMission()
 			This is called at the end of every campaign mission
 		*/
 		break;
-	case LDS_MKEEP:
+	case LEVEL_TYPE::LDS_MKEEP:
 		//any transporters that are flying in need to be emptied
 		emptyTransporters(true);
 		endMissionOffKeep();
 		break;
-	case LDS_EXPAND:
-	case LDS_BETWEEN:
+	case LEVEL_TYPE::LDS_EXPAND:
+	case LEVEL_TYPE::LDS_BETWEEN:
 		/*
 			This is called at the end of every campaign mission
 		*/
 
 		break;
-	case LDS_CAMCHANGE:
+	case LEVEL_TYPE::LDS_CAMCHANGE:
 		//any transporters that are flying in need to be emptied
 		emptyTransporters(false);
 		endMissionCamChange();
 		break;
 	/* left in so can skip the mission for testing...*/
-	case LDS_EXPAND_LIMBO:
+	case LEVEL_TYPE::LDS_EXPAND_LIMBO:
 		//shouldn't be any transporters on this mission but...who knows?
 		endMissionExpandLimbo();
 		break;
-	case LDS_MCLEAR:
+	case LEVEL_TYPE::LDS_MCLEAR:
 		//any transporters that are flying in need to be emptied
 		emptyTransporters(true);
 		endMissionOffClear();
 		break;
-	case LDS_MKEEP_LIMBO:
+	case LEVEL_TYPE::LDS_MKEEP_LIMBO:
 		//any transporters that are flying in need to be emptied
 		emptyTransporters(true);
 		endMissionOffKeepLimbo();
@@ -1477,7 +1450,7 @@ void endMission()
 
 
 	//mission.type = MISSION_NONE;
-	mission.type = LDS_NONE;
+	mission.type = LEVEL_TYPE::LDS_NONE;
 
 	// reset the transporters
 	initTransporters();
@@ -1536,7 +1509,7 @@ void resetLimboMission()
 	//add the units that were moved into the mission list at the start of the mission
 	restoreMissionLimboData();
 	//set the mission type to plain old expand...
-	mission.type = LDS_EXPAND;
+	mission.type = LEVEL_TYPE::LDS_EXPAND;
 }
 
 /* The update routine for all droids left back at home base
@@ -1579,6 +1552,8 @@ static void missionResetDroids()
 {
 	debug(LOG_SAVE, "called");
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	for (unsigned int player = 0; player < MAX_PLAYERS; player++)
 	{
 		for (DROID *psDroid = apsDroidLists[player]; psDroid != nullptr; psDroid = psDroid->psNext)
@@ -1731,6 +1706,7 @@ void unloadTransporter(DROID *psTransporter, UDWORD x, UDWORD y, bool goingHome)
 	//unload all the droids from within the current Transporter
 	if (isTransporter(psTransporter))
 	{
+		ASSERT(psTransporter->psGroup != nullptr, "psTransporter->psGroup is null??");
 		for (psDroid = psTransporter->psGroup->psList; psDroid != nullptr && psDroid != psTransporter; psDroid = psNext)
 		{
 			psNext = psDroid->psGrpNext;
@@ -1777,7 +1753,6 @@ void unloadTransporter(DROID *psTransporter, UDWORD x, UDWORD y, bool goingHome)
 
 		/* trigger script callback detailing group about to disembark */
 		transporterSetScriptCurrent(psTransporter);
-		eventFireCallbackTrigger((TRIGGER_TYPE)CALL_TRANSPORTER_LANDED);
 		triggerEvent(TRIGGER_TRANSPORTER_LANDED, psTransporter);
 		transporterSetScriptCurrent(nullptr);
 
@@ -1814,12 +1789,11 @@ void missionMoveTransporterOffWorld(DROID *psTransporter)
 {
 	W_CLICKFORM     *psForm;
 	DROID           *psDroid;
-	// FIXME: When we convert campaign scripts, use DROID_SUPERTRANSPORTER
-	if (psTransporter->droidType == DROID_TRANSPORTER)
+
+	if (psTransporter->droidType == DROID_SUPERTRANSPORTER)
 	{
 		/* trigger script callback */
 		transporterSetScriptCurrent(psTransporter);
-		eventFireCallbackTrigger((TRIGGER_TYPE)CALL_TRANSPORTER_OFFMAP);
 		triggerEvent(TRIGGER_TRANSPORTER_EXIT, psTransporter);
 		transporterSetScriptCurrent(nullptr);
 
@@ -1849,6 +1823,7 @@ void missionMoveTransporterOffWorld(DROID *psTransporter)
 		//need a callback for when all the selectedPlayers' reinforcements have been delivered
 		if (psTransporter->player == selectedPlayer)
 		{
+			ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
 			psDroid = nullptr;
 			for (psDroid = mission.apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psDroid->psNext)
 			{
@@ -1859,7 +1834,6 @@ void missionMoveTransporterOffWorld(DROID *psTransporter)
 			}
 			if (psDroid == nullptr)
 			{
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_NO_REINFORCEMENTS_LEFT);
 				triggerEvent(TRIGGER_TRANSPORTER_DONE, psTransporter);
 			}
 		}
@@ -1971,7 +1945,7 @@ bool intAddTransporterTimer()
 	sLabInit = W_LABINIT();
 	sLabInit.formID = IDTRANTIMER_BUTTON;
 	sLabInit.id = IDTRANS_CAPACITY;
-	sLabInit.x = 65;
+	sLabInit.x = 70;
 	sLabInit.y = 1;
 	sLabInit.width = 16;
 	sLabInit.height = 16;
@@ -1996,7 +1970,7 @@ UDWORD  missionGetReinforcementTime()
 }
 
 //fills in a hours(if bHours = true), minutes and seconds display for a given time in 1000th sec
-static void fillTimeDisplay(WzString &text, UDWORD time, bool bHours)
+static void fillTimeDisplay(W_LABEL &Label, UDWORD time, bool bHours)
 {
 	char psText[100];
 	//this is only for the transporter timer - never have hours!
@@ -2006,16 +1980,10 @@ static void fillTimeDisplay(WzString &text, UDWORD time, bool bHours)
 	}
 	else
 	{
-		time_t secs = time / GAME_TICKS_PER_SEC;
-		struct tm tmp;
-#if defined(WZ_OS_WIN)
-		gmtime_s(&tmp, &secs);
-#else
-		gmtime_r(&secs, &tmp);
-#endif
+		struct tm tmp = getUtcTime(static_cast<time_t>(time / GAME_TICKS_PER_SEC));
 		strftime(psText, sizeof(psText), bHours ? "%H:%M:%S" : "%M:%S", &tmp);
 	}
-	text = WzString::fromUtf8(psText);
+	Label.setString(WzString::fromUtf8(psText));
 }
 
 
@@ -2049,7 +2017,7 @@ void intUpdateMissionTimer(WIDGET *psWidget, W_CONTEXT *psContext)
 		timeRemaining = timeElapsed;
 	}
 
-	fillTimeDisplay(Label->aText, timeRemaining, true);
+	fillTimeDisplay(*Label, timeRemaining, true);
 	Label->show();  // Make sure its visible
 
 	if (challengeActive)
@@ -2148,28 +2116,27 @@ void intUpdateTransporterTimer(WIDGET *psWidget, W_CONTEXT *psContext)
 					if (psTransporter->action == DACTION_TRANSPORTWAITTOFLYIN)
 					{
 						missionFlyTransportersIn(selectedPlayer, false);
-						eventFireCallbackTrigger((TRIGGER_TYPE)CALL_TRANSPORTER_REINFORCE);
 						triggerEvent(TRIGGER_TRANSPORTER_ARRIVED, psTransporter);
 					}
 				}
 			}
-			fillTimeDisplay(Label->aText, timeRemaining, false);
+			fillTimeDisplay(*Label, timeRemaining, false);
 		}
 		else
 		{
-			fillTimeDisplay(Label->aText, ETA, false);
+			fillTimeDisplay(*Label, ETA, false);
 		}
 	}
 	else
 	{
 		if (missionCanReEnforce())  // ((mission.type == LDS_MKEEP) || (mission.type == LDS_MCLEAR)) & (mission.ETA >= 0) ) {
 		{
-			fillTimeDisplay(Label->aText, ETA, false);
+			fillTimeDisplay(*Label, ETA, false);
 		}
 		else
 		{
 
-			fillTimeDisplay(Label->aText, 0, false);
+			fillTimeDisplay(*Label, 0, false);
 
 		}
 	}
@@ -2230,20 +2197,28 @@ static void intDisplayMissionBackDrop(WIDGET *psWidget, UDWORD xOffset, UDWORD y
 
 static void missionResetInGameState()
 {
+	// Add the background
+	// get rid of reticule etc..
+	intResetScreen(false);
+
 	//stop the game if in single player mode
 	setMissionPauseState();
 
 	// reset the input state
 	resetInput();
 
-	// Add the background
-	// get rid of reticule etc..
-	intResetScreen(false);
 	forceHidePowerBar();
 	intRemoveReticule();
 	intRemoveMissionTimer();
 }
 
+static void intDestroyMissionResultWidgets()
+{
+	widgDelete(psWScreen, IDMISSIONRES_TITLE);
+	widgDelete(psWScreen, IDMISSIONRES_FORM);
+	widgDelete(psWScreen, IDMISSIONRES_BACKFORM);
+}
+
 static bool _intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDrop)
 {
 	missionResetInGameState();
@@ -2262,6 +2237,9 @@ static bool _intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDr
 		screen_RestartBackDrop();
 	}
 
+	// ensure these widgets are deleted before attempting to create
+	intDestroyMissionResultWidgets();
+
 	sFormInit.formID		= 0;
 	sFormInit.id			= IDMISSIONRES_BACKFORM;
 	sFormInit.style			= WFORM_PLAIN;
@@ -2273,19 +2251,22 @@ static bool _intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDr
 		psWidget->pUserData = nullptr;
 	};
 	W_FORM *missionResBackForm = widgAddForm(psWScreen, &sFormInit);
+	ASSERT_OR_RETURN(false, missionResBackForm != nullptr, "Failed to create IDMISSIONRES_BACKFORM");
 	missionResBackForm->setCalcLayout(LAMBDA_CALCLAYOUT_SIMPLE({
 		psWidget->setGeometry(0 + D_W, 0 + D_H, 640, 480);
 	}));
 
 	// TITLE
-	IntFormAnimated *missionResTitle = new IntFormAnimated(missionResBackForm);
+	auto missionResTitle = std::make_shared<IntFormAnimated>();
+	missionResBackForm->attach(missionResTitle);
 	missionResTitle->id = IDMISSIONRES_TITLE;
 	missionResTitle->setCalcLayout(LAMBDA_CALCLAYOUT_SIMPLE({
 		psWidget->setGeometry(MISSIONRES_TITLE_X, MISSIONRES_TITLE_Y, MISSIONRES_TITLE_W, MISSIONRES_TITLE_H);
 	}));
 
 	// add form
-	IntFormAnimated *missionResForm = new IntFormAnimated(missionResBackForm);
+	auto missionResForm = std::make_shared<IntFormAnimated>();
+	missionResBackForm->attach(missionResForm);
 	missionResForm->id = IDMISSIONRES_FORM;
 	missionResForm->setCalcLayout(LAMBDA_CALCLAYOUT_SIMPLE({
 		psWidget->setGeometry(MISSIONRES_X, MISSIONRES_Y, MISSIONRES_W, MISSIONRES_H);
@@ -2335,25 +2316,26 @@ static bool _intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDr
 	//if won
 	if (result || bMultiPlayer)
 	{
-		//continue
-		sButInit.x			= MISSION_2_X;
-		// Won the game, so display "Quit to main menu"
-		if (testPlayerHasWon() && !bMultiPlayer)
+		// Finished the mission, so display "Continue Game"
+		if (!testPlayerHasWon() || bMultiPlayer)
 		{
-			sButInit.id			= IDMISSIONRES_QUIT;
-			sButInit.y			= MISSION_2_Y - 8;
-			sButInit.pText		= _("Quit To Main Menu");
-			widgAddButton(psWScreen, &sButInit);
-		}
-		else
-		{
-			// Finished the mission, so display "Continue Game"
-			sButInit.y			= MISSION_2_Y;
+			sButInit.x			= MISSION_1_X;
+			sButInit.y			= MISSION_1_Y;
 			sButInit.id			= IDMISSIONRES_CONTINUE;
 			sButInit.pText		= _("Continue Game");//"Continue Game";
 			widgAddButton(psWScreen, &sButInit);
 		}
 
+		// Won the game, so display "Quit to main menu"
+		if (bMultiPlayer || (testPlayerHasWon() && !bMultiPlayer))
+		{
+			sButInit.x			= MISSION_2_X;
+			sButInit.y			= MISSION_2_Y;
+			sButInit.id			= IDMISSIONRES_QUIT;
+			sButInit.pText		= _("Quit To Main Menu");
+			widgAddButton(psWScreen, &sButInit);
+		}
+
 		// FIXME, We got serious issues with savegames at the *END* of some missions, and while they
 		// will load, they don't have the correct state information or other settings.
 		// See transition from CAM2->CAM3 for a example.
@@ -2405,14 +2387,13 @@ static bool _intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDr
 
 bool intAddMissionResult(bool result, bool bPlaySuccess, bool showBackDrop)
 {
+	ActivityManager::instance().completedMission(result, collectEndGameStatsData(), Cheated);
 	return _intAddMissionResult(result, bPlaySuccess, showBackDrop);
 }
 
 void intRemoveMissionResultNoAnim()
 {
-	widgDelete(psWScreen, IDMISSIONRES_TITLE);
-	widgDelete(psWScreen, IDMISSIONRES_FORM);
-	widgDelete(psWScreen, IDMISSIONRES_BACKFORM);
+	intDestroyMissionResultWidgets();
 
 	cdAudio_Stop();
 
@@ -2456,10 +2437,10 @@ void intRunMissionResult()
 
 static void missionContineButtonPressed()
 {
-	if (nextMissionType == LDS_CAMSTART
-	    || nextMissionType == LDS_BETWEEN
-	    || nextMissionType == LDS_EXPAND
-	    || nextMissionType == LDS_EXPAND_LIMBO)
+	if (nextMissionType == LEVEL_TYPE::LDS_CAMSTART
+	    || nextMissionType == LEVEL_TYPE::LDS_BETWEEN
+	    || nextMissionType == LEVEL_TYPE::LDS_EXPAND
+	    || nextMissionType == LEVEL_TYPE::LDS_EXPAND_LIMBO)
 	{
 		launchMission();
 	}
@@ -2540,7 +2521,7 @@ DROID *buildMissionDroid(DROID_TEMPLATE *psTempl, UDWORD x, UDWORD y, UDWORD pla
 	psNewDroid->pos.y = INVALID_XY;
 
 	//set all the droids to selected from when return
-	psNewDroid->selected = true;
+	psNewDroid->selected = isSelectable(psNewDroid);
 
 	// Set died parameter correctly
 	psNewDroid->died = NOT_CURRENT_LIST;
@@ -2552,7 +2533,7 @@ DROID *buildMissionDroid(DROID_TEMPLATE *psTempl, UDWORD x, UDWORD y, UDWORD pla
 void launchMission()
 {
 	//if (mission.type == MISSION_NONE)
-	if (mission.type == LDS_NONE)
+	if (mission.type == LEVEL_TYPE::LDS_NONE)
 	{
 		// tell the loop that a new level has to be loaded up
 		loopMissionState = LMS_NEWLEVEL;
@@ -2565,7 +2546,7 @@ void launchMission()
 
 
 //sets up the game to start a new mission
-bool setUpMission(UDWORD type)
+bool setUpMission(LEVEL_TYPE type)
 {
 	// Close the interface
 	intResetScreen(true);
@@ -2579,7 +2560,7 @@ bool setUpMission(UDWORD type)
 		return false;
 	}
 
-	if (type == LDS_CAMSTART)
+	if (type == LEVEL_TYPE::LDS_CAMSTART)
 	{
 		// Another one of those lovely hacks!!
 		bool    bPlaySuccess = true;
@@ -2596,9 +2577,9 @@ bool setUpMission(UDWORD type)
 		}
 		loopMissionState = LMS_SAVECONTINUE;
 	}
-	else if (type == LDS_MKEEP
-	         || type == LDS_MCLEAR
-	         || type == LDS_MKEEP_LIMBO)
+	else if (type == LEVEL_TYPE::LDS_MKEEP
+	         || type == LEVEL_TYPE::LDS_MCLEAR
+	         || type == LEVEL_TYPE::LDS_MKEEP_LIMBO)
 	{
 		launchMission();
 	}
@@ -2883,7 +2864,6 @@ void missionTimerUpdate()
 			if ((SDWORD)(gameTime - mission.startTime) > mission.time)
 			{
 				//the script can call the end game cos have failed!
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_MISSION_TIME);
 				triggerEvent(TRIGGER_MISSION_TIMEOUT);
 			}
 		}
@@ -2943,6 +2923,7 @@ void missionDestroyObjects()
 	}
 
 	// human player, check that we do not reference the cleared out data
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
 	Player = selectedPlayer;
 
 	psDroid = apsDroidLists[Player];
@@ -2998,6 +2979,8 @@ void processPreviousCampDroids()
 {
 	DROID           *psDroid, *psNext;
 
+	ASSERT(selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
 	// See if any are left
 	if (mission.apsDroidLists[selectedPlayer])
 	{
@@ -3042,6 +3025,7 @@ bool getPlayCountDown()
 //checks to see if the player has any droids (except Transporters left)
 bool missionDroidsRemaining(UDWORD player)
 {
+	ASSERT_OR_RETURN(false, player < MAX_PLAYERS, "invalid player: %" PRIu32 "", player);
 	for (DROID *psDroid = apsDroidLists[player]; psDroid != nullptr; psDroid = psDroid->psNext)
 	{
 		if (!isTransporter(psDroid))
@@ -3094,9 +3078,33 @@ void clearMissionWidgets()
 	intRemoveTransporterLaunch();
 }
 
+/**
+ * Try to find a transporter among the player's droids, or in the mission list (transporter waiting to come back).
+ */
+static DROID *find_transporter()
+{
+	ASSERT_OR_RETURN(nullptr, selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " exceeds MAX_PLAYERS", selectedPlayer);
+
+	for (auto droid_list : {apsDroidLists[selectedPlayer], mission.apsDroidLists[selectedPlayer]})
+	{
+		for (auto droid = droid_list; droid != nullptr; droid = droid->psNext)
+		{
+			if (isTransporter(droid))
+			{
+				return droid;
+			}
+		}
+	}
+
+	return nullptr;
+}
+
 void resetMissionWidgets()
 {
-	DROID       *psDroid;
+	if (mission.type == LEVEL_TYPE::LDS_NONE)
+	{
+		return;
+	}
 
 	//add back any widgets that should be up due to the missions
 	if (mission.time > 0)
@@ -3113,27 +3121,9 @@ void resetMissionWidgets()
 	//check not a typical reinforcement mission
 	else if (!missionForReInforcements())
 	{
-		for (psDroid = apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psDroid->psNext)
+		if (auto transporter = find_transporter())
 		{
-			if (isTransporter(psDroid))
-			{
-				intAddTransporterLaunch(psDroid);
-				break;
-			}
-		}
-		/*if we got to the end without adding a transporter - there might be
-		one sitting in the mission list which is waiting to come back in*/
-		if (!psDroid)
-		{
-			for (psDroid = mission.apsDroidLists[selectedPlayer]; psDroid != nullptr; psDroid = psDroid->psNext)
-			{
-				if (isTransporter(psDroid) &&
-				    psDroid->action == DACTION_TRANSPORTWAITTOFLYIN)
-				{
-					intAddTransporterLaunch(psDroid);
-					break;
-				}
-			}
+			intAddTransporterLaunch(transporter);
 		}
 	}
 
@@ -3150,12 +3140,55 @@ UDWORD	getCampaignNumber()
 	return camNumber;
 }
 
+std::string getCampaignName()
+{
+	UDWORD campaignNum = getCampaignNumber();
+	std::string campaignName;
+	std::vector<CAMPAIGN_FILE> list = readCampaignFiles();
+	if (list.size() >= campaignNum)
+	{
+		campaignName = list[campaignNum - 1].name.toStdString();
+	}
+	return campaignName;
+}
+
+std::vector<CAMPAIGN_FILE> readCampaignFiles()
+{
+	static std::vector<CAMPAIGN_FILE> result;
+	if (!result.empty())
+	{
+		return result;
+	}
+
+	WZ_PHYSFS_enumerateFiles("campaigns", [&](const char *i) -> bool {
+		CAMPAIGN_FILE c;
+		WzString filename("campaigns/");
+		filename += i;
+		if (!filename.endsWith(".json"))
+		{
+			return true; // continue;
+		}
+		WzConfig ini(filename, WzConfig::ReadOnlyAndRequired);
+		c.name = ini.value("name").toWzString();
+		c.level = ini.value("level").toWzString();
+		c.package = ini.value("package").toWzString();
+		c.loading = ini.value("loading").toWzString();
+		c.video = ini.value("video").toWzString();
+		c.captions = ini.value("captions").toWzString();
+		result.push_back(c);
+		return true; // continue
+	});
+	return result;
+}
+
 /*deals with any selectedPlayer's transporters that are flying in when the
 mission ends. bOffWorld is true if the Mission is currently offWorld*/
 void emptyTransporters(bool bOffWorld)
 {
 	DROID       *psTransporter, *psDroid, *psNext, *psNextTrans;
 
+	ASSERT_OR_RETURN(, selectedPlayer < MAX_PLAYERS, "selectedPlayer %" PRIu32 " >= MAX_PLAYERS", selectedPlayer);
+
 	//see if there are any Transporters in the world
 	for (psTransporter = apsDroidLists[selectedPlayer]; psTransporter != nullptr; psTransporter = psNextTrans)
 	{
