@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -27,7 +27,6 @@
 
 #include "droid.h"						// for droid sending and ordering.
 #include "droiddef.h"
-#include "keymap.h"
 #include "stats.h"
 #include "move.h"						// for ordering droids
 #include "objmem.h"
@@ -414,7 +413,8 @@ bool recvDroid(NETQUEUE queue)
 	}
 	NETend();
 
-	if (!getDebugMappingStatus() && bMultiPlayer)
+	const DebugInputManager& dbgInputManager = gInputManager.debugManager();
+	if (!dbgInputManager.debugMappingsAllowed() && bMultiPlayer)
 	{
 		debug(LOG_WARNING, "Failed to add droid for player %u.", NetPlay.players[queue.index].position);
 		return false;
@@ -480,7 +480,7 @@ static void NETQueuedDroidInfo(QueuedDroidInfo *info)
 		}
 		else
 		{
-			NETauto(&info->pos);
+			NETauto(info->pos);
 		}
 		if (info->order == DORDER_BUILD || info->order == DORDER_LINEBUILD)
 		{
@@ -489,11 +489,11 @@ static void NETQueuedDroidInfo(QueuedDroidInfo *info)
 		}
 		if (info->order == DORDER_LINEBUILD)
 		{
-			NETauto(&info->pos2);
+			NETauto(info->pos2);
 		}
 		if (info->order == DORDER_BUILDMODULE)
 		{
-			NETauto(&info->index);
+			NETauto(info->index);
 		}
 		NETbool(&info->add);
 		break;
@@ -507,36 +507,48 @@ static void NETQueuedDroidInfo(QueuedDroidInfo *info)
 // Actually send the droid info.
 void sendQueuedDroidInfo()
 {
-	// Sort queued orders, to group the same order to multiple droids.
-	std::sort(queuedOrders.begin(), queuedOrders.end());
-
-	std::vector<QueuedDroidInfo>::iterator eqBegin, eqEnd;
-	for (eqBegin = queuedOrders.begin(); eqBegin != queuedOrders.end(); eqBegin = eqEnd)
+	// Given an order type, we bring all other orders of the same type together
+	// WHILE KEEPING their relative order!! This is important because 
+	// sending all MOVE and then all HOLD_POSITION is clearly not the same as 
+	// sending all HOLD_POSITION and then MOVE
+	// yes, we need an ordered map here, not std::sort
+	static std::map<DROID_ORDER, std::vector<QueuedDroidInfo> > orderedMap; // static to avoid allocations
+	orderedMap.clear();
+	for (auto &info: queuedOrders)
 	{
-		// Find end of range of orders which differ only by the droid ID.
-		for (eqEnd = eqBegin + 1; eqEnd != queuedOrders.end() && eqEnd->orderCompare(*eqBegin) == 0; ++eqEnd)
-		{}
-
-		NETbeginEncode(NETgameQueue(selectedPlayer), GAME_DROIDINFO);
-		NETQueuedDroidInfo(&*eqBegin);
-
-		uint32_t num = eqEnd - eqBegin;
-		NETuint32_t(&num);
-
-		uint32_t prevDroidId = 0;
-		for (unsigned n = 0; n < num; ++n)
-		{
-			uint32_t droidId = (eqBegin + n)->droidId;
-
-			// Encode deltas between droid IDs, since the deltas are smaller than the actual droid IDs, and will encode to less bytes on average.
-			uint32_t deltaDroidId = droidId - prevDroidId;
-			NETuint32_t(&deltaDroidId);
-
-			prevDroidId = droidId;
-		}
-		NETend();
+		orderedMap[info.order].push_back(info);
 	}
+	std::vector<QueuedDroidInfo>::iterator eqBegin, eqEnd;
+	for (auto &pair: orderedMap)
+	{
+		auto qOrders = pair.second;
+		for (eqBegin = qOrders.begin(); eqBegin != qOrders.end(); eqBegin = eqEnd)
+		{
+			// Find end of range of orders which differ only by the droid ID.
+			for (eqEnd = eqBegin + 1; eqEnd != qOrders.end() && eqEnd->orderCompare(*eqBegin) == 0; ++eqEnd)
+			{}
 
+			NETbeginEncode(NETgameQueue(selectedPlayer), GAME_DROIDINFO);
+			NETQueuedDroidInfo(&*eqBegin);
+
+			uint32_t num = eqEnd - eqBegin;
+			NETuint32_t(&num);
+
+			uint32_t prevDroidId = 0;
+			for (unsigned n = 0; n < num; ++n)
+			{
+				uint32_t droidId = (eqBegin + n)->droidId;
+
+				// Encode deltas between droid IDs, since the deltas are smaller than the actual droid IDs, and will encode to less bytes on average.
+				uint32_t deltaDroidId = droidId - prevDroidId;
+				NETuint32_t(&deltaDroidId);
+
+				prevDroidId = droidId;
+			}
+			NETend();
+		}
+		qOrders.clear();
+	}	
 	// Sent the orders. Don't send them again.
 	queuedOrders.clear();
 }
@@ -563,6 +575,10 @@ void sendDroidInfo(DROID *psDroid, DroidOrder const &order, bool add)
 	{
 		return;
 	}
+	if (NETisReplay())
+	{
+		return;
+	}
 
 	QueuedDroidInfo info;
 
@@ -805,7 +821,8 @@ bool recvDestroyDroid(NETQUEUE queue)
 	}
 	NETend();
 
-	if (!getDebugMappingStatus() && bMultiPlayer)
+	const DebugInputManager& dbgInputManager = gInputManager.debugManager();
+	if (!dbgInputManager.debugMappingsAllowed() && bMultiPlayer)
 	{
 		debug(LOG_WARNING, "Failed to remove droid for player %u.", NetPlay.players[queue.index].position);
 		return false;
