@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -34,15 +34,11 @@
 #include "group.h"
 #include "hci.h"
 #include "order.h"
-#include "scriptfuncs.h"		// for objectinrange.
 #include "lib/gamelib/gtime.h"
 #include "effects.h"
 #include "lib/sound/audio.h"
 #include "lib/sound/audio_id.h"			// for samples.
 #include "wrappers.h"			// for gameover..
-#include "lib/script/script.h"
-#include "scripttabs.h"
-#include "scriptcb.h"
 #include "loop.h"
 #include "transporter.h"
 #include "mission.h" // for INVALID_XY
@@ -55,7 +51,6 @@
 #include "multimenu.h"			// for multimenu
 #include "multistat.h"
 #include "random.h"
-#include "keymap.h"
 
 ///////////////////////////////////////////////////////////////////////////////
 // prototypes
@@ -89,6 +84,23 @@ bool recvGift(NETQUEUE queue)
 		return false;
 	}
 
+	if (to >= MAX_PLAYERS)
+	{
+		debug(LOG_WARNING, "Gift (%d) from %d, to %d (invalid recipient player), queue.index %d", (int)type, (int)from, (int)to, (int)queue.index);
+		syncDebug("Invalid recipient player.");
+		return false;
+	}
+
+	if (bMultiPlayer && (to < NetPlay.players.size()))
+	{
+		if (NetPlay.players[to].isSpectator)
+		{
+			debug(LOG_WARNING, "Can't gift (%d) from %d, to %d (spectator player), queue.index %d", (int)type, (int)from, (int)to, (int)queue.index);
+			syncDebug("Can't gift to spectator.");
+			return false;
+		}
+	}
+
 	// Handle the gift depending on what it is
 	switch (type)
 	{
@@ -134,6 +146,8 @@ bool sendGift(uint8_t type, uint8_t to)
 {
 	int audioTrack;
 
+	ASSERT_OR_RETURN(false, selectedPlayer < MAX_PLAYERS, "Must be a player to send a gift (selectedPlayer: %" PRIu32 "", selectedPlayer);
+
 	switch (type)
 	{
 	case RADAR_GIFT:
@@ -260,7 +274,7 @@ static void recvGiftDroids(uint8_t from, uint8_t to, uint32_t droidID)
 	if (psDroid)
 	{
 		syncDebugDroid(psDroid, '<');
-		giftSingleDroid(psDroid, to);
+		giftSingleDroid(psDroid, to, false);
 		syncDebugDroid(psDroid, '>');
 		if (to == selectedPlayer)
 		{
@@ -436,9 +450,6 @@ void requestAlliance(uint8_t from, uint8_t to, bool prop, bool allowAudio)
 	alliances[to][from] = ALLIANCE_INVITATION;	// They've been invited
 
 
-	CBallFrom = from;
-	CBallTo = to;
-	eventFireCallbackTrigger((TRIGGER_TYPE) CALL_ALLIANCEOFFER);
 	triggerEventAllianceOffer(from, to);
 
 	if (to == selectedPlayer)
@@ -579,14 +590,41 @@ bool recvAlliance(NETQUEUE queue, bool allowAudio)
 		return false;
 	}
 
+	if (to >= MAX_PLAYERS)
+	{
+		debug(LOG_WARNING, "Invalid recipient player (%d), queue.index %d", (int)to, (int)queue.index);
+		return false;
+	}
+
+	auto prohibitedNewAlliance = [](uint8_t from, uint8_t to) -> bool {
+		if (bMultiPlayer)
+		{
+			if ((static_cast<size_t>(from) < NetPlay.players.size()) && NetPlay.players[from].isSpectator)
+			{
+				debug(LOG_WARNING, "Can't enable alliance from %d (spectator), to %d", (int)from, (int)to);
+				syncDebug("Can't enable alliance from spectator.");
+				return true;
+			}
+			if ((static_cast<size_t>(to) < NetPlay.players.size()) && NetPlay.players[to].isSpectator)
+			{
+				debug(LOG_WARNING, "Can't enable alliance from %d, to %d (spectator)", (int)from, (int)to);
+				syncDebug("Can't enable alliance to spectator.");
+				return true;
+			}
+		}
+		return false;
+	};
+
 	switch (state)
 	{
 	case ALLIANCE_NULL:
 		break;
 	case ALLIANCE_REQUESTED:
+		if (prohibitedNewAlliance(from, to)) { return false; }
 		requestAlliance(from, to, false, allowAudio);
 		break;
 	case ALLIANCE_FORMED:
+		if (prohibitedNewAlliance(from, to)) { return false; }
 		formAlliance(from, to, false, allowAudio, true);
 		break;
 	case ALLIANCE_BROKEN:
@@ -674,7 +712,8 @@ void recvMultiPlayerFeature(NETQUEUE queue)
 	}
 	NETend();
 
-	if (!getDebugMappingStatus() && bMultiPlayer)
+	const DebugInputManager& dbgInputManager = gInputManager.debugManager();
+	if (!dbgInputManager.debugMappingsAllowed() && bMultiPlayer)
 	{
 		debug(LOG_WARNING, "Failed to add feature for player %u.", NetPlay.players[queue.index].position);
 		return;
@@ -699,6 +738,12 @@ bool pickupArtefact(int toPlayer, int fromPlayer)
 {
 	if (fromPlayer < MAX_PLAYERS && bMultiPlayer)
 	{
+		if (toPlayer == selectedPlayer && toPlayer == fromPlayer)
+		{
+			audio_QueueTrack(ID_SOUND_ARTIFACT_RECOVERED);
+			return true;
+		}
+
 		for (int topic = asResearch.size() - 1; topic >= 0; topic--)
 		{
 			if (IsResearchCompleted(&asPlayerResList[fromPlayer][topic])
@@ -710,14 +755,14 @@ bool pickupArtefact(int toPlayer, int fromPlayer)
 					MakeResearchPossible(&asPlayerResList[toPlayer][topic]);
 					if (toPlayer == selectedPlayer)
 					{
-						CONPRINTF(_("You Discover Blueprints For %s"), getName(&asResearch[topic]));
+						CONPRINTF(_("You Discover Blueprints For %s"), getStatsName(&asResearch[topic]));
 					}
 					break;
 				}
 				// Invalid topic
 				else
 				{
-					debug(LOG_WARNING, "%s is a invalid research topic?", getName(&asResearch[topic]));
+					debug(LOG_WARNING, "%s is a invalid research topic?", getStatsName(&asResearch[topic]));
 				}
 			}
 		}
@@ -736,19 +781,19 @@ bool pickupArtefact(int toPlayer, int fromPlayer)
 /* Ally team members with each other */
 void createTeamAlliances()
 {
-	int i, j;
-
 	debug(LOG_WZ, "Creating teams");
 
-	for (i = 0; i < MAX_PLAYERS; i++)
+	for (unsigned i = 0; i < MAX_PLAYERS; i++)
 	{
-		for (j = 0; j < MAX_PLAYERS; j++)
+		for (unsigned j = 0; j < MAX_PLAYERS; j++)
 		{
-			if (i != j
-			    && NetPlay.players[i].team == NetPlay.players[j].team	// two different players belonging to the same team
-			    && !aiCheckAlliances(i, j)
-			    && game.skDiff[i]
-			    && game.skDiff[j])	// Not allied and not ignoring teams
+			if (i != j														// two different players
+			    && NetPlay.players[i].team == NetPlay.players[j].team		// ...belonging to the same team
+			    && !aiCheckAlliances(i, j)									// ...not allied and not ignoring teams
+			    && NetPlay.players[i].difficulty != AIDifficulty::DISABLED
+			    && NetPlay.players[j].difficulty != AIDifficulty::DISABLED	// ...not disabled
+			    && !NetPlay.players[i].isSpectator
+			    && !NetPlay.players[j].isSpectator)							// ...not spectators
 			{
 				// Create silently
 				formAlliance(i, j, false, false, false);
