@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -28,8 +28,22 @@
 #include "lib/framework/types.h"
 #include "lib/framework/vector.h"
 #include "lib/framework/crc.h"
+#include "lib/netplay/nettypes.h"
 #include "orderdef.h"
 #include "stringdef.h"
+#include "messagedef.h"
+#include "levels.h"
+#include "console.h"
+#include "multirecv.h"
+#include <vector>
+#include <string>
+#include <chrono>
+
+#include <optional-lite/optional.hpp>
+using nonstd::optional;
+using nonstd::nullopt;
+
+#include <3rdparty/json/json_fwd.hpp>
 
 class DROID_GROUP;
 struct BASE_OBJECT;
@@ -44,8 +58,8 @@ struct STRUCTURE;
 // Also used for skirmish games. And sometimes for campaign. Should be moved somewhere else.
 struct MULTIPLAYERGAME
 {
-	uint8_t		type;						// DMATCH/CAMPAIGN/SKIRMISH/TEAMPLAY etc...
-	bool		scavengers;					// whether scavengers are on or off
+	LEVEL_TYPE	type;						// DMATCH/CAMPAIGN/SKIRMISH/TEAMPLAY etc...
+	uint8_t		scavengers;					// ultimate scavengers, scavengers, no scavengers
 	char		map[128];					// name of multiplayer map being used.
 	uint8_t		maxPlayers;					// max players to allow
 	char		name[128];					// game name   (to be used)
@@ -54,10 +68,14 @@ struct MULTIPLAYERGAME
 	uint32_t    power;						// power level for arena game
 	uint8_t		base;						// clean/base/base&defence
 	uint8_t		alliance;					// no/yes/locked
-	uint8_t		skDiff[MAX_PLAYERS];		// skirmish game difficulty settings. 0x0=OFF 0xff=HUMAN
 	bool		mapHasScavengers;
 	bool		isMapMod;					// if a map has mods
+	bool        isRandom;                   // If a map is non-static.
 	uint32_t	techLevel;					// what technology level is being used
+	uint32_t	inactivityMinutes;			// The number of minutes without active play before a player should be considered "inactive". (0 = disable activity alerts)
+
+	// NOTE: If adding to this struct, a lot of things probably require changing
+	// (send/recvOptions? to/from_json in multiint.h.cpp?)
 };
 
 struct MULTISTRUCTLIMITS
@@ -66,23 +84,58 @@ struct MULTISTRUCTLIMITS
 	uint32_t        limit;
 };
 
+// The side we are *configured* as. Used to indicate whether we are the server or the client. Note that when
+// playing singleplayer, we are running as a host, not a client. Values are booleans as this replaces the old
+// `bool bHostSetup`.
+enum class InGameSide : bool {
+	HOST_OR_SINGLEPLAYER = true,
+	MULTIPLAYER_CLIENT = false
+};
+
 // info used inside games.
 struct MULTIPLAYERINGAME
 {
-	UDWORD				PingTimes[MAX_PLAYERS];				// store for pings.
-	bool				localOptionsReceived;				// used to show if we have game options yet..
-	bool				localJoiningInProgress;				// used before we know our player number.
-	bool				JoiningInProgress[MAX_PLAYERS];
-	bool				DataIntegrity[MAX_PLAYERS];
-	bool				bHostSetup;
-	int32_t				TimeEveryoneIsInGame;
+	UDWORD				PingTimes[MAX_CONNECTED_PLAYERS];				// store for pings.
+	int 				LagCounter[MAX_CONNECTED_PLAYERS];
+	bool				VerifiedIdentity[MAX_CONNECTED_PLAYERS];		// if the multistats identity has been verified.
+	bool				localOptionsReceived;							// used to show if we have game options yet..
+	bool				localJoiningInProgress;							// used before we know our player number.
+	bool				JoiningInProgress[MAX_CONNECTED_PLAYERS];
+	bool				DataIntegrity[MAX_CONNECTED_PLAYERS];
+	InGameSide			side;
+	optional<int32_t>	TimeEveryoneIsInGame;
 	bool				isAllPlayersDataOK;
-	UDWORD				startTime;
-	UDWORD				numStructureLimits;					// number of limits
-	MULTISTRUCTLIMITS	*pStructureLimits;					// limits chunk.
+	std::chrono::steady_clock::time_point startTime;
+	optional<std::chrono::steady_clock::time_point> endTime;
+	std::chrono::steady_clock::time_point lastLagCheck;
+	optional<std::chrono::steady_clock::time_point> lastSentPlayerDataCheck2[MAX_CONNECTED_PLAYERS] = {};
+	std::chrono::steady_clock::time_point lastPlayerDataCheck2;
+	std::vector<MULTISTRUCTLIMITS> structureLimits;
 	uint8_t				flags;  ///< Bitmask, shows which structures are disabled.
-	UDWORD		skScores[MAX_PLAYERS][2];			// score+kills for local skirmish players.
-	char		phrases[5][255];					// 5 favourite text messages.
+#define MPFLAGS_NO_TANKS	0x01  		///< Flag for tanks disabled
+#define MPFLAGS_NO_CYBORGS	0x02  		///< Flag for cyborgs disabled
+#define MPFLAGS_NO_VTOLS	0x04  		///< Flag for VTOLs disabled
+#define MPFLAGS_NO_UPLINK	0x08  		///< Flag for Satellite Uplink disabled
+#define MPFLAGS_NO_LASSAT	0x10  		///< Flag for Laser Satellite Command Post disabled
+#define MPFLAGS_FORCELIMITS	0x20  		///< Flag to force structure limits
+#define MPFLAGS_MAX		0x3f
+	SDWORD		skScores[MAX_PLAYERS][2];			// score+kills for local skirmish players.
+	std::vector<MULTISTRUCTLIMITS> lastAppliedStructureLimits;	// a bit of a hack to retain the structureLimits used when loading / starting a game
+};
+
+struct NetworkTextMessage
+{
+	/**
+	 * Sender can be a player index, SYSTEM_MESSAGE or NOTIFY_MESSAGE.
+	 **/
+	int32_t sender;
+	char text[MAX_CONSOLE_STRING_LENGTH];
+	bool teamSpecific = false;
+
+	NetworkTextMessage() {}
+	NetworkTextMessage(int32_t messageSender, char const *messageText);
+	void enqueue(NETQUEUE queue);
+	bool receive(NETQUEUE queue);
 };
 
 enum STRUCTURE_INFO
@@ -102,9 +155,16 @@ extern MULTIPLAYERINGAME	ingame;						// the game description.
 
 extern bool bMultiPlayer;				// true when more than 1 player.
 extern bool bMultiMessages;				// == bMultiPlayer unless multi messages are disabled
-extern bool openchannels[MAX_PLAYERS];
+extern bool openchannels[MAX_CONNECTED_PLAYERS];
 extern UBYTE bDisplayMultiJoiningStatus;	// draw load progress?
 
+#define RUN_ONLY_ON_SIDE(_side) \
+	if (ingame.side != _side) \
+	{ \
+		return; \
+	}
+
+
 // ////////////////////////////////////////////////////////////////////////////
 // defines
 
@@ -117,17 +177,24 @@ extern UBYTE bDisplayMultiJoiningStatus;	// draw load progress?
 
 #define ANYPLAYER				99
 
+#define NO_SCAVENGERS			0
+#define SCAVENGERS				1
+#define ULTIMATE_SCAVENGERS		2
+
 #define CAMP_CLEAN				0			// campaign subtypes
 #define CAMP_BASE				1
 #define CAMP_WALLS				2
 
-#define PING_LIMIT                              4000                    // If ping is bigger than this, then worry and panic, and don't even try showing the ping.
+#define PING_LIMIT				4000		// If ping is bigger than this, then worry and panic, and don't even try showing the ping.
 
 #define LEV_LOW					0
 #define LEV_MED					1
 #define LEV_HI					2
 
-#define DIFF_SLIDER_STOPS		20			//max number of stops for the multiplayer difficulty slider
+#define TECH_1					1
+#define TECH_2					2
+#define TECH_3					3
+#define TECH_4					4
 
 #define MAX_KICK_REASON			80			// max array size for the reason your kicking someone
 // functions
@@ -153,11 +220,20 @@ Vector3i cameraToHome(UDWORD player, bool scroll);
 
 bool multiPlayerLoop();							// for loop.c
 
+
+enum class HandleMessageAction
+{
+	Process_Message,
+	Silently_Ignore,
+	Disallow_And_Kick_Sender
+};
+HandleMessageAction getMessageHandlingAction(NETQUEUE& queue, uint8_t type);
+bool shouldProcessMessage(NETQUEUE& queue, uint8_t type);
 bool recvMessage();
 bool SendResearch(uint8_t player, uint32_t index, bool trigger);
-bool SendDestroyFeature(FEATURE *pF);					// send a destruct feature message.
-bool sendTextMessage(const char *pStr, bool cast, uint32_t from = selectedPlayer);	// send a text message
-void sendTeamMessage(const char *pStr, uint32_t from = selectedPlayer);	// send a team chat message
+void printInGameTextMessage(NetworkTextMessage const &message);
+void sendInGameSystemMessage(const char *text);
+int32_t findPlayerIndexByPosition(uint32_t position);
 void printConsoleNameChange(const char *oldName, const char *newName);  ///< Print message to console saying a name changed.
 
 void turnOffMultiMsg(bool bDoit);
@@ -179,7 +255,6 @@ bool SendDroid(DROID_TEMPLATE *pTemplate, uint32_t x, uint32_t y, uint8_t player
 bool SendDestroyDroid(const DROID *psDroid);
 void sendQueuedDroidInfo();  ///< Actually sends the droid orders which were queued by SendDroidInfo.
 void sendDroidInfo(DROID *psDroid, DroidOrder const &order, bool add);
-bool SendCmdGroup(DROID_GROUP *psGroup, UWORD x, UWORD y, BASE_OBJECT *psObj);
 
 bool sendDroidSecondary(const DROID *psDroid, SECONDARY_ORDER sec, SECONDARY_STATE state);
 bool sendDroidDisembark(DROID const *psTransporter, DROID const *psDroid);
@@ -188,42 +263,65 @@ bool sendDroidDisembark(DROID const *psTransporter, DROID const *psDroid);
 bool multiShutdown();
 bool sendLeavingMsg();
 
-bool hostCampaign(char *sGame, char *sPlayer);
-bool joinGame(const char *host, uint32_t port);
+bool hostCampaign(const char *SessionName, char *hostPlayerName, bool spectatorHost, bool skipResetAIs);
+struct JoinConnectionDescription
+{
+public:
+	JoinConnectionDescription() { }
+	JoinConnectionDescription(const std::string& host, uint32_t port)
+	: host(host)
+	, port(port)
+	{ }
+public:
+	std::string host;
+	uint32_t port = 0;
+};
+std::vector<JoinConnectionDescription> findLobbyGame(const std::string& lobbyAddress, unsigned int lobbyPort, uint32_t lobbyGameId);
+enum class JoinGameResult {
+	FAILED,
+	JOINED,
+	PENDING_PASSWORD
+};
+JoinGameResult joinGame(const char *connectionString, bool asSpectator = false);
+JoinGameResult joinGame(const char *host, uint32_t port, bool asSpectator = false);
+JoinGameResult joinGame(const std::vector<JoinConnectionDescription>& connection_list, bool asSpectator = false);
 void playerResponding();
 bool multiGameInit();
 bool multiGameShutdown();
 
 // syncing.
 bool sendScoreCheck();							//score check only(frontend)
+void multiSyncResetAllChallenges();
+void multiSyncResetPlayerChallenge(uint32_t playerIdx);
+void multiSyncPlayerSwap(uint32_t playerIndexA, uint32_t playerIndexB);
 bool sendPing();							// allow game to request pings.
 void HandleBadParam(const char *msg, const int from, const int actual);
 // multijoin
-bool sendResearchStatus(STRUCTURE *psBuilding, UDWORD index, UBYTE player, bool bStart);
-
-void displayAIMessage(char *pStr, SDWORD from, SDWORD to);  //make AI process a message
-
-
-/* for multiplayer message stack */
-UDWORD	msgStackPush(SDWORD CBtype, SDWORD plFrom, SDWORD plTo, const char *tStr, SDWORD x, SDWORD y, DROID *psDroid);
-bool	isMsgStackEmpty();
-bool	msgStackGetFrom(SDWORD  *psVal);
-bool	msgStackGetTo(SDWORD  *psVal);
-bool	msgStackGetMsg(char  *psVal);
-bool	msgStackPop();
-SDWORD	msgStackGetCount();
-void	msgStackReset();
-bool msgStackGetDroid(DROID **ppsDroid);
+bool sendResearchStatus(const STRUCTURE *psBuilding, UDWORD index, UBYTE player, bool bStart);
 
 bool sendBeacon(int32_t locX, int32_t locY, int32_t forPlayer, int32_t sender, const char *pStr);
-bool msgStackFireTop();
 
-bool multiplayPlayersReady(bool bNotifyStatus);
 void startMultiplayerGame();
-void resetReadyStatus(bool bSendOptions);
+void resetReadyStatus(bool bSendOptions, bool ignoreReadyReset = false);
 
 STRUCTURE *findResearchingFacilityByResearchIndex(unsigned player, unsigned index);
 
-void sendSyncRequest(int32_t req_id, int32_t x, int32_t y, BASE_OBJECT *psObj, BASE_OBJECT *psObj2);
+void sendSyncRequest(int32_t req_id, int32_t x, int32_t y, const BASE_OBJECT *psObj, const BASE_OBJECT *psObj2);
+
+
+bool sendBeaconToPlayer(SDWORD locX, SDWORD locY, SDWORD forPlayer, SDWORD sender, const char *beaconMsg);
+MESSAGE *findBeaconMsg(UDWORD player, SDWORD sender);
+VIEWDATA *CreateBeaconViewData(SDWORD sender, UDWORD LocX, UDWORD LocY);
+
+bool makePlayerSpectator(uint32_t player_id, bool removeAllStructs = false, bool quietly = false);
+
+class WZGameReplayOptionsHandler : public ReplayOptionsHandler
+{
+	virtual bool saveOptions(nlohmann::json& object) const override;
+	virtual bool saveMap(EmbeddedMapData& mapData) const override;
+	virtual bool restoreOptions(const nlohmann::json& object, EmbeddedMapData&& embeddedMapData, uint32_t replay_netcodeMajor, uint32_t replay_netcodeMinor) override;
+	virtual size_t desiredBufferSize() const override;
+	virtual size_t maximumEmbeddedMapBufferSize() const override;
+};
 
 #endif // __INCLUDED_SRC_MULTIPLAY_H__
