@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -33,10 +33,6 @@
 #include "map.h"
 #include "power.h"
 #include "objects.h"
-#include "lib/script/script.h"
-#include "scriptvals.h"
-#include "scripttabs.h"
-#include "scriptcb.h"
 #include "mission.h"
 #include "structuredef.h"
 #include "structure.h"
@@ -175,12 +171,6 @@ void objmemUpdate()
 	objListIntegCheck();
 #endif
 
-	// tell the script system about any destroyed objects
-	if (psDestroyedObj != nullptr)
-	{
-		scrvUpdateBasePointers();
-	}
-
 	/* Go through the destroyed objects list looking for objects that
 	   were destroyed before this turn */
 
@@ -207,25 +197,7 @@ void objmemUpdate()
 		else
 		{
 			// do the object died callback
-			psCBObjDestroyed = psCurr;
-			eventFireCallbackTrigger((TRIGGER_TYPE)CALL_OBJ_DESTROYED);
-			switch (psCurr->type)
-			{
-			case OBJ_DROID:
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_DROID_DESTROYED);
-				break;
-			case OBJ_STRUCTURE:
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_STRUCT_DESTROYED);
-				break;
-			case OBJ_FEATURE:
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_FEATURE_DESTROYED);
-				break;
-			default:
-				break;
-			}
-			psCBObjDestroyed = nullptr;
 			triggerEventDestroyed(psCurr);
-
 			psPrev = psCurr;
 		}
 	}
@@ -673,11 +645,27 @@ bool createFlagPosition(FLAG_POSITION **ppsNew, UDWORD player)
 	return true;
 }
 
+static bool isFlagPositionInList(FLAG_POSITION *psFlagPosToAdd)
+{
+	ASSERT_OR_RETURN(false, psFlagPosToAdd != nullptr, "Invalid FlagPosition pointer");
+	ASSERT_OR_RETURN(false, psFlagPosToAdd->player < MAX_PLAYERS, "Invalid FlagPosition player: %u", psFlagPosToAdd->player);
+	for (FLAG_POSITION* psCurr = apsFlagPosLists[psFlagPosToAdd->player]; (psCurr != nullptr); psCurr = psCurr->psNext)
+	{
+		if (psCurr == psFlagPosToAdd)
+		{
+			return true;
+		}
+	}
+	return false;
+}
+
 /* add the Flag Position to the Flag Position Lists */
 void addFlagPosition(FLAG_POSITION *psFlagPosToAdd)
 {
 	ASSERT_OR_RETURN(, psFlagPosToAdd != nullptr, "Invalid FlagPosition pointer");
 	ASSERT_OR_RETURN(, psFlagPosToAdd->coords.x != ~0, "flag has invalid position");
+	ASSERT_OR_RETURN(, psFlagPosToAdd->player < MAX_PLAYERS, "Invalid FlagPosition player: %u", psFlagPosToAdd->player);
+	ASSERT_OR_RETURN(, !isFlagPositionInList(psFlagPosToAdd), "FlagPosition is already in the list!");
 
 	psFlagPosToAdd->psNext = apsFlagPosLists[psFlagPosToAdd->player];
 	apsFlagPosLists[psFlagPosToAdd->player] = psFlagPosToAdd;
@@ -965,14 +953,6 @@ UDWORD getRepairIdFromFlag(FLAG_POSITION *psFlag)
 	return UDWORD_MAX;
 }
 
-
-// check a base object exists for an ID
-bool checkValidId(UDWORD id)
-{
-	return getBaseObjFromId(id) != nullptr;
-}
-
-
 // integrity check the lists
 #ifdef DEBUG
 static void objListIntegCheck()
@@ -997,7 +977,7 @@ static void objListIntegCheck()
 			ASSERT(psCurr->type == OBJ_STRUCTURE &&
 			       (SDWORD)psCurr->player == player,
 			       "objListIntegCheck: misplaced %s(%p) in the structure list for player %d, is owned by %d",
-			       objInfo(psCurr), psCurr, player, (int)psCurr->player);
+			       objInfo(psCurr), (void*) psCurr, player, (int)psCurr->player);
 		}
 	}
 	for (psCurr = (BASE_OBJECT *)apsFeatureLists[0]; psCurr; psCurr = psCurr->psNext)
