@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -37,14 +37,12 @@
 #include "lib/sequence/sequence.h"
 #include "lib/sound/audio.h"
 #include "lib/sound/cdaudio.h"
-#include "lib/script/script.h"
 
 #include "seqdisp.h"
 
 #include "warzoneconfig.h"
 #include "hci.h"//for font
 #include "loop.h"
-#include "scripttabs.h"
 #include "design.h"
 #include "wrappers.h"
 #include "init.h" // For fileLoadBuffer
@@ -281,7 +279,7 @@ bool seq_UpdateFullScreenVideo(int *pbClear)
 	unsigned int subMax = SUBTITLE_BOX_MIN + D_H2;
 
 	//get any text lines over bottom of the video
-	double realTime = seq_GetFrameTime();
+	double frameTime = seq_GetFrameTime();
 	for (i = 0; i < MAX_TEXT_OVERLAYS; i++)
 	{
 		SEQTEXT seqtext = aSeqList[currentPlaySeq].aText[i];
@@ -289,7 +287,7 @@ bool seq_UpdateFullScreenVideo(int *pbClear)
 		{
 			if (seqtext.bSubtitle)
 			{
-				if (((realTime >= seqtext.startTime) && (realTime <= seqtext.endTime)) ||
+				if (((frameTime >= seqtext.startTime) && (frameTime <= seqtext.endTime)) ||
 				    aSeqList[currentPlaySeq].bSeqLoop) //if its a looped video always draw the text
 				{
 					if (subMin > seqtext.y && seqtext.y > SUBTITLE_BOX_MIN)
@@ -303,7 +301,7 @@ bool seq_UpdateFullScreenVideo(int *pbClear)
 				}
 			}
 
-			if (realTime >= seqtext.endTime && realTime < seqtext.endTime)
+			if (frameTime >= seqtext.endTime && frameTime < seqtext.endTime)
 			{
 				if (pbClear != nullptr)
 				{
@@ -333,14 +331,14 @@ bool seq_UpdateFullScreenVideo(int *pbClear)
 	//call sequence player to download last frame
 	stillPlaying = seq_Update();
 	//print any text over the video
-	realTime = seq_GetFrameTime();
+	frameTime = seq_GetFrameTime();
 
 	for (i = 0; i < MAX_TEXT_OVERLAYS; i++)
 	{
 		SEQTEXT currentText = aSeqList[currentPlaySeq].aText[i];
 		if (currentText.pText[0] != '\0')
 		{
-			if (((realTime >= currentText.startTime) && (realTime <= currentText.endTime)) ||
+			if (((frameTime >= currentText.startTime) && (frameTime <= currentText.endTime)) ||
 			    (aSeqList[currentPlaySeq].bSeqLoop)) //if its a looped video always draw the text
 			{
 				if (i >= wzCachedSeqText.size())
@@ -388,6 +386,12 @@ bool seq_UpdateFullScreenVideo(int *pbClear)
 	return true;
 }
 
+void seqReleaseAll()
+{
+	seq_Shutdown();
+	wzCachedSeqText.clear();
+}
+
 bool seq_StopFullScreenVideo()
 {
 	if (!seq_AnySeqLeft())
@@ -524,8 +528,8 @@ static bool seq_AddTextFromFile(const char *pTextName, SEQ_TEXT_POSITIONING text
 				// Since all the positioning was hardcoded to specific values, we now calculate the
 				// ratio of our screen, compared to what the game expects and multiply that to x, y.
 				// This makes the text always take up the full screen, instead of original style.
-				xOffset = (double)pie_GetVideoBufferWidth() / 640. * (double)xOffset;
-				yOffset = (double)pie_GetVideoBufferHeight() / 480. * (double)yOffset;
+				xOffset = static_cast<SDWORD>((double)pie_GetVideoBufferWidth() / 640. * (double)xOffset);
+				yOffset = static_cast<SDWORD>((double)pie_GetVideoBufferHeight() / 480. * (double)yOffset);
 				//get the text
 				pText = strrchr(pCurrentLine, '"');
 				ASSERT(pText != nullptr, "error parsing text file");
@@ -623,17 +627,7 @@ void seq_StartNextFullScreenVideo()
 	if (bPlayedOK == false)
 	{
 		//don't do the callback if we're playing the win/lose video
-		if (!getScriptWinLoseVideo())
-		{
-			debug(LOG_SCRIPT, "*** Called video quit trigger!");
-			// Not sure this is correct... CHECK, since the callback should ONLY
-			// be called when a video is playing (always?)
-			if (seq_Playing())
-			{
-				eventFireCallbackTrigger((TRIGGER_TYPE)CALL_VIDEO_QUIT);
-			}
-		}
-		else
+		if (getScriptWinLoseVideo())
 		{
 			displayGameOver(getScriptWinLoseVideo() == PLAY_WIN, false);
 		}
