@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -43,6 +43,8 @@
 
 #define WEAPON_TIME		100
 
+#define DEFAULT_DROID_RESISTANCE	150
+
 /* The stores for the different stats */
 BODY_STATS		*asBodyStats;
 BRAIN_STATS		*asBrainStats;
@@ -69,20 +71,6 @@ UDWORD		numRepairStats;
 UDWORD		numWeaponStats;
 UDWORD		numConstructStats;
 
-//the max values of the stats used in the design screen
-static UDWORD	maxComponentWeight;
-static UDWORD	maxBodyArmour;
-static UDWORD	maxBodyPower;
-static UDWORD	maxBodyPoints;
-static UDWORD	maxSensorRange;
-static UDWORD	maxECMRange;
-static UDWORD	maxConstPoints;
-static UDWORD	maxRepairPoints;
-static UDWORD	maxWeaponRange;
-static UDWORD	maxWeaponDamage;
-static UDWORD	maxWeaponROF;
-static UDWORD	maxPropulsionSpeed;
-
 //stores for each players component states - can be either UNAVAILABLE, REDUNDANT, FOUND or AVAILABLE
 UBYTE		*apCompLists[MAX_PLAYERS][COMP_NUMCOMPONENTS];
 
@@ -90,32 +78,11 @@ UBYTE		*apCompLists[MAX_PLAYERS][COMP_NUMCOMPONENTS];
 UBYTE		*apStructTypeLists[MAX_PLAYERS];
 
 static std::unordered_map<WzString, BASE_STATS *> lookupStatPtr;
+static std::unordered_map<WzString, COMPONENT_STATS *> lookupCompStatPtr;
 
 static bool getMovementModel(const WzString &movementModel, MOVEMENT_MODEL *model);
 static bool statsGetAudioIDFromString(const WzString &szStatName, const WzString &szWavName, int *piWavID);
 
-//Access functions for the max values to be used in the Design Screen
-static void setMaxComponentWeight(UDWORD weight);
-static void setMaxBodyArmour(UDWORD armour);
-static void setMaxBodyPower(UDWORD power);
-static void setMaxBodyPoints(UDWORD points);
-static void setMaxSensorRange(UDWORD range);
-static void setMaxECMRange(UDWORD power);
-static void setMaxConstPoints(UDWORD points);
-static void setMaxRepairPoints(UDWORD repair);
-static void setMaxWeaponRange(UDWORD range);
-static void setMaxWeaponDamage(UDWORD damage);
-static void setMaxWeaponROF(UDWORD rof);
-static void setMaxPropulsionSpeed(UDWORD speed);
-
-//determine the effect this upgrade would have on the max values
-static void updateMaxWeaponStats(UWORD maxValue);
-static void updateMaxSensorStats(UWORD maxRange);
-static void updateMaxRepairStats(UWORD maxValue);
-static void updateMaxECMStats(UWORD maxValue);
-static void updateMaxBodyStats(UWORD maxBody, UWORD maxPower, UWORD maxArmour);
-static void updateMaxConstStats(UWORD maxValue);
-
 /***********************************************************************************
 *	Dealloc the extra storage tables
 ***********************************************************************************/
@@ -138,7 +105,7 @@ static void deallocTerrainTable()
 
 /* Macro to allocate memory for a set of stats */
 #define ALLOC_STATS(numEntries, list, listSize, type) \
-	ASSERT((numEntries) < REF_RANGE, "Number of stats entries too large for " #type);\
+	ASSERT((numEntries) < ~STAT_MASK + 1, "Number of stats entries too large for " #type);\
 	if ((list))	delete [] (list);	\
 	(list) = new type[numEntries]; \
 	(listSize) = (numEntries); \
@@ -161,18 +128,13 @@ void statsInitVars()
 	numRepairStats = 0;
 	numWeaponStats = 0;
 	numConstructStats = 0;
-
-	// init the max values
-	maxComponentWeight = maxBodyArmour = maxBodyPower =
-	        maxBodyPoints = maxSensorRange = maxECMRange =
-	                            maxConstPoints = maxRepairPoints = maxWeaponRange = maxWeaponDamage =
-	                                        maxPropulsionSpeed = 0;
 }
 
 /*Deallocate all the stats assigned from input data*/
 bool statsShutDown()
 {
 	lookupStatPtr.clear();
+	lookupCompStatPtr.clear();
 
 	STATS_DEALLOC(asWeaponStats, numWeaponStats);
 	STATS_DEALLOC(asBrainStats, numBrainStats);
@@ -258,12 +220,12 @@ static iIMDShape *statsGetIMD(WzConfig &json, BASE_STATS *psStats, const WzStrin
 		WzString filename = json_variant(value).toWzString();
 		retval = modelGet(filename);
 		ASSERT(retval != nullptr, "Cannot find the PIE model %s for stat %s in %s",
-		       filename.toUtf8().c_str(), getName(psStats), json.fileName().toUtf8().c_str());
+		       filename.toUtf8().c_str(), getStatsName(psStats), json.fileName().toUtf8().c_str());
 	}
 	return retval;
 }
 
-void loadStats(WzConfig &json, BASE_STATS *psStats, size_t index)
+static void loadStats(WzConfig &json, BASE_STATS *psStats, size_t index)
 {
 	psStats->id = json.group();
 	psStats->name = json.string("name");
@@ -272,15 +234,26 @@ void loadStats(WzConfig &json, BASE_STATS *psStats, size_t index)
 	lookupStatPtr.insert(std::make_pair(psStats->id, psStats));
 }
 
+void loadStructureStats_BaseStats(WzConfig &json, STRUCTURE_STATS *psStats, size_t index)
+{
+	loadStats(json, psStats, index);
+}
+
+void unloadStructureStats_BaseStats(const STRUCTURE_STATS &psStats)
+{
+	lookupStatPtr.erase(psStats.id);
+}
+
 static void loadCompStats(WzConfig &json, COMPONENT_STATS *psStats, size_t index)
 {
 	loadStats(json, psStats, index);
+	lookupCompStatPtr.insert(std::make_pair(psStats->id, psStats));
 	psStats->buildPower = json.value("buildPower", 0).toUInt();
 	psStats->buildPoints = json.value("buildPoints", 0).toUInt();
 	psStats->designable = json.value("designable", false).toBool();
 	psStats->weight = json.value("weight", 0).toUInt();
-	psStats->pBase->hitpoints = json.value("hitpoints", 0).toUInt();
-	psStats->pBase->hitpointPct = json.value("hitpointPct", 100).toUInt();
+	psStats->getBase().hitpoints = json.value("hitpoints", 0).toUInt();
+	psStats->getBase().hitpointPct = json.value("hitpointPct", 100).toUInt();
 
 	WzString dtype = json.value("droidType", "DROID").toWzString();
 	psStats->droidTypeOverride = DROID_ANY;
@@ -292,6 +265,10 @@ static void loadCompStats(WzConfig &json, COMPONENT_STATS *psStats, size_t index
 	{
 		psStats->droidTypeOverride = DROID_TRANSPORTER;
 	}
+	else if (dtype.compare("SUPERTRANSPORTER") == 0)
+	{
+		psStats->droidTypeOverride = DROID_SUPERTRANSPORTER;
+	}
 	else if (dtype.compare("CYBORG") == 0)
 	{
 		psStats->droidTypeOverride = DROID_CYBORG;
@@ -405,7 +382,7 @@ bool loadWeaponStats(WzConfig &ini)
 
 		ASSERT(psStats->flightSpeed > 0, "Invalid flight speed for %s", list[i].toUtf8().c_str());
 
-		psStats->ref = REF_WEAPON_START + i;
+		psStats->ref = STAT_WEAPON + i;
 
 		//get the IMD for the component
 		psStats->pIMD = statsGetIMD(ini, psStats, "model");
@@ -424,7 +401,7 @@ bool loadWeaponStats(WzConfig &ini)
 		//set the weapon class
 		if (!getWeaponClass(ini.value("weaponClass").toWzString(), &psStats->weaponClass))
 		{
-			debug(LOG_ERROR, "Invalid weapon class for weapon %s - assuming KINETIC", getName(psStats));
+			debug(LOG_ERROR, "Invalid weapon class for weapon %s - assuming KINETIC", getStatsName(psStats));
 			psStats->weaponClass = WC_KINETIC;
 		}
 
@@ -451,7 +428,7 @@ bool loadWeaponStats(WzConfig &ini)
 		//set the weapon effect
 		if (!getWeaponEffect(ini.value("weaponEffect").toWzString(), &psStats->weaponEffect))
 		{
-			debug(LOG_FATAL, "loadWepaonStats: Invalid weapon effect for weapon %s", getName(psStats));
+			debug(LOG_FATAL, "loadWepaonStats: Invalid weapon effect for weapon %s", getStatsName(psStats));
 			return false;
 		}
 
@@ -464,7 +441,7 @@ bool loadWeaponStats(WzConfig &ini)
 		}
 		else if (!getWeaponClass(periodicalDamageWeaponClass, &psStats->periodicalDamageWeaponClass))
 		{
-			debug(LOG_ERROR, "Invalid periodicalDamageWeaponClass for weapon %s - assuming same class as weapon", getName(psStats));
+			debug(LOG_ERROR, "Invalid periodicalDamageWeaponClass for weapon %s - assuming same class as weapon", getStatsName(psStats));
 			psStats->periodicalDamageWeaponClass = psStats->weaponClass;
 		}
 
@@ -477,7 +454,7 @@ bool loadWeaponStats(WzConfig &ini)
 		}
 		else if (!getWeaponSubClass(periodicalDamageWeaponSubClass.toUtf8().c_str(), &psStats->periodicalDamageWeaponSubClass))
 		{
-			debug(LOG_ERROR, "Invalid periodicalDamageWeaponSubClass for weapon %s - assuming same subclass as weapon", getName(psStats));
+			debug(LOG_ERROR, "Invalid periodicalDamageWeaponSubClass for weapon %s - assuming same subclass as weapon", getStatsName(psStats));
 			psStats->periodicalDamageWeaponSubClass = psStats->weaponSubClass;
 		}
 
@@ -490,7 +467,7 @@ bool loadWeaponStats(WzConfig &ini)
 		}
 		else if (!getWeaponEffect(periodicalDamageWeaponEffect.toUtf8().c_str(), &psStats->periodicalDamageWeaponEffect))
 		{
-			debug(LOG_ERROR, "Invalid periodicalDamageWeaponEffect for weapon %s - assuming same effect as weapon", getName(psStats));
+			debug(LOG_ERROR, "Invalid periodicalDamageWeaponEffect for weapon %s - assuming same effect as weapon", getStatsName(psStats));
 			psStats->periodicalDamageWeaponEffect = psStats->weaponEffect;
 		}
 
@@ -506,11 +483,11 @@ bool loadWeaponStats(WzConfig &ini)
 		unsigned int longHit = psStats->upgrade[0].hitChance;
 		if (shortRange > longRange)
 		{
-			debug(LOG_ERROR, "%s, Short range (%d) is greater than long range (%d)", getName(psStats), shortRange, longRange);
+			debug(LOG_ERROR, "%s, Short range (%d) is greater than long range (%d)", getStatsName(psStats), shortRange, longRange);
 		}
 		if (shortRange == longRange && shortHit != longHit)
 		{
-			debug(LOG_ERROR, "%s, shortHit and longHit should be equal if the ranges are the same", getName(psStats));
+			debug(LOG_ERROR, "%s, shortHit and longHit should be equal if the ranges are the same", getStatsName(psStats));
 		}
 
 		// set the face Player value
@@ -546,21 +523,12 @@ bool loadWeaponStats(WzConfig &ini)
 		WzString szWeaponWav = ini.value("weaponWav", "-1").toWzString();
 		WzString szExplosionWav = ini.value("explosionWav", "-1").toWzString();
 		bool result = statsGetAudioIDFromString(list[i], szWeaponWav, &weaponSoundID);
-		ASSERT_OR_RETURN(false, result, "Weapon sound %s not found for %s", szWeaponWav.toUtf8().c_str(), getName(psStats));
+		ASSERT_OR_RETURN(false, result, "Weapon sound %s not found for %s", szWeaponWav.toUtf8().c_str(), getStatsName(psStats));
 		result = statsGetAudioIDFromString(list[i], szExplosionWav, &explosionSoundID);
-		ASSERT_OR_RETURN(false, result, "Explosion sound %s not found for %s", szExplosionWav.toUtf8().c_str(), getName(psStats));
+		ASSERT_OR_RETURN(false, result, "Explosion sound %s not found for %s", szExplosionWav.toUtf8().c_str(), getStatsName(psStats));
 		psStats->iAudioFireID = weaponSoundID;
 		psStats->iAudioImpactID = explosionSoundID;
 
-		// Set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxWeaponRange(psStats->base.maxRange);
-			setMaxWeaponDamage(psStats->base.damage);
-			setMaxWeaponROF(weaponROF(psStats, 0));
-			setMaxComponentWeight(psStats->weight);
-		}
-
 		ini.endGroup();
 	}
 
@@ -589,30 +557,20 @@ bool loadBodyStats(WzConfig &ini)
 		psStats->base.thermal = ini.value("armourHeat").toInt();
 		psStats->base.armour = ini.value("armourKinetic").toInt();
 		psStats->base.power = ini.value("powerOutput").toInt();
-		psStats->base.resistance = ini.value("resistance", 30).toInt();
+		psStats->base.resistance = ini.value("resistance", DEFAULT_DROID_RESISTANCE).toInt();
 		for (int j = 0; j < MAX_PLAYERS; j++)
 		{
 			psStats->upgrade[j] = psStats->base;
 		}
-		psStats->ref = REF_BODY_START + i;
+		psStats->ref = STAT_BODY + i;
 		if (!getBodySize(ini.value("size").toWzString(), &psStats->size))
 		{
-			ASSERT(false, "Unknown body size for %s", getName(psStats));
+			ASSERT(false, "Unknown body size for %s", getStatsName(psStats));
 			return false;
 		}
 		psStats->pIMD = statsGetIMD(ini, psStats, "model");
 
 		ini.endGroup();
-
-		//set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxBodyArmour(psStats->base.armour);
-			setMaxBodyArmour(psStats->base.thermal);
-			setMaxBodyPower(psStats->base.power);
-			setMaxBodyPoints(psStats->base.hitpoints);
-			setMaxComponentWeight(psStats->weight);
-		}
 	}
 
 	// now get the extra stuff ... hack it together with above later, moved here from
@@ -707,14 +665,14 @@ bool loadBrainStats(WzConfig &ini)
 		ASSERT(rankNames.is_array(), "ranks is not an array");
 		for (const auto& v : rankNames)
 		{
-			psStats->rankNames.push_back(v);
+			psStats->rankNames.push_back(v.get<std::string>());
 		}
 		auto rankThresholds = ini.json("thresholds");
 		for (const auto& v : rankThresholds)
 		{
-			psStats->base.rankThresholds.push_back(v);
+			psStats->base.rankThresholds.push_back(v.get<int>());
 		}
-		psStats->ref = REF_BRAIN_START + i;
+		psStats->ref = STAT_BRAIN + i;
 
 		for (int j = 0; j < MAX_PLAYERS; j++)
 		{
@@ -726,7 +684,7 @@ bool loadBrainStats(WzConfig &ini)
 		if (ini.contains("turret"))
 		{
 			int weapon = getCompFromName(COMP_WEAPON, ini.value("turret").toWzString());
-			ASSERT_OR_RETURN(false, weapon >= 0, "Unable to find weapon for brain %s", getName(psStats));
+			ASSERT_OR_RETURN(false, weapon >= 0, "Unable to find weapon for brain %s", getStatsName(psStats));
 			psStats->psWeaponStat = asWeaponStats + weapon;
 		}
 		psStats->designable = ini.value("designable", false).toBool();
@@ -771,6 +729,8 @@ bool getPropulsionType(const char *typeName, PROPULSION_TYPE *type)
 	{
 		debug(LOG_ERROR, "getPropulsionType: Invalid Propulsion type %s - assuming Hover", typeName);
 		*type = PROPULSION_TYPE_HOVER;
+
+		return false;
 	}
 
 	return true;
@@ -796,7 +756,7 @@ bool loadPropulsionStats(WzConfig &ini)
 
 		psStats->base.hitpointPctOfBody = ini.value("hitpointPctOfBody", 0).toInt();
 		psStats->maxSpeed = ini.value("speed").toInt();
-		psStats->ref = REF_PROPULSION_START + i;
+		psStats->ref = STAT_PROPULSION + i;
 		psStats->turnSpeed = ini.value("turnSpeed", DEG(1) / 3).toInt();
 		psStats->spinSpeed = ini.value("spinSpeed", DEG(3) / 4).toInt();
 		psStats->spinAngle = ini.value("spinAngle", 180).toInt();
@@ -811,38 +771,10 @@ bool loadPropulsionStats(WzConfig &ini)
 		}
 		if (!getPropulsionType(ini.value("type").toWzString().toUtf8().c_str(), &psStats->propulsionType))
 		{
-			debug(LOG_FATAL, "loadPropulsionStats: Invalid Propulsion type for %s", getName(psStats));
+			debug(LOG_FATAL, "loadPropulsionStats: Invalid Propulsion type for %s", getStatsName(psStats));
 			return false;
 		}
 		ini.endGroup();
-
-		// set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxPropulsionSpeed(psStats->maxSpeed);
-		}
-	}
-
-	/* since propulsion weight is a multiple of body weight we may need to adjust the max component weight value */
-	if (asBodyStats && asPropulsionStats)	// check we've loaded them both in
-	{
-		//check against each body stat
-		for (int i = 0; i < numBodyStats; ++i)
-		{
-			//check stat is designable
-			if (asBodyStats[i].designable)
-			{
-				//check against each propulsion stat
-				for (int j = 0; j < numPropulsionStats; ++j)
-				{
-					//check stat is designable
-					if (asPropulsionStats[j].designable)
-					{
-						setMaxComponentWeight(asPropulsionStats[j].weight * asBodyStats[i].weight / 100);
-					}
-				}
-			}
-		}
 	}
 
 	return true;
@@ -872,7 +804,7 @@ bool loadSensorStats(WzConfig &ini)
 			psStats->upgrade[j] = psStats->base;
 		}
 
-		psStats->ref = REF_SENSOR_START + i;
+		psStats->ref = STAT_SENSOR + i;
 
 		WzString location = ini.value("location").toWzString();
 		if (location.compare("DEFAULT") == 0)
@@ -922,14 +854,6 @@ bool loadSensorStats(WzConfig &ini)
 		psStats->pMountGraphic = statsGetIMD(ini, psStats, "mountModel");
 
 		ini.endGroup();
-
-		// set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxSensorRange(psStats->base.range);
-			setMaxComponentWeight(psStats->weight);
-		}
-
 	}
 	return true;
 }
@@ -958,7 +882,7 @@ bool loadECMStats(WzConfig &ini)
 			psStats->upgrade[j] = psStats->base;
 		}
 
-		psStats->ref = REF_ECM_START + i;
+		psStats->ref = STAT_ECM + i;
 
 		WzString location = ini.value("location").toWzString();
 		if (location.compare("DEFAULT") == 0)
@@ -979,13 +903,6 @@ bool loadECMStats(WzConfig &ini)
 		psStats->pMountGraphic = statsGetIMD(ini, psStats, "mountModel");
 
 		ini.endGroup();
-
-		// Set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxECMRange(psStats->base.range);
-			setMaxComponentWeight(psStats->weight);
-		}
 	}
 	return true;
 }
@@ -1015,7 +932,7 @@ bool loadRepairStats(WzConfig &ini)
 		}
 		psStats->time = ini.value("time", 0).toInt() * WEAPON_TIME;
 
-		psStats->ref = REF_REPAIR_START + i;
+		psStats->ref = STAT_REPAIR + i;
 
 		WzString location = ini.value("location").toWzString();
 		if (location.compare("DEFAULT") == 0)
@@ -1032,20 +949,13 @@ bool loadRepairStats(WzConfig &ini)
 		}
 
 		//check its not 0 since we will be dividing by it at a later stage
-		ASSERT_OR_RETURN(false, psStats->time > 0, "Repair delay cannot be zero for %s", getName(psStats));
+		ASSERT_OR_RETURN(false, psStats->time > 0, "Repair delay cannot be zero for %s", getStatsName(psStats));
 
 		//get the IMD for the component
 		psStats->pIMD = statsGetIMD(ini, psStats, "model");
 		psStats->pMountGraphic = statsGetIMD(ini, psStats, "mountModel");
 
 		ini.endGroup();
-
-		//set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxRepairPoints(psStats->base.repairPoints);
-			setMaxComponentWeight(psStats->weight);
-		}
 	}
 	return true;
 }
@@ -1073,20 +983,13 @@ bool loadConstructStats(WzConfig &ini)
 		{
 			psStats->upgrade[j] = psStats->base;
 		}
-		psStats->ref = REF_CONSTRUCT_START + i;
+		psStats->ref = STAT_CONSTRUCT + i;
 
 		//get the IMD for the component
 		psStats->pIMD = statsGetIMD(ini, psStats, "sensorModel");
 		psStats->pMountGraphic = statsGetIMD(ini, psStats, "mountModel");
 
 		ini.endGroup();
-
-		// Set the max stat values for the design screen
-		if (psStats->designable)
-		{
-			setMaxConstPoints(psStats->base.constructPoints);
-			setMaxComponentWeight(psStats->weight);
-		}
 	}
 	return true;
 }
@@ -1314,110 +1217,6 @@ UDWORD getSpeedFactor(UDWORD type, UDWORD propulsionType)
 	return asTerrainTable[type * PROPULSION_TYPE_NUM + propulsionType];
 }
 
-//return the type of stat this stat is!
-UDWORD statType(UDWORD ref)
-{
-	if (ref >= REF_BODY_START && ref < REF_BODY_START +
-	    REF_RANGE)
-	{
-		return COMP_BODY;
-	}
-	if (ref >= REF_BRAIN_START && ref < REF_BRAIN_START +
-	    REF_RANGE)
-	{
-		return COMP_BRAIN;
-	}
-	if (ref >= REF_PROPULSION_START && ref <
-	    REF_PROPULSION_START + REF_RANGE)
-	{
-		return COMP_PROPULSION;
-	}
-	if (ref >= REF_SENSOR_START && ref < REF_SENSOR_START +
-	    REF_RANGE)
-	{
-		return COMP_SENSOR;
-	}
-	if (ref >= REF_ECM_START && ref < REF_ECM_START +
-	    REF_RANGE)
-	{
-		return COMP_ECM;
-	}
-	if (ref >= REF_REPAIR_START && ref < REF_REPAIR_START +
-	    REF_RANGE)
-	{
-		return COMP_REPAIRUNIT;
-	}
-	if (ref >= REF_WEAPON_START && ref < REF_WEAPON_START +
-	    REF_RANGE)
-	{
-		return COMP_WEAPON;
-	}
-	if (ref >= REF_CONSTRUCT_START && ref < REF_CONSTRUCT_START +
-	    REF_RANGE)
-	{
-		return COMP_CONSTRUCT;
-	}
-	//else
-	ASSERT(false, "Invalid stat pointer - cannot determine Stat Type");
-	return COMP_NUMCOMPONENTS;
-}
-
-//return the REF_START value of this type of stat
-UDWORD statRefStart(UDWORD stat)
-{
-	UDWORD start;
-
-	switch (stat)
-	{
-	case COMP_BODY:
-		{
-			start = REF_BODY_START;
-			break;
-		}
-	case COMP_BRAIN:
-		{
-			start = REF_BRAIN_START;
-			break;
-		}
-	case COMP_PROPULSION:
-		{
-			start = REF_PROPULSION_START;
-			break;
-		}
-	case COMP_SENSOR:
-		{
-			start = REF_SENSOR_START;
-			break;
-		}
-	case COMP_ECM:
-		{
-			start = REF_ECM_START;
-			break;
-		}
-	case COMP_REPAIRUNIT:
-		{
-			start = REF_REPAIR_START;
-			break;
-		}
-	case COMP_WEAPON:
-		{
-			start = REF_WEAPON_START;
-			break;
-		}
-	case COMP_CONSTRUCT:
-		{
-			start = REF_CONSTRUCT_START;
-			break;
-		}
-	default:
-		{
-			debug(LOG_FATAL, "Invalid stat type");
-			start = 0;
-		}
-	}
-	return start;
-}
-
 int getCompFromName(COMPONENT_TYPE compType, const WzString &name)
 {
 	return getCompFromID(compType, name);
@@ -1426,10 +1225,10 @@ int getCompFromName(COMPONENT_TYPE compType, const WzString &name)
 int getCompFromID(COMPONENT_TYPE compType, const WzString &name)
 {
 	COMPONENT_STATS *psComp = nullptr;
-	auto it = lookupStatPtr.find(WzString::fromUtf8(name.toUtf8().c_str()));
-	if (it != lookupStatPtr.end())
+	auto it = lookupCompStatPtr.find(WzString::fromUtf8(name.toUtf8().c_str()));
+	if (it != lookupCompStatPtr.end())
 	{
-		psComp = (COMPONENT_STATS *)it->second;
+		psComp = it->second;
 	}
 	ASSERT_OR_RETURN(-1, psComp, "No such component ID [%s] found", name.toUtf8().c_str());
 	ASSERT_OR_RETURN(-1, compType == psComp->compType, "Wrong component type for ID %s", name.toUtf8().c_str());
@@ -1442,22 +1241,33 @@ int getCompFromID(COMPONENT_TYPE compType, const WzString &name)
 COMPONENT_STATS *getCompStatsFromName(const WzString &name)
 {
 	COMPONENT_STATS *psComp = nullptr;
-	auto it = lookupStatPtr.find(name);
-	if (it != lookupStatPtr.end())
+	auto it = lookupCompStatPtr.find(name);
+	if (it != lookupCompStatPtr.end())
 	{
-		psComp = (COMPONENT_STATS *)it->second;
+		psComp = it->second;
 	}
 	/*if (!psComp)
 	{
-		debug(LOG_ERROR, "Not found: %s", name.toUtf8().constData());
-		foreach(BASE_STATS *psStat, lookupStatPtr)
+		debug(LOG_ERROR, "Not found: %s", name.toUtf8().c_str());
+		for (auto& it2 : lookupCompStatPtr)
 		{
-			debug(LOG_ERROR, "    %s", psStat->name.toUtf8().constData());
+			debug(LOG_ERROR, "    %s", it2.second->id.toUtf8().c_str());
 		}
 	}*/
 	return psComp;
 }
 
+BASE_STATS *getBaseStatsFromName(const WzString &name)
+{
+	BASE_STATS *psStat = nullptr;
+	auto it = lookupStatPtr.find(name);
+	if (it != lookupStatPtr.end())
+	{
+		psStat = it->second;
+	}
+	return psStat;
+}
+
 /*sets the store to the body size based on the name passed in - returns false
 if doesn't compare with any*/
 bool getBodySize(const WzString &size, BODY_SIZE *pStore)
@@ -1671,6 +1481,23 @@ bool getWeaponEffect(const WzString& weaponEffect, WEAPON_EFFECT *effect)
 	return true;
 }
 
+/*returns the weapon effect string based on the enum passed in */
+const char *getWeaponEffect(WEAPON_EFFECT effect)
+{
+	switch (effect)
+	{
+	case WE_ANTI_PERSONNEL: return "ANTI PERSONNEL";
+	case WE_ANTI_TANK: return "ANTI TANK";
+	case WE_BUNKER_BUSTER: return "BUNKER BUSTER";
+	case WE_ARTILLERY_ROUND: return "ARTILLERY ROUND";
+	case WE_FLAMER: return "FLAMER";
+	case WE_ANTI_AIRCRAFT: return "ANTI AIRCRAFT";
+	case WE_NUMEFFECTS: break;
+	}
+	ASSERT(false, "No such weapon effect");
+	return "Bad weapon effect";
+}
+
 bool getWeaponClass(const WzString& weaponClassStr, WEAPON_CLASS *weaponClass)
 {
 	if (weaponClassStr.compare("KINETIC") == 0)
@@ -1689,70 +1516,86 @@ bool getWeaponClass(const WzString& weaponClassStr, WEAPON_CLASS *weaponClass)
 	return true;
 }
 
+#define ASSERT_PLAYER_OR_RETURN(retVal, player) \
+	ASSERT_OR_RETURN(retVal, player >= 0 && player < MAX_PLAYERS, "Invalid player: %" PRIu32 "", player);
+
 /*Access functions for the upgradeable stats of a weapon*/
 int weaponFirePause(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].firePause;
 }
 
 /* Reload time is reduced for weapons with salvo fire */
 int weaponReloadTime(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].reloadTime;
 }
 
 int weaponLongHit(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].hitChance;
 }
 
 int weaponShortHit(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].shortHitChance;
 }
 
 int weaponDamage(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].damage;
 }
 
 int weaponRadDamage(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].radiusDamage;
 }
 
 int weaponPeriodicalDamage(const WEAPON_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].periodicalDamage;
 }
 
 int sensorRange(const SENSOR_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].range;
 }
 
 int ecmRange(const ECM_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].range;
 }
 
 int repairPoints(const REPAIR_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].repairPoints;
 }
 
 int constructorPoints(const CONSTRUCT_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].constructPoints;
 }
 
 int bodyPower(const BODY_STATS *psStats, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	return psStats->upgrade[player].power;
 }
 
 int bodyArmour(const BODY_STATS *psStats, int player, WEAPON_CLASS weaponClass)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	switch (weaponClass)
 	{
 	case WC_KINETIC:
@@ -1769,6 +1612,7 @@ int bodyArmour(const BODY_STATS *psStats, int player, WEAPON_CLASS weaponClass)
 //calculates the weapons ROF based on the fire pause and the salvos
 int weaponROF(const WEAPON_STATS *psStat, int player)
 {
+	ASSERT_PLAYER_OR_RETURN(0, player);
 	int rof = 0;
 	// if there are salvos
 	if (player >= 0
@@ -1791,288 +1635,6 @@ int weaponROF(const WEAPON_STATS *psStat, int player)
 	return rof;
 }
 
-//Access functions for the max values to be used in the Design Screen
-void setMaxComponentWeight(UDWORD weight)
-{
-	if (weight > maxComponentWeight)
-	{
-		maxComponentWeight = weight;
-	}
-}
-UDWORD getMaxComponentWeight()
-{
-	return maxComponentWeight;
-}
-
-void setMaxBodyArmour(UDWORD armour)
-{
-	if (armour > maxBodyArmour)
-	{
-		maxBodyArmour = armour;
-	}
-}
-UDWORD getMaxBodyArmour()
-{
-	return maxBodyArmour;
-}
-
-void setMaxBodyPower(UDWORD power)
-{
-	if (power > maxBodyPower)
-	{
-		maxBodyPower = power;
-	}
-}
-UDWORD getMaxBodyPower()
-{
-	return maxBodyPower;
-}
-
-void setMaxBodyPoints(UDWORD points)
-{
-	if (points > maxBodyPoints)
-	{
-		maxBodyPoints = points;
-	}
-}
-UDWORD getMaxBodyPoints()
-{
-	return maxBodyPoints;
-}
-
-void setMaxSensorRange(UDWORD range)
-{
-	if (range > maxSensorRange)
-	{
-		maxSensorRange = range;
-	}
-}
-
-UDWORD getMaxSensorRange()
-{
-	return maxSensorRange;
-}
-
-void setMaxECMRange(UDWORD range)
-{
-	if (range > maxECMRange)
-	{
-		maxECMRange = range;
-	}
-}
-
-UDWORD getMaxECMRange()
-{
-	return maxECMRange;
-}
-
-void setMaxConstPoints(UDWORD points)
-{
-	if (points > maxConstPoints)
-	{
-		maxConstPoints = points;
-	}
-}
-UDWORD getMaxConstPoints()
-{
-	return maxConstPoints;
-}
-
-void setMaxRepairPoints(UDWORD repair)
-{
-	if (repair > maxRepairPoints)
-	{
-		maxRepairPoints = repair;
-	}
-}
-UDWORD getMaxRepairPoints()
-{
-	return maxRepairPoints;
-}
-
-void setMaxWeaponRange(UDWORD range)
-{
-	if (range > maxWeaponRange)
-	{
-		maxWeaponRange = range;
-	}
-}
-UDWORD getMaxWeaponRange()
-{
-	return maxWeaponRange;
-}
-
-void setMaxWeaponDamage(UDWORD damage)
-{
-	if (damage > maxWeaponDamage)
-	{
-		maxWeaponDamage = damage;
-	}
-}
-UDWORD getMaxWeaponDamage()
-{
-	return maxWeaponDamage;
-}
-
-void setMaxWeaponROF(UDWORD rof)
-{
-	if (rof > maxWeaponROF)
-	{
-		maxWeaponROF = rof;
-	}
-}
-UDWORD getMaxWeaponROF()
-{
-	return maxWeaponROF;
-}
-
-void setMaxPropulsionSpeed(UDWORD speed)
-{
-	if (speed > maxPropulsionSpeed)
-	{
-		maxPropulsionSpeed = speed;
-	}
-}
-UDWORD getMaxPropulsionSpeed()
-{
-	return maxPropulsionSpeed;
-}
-
-//determine the effect this upgrade would have on the max values
-void updateMaxWeaponStats(UWORD maxValue)
-{
-	UDWORD currentMaxValue = getMaxWeaponDamage();
-
-	if (currentMaxValue < (currentMaxValue + maxValue / 100))
-	{
-		currentMaxValue += currentMaxValue * maxValue / 100;
-		setMaxWeaponDamage(currentMaxValue);
-	}
-
-	//the fire pause is dealt with differently
-}
-
-void updateMaxSensorStats(UWORD maxRange)
-{
-	UDWORD currentMaxValue = getMaxSensorRange();
-
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxRange / 100))
-	{
-		currentMaxValue += currentMaxValue * maxRange / 100;
-		setMaxSensorRange(currentMaxValue);
-	}
-}
-
-void updateMaxRepairStats(UWORD maxValue)
-{
-	UDWORD currentMaxValue = getMaxRepairPoints();
-
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxValue / 100))
-	{
-		currentMaxValue += currentMaxValue * maxValue / 100;
-		setMaxRepairPoints(currentMaxValue);
-	}
-}
-
-void updateMaxECMStats(UWORD maxValue)
-{
-	int currentMaxValue = getMaxECMRange();
-
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxValue / 100))
-	{
-		currentMaxValue += currentMaxValue * maxValue / 100;
-		setMaxECMRange(currentMaxValue);
-	}
-}
-
-void updateMaxBodyStats(UWORD maxBody, UWORD maxPower, UWORD maxArmour)
-{
-	UDWORD currentMaxValue = getMaxBodyPoints();
-
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxBody / 100))
-	{
-		currentMaxValue += currentMaxValue * maxBody / 100;
-		setMaxBodyPoints(currentMaxValue);
-	}
-
-	currentMaxValue = getMaxBodyPower();
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxPower / 100))
-	{
-		currentMaxValue += currentMaxValue * maxPower / 100;
-		setMaxBodyPower(currentMaxValue);
-	}
-
-	currentMaxValue = getMaxBodyArmour();
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxArmour / 100))
-	{
-		currentMaxValue += currentMaxValue * maxArmour / 100;
-		setMaxBodyArmour(currentMaxValue);
-	}
-}
-
-void updateMaxConstStats(UWORD maxValue)
-{
-	UDWORD currentMaxValue = getMaxConstPoints();
-
-	if (currentMaxValue < (currentMaxValue + currentMaxValue * maxValue / 100))
-	{
-		currentMaxValue += currentMaxValue * maxValue / 100;
-		setMaxConstPoints(currentMaxValue);
-	}
-}
-
-void adjustMaxDesignStats()
-{
-	UWORD       weaponDamage, sensorRange, repairPoints,
-	            ecmRange, constPoints, bodyPoints, bodyPower, bodyArmour;
-
-	// init all the values
-	weaponDamage = sensorRange = repairPoints = ecmRange = constPoints = bodyPoints = bodyPower = bodyArmour = 0;
-
-	//go thru' all the functions getting the max upgrade values for the stats
-
-	for (int j = 0; j < numBodyStats; j++)
-	{
-		BODY_STATS *psStats = asBodyStats + j;
-		bodyPoints = MAX(bodyPoints, psStats->upgrade[selectedPlayer].hitpoints);
-		bodyArmour = MAX(bodyArmour, psStats->upgrade[selectedPlayer].armour);
-		bodyPower = MAX(bodyPower, psStats->upgrade[selectedPlayer].power);
-	}
-	for (int j = 0; j < numSensorStats; j++)
-	{
-		SENSOR_STATS *psStats = asSensorStats + j;
-		sensorRange = MAX(sensorRange, psStats->upgrade[selectedPlayer].range);
-	}
-	for (int j = 0; j < numECMStats; j++)
-	{
-		ECM_STATS *psStats = asECMStats + j;
-		ecmRange = MAX(ecmRange, psStats->upgrade[selectedPlayer].range);
-	}
-	for (int j = 0; j < numRepairStats; j++)
-	{
-		REPAIR_STATS *psStats = asRepairStats + j;
-		repairPoints = MAX(repairPoints, psStats->upgrade[selectedPlayer].repairPoints);
-	}
-	for (int j = 0; j < numConstructStats; j++)
-	{
-		CONSTRUCT_STATS *psStats = asConstructStats + j;
-		constPoints = MAX(constPoints, psStats->upgrade[selectedPlayer].constructPoints);
-	}
-	for (int j = 0; j < numWeaponStats; j++)
-	{
-		WEAPON_STATS *psStats = asWeaponStats + j;
-		weaponDamage = MAX(weaponDamage, psStats->upgrade[selectedPlayer].damage);
-	}
-
-	//determine the effect on the max values for the stats
-	updateMaxWeaponStats(weaponDamage);
-	updateMaxSensorStats(sensorRange);
-	updateMaxRepairStats(repairPoints);
-	updateMaxECMStats(ecmRange);
-	updateMaxBodyStats(bodyPoints, bodyPower, bodyArmour);
-	updateMaxConstStats(constPoints);
-}
-
 /* Check if an object has a weapon */
 bool objHasWeapon(const BASE_OBJECT *psObj)
 {
