@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -71,12 +71,12 @@ extern STRUCTSTRENGTH_MODIFIER		asStructStrengthModifier[WE_NUMEFFECTS][NUM_STRU
 
 void handleAbandonedStructures();
 
-int getMaxDroids(int player);
-int getMaxCommanders(int player);
-int getMaxConstructors(int player);
-void setMaxDroids(int player, int value);
-void setMaxCommanders(int player, int value);
-void setMaxConstructors(int player, int value);
+int getMaxDroids(UDWORD player);
+int getMaxCommanders(UDWORD player);
+int getMaxConstructors(UDWORD player);
+void setMaxDroids(UDWORD player, int value);
+void setMaxCommanders(UDWORD player, int value);
+void setMaxConstructors(UDWORD player, int value);
 
 bool structureExists(int player, STRUCTURE_TYPE type, bool built, bool isMission);
 
@@ -97,17 +97,16 @@ void structureDemolish(STRUCTURE *psStructure, DROID *psDroid, int buildPoints);
 void structureRepair(STRUCTURE *psStruct, DROID *psDroid, int buildRate);
 /* Set the type of droid for a factory to build */
 bool structSetManufacture(STRUCTURE *psStruct, DROID_TEMPLATE *psTempl, QUEUE_MODE mode);
-
-//temp test function for creating structures at the start of the game
-void createTestStructures();
+uint32_t structureBuildPointsToCompletion(const STRUCTURE & structure);
+float structureCompletionProgress(const STRUCTURE & structure);
 
 //builds a specified structure at a given location
 STRUCTURE *buildStructure(STRUCTURE_STATS *pStructureType, UDWORD x, UDWORD y, UDWORD player, bool FromSave);
 STRUCTURE *buildStructureDir(STRUCTURE_STATS *pStructureType, UDWORD x, UDWORD y, uint16_t direction, UDWORD player, bool FromSave);
 /// Create a blueprint structure, with just enough information to render it
-STRUCTURE *buildBlueprint(STRUCTURE_STATS const *psStats, Vector2i xy, uint16_t direction, unsigned moduleIndex, STRUCT_STATES state);
+STRUCTURE *buildBlueprint(STRUCTURE_STATS const *psStats, Vector3i xy, uint16_t direction, unsigned moduleIndex, STRUCT_STATES state, uint8_t ownerPlayer);
 /* The main update routine for all Structures */
-void structureUpdate(STRUCTURE *psBuilding, bool mission);
+void structureUpdate(STRUCTURE *psBuilding, bool bMission);
 
 /* Remove a structure and free it's memory */
 bool destroyStruct(STRUCTURE *psDel, unsigned impactTime);
@@ -118,7 +117,7 @@ bool destroyStruct(STRUCTURE *psDel, unsigned impactTime);
 bool removeStruct(STRUCTURE *psDel, bool bDestroy);
 
 //fills the list with Structures that can be built
-UDWORD fillStructureList(STRUCTURE_STATS **ppList, UDWORD selectedPlayer, UDWORD limit);
+std::vector<STRUCTURE_STATS *> fillStructureList(UDWORD selectedPlayer, UDWORD limit, bool showFavorites);
 
 /// Checks if the two structures would be too close to build together.
 bool isBlueprintTooClose(STRUCTURE_STATS const *stats1, Vector2i pos1, uint16_t dir1, STRUCTURE_STATS const *stats2, Vector2i pos2, uint16_t dir2);
@@ -130,19 +129,6 @@ bool validLocation(BASE_STATS *psStats, Vector2i pos, uint16_t direction, unsign
 bool isWall(STRUCTURE_TYPE type);                                    ///< Structure is a wall. Not completely sure it handles all cases.
 bool isBuildableOnWalls(STRUCTURE_TYPE type);                        ///< Structure can be built on walls. Not completely sure it handles all cases.
 
-/* for a new structure, find a location along an edge which the droid can get
-to and return this as the destination for the droid */
-
-/* for a structure or feature, find a location along an edge which the droid can get
-to and return this as the destination for the droid*/
-bool getDroidDestination(BASE_STATS *psPositionStats, UDWORD structX, UDWORD structY, UDWORD *pDroidX, UDWORD *pDroidY);
-
-/* check along the width of a structure for an empty space */
-bool checkWidth(UDWORD maxRange, UDWORD x, UDWORD y, UDWORD *pDroidX, UDWORD *pDroidY);
-
-/* check along the length of a structure for an empty space */
-bool checkLength(UDWORD maxRange, UDWORD x, UDWORD y, UDWORD *pDroidX, UDWORD *pDroidY);
-
 void alignStructure(STRUCTURE *psBuilding);
 
 /* set the current number of structures of each type built */
@@ -150,9 +136,7 @@ void setCurrentStructQuantity(bool displayError);
 /* get a stat inc based on the name */
 int32_t getStructStatFromName(const WzString &name);
 /*check to see if the structure is 'doing' anything  - return true if idle*/
-bool  structureIdle(STRUCTURE *psBuilding);
-/*checks to see if any structure exists of a specified type with a specified status */
-bool checkStructureStatus(STRUCTURE_STATS *psStats, UDWORD player, UDWORD status);
+bool  structureIdle(const STRUCTURE *psBuilding);
 /*sets the point new droids go to - x/y in world coords for a Factory*/
 void setAssemblyPoint(FLAG_POSITION *psAssemblyPoint, UDWORD x, UDWORD y, UDWORD player, bool bCheck);
 
@@ -189,9 +173,6 @@ bool getSatUplinkExists(UDWORD player);
 /*sets the flag to indicate a Las Sat Exists - ONLY EVER WANT ONE*/
 void setLasSatExists(bool state, UDWORD player);
 
-/*returns the status of the flag*/
-bool getLasSatExists(UDWORD player);
-
 /* added int weapon_slot to fix the alway slot 0 hack */
 bool calcStructureMuzzleLocation(const STRUCTURE *psStructure, Vector3i *muzzle, int weapon_slot);
 bool calcStructureMuzzleBaseLocation(const STRUCTURE *psStructure, Vector3i *muzzle, int weapon_slot);
@@ -203,6 +184,8 @@ void buildingComplete(STRUCTURE *psBuilding);
 void checkForResExtractors(STRUCTURE *psPowerGen);
 void checkForPowerGen(STRUCTURE *psPowerGen);
 
+uint16_t countPlayerUnusedDerricks();
+
 // Set the command droid that factory production should go to struct _command_droid;
 void assignFactoryCommandDroid(STRUCTURE *psStruct, struct DROID *psCommander);
 
@@ -243,7 +226,6 @@ int32_t getStructureDamage(const STRUCTURE *psStructure);
 
 unsigned structureBodyBuilt(const STRUCTURE *psStruct);  ///< Returns the maximum body points of a structure with the current number of build points.
 UDWORD structureBody(const STRUCTURE *psStruct);
-UDWORD structureArmour(const STRUCTURE_STATS *psStats, UBYTE player);
 UDWORD structureResistance(const STRUCTURE_STATS *psStats, UBYTE player);
 
 void hqReward(UBYTE losingPlayer, UBYTE rewardPlayer);
@@ -292,10 +274,6 @@ void releaseProduction(STRUCTURE *psBuilding, QUEUE_MODE mode);
 /// Does the next item in the production list.
 void doNextProduction(STRUCTURE *psStructure, DROID_TEMPLATE *current, QUEUE_MODE mode);
 
-/*This function is called after a game is loaded so that any resource extractors
-that are active are initialised for when to start*/
-void checkResExtractorsActive();
-
 // Count number of factories assignable to a command droid.
 UWORD countAssignableFactories(UBYTE player, UWORD FactoryType);
 
@@ -350,21 +328,23 @@ bool IsStatExpansionModule(const STRUCTURE_STATS *psStats);
 bool structureIsBlueprint(const STRUCTURE *psStructure);
 bool isBlueprint(const BASE_OBJECT *psObject);
 
-/*returns the power cost to build this structure*/
-UDWORD structPowerToBuild(const STRUCTURE *psStruct);
+/*returns the power cost to build this structure, or to add its next module */
+UDWORD structPowerToBuildOrAddNextModule(const STRUCTURE *psStruct);
 
 // check whether a factory of a certain number and type exists
 bool checkFactoryExists(UDWORD player, UDWORD factoryType, UDWORD inc);
 
 /*checks the structure passed in is a Las Sat structure which is currently
 selected - returns true if valid*/
-bool lasSatStructSelected(STRUCTURE *psStruct);
+bool lasSatStructSelected(const STRUCTURE *psStruct);
 
 void cbNewDroid(STRUCTURE *psFactory, DROID *psDroid);
 
 StructureBounds getStructureBounds(const STRUCTURE *object);
 StructureBounds getStructureBounds(const STRUCTURE_STATS *stats, Vector2i pos, uint16_t direction);
 
+bool canStructureHaveAModuleAdded(const STRUCTURE* const structure);
+
 static inline int structSensorRange(const STRUCTURE *psObj)
 {
 	return objSensorRange((const BASE_OBJECT *)psObj);
@@ -509,4 +489,20 @@ static inline int getBuildingRearmPoints(STRUCTURE *psStruct)
 	return psStruct->pStructureType->upgrade[psStruct->player].rearm;
 }
 
+WzString getFavoriteStructs();
+void setFavoriteStructs(WzString list);
+
+struct LineBuild
+{
+	Vector2i back() const { return (*this)[count - 1]; }
+	Vector2i operator [](int i) const { return begin + i*step; }
+
+	Vector2i begin = {0, 0};
+	Vector2i step = {0, 0};
+	int count = 0;
+};
+
+LineBuild calcLineBuild(Vector2i size, STRUCTURE_TYPE type, Vector2i pos, Vector2i pos2);
+LineBuild calcLineBuild(STRUCTURE_STATS const *stats, uint16_t direction, Vector2i pos, Vector2i pos2);
+
 #endif // __INCLUDED_SRC_STRUCTURE_H__
