@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -55,6 +55,13 @@
 #include "hci.h"
 #include "loop.h"
 
+// TODO: Fix and remove after merging terrain rendering changes
+#if defined(__clang__)
+	#pragma clang diagnostic ignored "-Wfloat-conversion"
+#elif defined(__GNUC__)
+	#pragma GCC diagnostic ignored "-Wfloat-conversion"
+#endif
+
 /**
  * A sector contains all information to draw a square piece of the map.
  * The actual geometry and texture data is not stored in here but in large VBO's.
@@ -70,10 +77,10 @@ struct Sector
 	int waterSize;           ///< The size of the water geometry
 	int waterIndexOffset;    ///< The point in the water index VBO where the water triangles start
 	int waterIndexSize;      ///< The size of our water triangles
-	int *textureOffset;      ///< An array containing the offsets into the texture VBO for each terrain layer
-	int *textureSize;        ///< The size of the geometry for this layer for each layer
-	int *textureIndexOffset; ///< The offset into the index VBO for the texture for each layer
-	int *textureIndexSize;   ///< The size of the indices for each layer
+	std::unique_ptr<int[]> textureOffset;      ///< An array containing the offsets into the texture VBO for each terrain layer
+	std::unique_ptr<int[]> textureSize;        ///< The size of the geometry for this layer for each layer
+	std::unique_ptr<int[]> textureIndexOffset; ///< The offset into the index VBO for the texture for each layer
+	std::unique_ptr<int[]> textureIndexSize;   ///< The size of the indices for each layer
 	int decalOffset;         ///< Index into the decal VBO
 	int decalSize;           ///< Size of the part of the decal VBO we are going to use
 	bool draw;               ///< Do we draw this sector this frame?
@@ -90,14 +97,14 @@ struct DecalVertex
 };
 
 /// The lightmap texture
-static gfx_api::texture* lightmap_tex_num = nullptr;
+static gfx_api::texture* lightmap_texture = nullptr;
 /// When are we going to update the lightmap next?
 static unsigned int lightmapLastUpdate;
 /// How big is the lightmap?
-static int lightmapWidth;
-static int lightmapHeight;
+static size_t lightmapWidth;
+static size_t lightmapHeight;
 /// Lightmap image
-static GLubyte *lightmapPixmap;
+static std::unique_ptr<gfx_api::gfxUByte[]> lightmapPixmap;
 /// Ticks per lightmap refresh
 static const unsigned int LIGHTMAP_REFRESH = 80;
 
@@ -109,10 +116,10 @@ static gfx_api::buffer *waterVBO = nullptr, *waterIndexVBO = nullptr;
 static float waterOffset;
 
 /// These are properties of your videocard and hardware
-static GLint GLmaxElementsVertices, GLmaxElementsIndices;
+static int32_t GLmaxElementsVertices, GLmaxElementsIndices;
 
 /// The sectors are stored here
-static Sector *sectors;
+static std::unique_ptr<Sector[]> sectors;
 /// The default sector size (a sector is sectorSize x sectorSize)
 static int sectorSize = 15;
 /// What is the distance we can see
@@ -132,19 +139,17 @@ GLsizei dreCount;
 /// Are we actually drawing something using the DrawRangeElements functions?
 bool drawRangeElementsStarted = false;
 
+#define MIN_TERRAIN_TEXTURE_SIZE 512
+
 /// Pass all remaining triangles to OpenGL
+template<typename PSO>
 static void finishDrawRangeElements()
 {
 	if (drawRangeElementsStarted && dreCount > 0)
 	{
 		ASSERT(dreEnd - dreStart + 1 <= GLmaxElementsVertices, "too many vertices (%i)", (int)(dreEnd - dreStart + 1));
 		ASSERT(dreCount <= GLmaxElementsIndices, "too many indices (%i)", (int)dreCount);
-		glDrawRangeElements(GL_TRIANGLES,
-		                    dreStart,
-		                    dreEnd,
-		                    dreCount,
-		                    GL_UNSIGNED_INT,
-		                    BUFFER_OFFSET(sizeof(GLuint)*dreOffset));
+		PSO::get().draw_elements(dreCount, sizeof(GLuint)*dreOffset);
 	}
 	drawRangeElementsStarted = false;
 }
@@ -153,15 +158,12 @@ static void finishDrawRangeElements()
  * Either draw the elements or batch them to be sent to OpenGL later
  * This improves performance by reducing the amount of OpenGL calls.
  */
-static void addDrawRangeElements(GLenum mode,
-                                 GLuint start,
+template<typename PSO>
+static void addDrawRangeElements(GLuint start,
                                  GLuint end,
                                  GLsizei count,
-                                 GLenum type,
                                  GLuint offset)
 {
-	ASSERT(mode == GL_TRIANGLES, "not supported");
-	ASSERT(type == GL_UNSIGNED_INT, "not supported");
 
 	if (end - start + 1 > GLmaxElementsVertices)
 	{
@@ -188,9 +190,9 @@ static void addDrawRangeElements(GLenum mode,
 	    dreCount + count > GLmaxElementsIndices ||
 	    end - dreStart + 1 > GLmaxElementsVertices)
 	{
-		finishDrawRangeElements();
+		finishDrawRangeElements<PSO>();
 		// start anew
-		addDrawRangeElements(mode, start, end, count, type, offset);
+		addDrawRangeElements<PSO>(start, end, count, offset);
 	}
 	else
 	{
@@ -238,7 +240,7 @@ static Vector2f getTileTexCoords(Vector2f *uv, unsigned int tileNumber)
 		texsize = MAX_TILE_TEXTURE_SIZE;
 	}
 	const float centertile = 0.5f / texsize;	// compute center of tile
-	const float shiftamount = (texsize - 1.0) / texsize;	// 1 pixel border
+	const float shiftamount = (texsize - 1.0f) / texsize;	// 1 pixel border
 	// bump the texture coords, for 1 pixel border, so our range is [.5,(texsize - .5)]
 	const float one = 1.0f / (TILES_IN_PAGE_COLUMN * texsize) + centertile * shiftamount;
 
@@ -505,9 +507,11 @@ static void updateSectorGeometry(int x, int y)
 	ASSERT(waterSize    == sectors[x * ySectors + y].waterSize   , "something went seriously wrong updating the terrain");
 
 	geometryVBO->update(sizeof(RenderVertex)*sectors[x * ySectors + y].geometryOffset,
-	                    sizeof(RenderVertex)*sectors[x * ySectors + y].geometrySize, geometry);
+	                    sizeof(RenderVertex)*sectors[x * ySectors + y].geometrySize, geometry,
+						gfx_api::buffer::update_flag::non_overlapping_updates_promise);
 	waterVBO->update(sizeof(RenderVertex)*sectors[x * ySectors + y].waterOffset,
-	                 sizeof(RenderVertex)*sectors[x * ySectors + y].waterSize, water);
+	                 sizeof(RenderVertex)*sectors[x * ySectors + y].waterSize, water,
+					 gfx_api::buffer::update_flag::non_overlapping_updates_promise);
 
 	free(geometry);
 	free(water);
@@ -522,13 +526,23 @@ static void updateSectorGeometry(int x, int y)
 	setSectorDecals(x, y, decaldata, &decalSize);
 	ASSERT(decalSize == sectors[x * ySectors + y].decalSize   , "the amount of decals has changed");
 
-	decalVBO->update(sizeof(DecalVertex)*sectors[x * ySectors + y].decalOffset,
-	                 sizeof(DecalVertex)*sectors[x * ySectors + y].decalSize, decaldata);
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
+	if (decalSize > 0)
+	{
+		if (decalVBO)
+		{
+			decalVBO->update(sizeof(DecalVertex)*sectors[x * ySectors + y].decalOffset,
+							 sizeof(DecalVertex)*sectors[x * ySectors + y].decalSize, decaldata,
+							 gfx_api::buffer::update_flag::non_overlapping_updates_promise);
+		}
+		else
+		{
+			// didn't have decals, but now we do??
+			// code needs a refactoring if this is the case
+			ASSERT(false, "Didn't have decals, but now we do. Unsupported.");
+		}
+	}
 
 	free(decaldata);
-
-	glBindBuffer(GL_ARRAY_BUFFER, 0);  // HACK Must unbind GL_ARRAY_BUFFER (don't know if it has to be unbound everywhere), otherwise text rendering may mysteriously crash.
 }
 
 /**
@@ -575,6 +589,22 @@ void markTileDirty(int i, int j)
 	}
 }
 
+void loadTerrainTextures()
+{
+	ASSERT_OR_RETURN(, psGroundTypes.get(), "Ground type was not set, no textures will be seen.");
+
+	int32_t maxGfxTextureSize = gfx_api::context::get().get_context_value(gfx_api::context::context_value::MAX_TEXTURE_SIZE);
+	int maxTerrainTextureSize = std::max(std::min({getTextureSize(), maxGfxTextureSize}), MIN_TERRAIN_TEXTURE_SIZE);
+
+	// for each terrain layer
+	for (int layer = 0; layer < numGroundTypes; layer++)
+	{
+		// pre-load the texture
+		optional<size_t> texPage = iV_GetTexture(psGroundTypes[layer].textureName, true, maxTerrainTextureSize, maxTerrainTextureSize);
+		ASSERT(texPage.has_value(), "Failed to pre-load terrain texture: %s", psGroundTypes[layer].textureName);
+	}
+}
+
 /**
  * Check what the videocard + drivers support and divide the loaded map into sectors that can be drawn.
  * It also determines the lightmap size.
@@ -600,8 +630,8 @@ bool initTerrain()
 	bool decreasedSize = false;
 
 	// this information is useful to prevent crashes with buggy opengl implementations
-	glGetIntegerv(GL_MAX_ELEMENTS_VERTICES, &GLmaxElementsVertices);
-	glGetIntegerv(GL_MAX_ELEMENTS_INDICES,  &GLmaxElementsIndices);
+	GLmaxElementsVertices = gfx_api::context::get().get_context_value(gfx_api::context::context_value::MAX_ELEMENTS_VERTICES);
+	GLmaxElementsIndices = gfx_api::context::get().get_context_value(gfx_api::context::context_value::MAX_ELEMENTS_INDICES);
 
 	// testing for crappy cards
 	debug(LOG_TERRAIN, "GL_MAX_ELEMENTS_VERTICES: %i", (int)GLmaxElementsVertices);
@@ -651,7 +681,7 @@ bool initTerrain()
 	// Create the sectors
 	xSectors = (mapWidth + sectorSize - 1) / sectorSize;
 	ySectors = (mapHeight + sectorSize - 1) / sectorSize;
-	sectors = (Sector *)malloc(sizeof(Sector) * xSectors * ySectors);
+	sectors = std::unique_ptr<Sector[]> (new Sector[xSectors * ySectors]());
 
 	////////////////////
 	// fill the geometry part of the sectors
@@ -767,10 +797,16 @@ bool initTerrain()
 
 	if (waterIndexVBO)
 		delete waterIndexVBO;
-	waterIndexVBO = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::index_buffer);
-	waterIndexVBO->upload(sizeof(GLuint)*waterIndexSize, waterIndex);
+	if (waterIndexSize > 0)
+	{
+		waterIndexVBO = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::index_buffer);
+		waterIndexVBO->upload(sizeof(GLuint)*waterIndexSize, waterIndex);
+	}
+	else
+	{
+		waterIndexVBO = nullptr;
+	}
 	free(waterIndex);
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
 
 
 	////////////////////
@@ -787,10 +823,10 @@ bool initTerrain()
 			{
 				if (layer == 0)
 				{
-					sectors[x * ySectors + y].textureOffset = (int *)malloc(sizeof(int) * numGroundTypes);
-					sectors[x * ySectors + y].textureSize = (int *)malloc(sizeof(int) * numGroundTypes);
-					sectors[x * ySectors + y].textureIndexOffset = (int *)malloc(sizeof(int) * numGroundTypes);
-					sectors[x * ySectors + y].textureIndexSize = (int *)malloc(sizeof(int) * numGroundTypes);
+					sectors[x * ySectors + y].textureOffset = std::unique_ptr<int[]> (new int[numGroundTypes]());
+					sectors[x * ySectors + y].textureSize = std::unique_ptr<int[]> (new int[numGroundTypes]());
+					sectors[x * ySectors + y].textureIndexOffset = std::unique_ptr<int[]> (new int[numGroundTypes]());
+					sectors[x * ySectors + y].textureIndexSize = std::unique_ptr<int[]> (new int[numGroundTypes]());
 				}
 
 				sectors[x * ySectors + y].textureOffset[layer] = textureSize;
@@ -886,7 +922,6 @@ bool initTerrain()
 	textureIndexVBO = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::index_buffer);
 	textureIndexVBO->upload(sizeof(GLuint)*textureIndexSize, textureIndex);
 	free(textureIndex);
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
 
 	// and finally the decals
 	decaldata = (DecalVertex *)malloc(sizeof(DecalVertex) * mapWidth * mapHeight * 12);
@@ -904,12 +939,17 @@ bool initTerrain()
 	debug(LOG_TERRAIN, "%i decals found", decalSize / 12);
 	if (decalVBO)
 		delete decalVBO;
-	decalVBO = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::vertex_buffer, gfx_api::context::buffer_storage_hint::dynamic_draw);
-	decalVBO->upload(sizeof(DecalVertex)*decalSize, decaldata);
+	if (decalSize > 0)
+	{
+		decalVBO = gfx_api::context::get().create_buffer_object(gfx_api::buffer::usage::vertex_buffer, gfx_api::context::buffer_storage_hint::dynamic_draw);
+		decalVBO->upload(sizeof(DecalVertex)*decalSize, decaldata);
+	}
+	else
+	{
+		decalVBO = nullptr;
+	}
 	free(decaldata);
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
 
-	lightmap_tex_num = 0;
 	lightmapLastUpdate = 0;
 	lightmapWidth = 1;
 	lightmapHeight = 1;
@@ -917,33 +957,23 @@ bool initTerrain()
 	while (mapWidth > (lightmapWidth <<= 1)) {}
 	while (mapHeight > (lightmapHeight <<= 1)) {}
 	debug(LOG_TERRAIN, "the size of the map is %ix%i", mapWidth, mapHeight);
-	debug(LOG_TERRAIN, "lightmap texture size is %ix%i", lightmapWidth, lightmapHeight);
+	debug(LOG_TERRAIN, "lightmap texture size is %zu x %zu", lightmapWidth, lightmapHeight);
 
 	// Prepare the lightmap pixmap and texture
-	lightmapPixmap = (GLubyte *)calloc(lightmapWidth * lightmapHeight, 3 * sizeof(GLubyte));
+	lightmapPixmap = std::unique_ptr<gfx_api::gfxUByte[]> (new gfx_api::gfxUByte[lightmapWidth * lightmapHeight * 3]());
 	if (lightmapPixmap == nullptr)
 	{
 		debug(LOG_FATAL, "Out of memory!");
 		abort();
 		return false;
 	}
-	if (lightmap_tex_num)
-		delete lightmap_tex_num;
-	lightmap_tex_num = gfx_api::context::get().create_texture(lightmapWidth, lightmapHeight, gfx_api::pixel_format::rgb);
-
-	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
-	lightmap_tex_num->upload(0, 0, 0, lightmapWidth, lightmapHeight, gfx_api::pixel_format::rgb, lightmapPixmap);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-
-	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, lightmapWidth, lightmapHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, lightmapPixmap);
+	if (lightmap_texture)
+		delete lightmap_texture;
+	lightmap_texture = gfx_api::context::get().create_texture(1, lightmapWidth, lightmapHeight, gfx_api::pixel_format::FORMAT_RGB8_UNORM_PACK8);
 
+	lightmap_texture->upload(0, 0, 0, lightmapWidth, lightmapHeight, gfx_api::pixel_format::FORMAT_RGB8_UNORM_PACK8, lightmapPixmap.get());
 	terrainInitialised = true;
 
-	glBindBuffer(GL_ARRAY_BUFFER, 0);  // HACK Must unbind GL_ARRAY_BUFFER (in this function, at least), otherwise text rendering may mysteriously crash.
-
 	return true;
 }
 
@@ -975,17 +1005,15 @@ void shutdownTerrain()
 	{
 		for (int y = 0; y < ySectors; y++)
 		{
-			free(sectors[x * ySectors + y].textureOffset);
-			free(sectors[x * ySectors + y].textureSize);
-			free(sectors[x * ySectors + y].textureIndexOffset);
-			free(sectors[x * ySectors + y].textureIndexSize);
+			sectors[x * ySectors + y].textureOffset = nullptr;
+			sectors[x * ySectors + y].textureSize = nullptr;
+			sectors[x * ySectors + y].textureIndexOffset = nullptr;
+			sectors[x * ySectors + y].textureIndexSize = nullptr;
 		}
 	}
-	free(sectors);
 	sectors = nullptr;
-	delete lightmap_tex_num;
-	lightmap_tex_num = nullptr;
-	free(lightmapPixmap);
+	delete lightmap_texture;
+	lightmap_texture = nullptr;
 	lightmapPixmap = nullptr;
 
 	terrainInitialised = false;
@@ -1017,8 +1045,8 @@ static void updateLightMap()
 			if (!pie_GetFogStatus())
 			{
 				// fade to black at the edges of the visible terrain area
-				const float playerX = map_coordf(player.p.x);
-				const float playerY = map_coordf(player.p.z);
+				const float playerX = map_coordf(playerPos.p.x);
+				const float playerY = map_coordf(playerPos.p.z);
 
 				const float distA = i - (playerX - visibleTiles.x / 2);
 				const float distB = (playerX + visibleTiles.x / 2) - i;
@@ -1068,7 +1096,7 @@ static void cullTerrain()
 		{
 			float xPos = world_coord(x * sectorSize + sectorSize / 2);
 			float yPos = world_coord(y * sectorSize + sectorSize / 2);
-			float distance = pow(player.p.x - xPos, 2) + pow(player.p.z - yPos, 2);
+			float distance = pow(playerPos.p.x - xPos, 2) + pow(playerPos.p.z - yPos, 2);
 
 			if (distance > pow((double)world_coord(terrainDistance), 2))
 			{
@@ -1089,27 +1117,19 @@ static void cullTerrain()
 
 static void drawDepthOnly(const glm::mat4 &ModelViewProjection, const glm::vec4 &paramsXLight, const glm::vec4 &paramsYLight)
 {
-	const auto &program = pie_ActivateShader(SHADER_TERRAIN_DEPTH, ModelViewProjection, paramsXLight, paramsYLight, 1, glm::mat4(1.f), glm::mat4(1.f));
-	pie_SetTexturePage(TEXPAGE_NONE);
-	pie_SetRendMode(REND_OPAQUE);
-
-	// we only draw in the depth buffer of using fog of war, as the clear color is black then
-	if (!pie_GetFogStatus())
-	{
-		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
-	}
+	const auto &renderState = getCurrentRenderState();
+	
+	// bind the vertex buffer
+	gfx_api::TerrainDepth::get().bind();
+	gfx_api::TerrainDepth::get().bind_textures(lightmap_texture);
+	gfx_api::TerrainDepth::get().bind_vertex_buffers(geometryVBO);
+	gfx_api::TerrainDepth::get().bind_constants({ ModelViewProjection, paramsXLight, paramsYLight, glm::vec4(0.f), glm::vec4(0.f), glm::mat4(1.f), glm::mat4(1.f), 
+	glm::vec4(0.f), renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, 0, 0 });
+	gfx_api::context::get().bind_index_buffer(*geometryIndexVBO, gfx_api::index_type::u32);
 
 	// draw slightly higher distance than it actually is so it will not
 	// by accident obscure the actual terrain
-	glEnable(GL_POLYGON_OFFSET_FILL);
-	glPolygonOffset(0.1f, 1.0f);
-
-	// bind the vertex buffer
-	geometryIndexVBO->bind();
-	geometryVBO->bind();
-
-	glVertexAttribPointer(program.locVertex, 3, GL_FLOAT, GL_FALSE, sizeof(RenderVertex), BUFFER_OFFSET(0));
-	glEnableVertexAttribArray(program.locVertex);
+	gfx_api::context::get().set_polygon_offset(0.1f, 1.f);
 
 	for (int x = 0; x < xSectors; x++)
 	{
@@ -1117,28 +1137,18 @@ static void drawDepthOnly(const glm::mat4 &ModelViewProjection, const glm::vec4
 		{
 			if (sectors[x * ySectors + y].draw)
 			{
-				addDrawRangeElements(GL_TRIANGLES,
+				addDrawRangeElements<gfx_api::TerrainDepth>(
 					sectors[x * ySectors + y].geometryOffset,
 					sectors[x * ySectors + y].geometryOffset + sectors[x * ySectors + y].geometrySize,
 					sectors[x * ySectors + y].geometryIndexSize,
-					GL_UNSIGNED_INT,
 					sectors[x * ySectors + y].geometryIndexOffset);
 			}
 		}
 	}
-	finishDrawRangeElements();
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
-	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
-
-	if (!pie_GetFogStatus())
-	{
-		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
-	}
-
-	// disable the depth offset
-	glDisable(GL_POLYGON_OFFSET_FILL);
-	glDisableVertexAttribArray(program.locVertex);
-	pie_DeactivateShader();
+	finishDrawRangeElements<gfx_api::TerrainDepth>();
+	gfx_api::context::get().set_polygon_offset(0.f, 0.f);
+	gfx_api::TerrainDepth::get().unbind_vertex_buffers(geometryVBO);
+	gfx_api::context::get().unbind_index_buffer(*geometryIndexVBO);
 }
 
 static void drawTerrainLayers(const glm::mat4 &ModelViewProjection, const glm::vec4 &paramsXLight, const glm::vec4 &paramsYLight, const glm::mat4 &textureMatrix)
@@ -1150,43 +1160,31 @@ static void drawTerrainLayers(const glm::mat4 &ModelViewProjection, const glm::v
 		renderState.fogColour.vector[2] / 255.f,
 		renderState.fogColour.vector[3] / 255.f
 	);
-	const auto &program = pie_ActivateShader(SHADER_TERRAIN, ModelViewProjection, glm::vec4(0.f), glm::vec4(0.f), paramsXLight, paramsYLight, 0, 1,
-		glm::mat4(1.f), textureMatrix, renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, fogColor);
-
-	// additive blending
-	pie_SetRendMode(REND_ADDITIVE);
-
-	// only draw colors
-	glDepthMask(GL_FALSE);
-
-	textureIndexVBO->bind();
 
 	// load the vertex (geometry) buffer
-	geometryVBO->bind();
-	glVertexAttribPointer(program.locVertex, 3, GL_FLOAT, GL_FALSE, sizeof(RenderVertex), BUFFER_OFFSET(0));
-	glEnableVertexAttribArray(program.locVertex);
+	gfx_api::TerrainLayer::get().bind();
+	gfx_api::TerrainLayer::get().bind_vertex_buffers(geometryVBO, textureVBO);
+	gfx_api::context::get().bind_index_buffer(*textureIndexVBO, gfx_api::index_type::u32);
+	ASSERT_OR_RETURN(, psGroundTypes.get(), "Ground type was not set, no textures will be seen.");
 
-	textureVBO->bind();
-
-	ASSERT_OR_RETURN(, psGroundTypes, "Ground type was not set, no textures will be seen.");
+	int32_t maxGfxTextureSize = gfx_api::context::get().get_context_value(gfx_api::context::context_value::MAX_TEXTURE_SIZE);
+	int maxTerrainTextureSize = std::max(std::min({getTextureSize(), maxGfxTextureSize}), MIN_TERRAIN_TEXTURE_SIZE);
 
 	// draw each layer separately
 	for (int layer = 0; layer < numGroundTypes; layer++)
 	{
 		const glm::vec4 paramsX(0, 0, -1.0f / world_coord(psGroundTypes[layer].textureSize), 0 );
 		const glm::vec4 paramsY(1.0f / world_coord(psGroundTypes[layer].textureSize), 0, 0, 0 );
-		pie_ActivateShader(SHADER_TERRAIN, ModelViewProjection, paramsX, paramsY, paramsXLight, paramsYLight, 0, 1, glm::mat4(1.f), textureMatrix,
-			renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, fogColor);
+		gfx_api::TerrainLayer::get().bind_constants({ ModelViewProjection, paramsX, paramsY, paramsXLight, paramsYLight, glm::mat4(1.f), textureMatrix,
+			fogColor, renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, 0, 1 });
 
 		// load the texture
-		int texPage = iV_GetTexture(psGroundTypes[layer].textureName);
-		pie_SetTexturePage(texPage);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
+		optional<size_t> texPage = iV_GetTexture(psGroundTypes[layer].textureName, true, maxTerrainTextureSize, maxTerrainTextureSize);
+		ASSERT_OR_RETURN(, texPage.has_value(), "Failed to retrieve terrain texture: %s", psGroundTypes[layer].textureName);
+		gfx_api::TerrainLayer::get().bind_textures(&pie_Texture(texPage.value()), lightmap_texture);
 
 		// load the color buffer
-		glVertexAttribPointer(program.locColor, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(PIELIGHT), BUFFER_OFFSET(sizeof(PIELIGHT)*xSectors * ySectors * (sectorSize + 1) * (sectorSize + 1) * 2 * layer));
-		glEnableVertexAttribArray(program.locColor);
+		gfx_api::context::get().bind_vertex_buffers(1, { std::make_tuple(textureVBO, static_cast<size_t>(sizeof(PIELIGHT)*xSectors * ySectors * (sectorSize + 1) * (sectorSize + 1) * 2 * layer)) });
 
 		for (int x = 0; x < xSectors; x++)
 		{
@@ -1194,27 +1192,27 @@ static void drawTerrainLayers(const glm::mat4 &ModelViewProjection, const glm::v
 			{
 				if (sectors[x * ySectors + y].draw)
 				{
-					addDrawRangeElements(GL_TRIANGLES,
+					addDrawRangeElements<gfx_api::TerrainLayer>(
 						sectors[x * ySectors + y].geometryOffset,
 						sectors[x * ySectors + y].geometryOffset + sectors[x * ySectors + y].geometrySize,
 						sectors[x * ySectors + y].textureIndexSize[layer],
-						GL_UNSIGNED_INT,
 						sectors[x * ySectors + y].textureIndexOffset[layer]);
 				}
 			}
 		}
-		finishDrawRangeElements();
+		finishDrawRangeElements<gfx_api::TerrainLayer>();
 	}
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
-	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
-	// we don't need this one anymore
-	glDisableVertexAttribArray(program.locVertex);
-	glDisableVertexAttribArray(program.locColor);
-	pie_DeactivateShader();
+	gfx_api::TerrainLayer::get().unbind_vertex_buffers(geometryVBO, textureVBO);
+	gfx_api::context::get().unbind_index_buffer(*textureIndexVBO);
 }
 
 static void drawDecals(const glm::mat4 &ModelViewProjection, const glm::vec4 &paramsXLight, const glm::vec4 &paramsYLight, const glm::mat4 &textureMatrix)
 {
+	if (!decalVBO)
+	{
+		return; // no decals
+	}
+
 	const auto &renderState = getCurrentRenderState();
 	const glm::vec4 fogColor(
 		renderState.fogColour.vector[0] / 255.f,
@@ -1222,22 +1220,11 @@ static void drawDecals(const glm::mat4 &ModelViewProjection, const glm::vec4 &pa
 		renderState.fogColour.vector[2] / 255.f,
 		renderState.fogColour.vector[3] / 255.f
 	);
-	const auto &program = pie_ActivateShader(SHADER_DECALS, ModelViewProjection, paramsXLight, paramsYLight, 0, 1, textureMatrix,
-		renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, fogColor);
-	// select the terrain texture page
-	pie_SetTexturePage(terrainPage);
-
-	// use the alpha to blend
-	pie_SetRendMode(REND_ALPHA);
-
-	// and the texture coordinates buffer
-	decalVBO->bind();
-	glVertexAttribPointer(program.locVertex, 3, GL_FLOAT, GL_FALSE, sizeof(DecalVertex), BUFFER_OFFSET(0));
-	glEnableVertexAttribArray(program.locVertex);
-
-	glVertexAttribPointer(program.locTexCoord, 2, GL_FLOAT, GL_FALSE, sizeof(DecalVertex), BUFFER_OFFSET(12));
-	glEnableVertexAttribArray(program.locTexCoord);
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
+	gfx_api::TerrainDecals::get().bind();
+	gfx_api::TerrainDecals::get().bind_textures(&pie_Texture(terrainPage), lightmap_texture);
+	gfx_api::TerrainDecals::get().bind_vertex_buffers(decalVBO);
+	gfx_api::TerrainDecals::get().bind_constants({ ModelViewProjection, textureMatrix, paramsXLight, paramsYLight,
+		fogColor, renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, 0, 1 });
 
 	int size = 0;
 	int offset = 0;
@@ -1254,7 +1241,7 @@ static void drawDecals(const glm::mat4 &ModelViewProjection, const glm::vec4 &pa
 			// can't append, so draw what we have and start anew
 			if (size > 0)
 			{
-				glDrawArrays(GL_TRIANGLES, offset, size);
+				gfx_api::TerrainDecals::get().draw(size, offset);
 			}
 			size = 0;
 			if (y < ySectors && sectors[x * ySectors + y].draw)
@@ -1264,11 +1251,7 @@ static void drawDecals(const glm::mat4 &ModelViewProjection, const glm::vec4 &pa
 			}
 		}
 	}
-
-	glDepthMask(GL_TRUE);
-	glDisableVertexAttribArray(program.locTexCoord);
-	glDisableVertexAttribArray(program.locVertex);
-	pie_DeactivateShader();
+	gfx_api::TerrainDecals::get().unbind_vertex_buffers(decalVBO);
 }
 
 
@@ -1279,14 +1262,11 @@ static void drawDecals(const glm::mat4 &ModelViewProjection, const glm::vec4 &pa
  */
 void drawTerrain(const glm::mat4 &mvp)
 {
-	const glm::vec4 paramsXLight(1.0f / world_coord(mapWidth) *((float)mapWidth / lightmapWidth), 0, 0, 0);
-	const glm::vec4 paramsYLight(0, 0, -1.0f / world_coord(mapHeight) *((float)mapHeight / lightmapHeight), 0);
+	const glm::vec4 paramsXLight(1.0f / world_coord(mapWidth) *((float)mapWidth / (float)lightmapWidth), 0, 0, 0);
+	const glm::vec4 paramsYLight(0, 0, -1.0f / world_coord(mapHeight) *((float)mapHeight / (float)lightmapHeight), 0);
 
 	///////////////////////////////////
 	// set up the lightmap texture
-	glActiveTexture(GL_TEXTURE1);
-	// bind the texture
-	lightmap_tex_num->bind();
 
 	// we limit the framerate of the lightmap, because updating a texture is an expensive operation
 	if (realTime - lightmapLastUpdate >= LIGHTMAP_REFRESH)
@@ -1294,18 +1274,15 @@ void drawTerrain(const glm::mat4 &mvp)
 		lightmapLastUpdate = realTime;
 		updateLightMap();
 
-		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
-		lightmap_tex_num->upload(0, 0, 0, lightmapWidth, lightmapHeight, gfx_api::pixel_format::rgb, lightmapPixmap);
+		lightmap_texture->upload(0, 0, 0, lightmapWidth, lightmapHeight, gfx_api::pixel_format::FORMAT_RGB8_UNORM_PACK8, lightmapPixmap.get());
 	}
 
 	///////////////////////////////////
 	// terrain culling
 	cullTerrain();
 
-	glActiveTexture(GL_TEXTURE0);
-
 	// shift the lightmap half a tile as lights are supposed to be placed at the center of a tile
-	const glm::mat4 lightMatrix = glm::translate(glm::vec3(1.f / lightmapWidth / 2, 1.f / lightmapHeight / 2, 0.f));
+	const glm::mat4 lightMatrix = glm::translate(glm::vec3(1.f / (float)lightmapWidth / 2, 1.f / (float)lightmapHeight / 2, 0.f));
 
 	//////////////////////////////////////
 	// canvas to draw on
@@ -1318,15 +1295,6 @@ void drawTerrain(const glm::mat4 &mvp)
 	//////////////////////////////////
 	// decals
 	drawDecals(mvp, paramsXLight, paramsYLight, lightMatrix);
-
-	////////////////////////////////
-	// disable the lightmap texture
-	glActiveTexture(GL_TEXTURE1);
-	glActiveTexture(GL_TEXTURE0);
-
-	// leave everything in a sane state so it won't mess up somewhere else
-
-	//glBindBuffer(GL_ARRAY_BUFFER, 0);  // HACK Must unbind GL_ARRAY_BUFFER (don't know if it has to be unbound everywhere), otherwise text rendering may mysteriously crash.
 }
 
 /**
@@ -1334,6 +1302,11 @@ void drawTerrain(const glm::mat4 &mvp)
  */
 void drawWater(const glm::mat4 &viewMatrix)
 {
+	if (!waterIndexVBO)
+	{
+		return; // no water
+	}
+
 	int x, y;
 	const glm::vec4 paramsX(0, 0, -1.0f / world_coord(4), 0);
 	const glm::vec4 paramsY(1.0f / world_coord(4), 0, 0, 0);
@@ -1341,30 +1314,19 @@ void drawWater(const glm::mat4 &viewMatrix)
 	const glm::vec4 paramsY2(1.0f / world_coord(5), 0, 0, 0);
 	const auto &renderState = getCurrentRenderState();
 
-	const auto &program = pie_ActivateShader(SHADER_WATER, viewMatrix, paramsX, paramsY, paramsX2, paramsY2, 0, 1,
-		glm::translate(glm::vec3(waterOffset, 0.f, 0.f)), glm::mat4(1.f), renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd);
+	int32_t maxGfxTextureSize = gfx_api::context::get().get_context_value(gfx_api::context::context_value::MAX_TEXTURE_SIZE);
+	int maxTerrainTextureSize = std::max(std::min({getTextureSize(), maxGfxTextureSize}), MIN_TERRAIN_TEXTURE_SIZE);
 
-	glDepthMask(GL_FALSE);
-
-	// first texture unit
-	pie_SetTexturePage(iV_GetTexture("page-80-water-1.png"));
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-
-	// multiplicative blending
-	pie_SetRendMode(REND_MULTIPLICATIVE);
-
-	// second texture unit
-	glActiveTexture(GL_TEXTURE1);
-	pie_Texture(iV_GetTexture("page-81-water-2.png")).bind();
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-
-	// bind the vertex buffer
-	waterIndexVBO->bind();
-	waterVBO->bind();
-	glVertexAttribPointer(program.locVertex, 3, GL_FLOAT, GL_FALSE, sizeof(RenderVertex), BUFFER_OFFSET(0));
-	glEnableVertexAttribArray(program.locVertex);
+	optional<size_t> water1_texPage = iV_GetTexture("page-80-water-1.png", true, maxTerrainTextureSize, maxTerrainTextureSize);
+	optional<size_t> water2_texPage = iV_GetTexture("page-81-water-2.png", true, maxTerrainTextureSize, maxTerrainTextureSize);
+	ASSERT_OR_RETURN(, water1_texPage.has_value() && water2_texPage.has_value(), "Failed to load water texture");
+	gfx_api::WaterPSO::get().bind();
+	gfx_api::WaterPSO::get().bind_textures(&pie_Texture(water1_texPage.value()), &pie_Texture(water2_texPage.value()));
+	gfx_api::WaterPSO::get().bind_vertex_buffers(waterVBO);
+	gfx_api::WaterPSO::get().bind_constants({ viewMatrix, paramsX, paramsY, paramsX2, paramsY2,
+		glm::translate(glm::vec3(waterOffset, 0.f, 0.f)), glm::mat4(1.f), glm::vec4(0.f), renderState.fogEnabled, renderState.fogBegin, renderState.fogEnd, 0, 1
+	});
+	gfx_api::context::get().bind_index_buffer(*waterIndexVBO, gfx_api::index_type::u32);
 
 	for (x = 0; x < xSectors; x++)
 	{
@@ -1372,33 +1334,21 @@ void drawWater(const glm::mat4 &viewMatrix)
 		{
 			if (sectors[x * ySectors + y].draw)
 			{
-				addDrawRangeElements(GL_TRIANGLES,
+				addDrawRangeElements<gfx_api::WaterPSO>(
 				                     sectors[x * ySectors + y].geometryOffset,
 				                     sectors[x * ySectors + y].geometryOffset + sectors[x * ySectors + y].geometrySize,
 				                     sectors[x * ySectors + y].waterIndexSize,
-				                     GL_UNSIGNED_INT,
 				                     sectors[x * ySectors + y].waterIndexOffset);
 			}
 		}
 	}
-	finishDrawRangeElements();
-	glBindBuffer(GL_ARRAY_BUFFER, 0);
-	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
-
-	glDisableVertexAttribArray(program.locVertex);
+	finishDrawRangeElements<gfx_api::WaterPSO>();
+	gfx_api::WaterPSO::get().unbind_vertex_buffers(waterVBO);
+	gfx_api::context::get().unbind_index_buffer(*waterIndexVBO);
 
 	// move the water
 	if (!gamePaused())
 	{
 		waterOffset += graphicsTimeAdjustedIncrement(0.1f);
 	}
-
-	// disable second texture
-	glActiveTexture(GL_TEXTURE0);
-
-	// clean up
-	glDepthMask(GL_TRUE);
-	pie_DeactivateShader();
-
-	//glBindBuffer(GL_ARRAY_BUFFER, 0);  // HACK Must unbind GL_ARRAY_BUFFER (don't know if it has to be unbound everywhere), otherwise text rendering may mysteriously crash.
 }
