@@ -1,7 +1,7 @@
 /*
 	This file is part of Warzone 2100.
 	Copyright (C) 1999-2004  Eidos Interactive
-	Copyright (C) 2005-2019  Warzone 2100 Project
+	Copyright (C) 2005-2020  Warzone 2100 Project
 
 	Warzone 2100 is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -30,6 +30,8 @@
 #include "lib/sound/audio_id.h"
 #include "lib/ivis_opengl/ivisdef.h"
 
+#include <limits>
+
 #include "visibility.h"
 
 #include "objects.h"
@@ -40,7 +42,6 @@
 #include "hci.h"
 #include "mapgrid.h"
 #include "research.h"
-#include "scriptextern.h"
 #include "structure.h"
 #include "projectile.h"
 #include "display.h"
@@ -79,9 +80,6 @@ public:
 };
 static std::vector<SPOTTER *> apsInvisibleViewers;
 
-// horrible hack because this code is full of them and I ain't rewriting it all - Per
-#define MAX_SEEN_TILES (29*29 * 355/113)  // Increased hack to support 28 tile sensor radius. - Cyp
-
 #define MIN_VIS_HEIGHT 80
 
 struct VisibleObjectHelp_t
@@ -136,6 +134,7 @@ static inline void updateTileVis(MAPTILE *psTile)
 
 uint32_t addSpotter(int x, int y, int player, int radius, bool radar, uint32_t expiry)
 {
+	ASSERT_OR_RETURN(0, player >= 0 && player < MAX_PLAYERS, "invalid player: %d", player);
 	SPOTTER *psSpot = new SPOTTER(x, y, player, radius, (int)radar, expiry);
 	size_t size;
 	const WavecastTile *tiles = getWavecastTable(radius, &size);
@@ -228,9 +227,8 @@ SPOTTER::~SPOTTER()
 
 /* Record all tiles that some object confers visibility to. Only record each tile
  * once. Note that there is both a limit to how many objects can watch any given
- * tile, and a limit to how many tiles each object can watch. Strange but non fatal
- * things will happen if these limits are exceeded. This function uses icky globals. */
-static inline void visMarkTile(const BASE_OBJECT *psObj, int mapX, int mapY, MAPTILE *psTile, TILEPOS *recordTilePos, int *lastRecordTilePos)
+ * tile. Strange but non fatal things will happen if these limits are exceeded. */
+static inline void visMarkTile(const BASE_OBJECT *psObj, int mapX, int mapY, MAPTILE *psTile, std::vector<TILEPOS> &watchedTiles)
 {
 	const int rayPlayer = psObj->player;
 	const int xdiff = map_coord(psObj->pos.x) - mapX;
@@ -239,7 +237,7 @@ static inline void visMarkTile(const BASE_OBJECT *psObj, int mapX, int mapY, MAP
 	const bool inRange = (distSq < 16);
 	uint8_t *visionType = inRange ? psTile->watchers : psTile->sensors;
 
-	if (visionType[rayPlayer] < UBYTE_MAX && *lastRecordTilePos < MAX_SEEN_TILES)
+	if (visionType[rayPlayer] < UBYTE_MAX)
 	{
 		TILEPOS tilePos = {uint8_t(mapX), uint8_t(mapY), uint8_t(inRange)};
 
@@ -250,13 +248,12 @@ static inline void visMarkTile(const BASE_OBJECT *psObj, int mapX, int mapY, MAP
 			psTile->jammerBits |= (1 << rayPlayer); // mark it as being jammed
 		}
 		updateTileVis(psTile);
-		recordTilePos[*lastRecordTilePos] = tilePos;    // record having seen it
-		++*lastRecordTilePos;
+		watchedTiles.push_back(tilePos);  // record having seen it
 	}
 }
 
 /* The terrain revealing ray callback */
-static void doWaveTerrain(const BASE_OBJECT *psObj, TILEPOS *recordTilePos, int *lastRecordTilePos)
+static void doWaveTerrain(BASE_OBJECT *psObj)
 {
 	const int sx = psObj->pos.x;
 	const int sy = psObj->pos.y;
@@ -278,6 +275,7 @@ static void doWaveTerrain(const BASE_OBJECT *psObj, TILEPOS *recordTilePos, int
 	angles[!readList][writeListPos] = 0;               // Smallest angle.
 	++writeListPos;
 
+	psObj->watchedTiles.clear();
 	for (size_t i = 0; i < size; ++i)
 	{
 		const int mapX = map_coord(sx) + tiles[i].dx;
@@ -333,7 +331,7 @@ static void doWaveTerrain(const BASE_OBJECT *psObj, TILEPOS *recordTilePos, int
 		{
 			// Can see this tile.
 			psTile->tileExploredBits |= alliancebits[rayPlayer];                        // Share exploration with allies too
-			visMarkTile(psObj, mapX, mapY, psTile, recordTilePos, lastRecordTilePos);   // Mark this tile as seen by our sensor
+			visMarkTile(psObj, mapX, mapY, psTile, psObj->watchedTiles);   // Mark this tile as seen by our sensor
 		}
 	}
 }
@@ -372,9 +370,13 @@ static bool rayLOSCallback(Vector2i pos, int32_t dist, void *data)
 			MAPTILE *psTile = mapTile(tile);
 			if (TileHasWall(psTile) && !TileHasSmallStructure(psTile))
 			{
-				help->lastHeight = 2 * UBYTE_MAX * ELEVATION_SCALE;
-				help->wall = pos.xy();
-				help->numWalls++;
+				STRUCTURE *psStruct = (STRUCTURE *)psTile->psObject;
+				if (psStruct->pStructureType->type != REF_GATE || psStruct->state != SAS_OPEN)
+				{
+					help->lastHeight = 2 * TILE_MAX_HEIGHT;
+					help->wall = pos.xy();
+					help->numWalls++;
+				}
 			}
 		}
 	}
@@ -386,17 +388,16 @@ static bool rayLOSCallback(Vector2i pos, int32_t dist, void *data)
 /* Remove tile visibility from object */
 void visRemoveVisibility(BASE_OBJECT *psObj)
 {
-	if (psObj->watchedTiles && mapWidth && mapHeight)
+	if (mapWidth && mapHeight)
 	{
-		for (int i = 0; i < psObj->numWatchedTiles; i++)
+		for (TILEPOS pos : psObj->watchedTiles)
 		{
-			const TILEPOS pos = psObj->watchedTiles[i];
 			// FIXME: the mapTile might have been swapped out, see swapMissionPointers()
 			MAPTILE *psTile = mapTile(pos.x, pos.y);
 
 			ASSERT(pos.type < 2, "Invalid visibility type %d", (int)pos.type);
 			uint8_t *visionType = (pos.type == 0) ? psTile->sensors : psTile->watchers;
-			if (visionType[psObj->player] == 0 && game.type == CAMPAIGN)	// hack
+			if (visionType[psObj->player] == 0 && game.type == LEVEL_TYPE::CAMPAIGN)	// hack
 			{
 				continue;
 			}
@@ -415,25 +416,18 @@ void visRemoveVisibility(BASE_OBJECT *psObj)
 			updateTileVis(psTile);
 		}
 	}
-	free(psObj->watchedTiles);
-	psObj->watchedTiles = nullptr;
-	psObj->numWatchedTiles = 0;
+	psObj->watchedTiles.clear();
 	psObj->flags.set(OBJECT_FLAG_JAMMED_TILES, false);
 }
 
 void visRemoveVisibilityOffWorld(BASE_OBJECT *psObj)
 {
-	free(psObj->watchedTiles);
-	psObj->watchedTiles = nullptr;
-	psObj->numWatchedTiles = 0;
+	psObj->watchedTiles.clear();
 }
 
 /* Check which tiles can be seen by an object */
 void visTilesUpdate(BASE_OBJECT *psObj)
 {
-	TILEPOS recordTilePos[MAX_SEEN_TILES];
-	int lastRecordTilePos = 0;
-
 	ASSERT(psObj->type != OBJ_FEATURE, "visTilesUpdate: visibility updates are not for features!");
 
 	// Remove previous map visibility provided by object
@@ -452,23 +446,20 @@ void visTilesUpdate(BASE_OBJECT *psObj)
 
 	// Do the whole circle in ∞ steps. No more pretty moiré patterns.
 	psObj->flags.set(OBJECT_FLAG_JAMMED_TILES, objJammerPower(psObj) > 0);
-	doWaveTerrain(psObj, recordTilePos, &lastRecordTilePos);
-
-	// Record new map visibility provided by object
-	if (lastRecordTilePos > 0)
-	{
-		psObj->watchedTiles = (TILEPOS *)malloc(lastRecordTilePos * sizeof(*psObj->watchedTiles));
-		psObj->numWatchedTiles = lastRecordTilePos;
-		memcpy(psObj->watchedTiles, recordTilePos, lastRecordTilePos * sizeof(*psObj->watchedTiles));
-	}
+	doWaveTerrain(psObj);
 }
 
 /*reveals all the terrain in the map*/
 void revealAll(UBYTE player)
 {
-	UWORD   i, j;
+	SDWORD   i, j;
 	MAPTILE	*psTile;
 
+	if (player >= MAX_PLAYERS)
+	{
+		return;
+	}
+
 	//reveal all tiles
 	for (i = 0; i < mapWidth; i++)
 	{
@@ -593,10 +584,6 @@ int visibleObject(const BASE_OBJECT *psViewer, const BASE_OBJECT *psTarget, bool
 		*gNumWalls = help.numWalls;
 	}
 
-	// See if the target can be seen
-	int top = psTarget->pos.z + map_Height(psViewer->pos.x, psViewer->pos.y) - help.startHeight;
-	int targetGrad = top * GRAD_MUL / MAX(1, help.lastDist);
-
 	bool tileWatched = psTile->watchers[psViewer->player] > 0;
 	bool tileWatchedSensor = psTile->sensors[psViewer->player] > 0;
 
@@ -609,7 +596,7 @@ int visibleObject(const BASE_OBJECT *psViewer, const BASE_OBJECT *psTarget, bool
 		}
 	}
 	// Show objects that are seen directly
-	if ((tileWatched && targetGrad >= help.currGrad) || (!jammed && tileWatchedSensor))
+	if (tileWatched || (!jammed && tileWatchedSensor))
 	{
 		return UBYTE_MAX;
 	}
@@ -661,13 +648,16 @@ STRUCTURE *visGetBlockingWall(const BASE_OBJECT *psViewer, const BASE_OBJECT *ps
 
 bool hasSharedVision(unsigned viewer, unsigned ally)
 {
-	ASSERT_OR_RETURN(false, viewer < MAX_PLAYERS && ally < MAX_PLAYERS, "Bad viewer %u or ally %u.", viewer, ally);
+	if (viewer >= MAX_PLAYERS) { return false; }
+	ASSERT_OR_RETURN(false, ally < MAX_PLAYERS, "Bad ally %u (viewer: %u)", ally, viewer);
 
 	return viewer == ally || (bMultiPlayer && alliancesSharedVision(game.alliance) && aiCheckAlliances(viewer, ally));
 }
 
 static void setSeenBy(BASE_OBJECT *psObj, unsigned viewer, int val /*= UBYTE_MAX*/)
 {
+	if (viewer >= MAX_PLAYERS) { return; }
+
 	//forward out vision to our allies
 	for (int ally = 0; ally < MAX_PLAYERS; ++ally)
 	{
@@ -680,6 +670,8 @@ static void setSeenBy(BASE_OBJECT *psObj, unsigned viewer, int val /*= UBYTE_MAX
 
 static void setSeenByInstantly(BASE_OBJECT *psObj, unsigned viewer, int val /*= UBYTE_MAX*/)
 {
+	if (viewer >= MAX_PLAYERS) { return; }
+
 	//forward out vision to our allies
 	for (int ally = 0; ally < MAX_PLAYERS; ++ally)
 	{
@@ -717,7 +709,7 @@ static void processVisibilitySelf(BASE_OBJECT *psObj)
 	// you can always see anything that a CB sensor is targetting
 	// Anyone commenting this out again will get a knee capping from John.
 	// You have been warned!!
-	if (psStruct != nullptr && (structCBSensor(psStruct) || structVTOLCBSensor(psStruct)) && psStruct->psTarget[0] != nullptr)
+	if (psStruct != nullptr && psStruct->status == SS_BUILT && (structCBSensor(psStruct) || structVTOLCBSensor(psStruct)) && psStruct->psTarget[0] != nullptr)
 	{
 		setSeenByInstantly(psStruct->psTarget[0], psObj->player, UBYTE_MAX);
 	}
@@ -762,7 +754,7 @@ static void processVisibilityVision(BASE_OBJECT *psViewer)
 
 /* Find out what can see this object */
 // Fade in/out of view. Must be called after calculation of which objects are seen.
-static void processVisibilityLevel(BASE_OBJECT *psObj)
+static void processVisibilityLevel(BASE_OBJECT *psObj, bool& addedMessage)
 {
 	// update the visibility levels
 	for (unsigned player = 0; player < MAX_PLAYERS; player++)
@@ -826,6 +818,7 @@ static void processVisibilityLevel(BASE_OBJECT *psObj)
 					{
 						psMessage->psObj = psObj;
 						debug(LOG_MSG, "Added message for oil well or artefact, pViewData=%p", static_cast<void *>(psMessage->pViewData));
+						addedMessage = true;
 					}
 					if (!bInTutorial && player == selectedPlayer)
 					{
@@ -880,6 +873,7 @@ void processVisibility()
 			}
 		}
 	}
+	bool addedMessage = false;
 	for (int player = 0; player < MAX_PLAYERS; ++player)
 	{
 		BASE_OBJECT *lists[] = {apsDroidLists[player], apsStructLists[player], apsFeatureLists[player]};
@@ -888,10 +882,14 @@ void processVisibility()
 		{
 			for (BASE_OBJECT *psObj = lists[list]; psObj != nullptr; psObj = psObj->psNext)
 			{
-				processVisibilityLevel(psObj);
+				processVisibilityLevel(psObj, addedMessage);
 			}
 		}
 	}
+	if (addedMessage)
+	{
+		jsDebugMessageUpdate();
+	}
 }
 
 void	setUnderTilesVis(BASE_OBJECT *psObj, UDWORD player)
@@ -903,6 +901,11 @@ void	setUnderTilesVis(BASE_OBJECT *psObj, UDWORD player)
 	FEATURE_STATS const *psStats;
 	MAPTILE		*psTile;
 
+	if (player >= MAX_PLAYERS)
+	{
+		return;
+	}
+
 	if (psObj->type == OBJ_FEATURE)
 	{
 		psFeature = (FEATURE *)psObj;
